#!/usr/bin/env python
# -*- coding: utf8 -*-

from __future__ import print_function
from __future__ import unicode_literals

import os
import sys
import time
import datetime
import socket
import Queue
import multiprocessing
import logging
import logging.handlers
import traceback
import signal
import threading
import uuid
import argparse
import setproctitle

from driverng.protocol.message_pb2 import MCAMessage, MonitoringMessage
from driverng.protocol import get_protocole
from driverng.protocol.mca import MCAProtocol
from driverng.transport import get_transport
from driverng.transport.base import TransportException

if getattr(sys, "frozen", False):
    from driverng.version import VERSION
else:
    import pkg_resources
    VERSION = {
        "Version": pkg_resources.require("driverng")[0].version,
    }

LOG_FORMAT = "%(asctime)-15s %(processName)s[pid=%(process)d] %(message)s"

logger = logging.getLogger("driverng")
log_formatter = logging.Formatter(LOG_FORMAT)


class ConnexionAutomate(multiprocessing.Process):

    def __init__(self, id_connexion, options):
        self.rqueue = multiprocessing.Queue()
        self.wqueue = multiprocessing.Queue()
        self.id_connexion = id_connexion
        self.options = options
        self.status = MonitoringMessage.STATUS_STARTED
        multiprocessing.Process.__init__(self, name=options["label"])
        self.logger = None
        self.unstable_event = multiprocessing.Event()
        self.ack_queue = {}

    @staticmethod
    def check_parent():
        me = os.getpid()
        while True:
            if os.getppid() == 1:
                os.kill(me, signal.SIGKILL)
            time.sleep(1)

    def run_driver(self):
        transport = None
        while True:
            try:
                if setproctitle:
                    setproctitle.setproctitle("driver [%s]: running" % (
                        self.options["label"],
                    ))
                proto = get_protocole(self.options, self.logger)
                transport = get_transport(self.options, self.logger)
                proto.connect(transport)
                proto.start(self.rqueue, self.wqueue, background=False)
            except SystemExit:
                self.logger.info("Connexion stopped")
                break
            except Exception as e:
                if isinstance(e, TransportException):
                    self.unstable_event.set()
                    self.logger.error(
                        "TransportException: %s restart in 3 seconds", e)
                    if transport is not None:
                        transport.close()
                    time.sleep(3)
                else:
                    for line in traceback.format_exc().decode("ISO-8859-15").splitlines():
                        self.logger.error(line)
                    raise

    def run(self):
        signal.signal(signal.SIGTERM, signal.SIG_DFL)
        self.logger = logging.getLogger(self.options["label"])
        self.logger.setLevel(logging.DEBUG)
        if "--daemon" in sys.argv:
            log_handler = logging.handlers.TimedRotatingFileHandler(
                os.path.expanduser("~/logs/%s.log" % self.options["label"]),
                when="midnight",
                interval=1,
                backupCount=30,
            )
        else:
            log_handler = logging.StreamHandler()
        log_handler.setFormatter(log_formatter)
        self.logger.addHandler(log_handler)
        t = threading.Thread(target=self.check_parent)
        t.daemon = True
        t.start()
        self.run_driver()


class MCAClientError(Exception):
    pass


class MCAClient(MCAProtocol):
    managed_connexions = {}

    def __init__(self, addr, node_name):
        self.node_name = node_name
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            sock.connect(addr)
        except socket.error:
            self.handle_traceback()
            try:
                sock.close()
            except socket.error:
                self.handle_traceback()
            raise
        sock.setblocking(0)

        super(MCAClient, self).__init__(sock, logger)

        self.last_status = datetime.datetime.now()
        self.sig_hup = False
        signal.signal(signal.SIGHUP, self.reload_config)
        signal.signal(signal.SIGTERM, self.stop)

        if setproctitle:
            setproctitle.setproctitle("driver [manager]: running")

        self.configure()
        for cnx in MCAClient.managed_connexions.values():
            for msg in cnx.ack_queue.values():
                self.send_message(msg)

    def stop(self, signum, frame):
        self.global_exit = True
        for connexion in MCAClient.managed_connexions.itervalues():
            logger.info(
                "Try to stop connexion %s", connexion.options["label"])
            connexion.terminate()
            connexion.join()

    def reload_config(self, signum, frame):
        logger.info("SIGHUP recevied, force reloading")
        self.sig_hup = True

    def force_check(self):
        self.last_status -= datetime.timedelta(minutes=1)

    def configure(self):
        logger.info("Send config request")
        msg = MCAMessage()
        msg.type = MCAMessage.CONFIG_REQUEST
        msg.config_request.hostname = self.node_name
        msg.config_request.version = VERSION["Version"]
        self.send_message(msg)

    def do_something(self):

        if self.sig_hup:
            self.sig_hup = False
            self.configure()

        for connexion in MCAClient.managed_connexions.itervalues():

            if connexion.status == MonitoringMessage.STATUS_STOPPED:
                continue

            if not connexion.is_alive():
                logger.warning(
                    "Connexion %s (pid=%d) died",
                    connexion.name,
                    connexion.pid,
                )
                connexion.status = MonitoringMessage.STATUS_STOPPED
                self.force_check()
                continue

            if connexion.unstable_event.is_set():
                connexion.unstable_event.clear()
                if connexion.status != MonitoringMessage.STATUS_UNSTABLE:
                    connexion.status = MonitoringMessage.STATUS_UNSTABLE
                    self.force_check()
                    logger.warning(
                        "Connexion %s (pid=%d) is unstable",
                        connexion.name,
                        connexion.pid,
                    )

            while True:
                try:
                    data = connexion.rqueue.get_nowait()
                except Queue.Empty:
                    break

                msg = None

                if data["data"]:
                    msg = MCAMessage()
                    msg.type = MCAMessage.DATA
                    msg.data.id = connexion.id_connexion
                    msg.data.data = data["data"]
                    msg.data.uuid = uuid.uuid4().hex
                    connexion.ack_queue[msg.data.uuid] = msg
                elif data["uuid"]:  # ack
                    msg = MCAMessage()
                    msg.type = MCAMessage.ACKDATA
                    msg.ackdata.uuid = data["uuid"]

                if msg is not None:
                    if connexion.status == MonitoringMessage.STATUS_UNSTABLE:
                        logger.info(
                            "Connexion %s (pid=%s) become stable",
                            connexion.name,
                            connexion.pid,
                        )
                        connexion.status = MonitoringMessage.STATUS_STARTED
                        self.force_check()
                    self.send_message(msg)

        now = datetime.datetime.now()
        if now - self.last_status > datetime.timedelta(minutes=1):
            msg = self.get_status_message()
            self.last_status = now
            self.send_message(msg)

    def handle_message(self, msg):
        type_func_map = {
            MCAMessage.CONFIG_REPLY: self.handle_config_reply_message,
            MCAMessage.DATA: self.handle_data_message,
            MCAMessage.ACKDATA: self.handle_ackdata_message,
        }

        try:
            func = type_func_map[msg.type]
        except KeyError:
            raise MCAClientError("Unknown message type %d" % (msg.type,))

        func(msg)

    def handle_config_reply_message(self, msg):
        logger.info("Configuring connexions")
        active_connexions = []
        if msg.config.version:
            logger.info("Server version: %s", msg.config.version)
        for config in msg.config.config:
            active_connexions.append(config.id)
            options = {}
            for opt in config.options:
                options[opt.key] = opt.value

            if config.id in MCAClient.managed_connexions:
                connexion = MCAClient.managed_connexions[config.id]
                if connexion.is_alive() and options == connexion.options:
                    connexion.status = MonitoringMessage.STATUS_STARTED
                    logger.info("Connexion %s, no changes", options["label"])
                    continue
                else:
                    if connexion.is_alive():
                        logger.info(
                            "Connexion %s killed (for update)",
                            connexion.options["label"],
                        )
                        connexion.terminate()
                        connexion.join()

            connexion = ConnexionAutomate(config.id, options)
            MCAClient.managed_connexions[config.id] = connexion
            logger.info("Connexion %s started", options["label"])
            connexion.start()

        for pk, connexion in MCAClient.managed_connexions.items():
            if pk not in active_connexions:
                logger.info(
                    "Connexion %s killed (inactive)",
                    connexion.options["label"],
                )
                if connexion.is_alive():
                    connexion.terminate()
                    connexion.join()
                MCAClient.managed_connexions[pk].status = MonitoringMessage.STATUS_STOPPED

        self.force_check()

    @staticmethod
    def handle_data_message(msg):
        MCAClient.managed_connexions[msg.data.id].wqueue.put({
            "data": msg.data.data,
            "uuid": msg.data.uuid,
            "meta": dict((m.key, m.value) for m in msg.data.meta),
        })

    @staticmethod
    def handle_ackdata_message(msg):
        if msg.ackdata.id:
            try:
                del MCAClient.managed_connexions[msg.ackdata.id].ack_queue[msg.ack_queue.uuid]
            except KeyError:
                logger.warning("Unknown uuid %s", uuid)
        else:
            # LEGACY
            for cnx in MCAClient.managed_connexions.values():
                try:
                    del cnx.ack_queue[msg.ackdata.uuid]
                    break
                except KeyError:
                    pass

    @staticmethod
    def get_status_message():
        msg = MCAMessage()
        msg.type = MCAMessage.MONITORING
        for pk, connexion in MCAClient.managed_connexions.items():
            status = msg.monitoring.status.add()
            status.id = pk
            status.status = connexion.status

            if connexion.status == MonitoringMessage.STATUS_STOPPED:
                # garbage
                del MCAClient.managed_connexions[pk]

        return msg


class Driver(object):
    def __init__(self, args):

        try:
            self.server, port = args.server.split(":")
            self.port = int(port)
        except ValueError:
            self.server = args.server
            self.port = 6666
        self.node_name = args.node_name

        self.global_exit = False
        signal.signal(signal.SIGTERM, self.stop)

    def stop(self, signum, frame):
        self.global_exit = True

    def run(self):
        mca = None
        was_connected = True
        while not self.global_exit:
            try:
                mca = MCAClient((self.server, self.port), self.node_name)
                logger.info("Connected to MCA")
                was_connected = True
                mca.run_protocol()
            except Exception:
                pass

            if setproctitle:
                setproctitle.setproctitle("driver [manager]: disconnected")

            if mca is not None and mca.global_exit:
                self.global_exit = True
            if was_connected:
                logger.info("Disconnected from MCA, "
                            "try reconnect in 3 seconds")
                was_connected = False
            time.sleep(3)

    @staticmethod
    def start_driver():
        logger.setLevel(logging.DEBUG)
        hostname = socket.gethostname()
        parser = argparse.ArgumentParser(description="Driver NG")
        parser.add_argument(
            "--server",
            default="mca1",
            help="Server IP host:port or host (default port 6666)",
        )
        parser.add_argument(
            "--node-name",
            default=hostname,
            help="Node name",
        )
        parser.add_argument(
            "--daemon",
            action="store_true",
            help="Run in daemon mode",
        )
        parser.add_argument(
            "--version",
            action="store_true",
            help="Show version",
        )
        args = parser.parse_args()

        if args.version:
            print("Driver NG")
            print("\n".join(["%s: %s" % (k, v) for k, v in VERSION.items()]))
            sys.exit(0)

        if args.daemon:
            log_handler = logging.handlers.TimedRotatingFileHandler(
                os.path.expanduser("~/logs/driver.log"),
                when="midnight",
                backupCount=30,
            )
        else:
            log_handler = logging.StreamHandler()

        log_handler.setFormatter(log_formatter)
        logger.addHandler(log_handler)
        Driver(args).run()


def main():
    Driver.start_driver()

if __name__ == "__main__":
    main()
