# -*- coding: utf8 -*-

from __future__ import unicode_literals

from driverng.astm import (
    AstmMessage,
    HeaderSegment,
    Segment,
    HprimString,
    HprimMultiple,
    HprimList,
    HprimDate,
    HprimForeign,
)


class AblSegmentH(HeaderSegment):

    sender_type = HprimString((5, 1), default="MCA")
    sender_name = HprimString((5, 2))
    comment = HprimString(11)
    version = HprimString(13, default='1')
    date = HprimDate(14)
    patients = HprimForeign("P", "header")
    query = HprimForeign("Q", "header")
    end = HprimForeign(
        "L", "header", one2one=True, required=True
    )


class AblSegmentP(Segment):

    patient_lab_id = HprimString(4)
    last_name = HprimString((6, 1))
    first_name = HprimString((6, 2))
    birthdate = HprimDate(8, fmt_date="%Y%m%d")
    # M/F/U
    sex = HprimString(9)
    age = HprimString((15, 1))
    age_unit = HprimString((15, 2))
    gestation_age = HprimString((16, 1))
    gestation_age_unit = HprimString((16, 2))
    gestation_weight = HprimString((16, 3))
    gestation_weight_unit = HprimString((16, 4))
    height = HprimString((17, 1))
    height_unit = HprimString((17, 2))
    weight = HprimString((18, 1))
    weight_unit = HprimString((18, 2))
    location = HprimString(26)
    orders = HprimForeign("O", "patient")


class AblSegmentQ(Segment):
    patient_id = HprimString((3, 1))
    sample_id = HprimString((3, 2))
    location = HprimString((11, 1))
    patient_dept = HprimString((11, 2))
    status_code = HprimString(13)


class AblSegmentO(Segment):

    specimen_id = HprimString(3)
    # QC #: QC
    # Cal #: Calibration
    # Sample #: Patient
    sample_type = HprimString((4, 1))
    sample_no = HprimString((4, 2))
    tests = HprimList(5, basefield=HprimMultiple(
        0,
        code=HprimString(4),
    ))

    sample_time = HprimDate(8)

    specimen_type = HprimString((16, 1))
    specimen_sample = HprimString((16, 2))
    physician = HprimString(17)
    # F: Final
    # C: Corrected
    report_type = HprimString(27)

    results = HprimForeign("R", "order")


class AblSegmentR(Segment):

    test = HprimString((3, 4))
    type = HprimString((3, 5))
    value = HprimString(4)
    units = HprimString(5)

    # L: Below low normal range
    # H: Above high normal range
    # LL: Below low critical range
    # HH: Above high critical range
    # <: under range
    # >: over range
    # N: Normal value
    abnormal_flag = HprimList(7)

    # F: Final
    # R: retransmission
    # C: Correction
    status = HprimString(9)
    operator = HprimString(11)
    date_start = HprimDate(12)
    date_completed = HprimDate(13)


class AblSegmentC(Segment):

    # I: Clinical instrument system
    # L: Computer system
    source = HprimString(3)
    code = HprimString((4, 1))
    libelle = HprimString((4, 2))
    # I: Instrument flag comments
    # G: Générals comments
    c_type = HprimString(5)


class AblSegmentM(Segment):

    command = HprimString(3)


class AblSegmentL(Segment):

    # N: Normal termination
    # I: No info available
    code = HprimString(3, default="N")


class Driver(AstmMessage):
    SPEC = {
        "code": "abl",
        "label": "ABL 7XX, 8XX",
    }
    SEGMENT_CLASS_MAP = {
        "H": AblSegmentH,
        "P": AblSegmentP,
        "O": AblSegmentO,
        "Q": AblSegmentQ,
        "C": AblSegmentC,
        "R": AblSegmentR,
        "M": AblSegmentM,
        "L": AblSegmentL,
    }
    DEFAULT_SEPARATORS = "|\\^&"

    def create(self, **kwargs):
        message = AblSegmentH(type_segment="H", **kwargs)
        message.end = AblSegmentL(type_segment="L")
        return message

    def get_message_type(self, message):
        if message.query:
            return self.QUERY
        else:
            return self.ORDER

    def get_queries(self, message):
        queries = []
        for seg_q in message.query:
            queries.append(seg_q.sample_id)
        return queries

    def get_order_message(self, worklist, context):
        msg = self.create(date=context["now"])

        for patient in worklist.get_astm_worklist(
            unique_together=["action"],
            action={"del": worklist.IGNORE, worklist.ELSE: None},
        ):
            seg_p = msg.patients.add(
                patient_lab_id=patient["lab_patient_id"],
                last_name=patient["last_name"],
                first_name=patient["first_name"],
                birthdate=patient["birthdate"],
                sex=patient["sex"],
            )
            resultats = set()
            for sample, action in patient["samples"]:
                seg_o = seg_p.orders.add(
                    specimen_id=sample["sample_id"],
                    sample_time=sample["request_date"],
                    physician=patient["physician"],
                )
                for order in sample["orders"]:
                    test = seg_o.tests.add()
                    test.code = order["code"]
                for result in sample.results:
                    if not result.anterior:
                        resultats.add("%s:%s" % (result.code, result.value))
            if resultats:
                seg_p.location = "%s(%s)" % (
                    patient["location"],
                    " ".join(sorted(resultats)),
                )
            else:
                seg_p.location = patient["location"]

        if msg.patients:
            return msg
        elif context["query"]:
            seg_p = msg.patients.add()
            for sample_id in self.get_queries(context["query"]):
                seg_p.orders.add(specimen_id=sample_id)
            return msg

    def get_result_dict(self, message):
        results = {}
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                if seg_o.sample_type != 'Sample #':
                    continue
                result = results.setdefault(
                    seg_o.specimen_id,
                    {"resultats": []}
                )["resultats"]

                date = None
                operator = None

                for seg_r in seg_o.results:
                    if seg_r.status not in ("F", "C"):
                        continue
                    if seg_r.date_start:
                        date = seg_r.date_start
                    if seg_r.operator:
                        operator = seg_r.operator

                    alarmes = set(seg_r.abnormal_flag[:])
                    for c in seg_r.commentaires:
                        if c.source == 'I' and c.c_type == 'I' and c.code:
                            alarmes.add(c.code)

                    commentaires = [
                        seg_c.code for seg_c in
                        seg_r.commentaires if seg_c.source == 'G'
                    ]
                    result.append({
                        "test": seg_r.test,
                        "resu": seg_r.value,
                        "date": date,
                        "alarmes": sorted(alarmes),
                        "commentaires": commentaires,
                    })
                result.append({
                    "test": "PRELEV",
                    "resu": seg_o.specimen_type,
                    "date": date,
                })
                if seg_o.tests and seg_o.tests[0].code:
                    result.append({
                        "test": "ORIGPRL",
                        "resu": seg_o.tests[0].code,
                        "date": date,
                    })
                if operator:
                    result.append({
                        "test": "OPE",
                        "resu": operator,
                        "date": date,
                    })

        return results

    def get_qc_result_list(self, message):
        result = []
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                if seg_o.sample_type != 'QC #':
                    continue
                lot = seg_o.specimen_sample
                date = None
                for seg_r in seg_o.results:
                    if seg_r.date_start:
                        date = seg_r.date_start
                    test = seg_r.test
                    result.append({
                        "lot": lot,
                        "test": test,
                        "resu": seg_r.value,
                        "date": date,
                    })
        return result
