# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng.astm import (
    AstmMessage,
    HeaderSegment,
    Segment,
    HprimString,
    HprimList,
    HprimDate,
    HprimForeign,
    HprimMultiple,
)


class AcltopSegmentH(HeaderSegment):

    message_id = HprimString(3)
    emetteur = HprimString(5)
    recepteur = HprimString(10)
    # P: Production
    id_traitement = HprimString(12, default="P")
    version = HprimString(13, default="1394-97")
    date = HprimDate(14)

    patients = HprimForeign("P", "header")
    queries = HprimForeign("Q", "header")
    end = HprimForeign("L", "header", one2one=True, required=True)


class AcltopSegmentP(Segment):

    no_patient = HprimString(4)
    nom = HprimString((6, 1), max_length=20, trunc=True)
    prenom = HprimString((6, 2), max_length=20, trunc=True)
    date_naissance = HprimDate(8, fmt_date="%Y%m%d")

    # M/F/U
    sex = HprimString(9)
    code_correspondant = HprimString((14, 1))
    libelle_correspondant = HprimString((14, 2))

    orders = HprimForeign("O", "patient")


class AcltopSegmentO(Segment):

    tube = HprimString(3)
    instrument_id = HprimString(4)
    # /!\ Un seul test par segment O
    test = HprimString((5, 4))

    # Priorite :
    #  S : Stat Urgence
    #  R : Normal Routine
    urgence = HprimString(6)
    date_prelevement = HprimDate(7)

    # action code
    # A: Added
    # C: Cancel
    # Q: QC
    code_action = HprimString(12)

    # P: Plasma
    specimen_descriptor = HprimString(16)

    # Q: Response to query
    report_type = HprimList(26)

    results = HprimForeign("R", "order")


class AcltopSegmentR(Segment):

    test = HprimString((3, 4))
    valeur = HprimString(4)
    unite = HprimString(5)

    # Result Abnormal Flags
    #  L : Below low normal
    #  H : Above high normal
    #  N : Normal
    #  < : Below absolute low
    #  > : Above absolute high
    flag = HprimString(7)

    # Result status
    #  F : Final result
    #  P : Preliminary result
    #  V : result validated
    #  X : result cannot be done
    status = HprimList(9)
    operator_id = HprimString((11, 1))
    operator_passwd = HprimString((11, 2))

    # date et heure du resultat complete
    date = HprimDate(13)
    instrument_name = HprimString((14, 1))
    instrument_id = HprimString((14, 2))
    instrument_no = HprimString((14, 3))


class AcltopSegmentC(Segment):

    commentaire = HprimString(3)

    # numero du commentaire
    no_code = HprimString((4, 1))
    # libelle du commentaire
    libelle = HprimString((4, 2))
    # code du commentaire
    code = HprimString((4, 3), max_length=2)
    # type du commentaire
    #  I : instrument flag comment
    #  N : negative comment
    type_commentaire = HprimString(5)


class AcltopSegmentQ(Segment):

    # Starting Range ID Number
    # Multiple numero d'echantillon d'un rack
    # range_id = HprimString(3)
    range_id = HprimList(
        3,
        basefield=HprimMultiple(
            0,
            tube=HprimString(2),
        )
    )
    # request information status codes
    #  O : Request test order and demogs
    #  N : request new test orders still not downloaded
    request = HprimString(13)


class AcltopSegmentL(Segment):
    # code terminaison
    code = HprimString(3, default="N")


class Driver(AstmMessage):
    SPEC = {
        "code": "acltop",
        "label": "Acltop",
    }
    SEGMENT_CLASS_MAP = {
        "H": AcltopSegmentH,
        "P": AcltopSegmentP,
        "O": AcltopSegmentO,
        "R": AcltopSegmentR,
        "C": AcltopSegmentC,
        "Q": AcltopSegmentQ,
        "L": AcltopSegmentL,
    }
    DEFAULT_SEPARATORS = "|\^&"

    def __init__(self,
                 code_emetteur="LIS", code_recepteur="ACL TOP",
                 qc_prefix=None,
                 **kwargs):
        self.code_emetteur = code_emetteur
        self.code_recepteur = code_recepteur
        self.qc_prefix = qc_prefix
        super(Driver, self).__init__(**kwargs)

    def create(self, **kwargs):
        message = AcltopSegmentH(type_segment="H", **kwargs)
        message.end = AcltopSegmentL(type_segment="L")
        return message

    def get_message_type(self, message):
        if message.queries:
            return self.QUERY
        else:
            return self.ORDER

    def get_queries(self, message):
        queries = []
        for msg in message.queries:
            queries.extend([s.tube for s in msg.range_id])
        return queries

    def get_order_message(self, worklist, context):
        msg = self.create(
            emetteur=self.code_emetteur,
            recepteur=self.code_recepteur,
            date=context["now"],
        )
        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority"],
            action={"del": "C", worklist.ELSE: "A"},
            priority={"S": "S", worklist.ELSE: "R"},
        ):
            seg_p = msg.patients.add(
                no_patient=patient["patient_id"],
                nom=patient["last_name"],
                prenom=patient["first_name"],
                date_naissance=patient["birthdate"],
                sex=patient["sex"],
                code_correspondant=patient["location"],
            )
            # Un segment O par test
            for sample, action, priority in patient["samples"]:
                for order in sample["orders"]:
                    if context["query"]:
                        report_type = ["Q"]
                    else:
                        report_type = ["O"]
                    seg_p.orders.add(
                        tube=sample["sample_id"],
                        urgence=priority,
                        date_prelevement=sample["collection_date"],
                        code_action=action,
                        report_type=report_type,
                        specimen_descriptor="P",
                        test=order["code"],
                    )
        if msg.patients:
            return msg
        elif context["query"]:
            msg.end.code = "I"
            return msg

    def get_result_dict(self, message):
        result = {}
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                if seg_o.code_action == "Q":
                    continue
                if self.qc_prefix and seg_o.tube.startswith(
                        self.qc_prefix):
                    continue
                for seg_r in seg_o.results:
                    if seg_r.status[0] not in ("F", "X"):
                        continue
                    alarmes = set()
                    for seg_c in seg_r.commentaires:
                        if seg_c.type_commentaire == "I":
                            alarmes.add(seg_c.no_code)
                            alarmes.add(seg_c.code)

                    # le code resultat est la concatenation
                    # du test avec l'unite
                    if seg_r.unite:
                        test = "%s%s" % (seg_r.test, seg_r.unite)
                    else:
                        test = seg_r.test

                    if seg_r.flag:
                        alarmes.add(seg_r.flag)

                    result.setdefault(seg_o.tube, {
                        "resultats": []})["resultats"].append(
                        {
                            "test": test,
                            "resu": seg_r.valeur,
                            "alarmes": sorted(alarmes),
                            "date": seg_r.date,
                        })
        return result

    def get_qc_result_list(self, message):

        result = []
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                if seg_o.code_action != "Q" and (
                    not self.qc_prefix
                        or not seg_o.tube.startswith(self.qc_prefix)):
                    continue
                for seg_r in seg_o.results:
                    if "F" not in seg_r.status:
                        continue

                    lot = seg_o.tube
                    for seg_c in seg_r.commentaires:
                        if seg_c.no_code == "5653" and seg_c.code == "ME":
                            # Lot de qualification
                            lot = "QUALIF-%s" % (seg_o.tube,)
                    result.append({
                        "lot": lot,
                        "test": "%s%s" % (
                            seg_r.test, seg_r.unite,
                        ),
                        "resu": seg_r.valeur,
                        "date": seg_r.date,
                    })
        return result
