# -*- coding: utf8 -*-
from __future__ import unicode_literals

import datetime
import logging

from driverng import Message
from driverng.exceptions import ProgrammingError, ParsingError

logger = logging.getLogger("driverng.drivers.advia")


class AdviaSegment(object):
    TRAME_DESC = ()
    TRAME_PAR = {
        # KEY: (
        #   Number of characters transmitter for this item,
        #   Left or Right aligned,
        # Each unused character position is filled with S (space) or Z (zero
        # "0"))
        "URGENT": (1, "L", "S"),
        "IDEE": (14, "L", "S"),
        # For Multispecies system only
        "SPCODE": (4, "R", "S"),
        "PATNUM": (14, "L", "S"),
        "PATNAM": (30, "L", "S"),
        "BIRYEAR": (4, "R", "S"),
        "BIRMONT": (2, "R", "S"),
        "BIRDAY": (2, "R", "S"),
        "SEX": (1, "R", "S"),
        "COLYEAR": (2, "R", "Z"),
        "COLMONT": (2, "R", "Z"),
        "COLDAY": (2, "R", "Z"),
        "COLHOUR": (2, "R", "Z"),
        "COLMIN": (2, "R", "Z"),
        "LOCATIO": (6, "L", "S"),
        "DOCTOR": (6, "L", "S"),
        "POSTUBE": (6, "R", "S"),
        "ASPYEAR": (2, "R", "Z"),
        "ASPMONT": (2, "R", "Z"),
        "ASPDAY": (2, "R", "Z"),
        "ASPHOUR": (2, "R", "Z"),
        "ASPMIN": (2, "R", "Z"),
        "ASPSEC": (2, "R", "Z"),
        "UPDT": (1, "R", "S"),
        "SLIDE": (1, "R", "S"),
        "COCOM": (5, "L", "S"),
        "CPCOM": (5, "L", "S"),
        "SPTEXT": (13, "L", "S"),
    }

    def __init__(self):
        self.data = {}
        super(AdviaSegment, self).__init__()

    @classmethod
    def parse(cls, data):
        # On parse le message de gauche à droite
        cur = data
        obj = cls()
        for key in cls.TRAME_DESC:
            if key in cls.TRAME_PAR:
                num, justified, fill = cls.TRAME_PAR[key]

                if justified == "R":
                    left_justified = False
                elif justified == "L":
                    left_justified = True
                else:
                    raise ProgrammingError(
                        "Wrong justified flag '%s' for key '%s'" % (
                            justified,
                            key,
                        ))

                value = cur[:num]
                if fill == "Z":
                    if left_justified:
                        while value and value[-1] == "0":
                            value = value[:-1]
                    else:
                        while value and value[0] == "0":
                            value = value[1:]
                elif fill == "S":
                    if left_justified:
                        value = value.rstrip()
                    else:
                        value = value.lstrip()
                else:
                    raise ProgrammingError(
                        "Wrong fill character '%s' for key '%s'" % (
                            fill, key))

                obj.data[key] = value
                cur = cur[num:]
            elif isinstance(key, basestring):
                # On verifie juste que ça correspond
                if cur[:len(key)] != key:
                    raise ParsingError("Key '%s' not found" % (key,))
                cur = cur[len(key):]
            elif isinstance(key, int):
                # Nombre d'espaces à sauter.
                # On verifie que ce sont bien des espaces
                if cur[:key] != key * " ":
                    raise ParsingError("No enough spaces here !")
                cur = cur[key:]
            else:
                raise ProgrammingError("Unknown key '%s'" % (key,))

        if cur != "":
            raise ParsingError("Remaining data after parsing: '%s'" % (cur,))

        return obj

    def to_string(self):
        data = ""
        for key in self.TRAME_DESC:
            if key in self.TRAME_PAR:
                num, justified, fill = self.TRAME_PAR[key]

                if justified == "R":
                    left_justified = False
                elif justified == "L":
                    left_justified = True
                else:
                    raise ProgrammingError(
                        "Wrong justified flag '%s' for key '%s'" % (
                            justified,
                            key,
                        ))

                try:
                    value = "%s" % (self.data[key],)
                except KeyError:
                    value = ""

                if fill == "Z" and " " in value:
                    raise ParsingError(
                        "Space are not allowed in '%s' for key '%s'" % (
                            value, key))

                if len(value) > num:
                    value = value[:num]
                    logger.warning(
                        "Value '%s' for key '%s' is too long (>%d)" % (
                            value, key, num))

                if left_justified:
                    fmt = "%%-%d.%ds" % (num, num)
                else:
                    fmt = "%%%d.%ds" % (num, num)
                value = fmt % (value,)

                if fill == "Z":
                    # TODO pas exactement ce qu'il faut faire mais on
                    # interdit les espaces dans la valeur brute
                    # donc ça ne doit pas poser de problèmes
                    value = value.replace(" ", "0")

                data += value
            elif isinstance(key, basestring):
                data += key
            elif isinstance(key, int):
                data += key * " "
            else:
                raise ProgrammingError("Unknown key '%s'" % (key,))

        return data


class AdviaInit(AdviaSegment):
    TRAME_DESC = ("I", 1)
    # TODO:
    # Avec les images: "I |ID|\ftproot\|"


class AdviaToken(AdviaSegment):
    TRAME_DESC = ("S", 10)


class AdviaQuery(AdviaSegment):
    TRAME_DESC = ("Q", 1, "IDEE")
    TRAME_PAR = dict(AdviaSegment.TRAME_PAR.items() + {
        "IDEE": (14, "R", "Z"),
    }.items())


class AdviaNoOrder(AdviaSegment):
    TRAME_DESC = ("N", 1, "NO_ORDER_CODE", 1, "IDEE")
    TRAME_PAR = dict(AdviaSegment.TRAME_PAR.items() + {
        # R for Request
        # W for Workorder
        "NO_ORDER_CODE": (1, "R", "S"),
        "IDEE": (14, "R", "Z"),
    }.items())


class AdviaOrder(AdviaSegment):
    # Le message est composé d'une première partie concernant les infos
    # patient (cf. TRAME_DESC)
    # et d'une deuxièmes avec les tests, les deux parties sont séparés par
    # un CRLF
    # La deuxième partie est une liste de trois composants:
    # - Date OU heure anterio (7 bytes) ! optionel
    # - Valeur anterio (5 bytes)        ! optionel
    # - test demandé (3 bytes)
    TRAME_DESC = ("Y", 2, "URGENT",
                  "UPDT", 1, "IDEE",
                  8, "SLIDE", 16,
                  "PATNUM", 3, "PATNAM",
                  1, "BIRMONT", "/",
                  "BIRDAY", "/", "BIRYEAR",
                  1, "SEX", 1,
                  "COLMONT", "/", "COLDAY",
                  "/", "COLYEAR",
                  1, "COLHOUR", "COLMIN",
                  1, "LOCATIO", 1, "DOCTOR", 1,
                  )

    def __init__(self):
        super(AdviaOrder, self).__init__()
        # Manipulation de BIR{MONT,YEAR,DAY}
        self.date_naissance = None
        # Manipulation de COL{MONT,YEAR,DAY,HOUR,MIN}
        self.date_demande = None
        # Liste de code demandes
        self.tests = []

    @classmethod
    def parse(cls, data):
        data = data.splitlines()
        if len(data) > 2:
            raise ParsingError("Too much lines")
        obj = super(AdviaOrder, cls).parse(data[0])

        # Gestion des @#!é}#]@ dates
        if all([key in obj.data and obj.data[key] for key in (
                "BIRYEAR", "BIRMONT", "BIRDAY")]):
            obj.date_naissance = datetime.datetime(
                int(obj.data["BIRYEAR"]),
                int(obj.data["BIRMONT"]),
                int(obj.data["BIRDAY"]),
            )
        if all([key in obj.data and obj.data[key] for key in (
                "COLMONT", "COLDAY")]):
            obj.date_demande = datetime.datetime(
                datetime.datetime.strptime(("%2.2s" % (
                    obj.data["COLYEAR"],)).replace(" ", "0"), "%y").year,
                int(obj.data["COLMONT"]),
                int(obj.data["COLDAY"]),
                int(obj.data["COLHOUR"] or 0),
                int(obj.data["COLMIN"] or 0),
            )

        if len(data) == 2:
            line = data[1]
            while line:
                # /!\ Bien lire la doc avant de modifier.
                # Cette partie est complétement tirrée par les cheuveux.
                # Le code sur 3 bytes justifié à droite avec des "0" pour
                # remplir.
                code = line[:3]
                while code and code[0] == "0":
                    code = code[1:]
                obj.tests.append(code)
                line = line[15:]

        return obj

    def to_string(self):
        # Gestion des @#!é}#]@ dates
        if self.date_naissance:
            self.data["BIRMONT"] = self.date_naissance.month
            self.data["BIRYEAR"] = self.date_naissance.year
            self.data["BIRDAY"] = self.date_naissance.day
        if self.date_demande:
            self.data["COLMONT"] = self.date_demande.month
            self.data["COLYEAR"] = self.date_demande.strftime("%y")
            self.data["COLDAY"] = self.date_demande.day
            self.data["COLHOUR"] = self.date_demande.hour
            self.data["COLMIN"] = self.date_demande.minute

        data = super(AdviaOrder, self).to_string()

        order_data = ""
        for test in self.tests:
            code = "%3s" % (test,)
            code = code.replace(" ", "0")
            assert len(code) == 3
            order_data += "%-15.3s" % (code,)

        if order_data:
            return "\r\n".join([data, order_data])
        else:
            return data


class AdviaOrderValidation(AdviaSegment):
    TRAME_DESC = ("E", 8, "ERROR_FLAG")
    TRAME_PAR = dict(AdviaSegment.TRAME_PAR.items() + {
        "ERROR_FLAG": (2, "R", "S"),
    }.items())

    # Suivant le mode (download ou query):
    # " 0" ou "10" => Workorder Valid
    # " 4" ou "14" => Invalid test number (not defined)


class AdviaResultValidation(AdviaSegment):
    TRAME_DESC = ("Z", 17, "RESULT_VALIDATION_ERROR_FLAG")
    TRAME_PAR = dict(AdviaSegment.TRAME_PAR.items() + {
        "RESULT_VALIDATION_ERROR_FLAG": (2, "R", "S"),
        # " 0": Valid
        # " 1": Host computer results file is full.
        # " 2": Host computer does not want to receive any more results.
    }.items())


class _AdviaResultat(object):

    def __init__(self):
        self.code = None
        self.value = None
        self.flags = []
        self.garbage = None
        super(_AdviaResultat, self).__init__()

    def is_valid(self):
        try:
            assert self.code is not None, "Pas de code"
            assert self.value is not None, "Pas de valeur"
            assert self.garbage is not None, "Pas de garbage"
            assert len(self.code) <= 3, "Code trop long (max 3)"
            assert len(self.value) <= 5, "Valeur trop longue (max 5)"
            assert len(self.garbage) <= 5, "Garbage trop long (max 5)"
        except AssertionError as error:
            raise ParsingError("Resultat invalide: %s" % (error,))

    def to_string(self):
        self.is_valid()
        data = "%3.3s%5.5s%5.5s" % (
            self.code,
            self.value,
            self.garbage,
        )
        if self.flags:
            data += "|%s|" % (" ".join(self.flags))

        return data


class AdviaResult(AdviaSegment):
    TRAME_DESC = ("R", 1, "IDEE",
                  1, "POSTUBE", 6,
                  "SPCODE", 1, "ASPMONT",
                  "/", "ASPDAY", "/",
                  "ASPYEAR", 1, "ASPHOUR",
                  ":", "ASPMIN", ":",
                  "ASPSEC", 3,
                  )
    # Les flags de l'advia
    FLAGS = {
        "BC": "Baso Count Suspect",
        "BR": "Baso Irregular Flow Rate",
        "VB": "Baso No Valley",
        "NB": "Baso Noise",
        "BS": "Baso Saturation",
        "TB": "Baso Temperature Out of Range ",
        "CC": "Comparison Error MCHC/CHCM",
        "WC": "Comparison Error WBCB/WBCP",
        "HR": "Hgb Irregular Flow Rate",
        "PH": "Hgb Power Low",
        "PL": "Laser Power Low",
        "NC": "Suspect Cellular Interference",
        "NL": "Suspect Lipid Interference",
        "NP": "Suspect Large Platelets",
        "NV": "No Perox NRBC / Lymph Valley flag ",
        "XR": "Perox Irregular Flow Rate ",
        "VX": "Perox No Valley",
        "NX": "Perox Noise",
        "PX": "Perox Power Low",
        "XS": "Perox Saturation",
        "TX": "Perox Temperature Out of Range ",
        "NW": "Platelet Clumps",
        "NT": "Platelet Noise",
        "OT": "Platelet Origin Noise",
        "RR": "RBC Irregular Flow Rate",
        "CT": "Retic - Plt Interference",
        "CA": "Retic Abs Dist Abnorm",
        "RF": "Retic Absorption Flatness",
        "FC": "Retic Fit Suspect",
        "CR": "Retic Irregular Flow Rate",
        "NO": "Retic Noise origin",
        "CL": "Retic RBC Count Low",
        "CS": "Retic Saturation Cells",
        "SE": "Retic Slope Error",
        "WS": "WBC Substitution",
    }

    def __init__(self):
        super(AdviaResult, self).__init__()
        # Liste de _AdviaResultat
        self.results = []
        # Date des resultats
        self.date = None

    def add_result(self):
        resu = _AdviaResultat()
        self.results.append(resu)
        return resu

    @classmethod
    def parse(cls, data):
        data = data.splitlines()
        if len(data) > 2:
            raise ParsingError("Too much lines")
        obj = super(AdviaResult, cls).parse(data[0])

        obj.date = datetime.datetime(
            datetime.datetime.strptime(("%2.2s" % (
                obj.data["ASPYEAR"],
            )).replace(" ", "0"), "%y").year,
            int(obj.data["ASPMONT"]),
            int(obj.data["ASPDAY"]),
            int(obj.data["ASPHOUR"] or 0),
            int(obj.data["ASPMIN"] or 0),
            int(obj.data["ASPSEC"] or 0),
        )

        if len(data) == 2:
            line = data[1]
            while line:
                # Principalement deux cas:
                # <code 3 (R,S)><resultat 5 (R,S)><garbage ??? 5>
                # puis <|FL AG S| optionel>
                resu = _AdviaResultat()
                resu.code = line[:3].lstrip()
                resu.value = line[3:8].lstrip()
                resu.garbage = line[8:13]

                if len(line) > 13 and line[13] == "|":
                    # Récupération des flags et on va au résultat suivant
                    eoflags = line[14:].find('|')
                    resu.flags = line[14:][:eoflags].split(' ')
                    line = line[14 + eoflags + 1:]
                else:
                    # On va au résultat suivant
                    line = line[13:]

                obj.results.append(resu)

        return obj

    def to_string(self):
        if self.date is not None:
            self.data["ASPYEAR"] = self.date.strftime("%y")
            self.data["ASPMONT"] = self.date.month
            self.data["ASPDAY"] = self.date.day
            self.data["ASPHOUR"] = self.date.hour
            self.data["ASPMIN"] = self.date.minute
            self.data["ASPSEC"] = self.date.second
        data = super(AdviaResult, self).to_string()

        resu_data = "".join([r.to_string() for r in self.results])

        if resu_data:
            return "\r\n".join([data, resu_data])
        else:
            return data


class DriverAdvia2120(Message):
    SPEC = {
        "code": "advia",
        "label": "Advia 2120",
    }

    INITIALISATION = 100
    TOKEN = 101
    ORDER_VALIDATION = 102
    NO_ORDER = 103
    RESULT_VALIDATION = 104

    _MESSAGE_TYPE_MAP = {
        Message.QUERY: AdviaQuery,
        Message.ORDER: AdviaOrder,
        Message.RESULT: AdviaResult,
        INITIALISATION: AdviaInit,
        TOKEN: AdviaToken,
        ORDER_VALIDATION: AdviaOrderValidation,
        NO_ORDER: AdviaNoOrder,
        RESULT_VALIDATION: AdviaResultValidation,
    }

    def __init__(self, qc_prefix="9", **kwargs):
        self.qc_prefix = [e for e in qc_prefix.split(",") if e]
        super(DriverAdvia2120, self).__init__(**kwargs)

    def create(self):
        raise ProgrammingError(
            "You must call create_message() instead of create()")

    def create_message(self, message_type):
        try:
            cls = self._MESSAGE_TYPE_MAP[message_type]
        except KeyError:
            raise ProgrammingError("Unknow message type")

        return cls()

    def get_message_type(self, message):
        for m_type, cls in self._MESSAGE_TYPE_MAP.items():
            if isinstance(message, cls):
                return m_type

        raise ProgrammingError("Unknow message type")

    def to_string(self, message):
        return message.to_string()

    def parse(self, data):
        c_map = {
            "I": AdviaInit,
            "S": AdviaToken,
            "E": AdviaOrderValidation,
            "N": AdviaNoOrder,
            "Z": AdviaResultValidation,
            "Q": AdviaQuery,
            "Y": AdviaOrder,
            "R": AdviaResult,
        }
        try:
            cls = c_map[data[0]]
        except (IndexError, KeyError):
            raise ParsingError("Unknown message type")
        return cls.parse(data)

    def get_queries(self, message):
        if self.get_message_type(message) == self.QUERY:
            return [message.data["IDEE"]]
        else:
            return []

    def get_order_message(self, worklist, context):
        if len(worklist.samples) > 1:
            raise RuntimeError("Multi-sample is not implemented")
        elif len(worklist.samples) == 1:
            msg = self.create_message(self.ORDER)
            sample = worklist.samples[0]
            for order in sample["orders"]:
                if (
                    order["action"] != "del"
                    and order["priority"] in ("A", "S")
                ):
                    msg.data["URGENT"] = "U"
                if order["action"] != "del":
                    msg.tests.append(order["code"])
            msg.data["IDEE"] = sample["sample_id"]
            msg.data["PATNUM"] = sample["lab_patient_id"]
            msg.data["PATNAM"] = "%s %s" % (
                sample["last_name"],
                sample["first_name"],
            )
            msg.date_naissance = sample["birthdate"]
            msg.date_demande = sample["request_date"]
            msg.data["SEX"] = sample["sex"]
            msg.data["LOCATIO"] = sample["service"]
            msg.data["DOCTOR"] = sample["physician"]

            if msg.tests:
                return msg

        if context["query"]:
            msg = self.create_message(self.NO_ORDER)
            msg.data["IDEE"] = context["query"].data["IDEE"]
            msg.data["NO_ORDER_CODE"] = "W"
            return msg

    def _is_control(self, sample_id):
        for prefix in self.qc_prefix:
            if sample_id.startswith(prefix):
                return True
        return False

    def get_result_dict(self, message):
        sample_id = message.data["IDEE"]

        if self._is_control(sample_id):
            return {}

        result = {
            sample_id: {
                "resultats": [],
                "rack": message.data["POSTUBE"],
            }
        }

        alarmes_hemato = set()

        for resu in message.results:
            alarmes_hemato.update(resu.flags)
            result[sample_id]["resultats"].append({
                "test": resu.code,
                "resu": resu.value,
                "date": message.date,
                "alarmes": list(set(resu.flags)),
            })

        result[sample_id]["resultats"].append({
            "test": "ALARM",  # TODO: HACK
            "resu": ";".join(sorted(alarmes_hemato)),
            "date": message.date,
        })

        return result

    def get_qc_result_list(self, message):
        sample_id = message.data["IDEE"]

        if not self._is_control(sample_id):
            return []

        result = []

        for resu in message.results:
            result.append({
                "lot": sample_id,
                "test": resu.code,
                "resu": resu.value,
                "date": message.date,
            })

        return result
