# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng.astm import (
    AstmMessage,
    HeaderSegment,
    Segment,
    HprimString,
    HprimList,
    HprimDate,
    HprimForeign,
    HprimMultiple,
)


class AisSegmentH(HeaderSegment):
    sender_id = HprimString(5, default="MCA")
    processing_id = HprimString(12, default="P")
    date = HprimDate(14)

    patients = HprimForeign("P", "header")
    end = HprimForeign("L", "header", one2one=True, required=True)


class AisSegmentP(Segment):
    id_patient = HprimString(3)
    last_name = HprimString((6, 1))
    first_name = HprimString((6, 2))
    birthdate = HprimDate(8, fmt_date="%Y%m%d")
    # M/F/U ?
    sex = HprimString(9)
    orders = HprimForeign("O", "patient")


class AisSegmentO(Segment):
    specimen_id = HprimMultiple(
        3,
        tube=HprimString(1),
        rack=HprimString(3),
        pos=HprimString(4),
    )
    id_request = HprimString(4)
    tests = HprimList(5, basefield=HprimMultiple(
        0,
        code=HprimString(4),
    ))
    # Priorité :
    #  R: Routine
    #  S: STAT
    #  blanc : CQ
    priority = HprimString(6)

    # date used for send the result
    date = HprimDate(7)

    # date for the ordered test
    specimen_date = HprimDate(8)
    diagnostic_suspicion = HprimString(14)
    specimen_description = HprimString(16)

    # O pour order
    report_type = HprimString(26)

    results = HprimForeign("R", "order")


class AisSegmentR(Segment):
    test = HprimString((3, 4))
    # QUANT: numerical result
    # QUAL: text result
    type_resultat = HprimString((3, 5))
    value = HprimString(4)
    units = HprimString(5)

    # uniquement pour les resultats numeriques
    abnormal_flags = HprimList(7)

    # Result status :
    #  F: processed sample
    #  X: skipped sample
    status = HprimString(9)

    date_started = HprimDate(12)
    date_completed = HprimDate(13)


class AisSegmentC(Segment):
    comment_code = HprimString(3)
    comment_text = HprimString(4)


class AisSegmentL(Segment):
    termination_code = HprimString(3)


class Driver(AstmMessage):
    SPEC = {
        "code": "ais",
        "label": "Ais",
    }
    SEGMENT_CLASS_MAP = {
        "H": AisSegmentH,
        "P": AisSegmentP,
        "O": AisSegmentO,
        "R": AisSegmentR,
        "L": AisSegmentL,
    }
    DEFAULT_SEPARATORS = "|\\^&"

    def create(self, **kwargs):
        message = AisSegmentH(type_segment="H", **kwargs)
        message.end = AisSegmentL(type_segment="L")
        return message

    def get_message_type(self, message):
        return self.ORDER

    def get_order_message(self, worklist, context):
        msg = self.create(date=context["now"])

        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority"],
            action={"del": worklist.IGNORE, worklist.ELSE: ""},
            priority={"S": "S", "A": "S", worklist.ELSE: "R"},
        ):
            seg_p = msg.patients.add(
                id_patient=patient["patient_id"],
                last_name=patient["last_name"],
                first_name=patient["first_name"],
                birthdate=patient["birthdate"],
                sex=patient["sex"],
            )
            for sample, action, priority in patient["samples"]:
                seg_o = seg_p.orders.add(
                    priority=priority,
                    specimen_date=sample["request_date"],
                    report_type="O",
                )
                seg_o.specimen_id.tube = sample["sample_id"]
                for order in sample["orders"]:
                    t = seg_o.tests.add()
                    t.code = order["code"]

        if len(msg.patients):
            return msg

    def get_result_dict(self, message):
        results = {}
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                # exclusion des QC
                if seg_o.priority is '':
                    continue
                tube = seg_o.specimen_id.tube.strip()
                if seg_o.specimen_id.rack and seg_o.specimen_id.pos:
                    rack = "%s-%s" % (
                        seg_o.specimen_id.rack,
                        seg_o.specimen_id.pos,
                    )
                else:
                    rack = None
                results.setdefault(tube, {"resultats": [], "rack": rack, })
                for seg_r in seg_o.results:
                    # exclusion des resultats n'ayant pas le status du
                    # resultat F
                    if seg_r.status != "F":
                        continue
                    code_test = seg_r.test + seg_r.type_resultat
                    results[tube]["resultats"].append({
                        "test": code_test,
                        "resu": seg_r.value,
                        "date": seg_r.date_completed,
                        "alarmes": seg_r.abnormal_flags,
                    })

        return results

    def get_qc_result_list(self, message):
        results = []
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                # exclusion des resultats non CQ
                if seg_o.priority is not '':
                    continue

                # reference du lot
                lot = seg_o.specimen_id.tube.strip()
                for seg_r in seg_o.results:
                    code_test = seg_r.test + seg_r.type_resultat
                    results.append({
                        "lot": lot,
                        "test": code_test,
                        "resu": seg_r.value,
                        "date": seg_r.date_completed,
                    })

        return results
