# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng.astm import (
    AstmMessage,
    HeaderSegment,
    Segment,
    HprimString,
    HprimList,
    HprimDate,
    HprimForeign,
    HprimMultiple,
)


class Arch2000SegmentH(HeaderSegment):

    sender = HprimMultiple(5,
                           name=HprimString(1, default="MCA"),
                           version=HprimString(2),
                           serial=HprimString(3),
                           interface_version=HprimString(4),
                           )

    # Not supported for serial connexions
    receiver_hostname = HprimString((10, 1))
    receiver_ip = HprimString((10, 2))

    # P: Production
    # D: Debug
    # Q: Quality Control
    type_message = HprimString(12, default="P")

    version = HprimString(13, default="1")
    date = HprimDate(14)

    patients = HprimForeign("P", "header")
    queries = HprimForeign("Q", "header")
    end = HprimForeign("L", "header", one2one=True, required=True)


class Arch2000SegmentQ(Segment):

    tube = HprimString((3, 2), max_length=16)
    tests = HprimString((5, 4), max_length=16)
    status_code = HprimString((13, 1), max_length=1)


class Arch2000SegmentP(Segment):

    practice_pid = HprimString(3)

    # This field is the recommended field for an LIS to
    # communicate Patient IDs to an instrument
    laboratory_pid = HprimString(4)

    instrument_pid = HprimString(5)

    last_name = HprimString((6, 1), max_length=20, trunc=True)
    first_name = HprimString((6, 2), max_length=20, trunc=True)

    nomjf = HprimString(7)
    birthdate = HprimDate(8, fmt_date="%Y%m%d")
    # M/F/U
    sex = HprimString(9)

    # Location - Service
    location = HprimString((26, 2))

    orders = HprimForeign("O", "patient")


class ArchTestField(HprimMultiple):
    def __init__(self, *position, **kwargs):
        kwargs.update({
            "code": HprimString(4),
            "name": HprimString(5),
            # Dilution protocol name (empty for calculated test results)
            "protocol": HprimString(6),
            # Assay status
            # P: Assay installed as primary version
            # C: Array installed as correlation version
            "status": HprimString(7),
            # Empty for calculated results
            "reagent_lot": HprimString(8),
            "reagent_serial": HprimString(9),
            "control_lot": HprimString(10),
            # F: Final result concentration patient or control result
            # P: Preliminary instrument result
            # Interpretation of final result for patient test results
            "result_type": HprimString(11),
        })
        super(ArchTestField, self).__init__(*position, **kwargs)


class Arch2000SegmentO(Segment):

    tube = HprimString((3, 1))
    rack = HprimString((4, 2))
    position = HprimString((4, 3))

    tests = HprimList(5, basefield=ArchTestField(0))

    # (S)tat
    # (R)outine default value
    priority = HprimString(6, max_length=1)

    # Requested/Ordered Date and time
    date_requested = HprimDate(7)

    # Date et heure de prelevement
    date_prelevement = HprimDate(8)

    # C: Cancel
    # Q : Quality Control
    # A: Add
    # N: New. If the specimen is known by the instrument/system
    # this message is ignored as a duplicate transmission.
    code_action = HprimString(12, max_length=1)

    # O: Order, indicate a normal patient/order request from a host
    # F: Final report, normal report of results
    # Q: Query, Indicate that the Order is in response to a request:
    # X: Caneled (error)
    # I: Pending
    # Y: No such test ordered
    report_type = HprimString(26)

    results = HprimForeign("R", "order")


class Arch2000SegmentR(Segment):

    # result_qualifier:
    # F: Final
    # I: Interpreted
    # P: Preliminary
    # Favg: Average of a set of final results
    # Pavg: Average of a set of Preliminary results
    # Iavg: Interpretation associated with the average final result
    test = ArchTestField(3)

    # le resultat
    valeur = HprimString(4)

    units = HprimString(5)

    # For result type F this field can be blank or
    # contains one of the following flags:
    # LOW, HIGH, L, HH, <, >, EDIT, EXP, EXPC, CNTL,
    # Westgard flags, A#1, A#2, CORR, FLEX, PSHH, IUO,
    # INDX (post launch)
    # For result type P and I, this field is blank
    # NOTE: Multiple flags can be sent
    # when the Result Type in field 10.1.3 is
    # F. Multiple flags are sent separated by
    # component delimiters (which are used
    # as a repeat delimiter)
    abnormal_flags = HprimString(7)

    # statut du resultat
    # F: final
    # P: Preliminaire
    status = HprimString(9)

    # identification de l'operateur
    # information presente uniquement lors du retour du resultat
    operateur = HprimMultiple(11,
                              nom=HprimString(1),  # nom ou code
                              aprouver=HprimString(2),
                              )

    # date du resultat ou du QC
    date = HprimDate(13)

    # nom de l'analyseur
    nom_analyseur = HprimString(14)


class Arch2000SegmentC(Segment):

    # I: Instrument
    comment_source = HprimString(3, default="I")
    comment = HprimString(4)

    # G: Result comment
    # I: Exception string
    comment_type = HprimString(5)


class Driver(AstmMessage):
    SPEC = {
        "code": "arch2000",
        "label": "Arch2000",
    }

    SEGMENT_CLASS_MAP = {
        "H": Arch2000SegmentH,
        "P": Arch2000SegmentP,
        "Q": Arch2000SegmentQ,
        "O": Arch2000SegmentO,
        "R": Arch2000SegmentR,
        "C": Arch2000SegmentC,
        "L": Segment,
    }

    DEFAULT_SEPARATORS = "|\^&"

    def create(self, **kwargs):
        message = Arch2000SegmentH(type_segment="H", **kwargs)
        message.end = Segment(type_segment="L")
        return message

    def get_message_type(self, message):
        if message.queries:
            return self.QUERY
        else:
            return self.ORDER

    def get_queries(self, message):
        queries = []
        for seg_q in message.queries:
            queries.append(seg_q.tube)
        return queries

    def get_order_message(self, worklist, context):
        msg = self.create(data=context["now"])
        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority"],
            action={"del": "C", worklist.ELSE: "A"},
            priority={"S": "S", worklist.ELSE: "R"},
        ):
            seg_p = msg.patients.add(
                laboratory_pid=patient["patient_id"],
                last_name=patient["last_name"],
                first_name=patient["first_name"],
                nomjf=patient["maiden_name"],
                birthdate=patient["birthdate"],
                sex=patient["sex"],
                location=patient["location"],
            )
            for sample, action, priority in patient["samples"]:
                seg_o = seg_p.orders.add(
                    tube=sample["sample_id"],
                    urgence=priority,
                    date_prelevement=sample["collection_date"],
                    report_type="Q" if context["query"] else "O",
                    code_action=action,
                )
                for order in sample["orders"]:
                    t = seg_o.tests.add()
                    t.code = order["code"]
                    if order["dilution"]:
                        t.protocol = "1:%s" % (order["dilution"],)

        if msg.patients:
            return msg
        elif context["query"]:
            msg.end.code = "N"
            for sample_id in self.get_queries(context["query"]):
                seg_q = msg.queries.add()
                seg_q.tube = sample_id
                seg_q.tests = "ALL"
                seg_q.status_code = "X"
            return msg

    def get_result_dict(self, message):

        result = {}

        for seg_p in message.patients:

            for seg_o in seg_p.orders:

                if seg_o.code_action == "Q":
                    continue

                if seg_o.rack and seg_o.position:
                    rack = "%s-%s" % (seg_o.rack, seg_o.position)
                else:
                    rack = None

                result.setdefault(seg_o.tube, {"resultats": [], "rack": rack})

                for seg_r in seg_o.results:

                    if isinstance(seg_r.abnormal_flags, list):
                        alarmes = seg_r.abnormal_flags
                    elif seg_r.abnormal_flags:
                        alarmes = [seg_r.abnormal_flags]
                    else:
                        alarmes = []

                    commentaires = []
                    for seg_c in seg_r.commentaires:
                        if seg_c.comment_type == "G":
                            commentaires.append(seg_c.comment)
                        elif seg_c.comment_type == "I":
                            alarmes.append(seg_c.comment)

                    # On ajoute le "result_type"
                    if seg_r.test.result_type:
                        test = "%s%s" % (
                            seg_r.test.code,
                            seg_r.test.result_type,
                        )
                    else:
                        test = seg_r.test.code

                    result[seg_o.tube]["resultats"].append({
                        "test": test,
                        "resu": seg_r.valeur,
                        "alarmes": alarmes,
                        "date": seg_r.date,
                        "analyseur": seg_r.nom_analyseur,
                        "commentaires": commentaires,
                    })

        return result

    def get_qc_result_list(self, message):
        result = []
        for seg_p in message.patients:

            for seg_o in seg_p.orders:

                if seg_o.code_action != "Q":
                    continue

                for seg_r in seg_o.results:

                    if seg_r.test.result_type == "P":
                        # P: Preliminary instrument result
                        continue

                    # On ajoute le "result_type"
                    if seg_r.test.result_type:
                        test = "%s%s" % (
                            seg_r.test.code,
                            seg_r.test.result_type,
                        )
                    else:
                        test = seg_r.test.code

                    lot = "%s%s" % (seg_o.tube, seg_r.test.control_lot)

                    result.append({
                        "lot": lot,
                        "sample_id": seg_r.test.control_lot,
                        "control_level": seg_o.tube,
                        "test": test,
                        "resu": seg_r.valeur,
                        "date": seg_r.date,
                        "analyseur": seg_r.nom_analyseur,
                    })

        return result
