# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng import Message
from driverng.fixed import FixedLengthFormat, String, Multiple
from driverng.utils import num_years

SAMPLE_ID_LENGTH = 16


class Au680Query(FixedLengthFormat):
    FMT_DESCRIPTION = [
        # RB: Sample information request
        # R<space>: Normal Sample request
        # RH: Repeat run sample
        # Rh: Automatic repeat run sample
        # RE: Sample information request end
        String("code", 2),
        # 00 to 99
        String("system_no", 2),

        # 4 ou 5 digits
        # 0001 to 9999
        String("rack", 4),

        # Rack sample: 01 to 10
        # STAT sample: 01 to 22
        String("cup", 2),
        # <space>: Serum
        # U: Urine
        # X: Other
        # Y: Other-1
        # W: Whole blood
        # N: Not specified
        String("sample_type", 1),
        # Routine sample: 0001 to 9999
        # Emergency sample: E001 to E999
        # STAT sample: P001 to P999
        # QC sample: Q001 to Q999
        # Reagent blank sample: R001 to R999
        # Calibration sample: A001 to A999
        String("sample_no", 4),
        String("sample_id", SAMPLE_ID_LENGTH, right_justified=True),
    ]


class Au680Order(FixedLengthFormat):
    HEADER = [
        String("rack", 4),
        String("cup", 2),
        String("sample_type", 1),
        String("sample_no", 4),
        String("sample_id", SAMPLE_ID_LENGTH, right_justified=True),
        String("dummy", 4),
        # 0 to 9 or E
        String("data_classification", 1),
    ]
    PATIENT = [
        # M: Male
        # F: Female
        # <space>: No reply
        # 0: none
        String("sex", 1),
        # 000 to 150
        # <space>: Not specified
        String("age", 3),
        # 00 to 11
        # <space>: Not specified
        String("month_age", 2),

        # 6 champs entre 0 et 20
        # On utilise que les deux premier pour nom et prenom
        String("patient_info1", 20),
        String("patient_info2", 20, truncable=True),
    ]

    FMT_DESCRIPTION = [
        # S<space>: Normal sample reponse
        # SH: Repeat run sample response
        # Sh: Automatic repeat run sample
        # SE: Sample response stop
        String("code", 2),
    ] + HEADER + PATIENT + [
        Multiple("tests", [
            # 2 or 3 digits
            String("code", 3),
            # 0: Normal
            # 1: Diluent
            # 2: Condense
            String("diluent_type", 1),
        ]),
    ]


class Au680RerunOrder(FixedLengthFormat):
    FMT_DESCRIPTION = [
        String("code", 2),
    ] + Au680Order.HEADER + [
        Multiple("tests", [
            # 2 or 3 digits
            String("code", 3),
            # 0: Normal
            # 1: Diluent
            # 2: Condense
            String("diluent_type", 1),
        ]),
    ]


class Au680Result(FixedLengthFormat):
    TEST = [
        # 001 to 120
        # 3 ou 2 digits...
        String("test_number", 3),
        # Reagent (4x4)
        String("reagent", 16),
        # 9 ou 6 digits
        String("analysis_data", 9),
        # 2 ou 8 digits
        String("data_flag", 8),
    ]
    FMT_DESCRIPTION = [
        # DB: Analysis data start
        # D<space>: Normal sample data
        # DH: Repeat run sample data
        # DR: Reagent blank sample data
        # DA: Calibration sample data
        # d : STAT quick output data
        # dH: Repeat run STAT quick output data
        # DQ: QC sample data
        # DE: Analysis data transmission end
        String("code", 2),
        String("system_no", 2),
    ] + Au680Order.HEADER + Au680Order.PATIENT + [
        Multiple("tests", TEST),
    ]


class Au680RerunResult(FixedLengthFormat):
    FMT_DESCRIPTION = [
        String("code", 2),
        String("system_no", 2),
    ] + Au680Order.HEADER + [Multiple("tests", Au680Result.TEST)]


class Au680QC(FixedLengthFormat):
    FMT_DESCRIPTION = [
        # DQ: QC sample data
        String("code", 2),
        String("system_no", 2),
    ] + Au680Order.HEADER + [
        Multiple("tests", Au680Result.TEST),
    ]


class Driver(Message):
    SPEC = {
        "code": "au680",
        "label": "AU 680/480",
    }

    @staticmethod
    def get_klass(code):
        if code in ("Sh", "SH"):
            return Au680RerunOrder
        elif code[0] == "S":
            return Au680Order
        elif code[0] == "R":
            return Au680Query
        elif code in ("DQ", "DR", "DA"):
            return Au680QC
        elif code in ("Dh", "DH"):
            return Au680RerunResult
        elif code[0] == "D":
            return Au680Result
        else:
            raise NotImplementedError

    def create(self, code, **kwargs):
        kwargs.setdefault("code", code)
        return self.get_klass(code)(**kwargs)

    def parse(self, data):
        obj = self.get_klass(data[:2])()
        obj.parse(data)
        return obj

    def to_string(self, message):
        return message.to_string()

    def get_message_type(self, message):
        if isinstance(message, Au680Query):
            return self.QUERY
        else:
            return self.ORDER

    def get_queries(self, message):
        return [message.sample_id]

    def get_order_message(self, worklist, context):
        samples = worklist.get_worklist(
            unique_together=["action"],
            action={"del": worklist.IGNORE, worklist.ELSE: None},
        )
        if len(samples) == 0:
            if not context["query"]:
                return
            else:
                query = context["query"]
                return self.create(
                    "S%s" % query.code[1:],
                    sample_type=query.sample_type,
                    sample_id=query.sample_id,
                    data_classification="E",
                    rack=query.rack,
                    cup=query.cup,
                    sample_no=query.sample_no,
                )
        elif len(samples) > 1:
            raise RuntimeError("Multi-sample not implemented")
        else:
            sample = samples[0][0]

        if sample["specimen"] == "S":
            sample_type = " "
        else:
            sample_type = sample["specimen"]

        query = context["query"]
        if query:
            code = "S%s" % query.code[1:]
        else:
            code = "S"

        message = self.create(
            code,
            sample_type=sample_type,
            sample_id=sample["sample_id"],
            data_classification="E",
        )
        if isinstance(message, Au680Order):
            message.sex = sample["sex"] if sample["sex"] in ("M", "F") else ""
            message.age = num_years(
                sample["birthdate"], sample["collection_date"])
            message.patient_info1 = sample["patient_id"]
            message.patient_info2 = "%s %s" % (
                sample["last_name"],
                sample["first_name"],
            )
        if query:
            message.rack = query.rack
            message.cup = query.cup
            message.sample_no = query.sample_no

        for order in sample["orders"]:
            message.tests.add(
                code=order["code"],
                diluent_type="0",
            )
        return message

    @staticmethod
    def get_reagents(data):
        reagents = []
        for i in range(0, 16, 4):
            reagent = data[i:][:4]
            if reagent:
                reagents.append(reagent)
        return reagents

    def get_result_dict(self, message):
        if message.code == "DQ":
            return {}

        results = {}
        resultats = results.setdefault(message.sample_id, {
            "rack": "%s-%s" % (message.rack, message.cup),
            "resultats": [],
        })["resultats"]
        for test in message.tests:
            resultats.append({
                "test": test.test_number,
                "resu": test.analysis_data,
                "alarmes": [e for e in test.data_flag if e.strip()],
                "type_prlv": message.sample_type or "S",
                "lots_reactifs": self.get_reagents(test.reagent),
            })
        return results

    def get_qc_result_list(self, message):
        if message.code != "DQ":
            return []
        results = []
        for test in message.tests:
            results.append({
                "lot": message.sample_id,
                "test": test.test_number,
                "resu": test.analysis_data,
                "date": None,
                "type_prlv": message.sample_type or "S",
                "lots_reactifs": self.get_reagents(test.reagent),
            })
        return results
