# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng import Message
from driverng.fixed import (
    FixedLengthFormat,
    String,
    Multiple,
    Date,
)


class BctModeBase(FixedLengthFormat):
    pass


class Query(BctModeBase):
    FMT_DESCRIPTION = [
        String("message_type", 1),
        1,
        String("sample_id", 22),
    ]


class Order(BctModeBase):
    FMT_DESCRIPTION = [
        String("message_type", 1),
        1,
        String("sample_id", 22),
        1,
        String("emergency", 1),
        1,
        String("mode", 1),
        1,
        String("patient_name", 32),
        1,
        String("requester", 16),
        1,
        Multiple("tests", [
            String("code", 5),
            1,
        ]),
    ]


class Result(BctModeBase):
    FMT_DESCRIPTION = [
        String("message_type", 1),
        1,
        String("sample_id", 22),
        1,
        String("assay_number", 5),
        1,
        String("result", 10),
        1,
        String("unit", 10),
        1,
        # Bit masks
        # 1: Invalid
        # 2: Doubtful
        # 4: Above measuring range
        # 8: Below measuring range
        # 16: Above reference range
        # 32: Below reference range
        # 64: Max variation exceeded
        # 128: No clot
        # 256: Repeated
        # 512: Invalid control
        # 1024: New wave
        # 2048: AMaxMin
        String("flag", 4),
        1,
        Date("date", 15, fmt_date="%y%m%d %H:%M:%S"),
        1,
        # S: Stat
        # N: Normal
        String("stat", 1),
        1,
        # U: Updated
        # N: New
        String("mode", 1),
        1,
        # M: Manually
        # H: Host
        String("source", 2),
        String("analyser", 6),
    ]


class Control(BctModeBase):
    FMT_DESCRIPTION = [
        String("message_type", 1),
        1,
        String("control_id", 4),
        1,
        String("control_lot", 17),
        1,
        String("assay_number", 5),
        1,
        String("result", 10),
        1,
        String("unit", 10),
        1,
        String("flag", 4),
        1,
        Date("date", 15, fmt_date="%y%m%d %H:%M:%S"),
        1,
        String("stat", 1),
        1,
        String("analyser", 6),
    ]


class BctModeMultiple(list):
    MESSAGE_TYPE_MAP = {
        "J": Order,
        "R": Result,
        "I": Query,
        "Q": Control,
    }

    def add(self, message_type=None, **kwargs):
        obj = self.MESSAGE_TYPE_MAP[message_type](
            message_type=message_type, **kwargs)
        self.append(obj)
        return obj

    def parse(self, data):
        for line in data.splitlines():
            if line:
                obj = self.MESSAGE_TYPE_MAP[line[0]]()
                obj.parse(line)
                self.append(obj)

    def to_string(self):
        return "\n".join([
            e.to_string() for e in self
        ])


class Driver(Message):
    SPEC = {
        "code": "bcsxp_bctmode",
        "label": "BCS XP (BCTMODE)",
    }

    def create(self):
        return BctModeMultiple()

    def get_message_type(self, message):
        for msg in message:
            if msg.message_type == "I":
                return self.QUERY
        return self.ORDER

    def to_string(self, message):
        return message.to_string()

    def parse(self, data):
        obj = BctModeMultiple()
        obj.parse(data)
        return obj

    def get_queries(self, message):
        queries = []
        for msg in message:
            if msg.message_type == "I":
                queries.append(msg.sample_id)
        return queries

    def get_order_message(self, worklist, context):
        msg = self.create()

        for sample, action, priority in worklist.get_worklist(
            unique_together=["action", "priority"],
            priority={"A": "S", "S": "S", worklist.ELSE: "N"},
            action={"del": worklist.IGNORE, worklist.ELSE: ""},
        ):
            patient_name = "%s %s" % (
                sample["last_name"], sample["first_name"],
            )
            message = msg.add(
                message_type="J",
                sample_id=sample["sample_id"],
                emergency=priority,
                mode=action,
                patient_name=patient_name,
                requester=sample["service"],
            )
            for order in sample["orders"]:
                message.tests.add(code=order["code"])

        if msg:
            return msg

    def get_result_dict(self, message):
        results = {}
        for resultat in message:
            if resultat.message_type == "Q":
                continue
            results.setdefault(
                resultat.sample_id,
                {"resultats": []},
            )["resultats"].append({
                "test": resultat.assay_number.strip(),
                "resu": resultat.result.strip(),
                "date": resultat.date,
                "alarmes": [resultat.flag],
            })
        return results

    def get_qc_result_list(self, message):
        result = []
        for qc in message:
            if qc.message_type == "R":
                continue
            lot = "%s%s" % (qc.control_id, qc.control_lot,)
            result.append({
                "lot": lot,
                "test": qc.assay_number.strip(),
                "resu": qc.result.strip(),
                "date": qc.date,
            })
        return result
