# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng import MessageV2, Worklist
from driverng.fixed import (
    FixedLengthFormat,
    Multiple,
    String,
    Date,
)


class BN2Base(FixedLengthFormat):
    pass


class BN2Query(BN2Base):
    FMT_DESCRIPTION = [
        "J",
        String("sample_id", 29),
    ]


class BN2Order(BN2Base):
    FMT_DESCRIPTION = [
        "J",
        String("sample_id", 29),
        Date("birthdate", 8, fmt_date="%d%m%Y"),
        # m,M: Male
        # f,F: Female
        String("sex", 1),

        # 0: Normal
        # 1: Stat
        String("stat", 1),
        # 00: None
        # 01: Unspecific
        # 02: Manuel predelutd 1:5
        # 03: Unspecific and manuel predeluted 1:5
        # 04: Manuel predeluted 1:20
        # 05: Unspecific and predeluted 1:20
        String("remark", 2, default="00"),
        String("flag", 1, default="1"),
        String("name", 20),
        String("prename", 20),
        String("patient_id", 30),
        String("ordered_by", 20),
        String("kind_sample", 2),
        String("interpretation_id", 2),
        String("joblist_no", 3),
        Multiple("tests", [
            String("code", 2, right_justified=True),
            String("dilution", 1),
        ]),
    ]


class BN2Result(BN2Base):
    FMT_DESCRIPTION = [
        "D",
        String("sample_id", 29),
        String("code", 2, right_justified=True),

        # decimal 4 bits
        # bits 0-1:
        #   00: Ok
        #   01: Above range
        #   10: Below range
        #   11: Not valid
        # bit 2:
        #   0: Normal
        #   1: Turbid
        # bit 3:
        #   0: Default predilution
        #   1: Special predilution
        String("result_flags", 4, right_justified=True),
        String("result", 15, right_justified=True),
        String("unit", 9),
        String("emergency", 1),
        String("kind_sample", 2),
        String("interpretation_id", 2),
        String("result_no", 2),
        String("instrument_id", 5),
        Date("date", 12, fmt_date="%d%m%y%H%M%S"),
    ]


class BN2Control(BN2Base):
    FMT_DESCRIPTION = [
        "R",
        String("control", 29, right_justified=True),
        String("code", 2, right_justified=True),
        String("result_flags", 4, right_justified=True),
        String("result", 15, right_justified=True),
        String("unit", 9),
        String("deviation", 15, right_justified=True),
        String("charge_no", 2),
        String("instrument_id", 5),
        Date("date", 12, fmt_date="%d%m%y%H%M%S"),
    ]


class Driver(MessageV2):
    SPEC = {
        "code": "bn2_autohost",
        "label": "BN2 AutoHost",
    }

    def create(self, message_type, **kwargs):
        if message_type == self.ORDER:
            return BN2Order(**kwargs)
        elif message_type == self.QUERY:
            return BN2Query(**kwargs)
        elif message_type == self.RESULT:
            return BN2Result(**kwargs)
        elif message_type == self.QC:
            return BN2Control(**kwargs)
        else:
            raise NotImplementedError

    def parse(self, data):
        if data[0] == "J" and len(data) > 30:
            klass = BN2Order
        elif data[0] == "J" and len(data) == 30:
            klass = BN2Query
        elif data[0] == "D":
            klass = BN2Result
        elif data[0] == "R":
            klass = BN2Control
        else:
            raise NotImplementedError

        obj = klass()
        obj.parse(data)
        return obj

    def to_string(self, message):
        return message.to_string()

    def encode_worklist(self, worklist, context):
        if worklist.context != "order":
            raise NotImplementedError

        samples = worklist.get_worklist(
            unique_together=["action"],
            action={"del": worklist.IGNORE, worklist.ELSE: ""},
        )
        if len(samples) == 0:
            return
        elif len(samples) > 1:
            raise RuntimeError("Multi sample")
        else:
            sample, _ = samples[0]
            message = self.create(
                self.ORDER,
                sample_id=sample.sample_id,
                birthdate=sample.birthdate,
                sex=sample.sex if sample.sex in ("M", "F") else "",
                stat="0",
                name=sample.last_name,
                prename=sample.first_name,
                patient_id=sample.patient_id,
            )
            for physician in sample.physicians:
                message.ordered_by = physician.code
            for i, order in enumerate(sample.orders):
                message.tests.add(
                    code=order.code,
                    dilution="%.1s" % (order.dilution,) or "0"
                )
                if order.priority in ("S", "A"):
                    message.stat = "1"
            return message

    def decode_message(self, message):
        if isinstance(message, BN2Query):
            worklist = Worklist(context="query")
            worklist.get_or_create(type="patient", sample_id=message.sample_id.strip())
            return worklist

        worklist = Worklist(context="result")
        if isinstance(message, BN2Result):
            sample = worklist.get_or_create(
                type="patient",
                sample_id=message.sample_id,
            )
        elif isinstance(message, BN2Control):
            sample = worklist.get_or_create(
                type="control",
                control_lot=message.control,
                sample_id=message.control,
            )
        else:
            raise NotImplementedError

        flags = []
        status = int(message.result_flags or "0")

        if status & 3 == 3:
            flags.append("NOT_VALID")
        elif status & 1 == 1:
            flags.append("ABOVE_RANGE")
        elif status & 2 == 2:
            flags.append("BELLOW_RANGE")

        if status & 4 == 4:
            flags.append("TURBID")
        if status & 8 == 8:
            flags.append("SPECIAL_PREDILUTION")

        sample.results.add(
            code=message.code,
            value=message.result,
            flags=flags,
            date=message.date,
        )
        return worklist
