# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng import MessageV2, Worklist
from driverng.utils import num_years
from driverng.fixed import (
    FixedLengthFormat,
    Multiple,
    String,
)


class BN2Base(FixedLengthFormat):
    pass


class BN2Order(BN2Base):
    FMT_DESCRIPTION = [
        "J",
        String("sample_id", 29),
        1,
        String("age", 3, right_justified=True),
        # m,M: Male
        # f,F: Female
        String("sex", 1),

        # 0: Normal
        # 1: Stat
        String("stat", 1),

        # 00: None
        # 01: Unspecific
        # 02: Manuel predelutd 1:5
        # 03: Unspecific and manuel predeluted 1:5
        # 04: Manuel predeluted 1:20
        # 05: Unspecific and predeluted 1:20
        String("remark", 2),

        Multiple("tests", [
            String("code", 2, right_justified=True),
        ]),
    ]


class BN2Result(BN2Base):
    FMT_DESCRIPTION = [
        "D",
        String("sample_id", 29),
        2,
        String("code", 2, right_justified=True),
        1,

        # decimal 4 bits
        # bits 0-1:
        #   00: Ok
        #   01: Above range
        #   10: Below range
        #   11: Not valid
        # bit 2:
        #   0: Normal
        #   1: Turbid
        # bit 3:
        #   0: Default predilution
        #   1: Special predilution
        String("status", 4, right_justified=True),
        1,
        String("result", 15, right_justified=True),
        1,
        String("unit", 9),
    ]


class BN2Control(BN2Base):
    FMT_DESCRIPTION = [
        "R",
        String("control", 4, right_justified=True),
        1,
        String("code", 2, right_justified=True),
        1,
        String("status", 4, right_justified=True),
        1,
        String("result", 15, right_justified=True),
        1,
        String("unit", 9),
        1,
        String("deviation", 15, right_justified=True),
    ]


class Driver(MessageV2):
    SPEC = {
        "code": "bn2_serhost",
        "label": "BN2 SerHost",
    }

    def create(self, message_type, **kwargs):
        if message_type == self.ORDER:
            return BN2Order(**kwargs)
        elif message_type == self.RESULT:
            return BN2Result(**kwargs)
        elif message_type == self.QC:
            return BN2Control(**kwargs)
        else:
            raise NotImplementedError

    def parse(self, data):
        if data[0] == "J":
            klass = BN2Order
        elif data[0] == "D":
            klass = BN2Result
        elif data[0] == "R":
            klass = BN2Control
        else:
            raise NotImplementedError

        obj = klass()
        obj.parse(data)
        return obj

    def to_string(self, message):
        return message.to_string()

    def encode_worklist(self, worklist, context):
        if worklist.context != "order":
            raise NotImplementedError

        samples = worklist.get_worklist(
            unique_together=["action"],
            action={"del": worklist.IGNORE, worklist.ELSE: ""},
        )
        if len(samples) == 0:
            return
        elif len(samples) > 1:
            raise RuntimeError("Multi sample")
        else:
            sample, _ = samples[0]
            sample_id = "%s %s" % (sample.sample_id, sample.last_name.upper())
            message = self.create(
                self.ORDER,
                sample_id=sample_id[:19],
                sex=sample.sex if sample.sex in ("M", "F") else "",
                age=num_years(sample.birthdate, sample.collection_date),
                stat="0",
            )
            for i, order in enumerate(sample.orders):
                message.tests.add(code=order.code)
                if order.priority in ("S", "A"):
                    message.stat = "1"
            while i < 19:
                message.tests.add(code="0")
                i += 1
            return message

    def decode_message(self, message):
        worklist = Worklist(context="result")
        if isinstance(message, BN2Result):
            sample_id, last_name = message.sample_id.split(" ", 1)
            sample = worklist.get_or_create(
                type="patient",
                sample_id=sample_id,
                last_name=last_name,
            )
        elif isinstance(message, BN2Control):
            sample = worklist.get_or_create(
                type="control",
                control_lot=message.control,
                sample_id=message.control,
            )
        else:
            raise NotImplementedError

        flags = []
        status = int(message.status or "0")

        if status & 3 == 3:
            flags.append("NOT_VALID")
        elif status & 1 == 1:
            flags.append("ABOVE_RANGE")
        elif status & 2 == 2:
            flags.append("BELLOW_RANGE")

        if status & 4 == 4:
            flags.append("TURBID")
        if status & 8 == 8:
            flags.append("SPECIAL_PREDILUTION")

        sample.results.add(
            code=message.code,
            value=message.result,
            flags=sorted(flags),
        )
        return worklist
