# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng import Worklist
from driverng.astm import (
    AstmMessageV2,
    HeaderSegment,
    Segment,
    HprimString,
    HprimList,
    HprimDate,
    HprimForeign,
    HprimMultiple,
)


class BnProsSegmentH(HeaderSegment):
    password = HprimString(4)
    sender = HprimMultiple(
        5,
        name=HprimString(1),
        version=HprimString(2),
        serial_no=HprimString(3),
        terminal_version=HprimString(4),
    )
    result_type = HprimString(12, default="P")
    version = HprimString(13, default="1")
    date = HprimDate(14)

    patients = HprimForeign("P", "header")
    queries = HprimForeign("Q", "header")
    end = HprimForeign("L", "header", one2one=True, required=True)


class BnProsSegmentP(Segment):
    patient_id = HprimString(3)
    laboratory_patient_id = HprimString(4)
    last_name = HprimString((6, 1))
    first_name = HprimString((6, 2))
    midle_initial = HprimString((6, 3))
    date_birth = HprimDate(8, fmt_date="%Y%m%d")
    # M/F/U
    sex = HprimString(9)

    orders = HprimForeign("O", "patient")


class BnProsSegmentO(Segment):
    sample_id = HprimString(3)
    sample_id2 = HprimString(4)
    test_id = HprimList(5, basefield=HprimMultiple(
        0,
        assay_no=HprimString(4),
        test_name=HprimString(5),
        dilution=HprimString(6),
    ))
    # R/S
    priority = HprimString(6)
    action_code = HprimString(12)
    report_type = HprimString(26)

    results = HprimForeign("R", "order")


class BnProsSegmentR(Segment):
    universal_test_id = HprimMultiple(
        3,
        test=HprimString(4),
        name=HprimString(5),
        dilution=HprimString(6),
        status=HprimString(8),  # always F
    )
    result = HprimString(4)
    units = HprimString(5)
    reference_ranges = HprimString((6, 1))
    # L: result < range
    # H: result > range
    flag = HprimString(7)
    result_status = HprimString(9)
    date_result = HprimDate(13)


class BnProsSegmentC(Segment):
    source = HprimString(3)  # always I
    text = HprimString(4)
    # G: Text after result: G Otherwise: I
    origine = HprimString(5)


class BnProsSegmentQ(Segment):
    specimen_id = HprimString((3, 2))
    tests = HprimString((5, 4), default="ALL")  # always ALL
    request_status = HprimString(13)  # always O


class BnProsSegmentL(Segment):
    pass


class Driver(AstmMessageV2):
    SPEC = {
        "code": "bnpros",
        "label": "BN ProSpec",
    }
    SEGMENT_CLASS_MAP = {
        "H": BnProsSegmentH,
        "P": BnProsSegmentP,
        "O": BnProsSegmentO,
        "Q": BnProsSegmentQ,
        "R": BnProsSegmentR,
        "C": BnProsSegmentC,
        "L": BnProsSegmentL,
    }
    DEFAULT_SEPARATORS = "|\\^&"

    def create(self, **kwargs):
        message = BnProsSegmentH(type_segment="H", **kwargs)
        message.end = BnProsSegmentL(type_segment="L")
        return message

    def encode_worklist(self, worklist, context):
        if worklist.context != "order":
            raise NotImplementedError("context %s" % (worklist.context,))

        message = self.create(date=context["now"])
        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority"],
            action={"del": worklist.IGNORE, worklist.ELSE: ""},
            priority={"A": "S", "S": "S", worklist.ELSE: "R"},
        ):
            seg_p = message.patients.add(
                patient_id=patient.patient_id,
                last_name=patient.last_name,
                first_name=patient.first_name,
                birthdate=patient.birthdate,
                sex=patient.sex,
            )
            for sample, action, priority in patient.samples:
                seg_o = seg_p.orders.add(
                    sample_id=sample.sample_id,
                    sample_id2=sample.sample_id,
                    priority=priority,
                    action_code=action,
                    report_type="Q" if context["query"] else "O",
                )
                for order in sample.orders:
                    t = seg_o.test_id.add()
                    t.assay_no = order.code

        if len(message.patients):
            return message
        elif context["query"] and context["query"].queries:
            for seg_q in context["query"].queries:
                message.queries.add(
                    specimen_id=seg_q.specimen_id,
                    request_status="X",
                )
            return message

    def decode_message(self, message):

        if message.queries:
            worklist = Worklist(context="query")
            for seg_q in message.queries:
                worklist.get_or_create(type="patient", sample_id=seg_q.specimen_id.strip())
            return worklist

        worklist = Worklist(context="result")
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                sample_id = seg_o.sample_id.strip()
                if seg_o.action_code == "Q":
                    sample = worklist.get_or_create(
                        type="control",
                        control_lot=sample_id,
                        sample_id=sample_id,
                    )
                else:
                    sample = worklist.get_or_create(
                        type="patient",
                        sample_id=sample_id,
                    )
                for seg_r in seg_o.results:
                    result = sample.results.add(
                        code=seg_r.universal_test_id.test,
                        value=seg_r.result,
                        date=seg_r.date_result,
                    )
                    if seg_r.flag:
                        result.flags.append(seg_r.flag)

                    for seg_c in seg_r.commentaires:
                        result.comments.append(seg_c.text)

        return worklist
