# -*- coding: utf8 -*-
from __future__ import unicode_literals

import binascii
import base64
import struct
import logging
from StringIO import StringIO


from driverng import GraphContainer
from driverng.lazy import lazy
from driverng.astm import (
    AstmMessage,
    HeaderSegment,
    Segment,
    HprimString,
    HprimList,
    HprimDate,
    HprimForeign,
    HprimMultiple,
)

logger = logging.getLogger("driverng.drivers.ruby")


class RubyGraphContainer(GraphContainer):
    # TODO
    # pylint: disable=C0324,C0323,C0103,W0311,C0301
    SCATTOGRAM_COLOR_MAP = {
        "lymph": "c,",
        "mono": "m,",
        "baso": "w,",
        "neutro": "y,",
        "eos": "g,",
        "rbc": "r,",
    }

    def __init__(self, seg_m, wbc2=False, type_fichier="png"):
        self.seg_m = seg_m
        self.wbc2 = wbc2
        super(RubyGraphContainer, self).__init__(type_fichier)

    @staticmethod
    def generate_hemogram_graph(seg_m):
        pylab = lazy.pylab
        pylab.figure(figsize=(4, 4), edgecolor = "r")
        l = seg_m.discriminants
        pylab.axvline(x=l[0], ymin=0, ymax=1, color='green')

        if len(l) > 1:
            pylab.axvline(x=l[1], ymin=0, ymax=1, color='green')

        # cell data : decodage de la base 64 en valeur de 16 bits jusqu'a 256
        # chaque valeur de 16 bits correspond a un point de l'histogramme
        # representant la valeur Y et la valeur X associee comprises de 0 a 255
        cell_data = binascii.a2b_base64(seg_m.cell_data)
        fmt = (len(cell_data) / 2) * b'H'  # nb de short
        ty = struct.unpack(fmt, cell_data)
        pylab.plot(ty)

        if len(l) > 1:
            tx1 = list(range(1, int(l[0]) + 1))
            ty1 = ty[:int(l[0])]
            pylab.fill_between(tx1, ty1, color='#FF8000')

            tx2 = list(range(int(l[0]), int(l[1])))
            ty2 = ty[int(l[0]):int(l[1])]
            pylab.fill_between(tx2, ty2, color='blue')
            tx3 = list(range(int(l[1]), 250))
            ty3 = ty[int(l[1]): 250]
            pylab.fill_between(tx3, ty3, color='red')

        bufimg = StringIO()
        pylab.savefig(bufimg, dpi=100)
        pylab.close()
        return bufimg.getvalue()

    def generate_scattogram_graphs(self, seg_m, wbc2=False, colors=None):
        pylab = lazy.pylab
        if colors is None:
            colors = self.SCATTOGRAM_COLOR_MAP
        t_channel = seg_m.mesure_point
        type_data = seg_m.type_graphe

        lenVect = len(t_channel)
        t_data = base64.b64decode(seg_m.cell_data)
        cell_data = binascii.a2b_base64(seg_m.cell_data)  # .replace('\r',''))
        fmt = (len(cell_data)) * b'B'  # nb de uChar
        tvect = struct.unpack(fmt, t_data)
        figure = pylab.figure(figsize=(4, 4), edgecolor="r", facecolor='k')
        ax1 = figure.add_subplot(111, axisbg='k')
        pylab.title(type_data)
        l = seg_m.discriminants
        if len(l) > 2:
            tlig = l[2]. split('^')
        else:
            tlig = []
        pylab.xlabel(t_channel[0])
        pylab.ylabel(t_channel[1])
        tx, ty = [], []
        for t in range(len(tlig) / 2):
            if t == 0:
                continue
            tx.append(tlig[2 * t])
            ty.append(tlig[2 * t + 1])
        ax1.plot(tx, ty, 'w')
        txlymph, tylymph, tylymph2 = [], [], []
        txmono, tymono, tymono2 = [], [], []
        txbaso, tybaso, tybaso2 = [], [], []
        txneutro, tyneutro, tyneutro2 = [], [], []
        txeos, tyeos, tyeos2 = [], [], []
        txrbc, tyrbc, tyrbc2 = [], [], []
        dictcell = {}
        for i in range(len(tvect) / lenVect):  # les 5 valeurs des vecteurs
            typeCT = tvect[lenVect * i + lenVect - 1]
            if not dictcell.get(typeCT, None):
                dictcell[typeCT] = [[]] * 4
            for j in range(lenVect - 1):
                try:
                    dictcell[typeCT][j].append(tvect[i * lenVect + j])
                except IndexError:
                    pass
            x = tvect[i * lenVect + 1]
            y = tvect[i * lenVect + 0]
            y2 = tvect[i * lenVect + 2]
            if typeCT in (1, 11, 17):  # cyan
                txlymph.append(x)
                tylymph.append(y)
                tylymph2.append(y2)
            if typeCT in (2, 12):  # magenta
                txmono.append(x)
                tymono.append(y)
                tymono2.append(y2)
            if typeCT in (3, 13):  # white
                txbaso.append(x)
                tybaso.append(y)
                tybaso2.append(y2)
            if typeCT in (4, 10, 15):  # yellow
                txneutro.append(x)
                tyneutro.append(y)
                tyneutro2.append(y2)
            if typeCT in (5, ):  # green
                txeos.append(x)
                tyeos.append(y)
                tyeos2.append(y2)
            if typeCT in (9, ):  # red
                txrbc.append(x)
                tyrbc.append(y)
                tyrbc2.append(y2)

        ax1.plot(
            txlymph, tylymph, colors["lymph"],
            txmono, tymono, colors["mono"],
            txbaso, tybaso, colors["baso"],
            txneutro, tyneutro, colors["neutro"],
            txeos, tyeos, colors["eos"],
            txrbc, tyrbc, colors["rbc"],
            label=type_data)

        scattogram_graph, wbc_graph = StringIO(), None
        pylab.savefig(scattogram_graph, dpi=100)
        pylab.close()

        if not wbc2 or type_data != "WBC":
            return scattogram_graph.getvalue()
        else:
            figure = pylab.figure(
                figsize=(4, 4), edgecolor="r", facecolor='k')
            ax1 = figure.add_subplot(111, axisbg='k')
            pylab.title(type_data)
            pylab.xlabel(t_channel[0])
            pylab.ylabel(t_channel[2])
            ax1.plot(
                tylymph, tylymph2, colors["lymph"],
                tymono, tymono2, colors["mono"],
                tybaso, tybaso2, colors["baso"],
                tyneutro, tyneutro2, colors["neutro"],
                tyeos, tyeos2, colors["eos"],
                tyrbc, tyrbc2, colors["rbc"],
                label=type_data)
            wbc_graph = StringIO()
            pylab.savefig(wbc_graph, dpi=100)
            pylab.close()
            return wbc_graph.getvalue()

    def get_bytes(self):
        if self.seg_m.type_data == "H":
            return self.generate_hemogram_graph(self.seg_m)
        elif self.seg_m.type_data == "S":
            return self.generate_scattogram_graphs(self.seg_m, wbc2=self.wbc2)
        else:
            logger.warning("Unkwnon type_data %s", self.seg_m.type_data)
            return None


class RubySegmentH(HeaderSegment):

    # code emetteur
    emetteur = HprimString(5)

    # Processing ID
    # code de traitement par defaut P pour Production
    id_traitement = HprimString(12, default="P")

    # No de Version
    no_version = HprimString(13, default="LIS2-A")

    patients = HprimForeign("P", "header")
    queries = HprimForeign("Q", "header")
    end = HprimForeign("L", "header", one2one=True, required=True)


class RubySegmentP(Segment):

    # Patient ID ou numero de patient
    no_patient = HprimString(4)

    # nom et prenom patient
    nom = HprimString(6)

    # date de naissance  format : YYYYMMDD
    date_naissance = HprimDate(8, fmt_date="%Y%m%d")

    # sexe  M/F/U
    sexe = HprimString(9)

    # code du correspondant
    correspondant = HprimString(14)

    # Special Field
    user_field1 = HprimString((16, 1))
    user_field2 = HprimString((16, 2))

    orders = HprimForeign("O", "patient")


class RubySegmentO(Segment):

    # numero de tube
    tube = HprimString((3, 1))
    rack = HprimString((3, 2))

    # liste des tests
    tests = HprimList(
        5,
        basefield=HprimMultiple(
            0,
            code=HprimString(4),
            parameter_set_selection=HprimString(5),
            limit_set_selection=HprimString(6),
        ),
    )

    # date et heure de prelevement au format YYYYMMDDHHMMSS
    date_prelevement = HprimDate(8)

    # specimen descriptor
    # ST  = Specimen Type
    # QC  = Controle qualite
    # Lors de la demande, ce champ est optionnel,
    # si present, le specimen type prend la valeur "Patient"
    type_echantillon = HprimString((16, 1))
    # SST = Specimen Subtype
    specimen_subtype = HprimString((16, 2))

    # Report type
    report_type = HprimString(26)

    results = HprimForeign("R", "order")
    graphe = HprimForeign("M", "order")


class RubySegmentQ(Segment):

    # numero de tube
    tube = HprimString((3, 2))


class RubySegmentR(Segment):

    # code test
    # ^^^PROFIL^^^TEST
    test = HprimString((3, 7))

    # valeur
    valeur = HprimString(4)

    # status du resultat
    status = HprimString(9)

    # Operator ID
    operator = HprimString(11)

    # Date et heure du resultat
    date = HprimDate(12)

    # instrument identification
    # nom ou reference de l'analyseur
    analyseur = HprimString(14)


class RubySegmentM(Segment):

    # type de donnee
    type_data = HprimString(3)

    # type de graphe
    # H: Histogram
    # S: Scattergram
    type_graphe = HprimString(4)

    # segment M pour enregistrement des points
    # ATTENTION: C'est un champ composé
    # Mais mappé en String pour le script de graphes.
    mesure_point = HprimString(5)

    # discriminants
    # ATTENTION: C'est une liste de champs composés
    # mais mappés en liste de String pour le script de graphes.
    discriminants = HprimList(6)

    # cell data
    cell_data = HprimString(7)


class RubySegmentC(Segment):

    # comment source
    code_commentaire = HprimString(3)

    # texte du commentaire
    commentaire = HprimString(4)

    # type de commentaire
    type_commentaire = HprimString(5)


class RubySegmentL(Segment):

    # code terminaison
    # N terminaison normale
    code = HprimString(3, default="N")


class DriverRuby(AstmMessage):
    SPEC = {
        "code": "ruby",
        "label": "Celldyn Ruby",
    }

    SEGMENT_CLASS_MAP = {
        "H": RubySegmentH,
        "P": RubySegmentP,
        "Q": RubySegmentQ,
        "O": RubySegmentO,
        "R": RubySegmentR,
        "M": RubySegmentM,
        "C": RubySegmentC,
        "L": RubySegmentL,
    }
    DEFAULT_SEPARATORS = "|\^&"
    GRAPH_CLASS = RubyGraphContainer

    def create(self, **kwargs):
        message = RubySegmentH(type_segment="H", **kwargs)
        message.end = RubySegmentL(type_segment="L")
        return message

    def get_message_type(self, message):
        if message.queries:
            return self.QUERY
        else:
            return self.ORDER

    def get_queries(self, message):
        queries = []
        for seg_q in message.queries:
            queries.append(seg_q.tube)
        return queries

    def get_order_message(self, worklist, context):
        msg = self.create(date=context["now"])
        for patient in worklist.get_astm_worklist(
            unique_together=["action"],
            action={"del": worklist.IGNORE, worklist.ELSE: ""},
        ):
            seg_p = msg.patients.add(
                no_patient=patient["patient_id"],
                nom="%s %s" % (patient["last_name"], patient["first_name"]),
                date_naissance=patient["birthdate"],
                sexe=patient["sex"],
                correspondant=patient["physician"],
            )
            for sample, action in patient["samples"]:
                seg_o = seg_p.orders.add(
                    tube=sample["sample_id"],
                    date_prelevement=sample["collection_date"],
                    type_echantillon="Patient",
                    specimen_subtype="Human",
                    report_type="F",
                )
                for order in sample["orders"]:
                    t = seg_o.tests.add()
                    t.code = order["code"]

        if len(msg.patients):
            return msg

    @staticmethod
    def ignore_resultat(seg_r):
        # prendre uniquement le resultat final
        # ne pas prendre les alarmes hemato
        if seg_r.valeur == "FLAG" or seg_r.status != "F":
            return True
        else:
            return False

    def get_alarmes_hemato(self, message):
        alarmes_hemato = {}
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                if "QC" in seg_o.type_echantillon:
                    continue
                for seg_r in seg_o.results:
                    if seg_r.valeur != "FLAG":
                        # Pas une alarme hemato
                        continue
                    analyseur = self.get_analyseur(seg_r.analyseur)
                    alarmes_hemato.setdefault(seg_o.tube, {
                        "alarmes": [],
                        "analyseur": analyseur,
                    })["alarmes"].append(seg_r.test)
        return alarmes_hemato

    @staticmethod
    def get_rack(seg_o):
        rack = None
        if seg_o.rack:
            try:
                _rack, pos = seg_o.rack.split()
                rack = "%s-%s" % (_rack, pos)
            except ValueError:
                pass
        return rack

    def get_result_dict(self, message):
        graphics_available = lazy.import_all()
        result = {}
        alarmes_hemato = {}
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                # exclusion des CQ
                if "QC" in seg_o.type_echantillon:
                    continue

                rack = self.get_rack(seg_o)
                result.setdefault(seg_o.tube, {"resultats": [], "rack": rack})

                # le code test est la concatenation de GR_ et du type de graphe
                if graphics_available:
                    for seg_m in seg_o.graphe:
                        code_test = "GR_%s" % (seg_m.type_graphe,)
                        result[seg_o.tube]["resultats"].append({
                            "test": code_test,
                            "resu": None,
                            "graph": self.GRAPH_CLASS(seg_m)
                        })
                        if (
                            seg_m.type_data == "S"
                            and seg_m.type_graphe == "WBC"
                        ):
                            # On a deux graphes pour WBC (GR_WBC et GR_WBC2)
                            result[seg_o.tube]["resultats"].append({
                                "test": "GR_WBC2",
                                "resu": None,
                                "graph": self.GRAPH_CLASS(seg_m, wbc2=True)
                            })

                for seg_r in seg_o.results:
                    if self.ignore_resultat(seg_r):
                        continue

                    analyseur = self.get_analyseur(seg_r.analyseur)

                    result[seg_o.tube]["resultats"].append({
                        "test": seg_r.test,
                        "resu": seg_r.valeur,
                        "date": seg_r.date,
                        "analyseur": analyseur,
                    })

        alarmes_hemato = self.get_alarmes_hemato(message)
        for tube, values in alarmes_hemato.items():
            result.setdefault(tube, {
                "resultats": [],
                "rack": None,
            })["resultats"].append({
                "test": "ALARME",
                "resu": "",
                "analyseur": values["analyseur"],
                "commentaires": values["alarmes"],
            })
        return result

    @staticmethod
    def get_analyseur(analyseur):
        # Souvent des espaces qui poluent
        if analyseur:
            return analyseur.strip()
        else:
            return None

    def get_qc_result_list(self, message):
        result = []
        for seg_p in message.patients:

            for seg_o in seg_p.orders:

                # uniquement les CQ
                if "QC" not in seg_o.type_echantillon:
                    continue

                for seg_r in seg_o.results:
                    if self.ignore_resultat(seg_r):
                        continue
                    analyseur = self.get_analyseur(seg_r.analyseur)
                    result.append({
                        "lot": seg_o.tube,
                        "test": seg_r.test,
                        "resu": seg_r.valeur,
                        "date": seg_r.date,
                        "analyseur": analyseur,
                    })

        return result

    def enregistrer(self, data, message=None):
        # En full download on n'enregistre pas les
        # query ALL intempestifs
        if data:
            lines = data.splitlines()
            if len(lines) == 3 and lines[1] == "Q|1|^ALL":
                return False
        return True
