# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng import Message


class Field(object):
    def __init__(self, name):
        self.name = name
        super(Field, self).__init__()

    def to_python(self, value):
        raise NotImplementedError

    def to_string(self, value):
        raise NotImplementedError


class String(Field):
    def __init__(self, name):
        super(String, self).__init__(name)

    def to_python(self, value):
        return value.strip()

    def to_string(self, value):
        if value is None:
            return ""
        else:
            return value


class Cla1Base(object):
    FIELDS = ()

    def __init__(self, **kwargs):
        for field in self.FIELDS:
            setattr(self, field.name, kwargs.get(field.name, None))
        super(Cla1Base, self).__init__()

    def parse(self, data):
        fields = self.FIELDS
        maxEltLigne = len(fields)
        for idx, elt in enumerate(data.split(";")[:-1]):
            if idx >= maxEltLigne:
                break
            field = fields[idx]
            value = field.to_python(elt)
            setattr(self, field.name, value)

    def to_string(self):
        ligne = []
        for field in self.FIELDS:
            ligne.append(field.to_string(getattr(self, field.name)))
        return "%s;" % (";".join(ligne))


class Order(Cla1Base):
    FIELDS = (
        String("patient_code"),
        String("panel_code"),
        String("patient_first_name"),
        String("patient_last_name"),
    )


class Resultat(Cla1Base):
    FIELDS = (
        String("patient_code"),
        String("panel_code"),
        String("patient_first_name"),
        String("patient_last_name"),
    ) + tuple([String("allergen%d" % (i,)) for i in range(1, 39)])


class Cla1Multiple(list):

    _MESSAGE_TYPE_MAP = {
        Message.ORDER: Order,
        Message.RESULT: Resultat,
    }

    def add(self, type_message=None, **kwargs):
        obj = self._MESSAGE_TYPE_MAP[type_message](**kwargs)
        self.append(obj)
        return obj

    def parse(self, data):
        for line in data.splitlines():
            if not len(line):
                continue
            if len(line.split(";")) == 4:
                type_message = Message.ORDER
            else:
                type_message = Message.RESULT
            obj = self.add(type_message=type_message)
            obj.parse(line)

    def to_string(self):
        return "\r\n".join([e.to_string() for e in self]) + "\r\n"


class Driver(Message):
    SPEC = {
        "code": "cla1",
        "label": "Cla1 Luminometer",
    }

    def create(self):
        return Cla1Multiple()

    def get_message_type(self, message):
        return self.ORDER

    def to_string(self, message):
        return message.to_string()

    def parse(self, data):
        obj = Cla1Multiple()
        obj.parse(data)
        return obj

    def get_order_message(self, worklist, context):
        msg = self.create()
        for sample in worklist.samples:
            for order in sample["orders"]:
                if order["action"] == "del":
                    continue
                msg.add(
                    type_message=self.ORDER,
                    patient_code=sample["sample_id"],
                    panel_code=order["code"],
                    patient_first_name="%.20s" % (sample["first_name"],),
                    patient_last_name="%.15s" % (sample["last_name"],),
                )
        if len(msg):
            return msg

    def get_result_dict(self, message):
        results = {}
        for resultat in message:
            for i in range(1, 39):
                results.setdefault(
                    resultat.patient_code,
                    {"resultats": []},
                )["resultats"].append({
                    "test": "%s_%d" % (resultat.panel_code, i),
                    "resu": getattr(resultat, "allergen%d" % (i,)),
                })
        return results
