# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng.astm import (
    AstmMessage,
    HeaderSegment,
    Segment,
    HprimString,
    HprimMultiple,
    HprimList,
    HprimDate,
    HprimForeign,
)
from driverng.utils import num_years, num_month, num_days


class CobasSegmentH(HeaderSegment):

    sender_name = HprimString((5, 1))
    sender_cersion = HprimString((5, 2))
    recepteur = HprimString(10)

    # Roche/Hitachi cobas6000 system reception process
    # To indicate classification of the transmitted message from
    # the host, one of the following is transmitted.
    # TSDWN^REPLY
    # RSREQ^REAL
    # TSDWN^BATCH
    special1 = HprimString((11, 1), default="TSDWN")
    special2 = HprimString((11, 2), default="BATCH")
    processing = HprimString(12, default="1")
    version = HprimString(13, default="1")
    date = HprimDate(14)

    patients = HprimForeign("P", "header")
    queries = HprimForeign("Q", "header")
    infos = HprimForeign("M", "header")
    end = HprimForeign("L", "header", one2one=True, required=True)


class CobasSegmentP(Segment):

    no_patient = HprimString(5)
    nom = HprimString((6, 2))
    prenom = HprimString((6, 3))
    date_naissance = HprimDate(8, fmt_date="%Y%m%d")
    # M/F/U
    sexe = HprimString(9)
    age = HprimString((15, 1), max_length=3)
    # Y/M/D
    age_unit = HprimString((15, 2), max_length=1)

    orders = HprimForeign("O", "patient")


class CobasSegmentQ(Segment):

    # sample id analyseur
    sample_id = HprimString((3, 3), max_length=22)
    sample_no = HprimString((3, 4), max_length=5)
    rack_id = HprimString((3, 5), max_length=5)
    position = HprimString((3, 6), max_length=1)
    rack_type = HprimString((3, 8), max_length=2)
    container_type = HprimString((3, 9), max_length=2)

    # R1: For 1st measurement
    # R2: For rerun measurement
    # This information is sent to the host only when an option of "Send
    # 1st/Rerun information" is specified
    kind = HprimString((3, 10), max_length=2)
    tests = HprimString(5, default='ALL')
    status = HprimString(13, max_length=1)


class CobasSegmentO(Segment):

    # tube OU ref du controle
    tube = HprimString(3, max_length=22)
    # sample id analyseur
    sample_no = HprimString((4, 1), max_length=5)
    rack_id = HprimString((4, 2), max_length=5)
    position = HprimString((4, 3), max_length=1)
    rack_type = HprimString((4, 5), max_length=2)
    container_type = HprimString((4, 6), max_length=2)

    tests = HprimList(5, basefield=HprimMultiple(
        0,
        code=HprimString(4),
        dilution=HprimString(5, obligatoire=True),
    ))

    # S: Urgent
    # R: Routine
    urgence = HprimString(6, max_length=1, default="R")

    collection_date = HprimDate(8)

    # A: Ajout de tests
    # C: suppression de tests
    code_action = HprimString(12)

    # 1 serum
    # 2 urine
    # 3 CSF
    # 4 Supmt
    # 5 Other
    specimen_descriptor = HprimString(16, max_length=1)

    # Date et heure des resultats
    resu_date = HprimDate(23)

    # O: SIL
    # F: analyseur
    report_type = HprimString(26, max_length=1, default='O')
    results = HprimForeign("R", "order")


class CobasSegmentR(Segment):

    test = HprimString((3, 4), max_length=5)
    valeur_qt = HprimString((4, 1), max_length=6)
    valeur_ql = HprimString((4, 2), max_length=6)

    unit = HprimString(5)
    flags = HprimList(7)

    # F: final
    # C: rerun
    status = HprimString(9, max_length=1)

    operator = HprimString(11, max_length=6)
    date = HprimDate(12)

    numero_analyseur = HprimString(14, max_length=4)
    infos = HprimForeign("M", "result")


class CobasSegmentM(Segment):

    # use ECR
    type_sub = HprimString(3, max_length=3)
    operator = HprimString(4, max_length=6)
    test = HprimString((5, 4), max_length=6)
    method = HprimString(6)
    module = HprimString(7, max_length=3)
    lot_rack = HprimString(8, max_length=8)
    no_rack = HprimString(9, max_length=5)

    # lot réactif
    expired_flag = HprimString(10, max_length=1)

    calibrator_lot = HprimString(11, max_length=8)
    status = HprimString(12, max_length=1)
    caracteristics = HprimList(13, basefield=HprimMultiple(
        0,
        id=HprimString(1),
        criteria=HprimString(2),
        detail=HprimString(3),
    ))
    signal = HprimList(14, basefield=HprimMultiple(
        0,
        val1=HprimString(1),
        val2=HprimString(2),
        val3=HprimString(3),
        val4=HprimString(4),
    ))
    target1 = HprimString((15, 1))
    target2 = HprimString((15, 2))
    target3 = HprimString((15, 3))
    target4 = HprimString((15, 4))
    target5 = HprimString((15, 5))
    unit = HprimString(16)
    cut_off = HprimString(17)
    border_line = HprimString(18)
    date = HprimDate(19)


class CobasSegmentC(Segment):
    # C|1|L|XXXX^XXXXXX

    source = HprimString(3, max_length=1, default='L')
    comment1 = HprimString((4, 1))
    comment2 = HprimString((4, 2))
    comment3 = HprimString((4, 3))
    comment4 = HprimString((4, 4))
    comment5 = HprimString((4, 5))
    type = HprimString(5, max_length=1, default='G')


class CobasSegmentL(Segment):

    # N: Normal
    # T: Forced termination of the transmitting side
    # R: Forced termination of the receiving side
    # E: System error
    # Q: Request error for the last reception
    # I: No information
    # F: Last request processed
    termination_code = HprimString(3, default='N')


class Driver(AstmMessage):
    SPEC = {
        "code": "cobas",
        "label": "Cobas Roche",
    }
    SEGMENT_CLASS_MAP = {
        "H": CobasSegmentH,
        "P": CobasSegmentP,
        "O": CobasSegmentO,
        "R": CobasSegmentR,
        "M": CobasSegmentM,
        "C": CobasSegmentC,
        "Q": CobasSegmentQ,
        "L": CobasSegmentL,
    }
    DEFAULT_SEPARATORS = "|\^&"
    ENCODING = "latin1"

    def create(self, **kwargs):
        message = CobasSegmentH(type_segment="H", **kwargs)
        message.end = CobasSegmentL(type_segment="L")
        return message

    def get_message_type(self, message):
        if message.queries:
            return self.QUERY
        else:
            return self.ORDER

    def get_queries(self, message):
        return [seg_q.sample_id.strip() for seg_q in message.queries]

    def get_order_message(self, worklist, context):
        message = self.create(
            sender_name="MCA",
            date=context["now"],
        )

        if context["query"]:
            message.special2 = "REPLY"

        for patient in worklist.get_astm_worklist(
            unique_together=["action"],
            action={"del": "C", worklist.ELSE: "A"},
        ):
            age = num_years(patient["birthdate"], patient["collection_date"])
            unit_age = "Y"
            if age in (None, 0):
                age = num_month(patient["birthdate"],
                                patient["collection_date"])
                unit_age = "M"
            if age in (None, 0):
                age = num_days(patient["birthdate"],
                               patient["collection_date"])
                unit_age = "D"
            if age in (None, 0):
                age = unit_age = None
            seg_p = message.patients.add(
                no_patient=patient["lab_patient_id"],
                nom=patient["last_name"],
                prenom=patient["first_name"],
                date_naissance=patient["birthdate"],
                sexe=patient["sex"],
                age=age,
                age_unit=unit_age,
            )
            for sample, action in patient["samples"]:
                seg_o = seg_p.orders.add(
                    tube=sample["sample_id"],
                    collection_date=sample["collection_date"],
                    code_action=action,
                    specimen_descriptor=sample["specimen"],
                )

                if context["query"]:  # renvoie du rack et position
                    for seg_q in context["query"].queries:
                        if seg_q.sample_id.strip() == sample["sample_id"]:
                            seg_o.sample_no = seg_q.sample_no
                            seg_o.rack_id = seg_q.rack_id
                            seg_o.position = seg_q.position
                            seg_o.rack_type = seg_q.rack_type
                            seg_o.container_type = seg_q.container_type
                elif (sample.get("rack", "") or "").count("-") == 3:
                    seg_o.sample_no = '0'
                    (
                        seg_o.rack_id, seg_o.position,
                        seg_o.rack_type, seg_o.container_type
                    ) = sample["rack"].split("-")

                for order in sample["orders"]:
                    test = seg_o.tests.add()
                    test.code = order["code"]
                    test.dilution = order["dilution"]
                seg_o.commentaires.add(
                    source='I',
                    comment1="%s %s %s" % (
                        patient["last_name"],
                        patient["first_name"],
                        patient["maiden_name"],
                    ),
                    comment2=sample["physician"],
                    comment3=sample["service"],
                )
        if message.patients:
            return message
        elif context["query"]:
            message.end.termination_code = "I"
            return message

    def get_result_dict(self, message):
        results = {}
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                if seg_o.code_action == "Q":
                    continue
                rack = "%s-%s-%s-%s" % (
                    seg_o.rack_id,
                    seg_o.position,
                    seg_o.rack_type,
                    seg_o.container_type,
                )
                result = results.setdefault(
                    seg_o.tube.strip(), {
                        "resultats": [],
                        "rack": rack,
                    })["resultats"]

                for seg_r in seg_o.results:
                    lots_reactifs = []
                    lots_calibrations = []
                    for seg_m in seg_r.infos:
                        # Elecsys Calibration Result Record
                        # A verifier si cela correspond a des lot de reactif
                        if seg_m.type_sub == "ECR":
                            if seg_m.lot_rack and not (
                                    seg_m.lot_rack in lots_reactifs):
                                lots_reactifs.append(seg_m.lot_rack)
                            if seg_m.calibrator_lot and (
                                seg_m.calibrator_lot not in lots_calibrations
                            ):
                                lots_calibrations.append(seg_m.calibrator_lot)

                    alarmes = seg_r.flags[:]
                    for seg_c in seg_r.commentaires:
                        if seg_c.comment1 and seg_c.comment1 != '0':
                            alarmes.append(seg_c.comment1)

                    for test, resu in self.get_resultats(seg_r):
                        result.append({
                            "test": test,
                            "resu": resu,
                            "date": seg_r.date,
                            "lots_reactifs": lots_reactifs,
                            "lots_calibrations": lots_calibrations,
                            "analyseur": seg_r.numero_analyseur,
                            "alarmes": alarmes,
                        })
        return results

    @staticmethod
    def get_resultats(seg_r):
        test = seg_r.test.split("/", 1)[0]
        yield (test, seg_r.valeur_qt if seg_r.valeur_qt.strip() else "ERREUR")
        if seg_r.valeur_ql:
            yield (test + "Q", seg_r.valeur_ql)

    def get_qc_result_list(self, message):
        result = []
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                if seg_o.code_action != "Q":
                    continue
                for seg_r in seg_o.results:
                    lot = seg_o.tube.strip()
                    for test, resu in self.get_resultats(seg_r):
                        result.append({
                            "lot": lot,
                            "test": test,
                            "resu": resu,
                            "date": seg_r.date,
                            "analyseur": seg_r.numero_analyseur,
                        })
        return result
