# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng.astm import (
    AstmMessage,
    HeaderSegment,
    Segment,
    HprimString,
    HprimMultiple,
    HprimList,
    HprimDate,
    HprimForeign,
)


class CS2x00iSegmentH(HeaderSegment):

    sender_id = HprimMultiple(
        5,
        # Fixed as "CS-2000i” or”CS-2100i"
        analyzer_name=HprimString(1),
        software_version=HprimString(2),
        analyzer_serial=HprimString(3),
        interface_version=HprimString(4),
        logical_id=HprimString(5),
        nickname=HprimString(6),
        # “06317410” is output in CS-2000i. “06372511” is output in CS-2100i
        ps_code=HprimString(7),
    )
    version = HprimString(13, default="E1394-97")
    date = HprimDate(14)

    patients = HprimForeign("P", "header")
    queries = HprimForeign("Q", "header")
    end = HprimForeign(
        "L", "header", one2one=True, required=True)


class CS2x00iSegmentP(Segment):

    patient_id = HprimString(5, max_length=16)
    first_name = HprimString((6, 2), max_length=20)
    last_name = HprimString((6, 3), max_length=20)
    birthdate = HprimDate(8, fmt_date="%Y%m%d")
    # M/F/U
    sex = HprimString(9)
    physician_code = HprimString((14, 2), max_length=20)
    physician_name = HprimString((14, 1), max_length=20)
    location = HprimString((26, 4), max_length=20)
    orders = HprimForeign("O", "patient")


class CS2x00iSegmentQ(Segment):

    range_id = HprimMultiple(
        3,
        rack=HprimString(1),
        position=HprimString(2),
        sample_id=HprimString(3, max_length=15),
        # M: Sample is is manually entered
        # A: Analyser automatically aissigned number (begins with "ERR")
        # B: Barcode reader read number
        sample_id_attribute=HprimString(4),
    )
    tests = HprimList(5, basefield=HprimMultiple(
        0,
        code=HprimString(4),
        name=HprimString(5),
    ))
    # 0: Request date is not defined in the ASTM
    nature_of_request_day_limit = HprimString(6, default="0")
    request_date = HprimDate(7)


class CS2x00iSegmentO(Segment):

    specimen_id = HprimMultiple(
        3,
        rack=HprimString(1),
        position=HprimString(2),
        sample_id=HprimString(3),
        sample_id_attribute=HprimString(4),
    )
    instrument_specimen_id = HprimMultiple(
        4,
        rack=HprimString(1),
        position=HprimString(2),
        sample_id=HprimString(3),
        sample_id_attribute=HprimString(4),
        request_flag=HprimString(5),
    )

    tests = HprimList(5, basefield=HprimMultiple(
        0,
        code=HprimString(4),
        dilution_ratio=HprimString(6),
        option=HprimString(7),
    ))

    # S: Stat
    # R: Routine
    priority = HprimString(6)

    request_date = HprimDate(7)

    # N: Normal sample
    # Q: QC material
    action_code = HprimString(12)

    results = HprimForeign("R", "order")


class CS2x00iSegmentR(Segment):

    test = HprimMultiple(
        3,
        code=HprimString(4),
        # HIL flag
        parameter=HprimString(5),
        dilution_ratio=HprimString(6),
        analysis_type=HprimString(7),
        extended_order_request=HprimString(8),
        reflex_test_request=HprimString(9),
    )
    value = HprimString(4)
    units = HprimString(5)
    # A: Result with HIL flag or sample info flag
    # W: Above the analysis range of the HIL detector (only for HIL flag)
    # N: Sample volume check fell within the preset range (Only for sample
    # volume flag)
    abnormal_flags = HprimList(7)
    date = HprimDate(13)


class CS2x00iSegmentL(Segment):

    # N: Normal termination
    code = HprimString(3, default="N")


class Driver(AstmMessage):
    SPEC = {
        "code": "cs2x00i",
        "label": "Sysmex CS-2000i/CS-2100i ASTM",
    }
    SEGMENT_CLASS_MAP = {
        "H": CS2x00iSegmentH,
        "P": CS2x00iSegmentP,
        "O": CS2x00iSegmentO,
        "Q": CS2x00iSegmentQ,
        "R": CS2x00iSegmentR,
        "L": CS2x00iSegmentL,
    }
    DEFAULT_SEPARATORS = "|\\^&"

    def create(self, **kwargs):
        message = CS2x00iSegmentH(type_segment="H", **kwargs)
        message.end = CS2x00iSegmentL(type_segment="L")
        return message

    def get_message_type(self, message):
        if message.queries:
            return self.QUERY
        else:
            return self.ORDER

    def get_queries(self, message):
        queries = []
        for seg_q in message.queries:
            queries.append(seg_q.range_id.sample_id.strip())
        return queries

    def get_order_message(self, worklist, context):
        msg = self.create(date=context["now"])
        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority"],
            action={"del": worklist.IGNORE, worklist.ELSE: "N"},
            priority={"S": "S", "A": "A", worklist.ELSE: "R"},
        ):
            seg_p = msg.patients.add(
                patient_id=patient["patient_id"],
                last_name=patient["last_name"],
                first_name=patient["first_name"],
                birthdate=patient["birthdate"],
                sex=patient["sex"],
                location=patient["location"],
            )
            for sample, action, priority in patient["samples"]:
                seg_o = seg_p.orders.add(
                    priority=priority,
                    action_code=action,
                )
                seg_o.specimen_id.sample_id = "%15.15s" % (
                    sample["sample_id"],)
                if context["query"] is not None:
                    for seg_q in context["query"].queries:
                        if seg_q.range_id.sample_id.strip() == sample["sample_id"]:
                            for attr in (
                                "rack", "position", "sample_id_attribute",
                            ):
                                setattr(seg_o.specimen_id, attr, getattr(
                                    seg_q.range_id, attr))
                for order in sample["orders"]:
                    t = seg_o.tests.add()
                    t.code = order["code"]
                    t.dilution_ratio = order["dilution"]

        if msg.patients:
            return msg
        elif context["query"]:
            for seg_q in context["query"].queries:
                seg_p = msg.patients.add(sex="U")
                seg_o = seg_p.orders.add()
                for attr in (
                    "rack", "position", "sample_id", "sample_id_attribute",
                ):
                    setattr(seg_o.specimen_id, attr,
                            getattr(seg_q.range_id, attr))
            return msg

    def get_result_dict(self, message):
        results = {}
        hil_map = {
            "Hemolytic Sample": "H",
            "Icterous Sample": "I",
            "Lipemic Sample": "L",
            "Defective Sample Volume": "D",
        }
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                if seg_o.action_code == "Q":
                    continue
                if seg_o.instrument_specimen_id.rack and seg_o.instrument_specimen_id.position:
                    rack = "%s-%s" % (
                        seg_o.instrument_specimen_id.rack,
                        seg_o.instrument_specimen_id.position,
                    )
                else:
                    rack = None
                tube = seg_o.instrument_specimen_id.sample_id.strip()
                results.setdefault(tube, {
                    "rack": rack,
                    "resultats": [],
                })
                for seg_r in seg_o.results:
                    if not seg_r.test.code and seg_r.test.parameter in hil_map:
                        test = hil_map[seg_r.test.parameter]
                        resu = "YES"
                    else:
                        test = seg_r.test.code
                        resu = seg_r.value.strip()

                    results[tube]["resultats"].append({
                        "test": test,
                        "resu": resu,
                        "alarmes": seg_r.abnormal_flags,
                        "date": seg_r.date,
                    })
        return results

    def get_qc_result_list(self, message):
        results = []
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                if seg_o.action_code != "Q":
                    continue
                for seg_r in seg_o.results:
                    results.append({
                        "lot": seg_o.instrument_specimen_id.sample_id.strip(),
                        "test": seg_r.test.code,
                        "resu": seg_r.value,
                        "date": seg_r.date,
                    })
        return results
