# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng import Worklist
from driverng.astm import (
    AstmMessageV2,
    HeaderSegment,
    Segment,
    HprimString,
    HprimList,
    HprimDate,
    HprimForeign,
    HprimMultiple,
)


class CtlinkSegmentH(HeaderSegment):

    emetteur = HprimString(5)
    recepteur = HprimString(10)
    type_message = HprimString(12, default="P")
    version = HprimString(13, default="1")
    date = HprimDate(14)

    patients = HprimForeign("P", "header")
    infos = HprimForeign("M", "header")
    end = HprimForeign("L", "header", one2one=True, required=True)


class CtlinkSegmentP(Segment):

    no_patient = HprimString(3, max_length=20)
    no_patient_executant = HprimString(4)

    # nom patient
    nom = HprimString((6, 1), max_length=30)

    # prenom patient
    prenom = HprimString((6, 2), max_length=30)
    nomjf = HprimString(7, max_length=30)
    date_naissance = HprimDate(8)
    # M/F/U
    sexe = HprimString(9)

    # Special field
    species_code = HprimString((15, 1))
    species_name = HprimString((15, 2))

    # Location - Service
    service = HprimString(26)

    orders = HprimForeign("O", "patient")


class CtlinkSegmentO(Segment):

    # tube OU nom du controle
    tube = HprimString((3, 1), max_length=16)

    # Non présent dans la doc ??
    rack = HprimString((4, 2))
    position = HprimString((4, 3))

    # Liste des codes tests
    # ^^^Test^Request.DilutionCondition^Request.DilutionCoefficient
    tests = HprimList(5, basefield=HprimMultiple(
        0,
        code=HprimString(4),
        dilution_condition=HprimString(5),
        coeff_dilution=HprimString(6),
    ))

    # Priorité
    # S: stat
    # A: ASAP
    # R: Routine
    # Anything else if interpreted as Routine
    urgence = HprimString(6, max_length=1)

    # Requested/Ordered Date and time
    date_requested = HprimDate(7)

    # Date et heure de prelevement
    date_prelevement = HprimDate(8)

    # type de contenant de l'echantillon
    type_contenant = HprimString((10, 3))

    # Code action
    # A ou N ou vide: Ajout de tests
    # C: suppréssion de tests
    # Q : QC resultat
    code_action = HprimString(12, max_length=1)

    # Date et heure de l'envoi du resultat de l'echantillon et du QC - valeur
    # toujours vide
    order_date = HprimDate(15)

    # Sample type identifier
    # Code fluide  obligatoire avec la demande, present avec le resultat et
    # vide pour le QC
    # Urine/Serum/Plasma
    code_fluide = HprimString(16)

    # Prescripteur
    prescripteur = HprimString(17, max_length=25)

    # Type de resultat sur l'echantillon
    #  F: Final
    #  P: Preliminaire
    report_type = HprimString(26)

    results = HprimForeign("R", "order")


class CtlinkSegmentR(Segment):

    # code test
    # ^^^Test^Result.DilutionCondition
    # ^DilutioniCoefficient^^Aspect.Name^Test.Datatype
    test = HprimString((3, 4), max_length=15)
    coeff_dilution = HprimString((3, 6))
    aspect = HprimString((3, 8))

    # Test.Datatype
    # N: Numerique
    # S: String
    # E: Enumerated
    # I: Image
    type_data = HprimString((3, 9))

    # le resultat
    valeur = HprimString(4)

    # statut du resultat
    # F: final
    # P: Preliminaire
    status = HprimString(7)

    # identification de l'operateur
    # information presente uniquement lors du retour du resultat
    # TODO: Pas bien défini
    operateur = HprimMultiple(
        11,
        nom=HprimString(1),
        code=HprimString(2),
    )

    # date du resultat ou du QC
    date = HprimDate(12)

    # nom de l'analyseur
    nom_analyseur = HprimString(14)


class CtlinkSegmentM(Segment):
    # voir trame production
    # Code I
    code = HprimString(3)

    tube = HprimString((4, 2))

    # Date plateau
    date_plateau = HprimDate((4, 3))


class CtlinkSegmentC(Segment):
    # C|1|I|XXXXXX|I

    # TODO ???
    code_commentaire = HprimString(3)

    # TODO ??? Séparé par des & visiblement ...
    # Code Flag erreur ou autre commentaire
    commentaire = HprimString(4)

    # type de commentaire
    # I ou G
    type_commentaire = HprimString(5)


class CtlinkSegmentL(Segment):
    # Code de terminaison
    # N: Terminaison normale
    code = HprimString(3)


class Driver(AstmMessageV2):
    SPEC = {
        "code": "ctlink",
        "label": "Ctlink",
    }
    SEGMENT_CLASS_MAP = {
        "H": CtlinkSegmentH,
        "P": CtlinkSegmentP,
        "O": CtlinkSegmentO,
        "R": CtlinkSegmentR,
        "M": CtlinkSegmentM,
        "C": CtlinkSegmentC,
        "L": CtlinkSegmentL,
    }
    DEFAULT_SEPARATORS = "|\^&"

    def __init__(self, analyseurs_hemato=None, **kwargs):
        if analyseurs_hemato is not None:
            self.analyseurs_hemato = [
                e for e in analyseurs_hemato.split(";") if e
            ]
        else:
            self.analyseurs_hemato = []
        super(Driver, self).__init__(**kwargs)

    def create(self, **kwargs):
        message = CtlinkSegmentH(type_segment="H", **kwargs)
        message.end = CtlinkSegmentL(type_segment="L")
        return message

    def encode_worklist(self, worklist, context):
        if worklist.context != "order":
            raise NotImplementedError("context %s" % (worklist.context,))
        message = self.create(date=context["now"])
        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority"],
            action={"del": "C", worklist.ELSE: "A"},
            priority={"A": "A", "S": "S", worklist.ELSE: "R"},
        ):
            seg_p = message.patients.add(
                no_patient=patient.patient_id,
                nom=patient.last_name,
                prenom=patient.first_name,
                nomjf=patient.maiden_name,
                date_naissance=patient.birthdate,
                sexe=patient.sex,
                service=patient.location,
            )
            for sample, action, priority in patient.samples:
                seg_o = seg_p.orders.add(
                    tube=sample.sample_id,
                    urgence=priority,
                    date_prelevement=sample.collection_date,
                    code_action=action,
                    code_fluide=sample.specimen,
                )
                for order in sample.orders:
                    t = seg_o.tests.add()
                    t.code = order.code

        if not len(message.patients):
            return None
        else:
            return message

    def decode_message(self, message):
        if message.infos:
            if len(message.infos) > 1:
                raise NotImplementedError("Multiple segment M")
            return Worklist(context="result", samples=[{
                "type": "patient",
                "sample_id": message.infos[0].tube,
            }])

        # sample_id => alarmes
        alarmes_hemato = {}

        worklist = Worklist(context="result")
        for seg_p in message.patients:
            for seg_o in seg_p.orders:

                if seg_o.code_action == "Q":
                    defaults = {"type": "control", "control_lot": seg_o.tube}
                else:
                    defaults = {"type": "patient"}

                sample = worklist.get_or_create(
                    sample_id=seg_o.tube,
                    specimen=seg_o.code_fluide,
                    **defaults)

                if seg_o.rack and seg_o.position:
                    sample.rack = "%s-%s" % (seg_o.rack, seg_o.position)

                for seg_r in seg_o.results:
                    if (
                        sample.type == "patient"
                        and seg_r.nom_analyseur in self.analyseurs_hemato
                    ):
                        alarmes_hemato.setdefault(seg_o.tube, set())

                    comments = []
                    flags = []
                    for seg_c in seg_r.commentaires:
                        if seg_c.commentaire:
                            values = seg_c.commentaire.split("&")
                            if seg_c.type_commentaire == "I":
                                if (
                                    sample.type == "patient"
                                    and seg_r.nom_analyseur in self.analyseurs_hemato
                                ):
                                    for c in [e for e in values if e != "*"]:
                                        alarmes_hemato[seg_o.tube].add(c)
                                flags.extend(values)
                            elif seg_c.type_commentaire == "G":
                                if (
                                    # (c) 2015, Damien Bastit
                                    sample.rack is None
                                    and len(values) == 5
                                    and values[0] == "Portoir_: "
                                    and values[1] == values[3] == "XA"
                                ):
                                    sample.rack = "%s-%s" % (values[2], values[4])
                                else:
                                    comments.extend(values)

                    if seg_r.aspect:
                        code = seg_r.test + seg_r.aspect
                    else:
                        code = seg_r.test

                    sample.results.add(
                        code=code,
                        value=seg_r.valeur,
                        date=seg_r.date,
                        instrument=seg_r.nom_analyseur,
                        flags=flags,
                        comments=comments,
                    )

        for sample_id, flags in alarmes_hemato.items():
            sample = worklist.get_or_create(type="patient", sample_id=sample_id)
            sample.results.add(
                code="ALARME",
                value=";".join(sorted(flags)),
            )

        return worklist
