# -*- coding: utf8 -*-
from __future__ import unicode_literals

from unidecode import unidecode

from driverng import JSONGraphContainer
from driverng.astm import (
    AstmMessage,
    HeaderSegment,
    Segment,
    HprimString,
    HprimMultiple,
    HprimList,
    HprimDate,
    HprimForeign,
)


class DxhSegmentH(HeaderSegment):

    receiver_id = HprimString((10, 1))
    workcell_id = HprimString((10, 2))
    # P: Production
    # Q: QC
    processing_id = HprimString(12, default="P")
    version = HprimString(13, default="LIS2-A"),
    date = HprimDate(14)
    patients = HprimForeign("P", "header")
    queries = HprimForeign("Q", "header")
    end = HprimForeign(
        "L", "header", one2one=True, required=True
    )


class DxhPhysicianField(HprimMultiple):
    def __init__(self, *position, **kwargs):
        kwargs.update({
            "id": HprimString(1, max_length=25, trunc=True),
            "last_name": HprimString(2),
            "first_name": HprimString(3),
            "middle_name": HprimString(4),
            "suffix": HprimString(5),
        })
        super(DxhPhysicianField, self).__init__(*position, **kwargs)


class DxhSegmentP(Segment):

    lab_patient_id = HprimString(4)
    patient_id_flag = HprimString(5)
    last_name = HprimString((6, 1), max_length=20, trunc=True)
    first_name = HprimString((6, 2), max_length=20, trunc=True)
    middle_name = HprimString((6, 3))
    birthdate = HprimDate((8, 1), fmt_date="%Y%m%d")
    age = HprimString((8, 2))
    # H: Hours
    # D: Days
    # W: Weeks
    # M: Months
    # Y: Years
    age_units = HprimString((8, 3))
    # M/F/U
    sex = HprimString(9)
    physician = DxhPhysicianField(14)
    user_field1 = HprimString((15, 1))
    user_field2 = HprimString((15, 2))
    user_field3 = HprimString((15, 3))
    diagnosis = HprimString(19)
    location = HprimString((26, 1))
    patient_location_field = HprimString((26, 2))
    orders = HprimForeign("O", "patient")


class DxhSegmentO(Segment):

    # QC: Lot number
    specimen_id = HprimString((3, 1), max_length=22)

    # QC: Control level
    specimen_edit_flag = HprimString((3, 2))

    # QC: Control type
    requisition_number = HprimString((3, 3))

    # Only for QC
    expiration_date = HprimDate((3, 4), fmt="%Y%m%d")

    tube_position_id = HprimString((4, 1))
    tube_position_edit_flag = HprimString((4, 2))
    tests = HprimList(5, basefield=HprimMultiple(
        0,
        code=HprimString(4),
        number_slides=HprimString(5),
    ))

    # S: Stat
    # R: Routine
    # P: Preoperative
    # C: Callback
    # A: ASAP
    priority = HprimString(6)

    request_date = HprimDate(7)
    collection_date = HprimDate(8)

    # C: Cancel
    # A: Add
    # N: New
    # Q: QC
    action_code = HprimString(12)
    specimen_rcv_date = HprimDate(13)

    # Whole blood
    # Synovial
    # Pleural
    # CSF
    # Peritoneal
    # Pericardial
    specimen_type = HprimString((16, 1), required=True)
    physician = DxhPhysicianField(17)
    location = HprimString((28, 1))
    draw_location = HprimString((28, 2))
    graphes = HprimForeign("M", "order")
    results = HprimForeign("R", "order")


class DxhSegmentR(Segment):
    test_code = HprimString((3, 4))
    test_loinc_code = HprimString((3, 5))

    # +++++ Indicate the parameter if over the operating range
    # ..... Indicate an incomplete result
    # ::::: Indicate flow cell clogs
    # ===== Indicate an analysis disabled
    # xxxxx Indicate a parameter was disabled when the specimen was run
    #       or when results where released, but the parameter was
    #       subsequently enabled
    # ----- Indicate a voteout
    # ????? Indicate a result is outside of the range of values that can
    #       be formatted for display
    value = HprimString((4, 1))

    # FLAG 1
    # E: Parameter result edited flag
    # e: Parameter result has been recalculated based on an edited result
    # M: Parameter result amended flag
    # m: Parameter result has been recalculated based on an amended result
    # +: Result is higher than measuring range high limit
    # -: Result is lower than measuring range low limit
    # R: Review flag

    # FLAG 2 and 3
    # c and H: Result is higher than operator defined range for critical limits
    # c and L: Result is lower than operator defined range for critical limits
    # a and H: Result is higher than operator defined range for action limits
    # a and L: Result is lower than operator defined range for action limits
    # space and H: Result is higher than operator defined reference range
    # space and L: Result is lower than operator defined reference range

    # FLAG 4
    # P: Partial aspiration flag
    # N: Non blood specimen detected
    # D: Delta flag
    flags = HprimString((4, 2))
    units = HprimString(5)
    dilution = HprimString(6)
    reference_ranges = HprimString(7)

    # F: Final
    # C: Correction (amended results)
    # S: Partial
    # R: Préviously transmitted
    status = HprimString(10)
    operator = HprimString(12)
    date = HprimDate(14)
    instrument_id = HprimString(15)


class DxhSegmentM(Segment):
    test_code = HprimString((3, 4))
    test_loinc_code = HprimString((3, 5))
    value = HprimString((4, 1))
    status = HprimString(9)
    date = HprimDate(13)
    instrument_name = HprimString(14)


class DxhSegmentC(Segment):
    # I: Clinical instrument system
    # L: Computer system
    source = HprimString(3)

    text = HprimString(4)

    # G: Generic
    # I: Instrument
    # T: Test name comment
    # L: Lab action
    comment_type = HprimString((5, 1))

    # Y: System message
    # U: Suspect message
    flag_type = HprimString((5, 2))


class DxhSegmentQ(Segment):
    specimen_id = HprimString((3, 2))
    test = HprimString(5, default="ALL")
    status = HprimString(13, default="O")


class DxhSegmentL(Segment):

    # N: Normal termination
    code = HprimString(3, default="N")


class Driver(AstmMessage):
    SPEC = {
        "code": "dxh",
        "label": "DxH 800",
    }
    SEGMENT_CLASS_MAP = {
        "H": DxhSegmentH,
        "P": DxhSegmentP,
        "O": DxhSegmentO,
        "Q": DxhSegmentQ,
        "C": DxhSegmentC,
        "R": DxhSegmentR,
        "M": DxhSegmentM,
        "L": DxhSegmentL,
    }
    DEFAULT_SEPARATORS = "|\\!~"

    def create(self, **kwargs):
        message = DxhSegmentH(type_segment="H", **kwargs)
        message.end = DxhSegmentL(type_segment="L")
        return message

    def get_message_type(self, message):
        if message.queries:
            return self.QUERY
        else:
            return self.ORDER

    def get_queries(self, message):
        queries = []
        for seg_q in message.queries:
            queries.append(seg_q.specimen_id)
        return queries

    def get_order_message(self, worklist, context):
        msg = self.create(date=context["now"])
        for patient in worklist.get_astm_worklist(
                unique_together=["action", "priority"],
                action={"del": "C", "add": "A", "new": "N"},
                priority={
                    "S": "S",
                    "A": "A",
                    worklist.ELSE: "R",
                },
        ):
            seg_p = msg.patients.add(
                lab_patient_id=patient["patient_id"],
                last_name=unidecode(patient["last_name"]),
                first_name=unidecode(patient["first_name"]),
                birthdate=patient["birthdate"],
                sex=patient["sex"],
            )
            for sample, action, priority in patient["samples"]:
                seg_o = seg_p.orders.add(
                    specimen_id=sample["sample_id"],
                    priority=priority,
                    action_code=action,
                    request_date=sample["request_date"],
                    report_type="O",
                    collection_date=sample["collection_date"],
                    specimen_type=sample["specimen"],
                )
                seg_o.physician.id = sample["physician"]
                for order in sample["orders"]:
                    test = seg_o.tests.add()
                    if "!" in order["code"]:
                        test.code, test.number_slides = order["code"].split("!", 1)
                    else:
                        test.code = order["code"]

        if msg.patients:
            return msg

    def get_result_dict(self, message):
        results = {}
        specimen_alarmes = {}
        specimen_commentaires = {}
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                # exclusion des QC
                if seg_o.action_code == 'Q':
                    continue
                for seg_c in seg_o.commentaires:
                    if seg_c.comment_type == "I":
                        specimen_alarmes.setdefault(
                            seg_o.specimen_id,
                            []).append(seg_c.text)
                    elif seg_c.comment_type == "G":
                        specimen_commentaires.setdefault(
                            seg_o.specimen_id,
                            []).append(seg_c.text)
                results.setdefault(
                    seg_o.specimen_id, {
                        "resultats": [],
                        "rack": seg_o.tube_position_id,
                    },
                )
                for seg_m in seg_o.graphes:
                    if seg_m.test_code in (
                        "5PD1", "5PD2", "NRBC1", "NRBC2",
                        "RETIC1", "RETIC2",
                    ):
                        graph = JSONGraphContainer({
                            "graph_type": "dxh_scatter",
                            "blob": seg_m.value,
                        })
                    elif seg_m.test_code in (
                        "WBC.Histogram.Array",
                        "TNC.Histogram.Array",
                        "RBC.Histogram.Array",
                        "Plt.Histogram.Array",
                    ):
                        points = []
                        p = seg_m.value
                        while p:
                            points.append(int(p[0:2], 16))
                            p = p[2:]
                        graph = JSONGraphContainer({
                            "graph_type": "area",
                            "points": points,
                        })
                    else:
                        raise RuntimeError(
                            "Unkwnon graph type '%s'" % seg_m.test_code)
                    results[seg_o.specimen_id]["resultats"].append({
                        "test": seg_m.test_code,
                        "date": seg_m.date,
                        "resu": "GRAPHE",
                        "graph": graph,
                        "type_prlv": seg_o.specimen_type,
                    })
                for seg_r in seg_o.results:
                    commentaires = []
                    for seg_c in seg_r.commentaires:
                        if seg_c.comment_type == "I":
                            commentaires.append(seg_c.text)
                    if seg_r.status not in ("F", "S", "R"):
                        continue
                    alarmes = []
                    for flag in (
                        seg_r.flags[0:1].strip(),
                        seg_r.flags[1:3].strip(),
                        seg_r.flags[3:4].strip(),
                    ):
                        if flag:
                            alarmes.append(flag)
                    results[seg_o.specimen_id]["resultats"].append({
                        "test": seg_r.test_code,
                        "date": seg_r.date,
                        "alarmes": alarmes,
                        "resu": seg_r.value,
                        "type_prlv": seg_o.specimen_type,
                        "analyseur": seg_r.instrument_id,
                        "commentaires": commentaires,
                    })
                    # Hors LAME
                    if seg_r.test_code == 'SM':
                        specimen_commentaires[seg_o.specimen_id] = []
        for sample, values in specimen_alarmes.items():
            results[sample]["resultats"].append({
                "test": "ALARME",
                "resu": "",
                "commentaires": list(set(values)),
                "date": message.date,
            })
        for sample, values in specimen_commentaires.items():
            for i, value in enumerate(list(set(values)), 1):
                results[sample]["resultats"].append({
                    "test": "COM%d" % (i,),
                    "resu": value,
                    "date": message.date,
                })
        return results

    def get_qc_result_list(self, message):
        result = []
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                # uniquement les QC
                if seg_o.action_code != "Q":
                    continue
                for seg_r in seg_o.results:
                    if seg_r.status not in ("F", "S"):
                        continue
                    result.append({
                        "lot": seg_o.specimen_id,
                        "test": seg_r.test_code,
                        "resu": seg_r.value,
                        "date": seg_r.date,
                        "type_prlv": seg_o.specimen_type,
                    })
        return result
