# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng.astm import (
    AstmMessage,
    HeaderSegment,
    Segment,
    HprimString,
    HprimList,
    HprimDate,
    HprimForeign,
    HprimMultiple,
)


class ElecsysSegmentH(HeaderSegment):
    password = HprimString(4)
    sender_id = HprimMultiple(
        5,
        name=HprimString(1),
        manufacturer=HprimString(2),
        instrument_type=HprimString(3),
        software_version=HprimString(4),
    )
    receiver_id = HprimString(10)
    processing_id = HprimString(12, default="P")
    version = HprimString(13, default="1")
    date = HprimDate(14)

    patients = HprimForeign("P", "header")
    queries = HprimForeign("Q", "header")
    end = HprimForeign("L", "header", one2one=True, required=True)


class ElecsysSegmentP(Segment):
    patient_id = HprimString(3)
    laboratory_patient_id = HprimString(4)
    last_name = HprimString((6, 1))
    first_name = HprimString((6, 2))
    mother_maiden_name = HprimString(7)
    date_birth = HprimDate(8, fmt_date="%Y%m%d")
    # M/F/U
    sex = HprimString(9)
    location = HprimString(26)

    orders = HprimForeign("O", "patient")


class ElecsysSegmentO(Segment):
    specimen_id = HprimString(3)
    instrument_specimen_id = HprimMultiple(
        4,
        sequence=HprimString(1),
        rack=HprimString(2),
        position=HprimString(3),
        format_carrier=HprimString(4),
        data_carrier=HprimString(5),
        container_type=HprimString(6),
    )
    universal_test_id = HprimList(5, basefield=HprimMultiple(
        0,
        test_id=HprimString(1),
        test_name=HprimString(2),
        test_type=HprimString(3),
        code=HprimString(4),
        dilution=HprimString(5),
    ))
    priority = HprimString(6)
    request_date = HprimDate(7)
    collection_date = HprimDate(8)
    action_code = HprimList(12)
    info_clinical = HprimString(14)
    specimen_descriptor = HprimString(16)
    report_type = HprimString(26)

    results = HprimForeign("R", "order")


class ElecsysSegmentR(Segment):
    universal_test_id = HprimMultiple(
        3,
        test=HprimString(4),
        name=HprimString(5),
        test_type=HprimString(6),
        test_code=HprimString(7),
    )
    data = HprimString(4)
    units = HprimString(5)
    reference_ranges = HprimString((6, 1))
    upper_limit = HprimString((6, 2))
    abnormal_flag = HprimString(7)
    nature_abnormality = HprimString(8)
    result_status = HprimString(9)
    date_started = HprimDate(12)
    date_completed = HprimDate(13)
    instrument_identification = HprimString(14)


class ElecsysSegmentC(Segment):
    # I: Clinical instrument system
    comment_source = HprimString(3)
    comment_text = HprimString((4, 2))
    # I: Instrument flag
    # G: Manuel (For version > 1.35 'G' will be replaced by 'I')
    comment_type = HprimString(5)


class ElecsysSegmentQ(Segment):
    starting_range_id = HprimMultiple(
        3,
        patient_id=HprimString(1),
        specimen_id=HprimString(2),
        sequence=HprimString(3),
        rack=HprimString(4),
        position=HprimString(5),
        format_carrier=HprimString(6),
        data_carrier=HprimString(7),
        container_type=HprimString(8),
    )
    request_status = HprimString(13)


class ElecsysSegmentL(Segment):
    termination_code = HprimString(3, default="F")


class Driver(AstmMessage):
    SPEC = {
        "code": "elecsys",
        "label": "Elecsys",
    }
    SEGMENT_CLASS_MAP = {
        "H": ElecsysSegmentH,
        "P": ElecsysSegmentP,
        "O": ElecsysSegmentO,
        "Q": ElecsysSegmentQ,
        "R": ElecsysSegmentR,
        "C": ElecsysSegmentC,
        "L": ElecsysSegmentL,
    }
    DEFAULT_SEPARATORS = "|\\^&"

    def create(self, **kwargs):
        message = ElecsysSegmentH(type_segment="H", **kwargs)
        message.end = ElecsysSegmentL(type_segment="L")
        return message

    def get_message_type(self, message):
        if message.queries:
            return self.QUERY
        else:
            return self.ORDER

    def get_queries(self, message):
        queries = []
        for seg_q in message.queries:
            queries.append(seg_q.starting_range_id.specimen_id)
        return queries

    def get_order_message(self, worklist, context):
        msg = self.create(date=context["now"])
        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority"],
            action={"del": "C", worklist.ELSE: "N"},
            priority={"A": "S", "S": "S", worklist.ELSE: "R"},
        ):
            seg_p = msg.patients.add(
                patient_id=patient["patient_id"],
                last_name=patient["last_name"],
                first_name=patient["first_name"],
                mother_maiden_name=patient["maiden_name"],
                birthdate=patient["birthdate"],
                sex=patient["sex"],
            )
            for sample, action, priority in patient["samples"]:
                seg_o = seg_p.orders.add(
                    priority=priority,
                    request_date=sample["request_date"],
                    collection_date=sample["collection_date"],
                    action_code=[action],
                    report_type="Q" if context["query"] else "O",
                )
                seg_o.specimen_id = sample["sample_id"]
                # Renvoyer la sequence
                if context["query"]:
                    for seg_q in context["query"].queries:
                        seg_o.instrument_specimen_id.sequence = seg_q.starting_range_id.sequence
                for order in sample["orders"]:
                    t = seg_o.universal_test_id.add()
                    t.code = order["code"]

        if msg.patients or context["query"]:
            return msg

    def get_result_dict(self, message):
        results = {}
        for seg_p in message.patients:
            for seg_o in seg_p.orders:

                # exclusion des CQ
                if "Q" in seg_o.action_code:
                    continue
                if seg_o.instrument_specimen_id.rack and seg_o.instrument_specimen_id.position:
                    rack = "%s-%s" % (
                        seg_o.instrument_specimen_id.rack,
                        seg_o.instrument_specimen_id.position,
                    )
                else:
                    rack = None
                tube = seg_o.specimen_id
                results.setdefault(tube, {
                    "resultats": [],
                    "rack": rack,
                })

                for seg_r in seg_o.results:
                    if seg_r.abnormal_flag:
                        alarmes = [seg_r.abnormal_flag]
                    else:
                        alarmes = []

                    commentaires = []
                    for seg_c in seg_r.commentaires:
                        commentaires.append(seg_c.comment_text)

                    results[tube]["resultats"].append({
                        "test": seg_r.universal_test_id.test,
                        "resu": seg_r.data,
                        "date": seg_r.date_completed,
                        "alarmes": alarmes,
                        "commentaires": commentaires,
                    })
        return results

    def get_qc_result_list(self, message):
        results = []
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                # Exclusion des resultats non CQ
                if "Q" not in seg_o.action_code:
                    continue
                lot = seg_o.specimen_id.strip()
                for seg_r in seg_o.results:
                    results.append({
                        "lot": lot,
                        "test": seg_r.universal_test_id.test,
                        "resu": seg_r.data,
                        "date": seg_r.date_completed,
                    })
        return results
