# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng.astm import (
    AstmMessage,
    HeaderSegment,
    Segment,
    HprimList,
    HprimString,
    HprimDate,
    HprimForeign,
    HprimMultiple,
)


class ErytraSegmentH(HeaderSegment):
    emetteur = HprimString(5)
    destinataire = HprimString(10)
    processing = HprimString(12, default="P")
    version = HprimString(13, default="1")
    date = HprimDate(14)
    patients = HprimForeign("P", "header")
    queries = HprimForeign("Q", "header")
    end = HprimForeign("L", "header", one2one=True, required=True)


class ErytraSegmentP(Segment):
    ref_patient = HprimString(3)
    patient_name = HprimString((6, 1))
    prenom = HprimString((6, 2))
    jeune_fille = HprimString(7)
    birthdate = HprimDate(8, fmt_date="%Y%m%d")
    sexe = HprimString(9)
    orders = HprimForeign("O", "patient")


class ErytraSegmentO(Segment):
    sample_id = HprimString(3)
    test_code = HprimList(5, basefield=HprimMultiple(
        0,
        code=HprimString(4),
        analyse=HprimString(5),
    ))
    urgence = HprimString(6, default='R')
    action = HprimString(12, default='A')
    report_type = HprimString(26, default='O')
    results = HprimForeign("R", "order")


class ErytraSegmentQ(Segment):
    specimen_id = HprimString((3, 1))
    tests = HprimString((5, 1), default="ALL")
    request_status = HprimString(13)


class ErytraSegmentR(Segment):
    test = HprimString((3, 4))
    analyse = HprimString((3, 5))
    resu = HprimString(4)
    resu_code = HprimString(5)
    status = HprimList(9)
    operateur = HprimDate(11)
    version1 = HprimDate((14, 1))
    version2 = HprimDate((14, 2))
    version3 = HprimDate((14, 3))


class ErytraSegmentL(Segment):
    terminaison_code = HprimString(3, default="N")


class Driver(AstmMessage):
    SPEC = {
        "code": "erytra",
        "label": "Erytra Grifols",
    }
    SEGMENT_CLASS_MAP = {
        "H": ErytraSegmentH,
        "P": ErytraSegmentP,
        "Q": ErytraSegmentQ,
        "O": ErytraSegmentO,
        "R": ErytraSegmentR,
        "L": ErytraSegmentL,
    }
    DEFAULT_SEPARATORS = "|\\^&"

    def create(self, **kwargs):
        message = ErytraSegmentH(type_segment="H", **kwargs)
        message.end = ErytraSegmentL(type_segment="L")
        return message

    def get_message_type(self, message):
        if message.queries:
            return self.QUERY
        else:
            return self.ORDER

    def get_queries(self, message):
        queries = []
        for seg_q in message.queries:
            queries.append(seg_q.specimen_id.strip())
        return queries

    def get_order_message(self, worklist, context):
        msg = self.create(date=context["now"])
        for patient in worklist.get_astm_worklist(
            unique_together=["action"],
            action={"del": worklist.IGNORE, worklist.ELSE: "A"},
        ):
            seg_p = msg.patients.add(
                ref_patient=patient["patient_id"],
                patient_name=patient["last_name"],
                prenom=patient["first_name"],
                jeune_fille=patient["maiden_name"],
                birthdate=patient["birthdate"],
                sexe=patient["sex"],
            )
            for sample, action in patient["samples"]:
                for order in sample["orders"]:
                    seg_o = seg_p.orders.add(
                        sample_id=sample["sample_id"],
                    )
                    t = seg_o.test_code.add()
                    tab = order["code"].split('^')
                    t.code = tab[0]
                    if len(tab) > 1:
                        t.analyse = tab[1]

        if msg.patients:
            return msg

    def get_result_dict(self, message):
        results = {}
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                tube = seg_o.sample_id.strip()
                resultats = results.setdefault(
                    tube,
                    {"resultats": []},
                )["resultats"]
                for seg_r in seg_o.results:
                    for n, resu in enumerate(seg_r.resu, 1):
                        if resu:
                            resultats.append({
                                "test": "%s_%d" % (seg_r.test, n),
                                "resu": resu,
                            })
        return results
