# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng import Worklist
from driverng.astm import (
    AstmMessageV2,
    HeaderSegment,
    Segment,
    HprimString,
    HprimList,
    HprimDate,
    HprimForeign,
)


class EtimaxSegmentH(HeaderSegment):
    sender_name = HprimString(5)

    patients = HprimForeign("P", "header")
    queries = HprimForeign("Q", "header")
    end = HprimForeign(
        "L", "header", one2one=True, required=True
    )


class EtimaxSegmentP(Segment):
    laboratory_patient_id = HprimString(4)
    patient_name = HprimString((6, 1))
    first_name = HprimString((6, 2))
    birthdate = HprimDate(8, fmt_date="%Y%m%d")
    sex = HprimString(9)

    orders = HprimForeign("O", "patient")


class EtimaxSegmentO(Segment):
    universal_test_id = HprimString((5, 4))
    dilution = HprimString((5, 5))
    collection_date = HprimDate(8)

    results = HprimForeign("R", "order")


class EtimaxSegmentR(Segment):
    test = HprimString((3, 4))
    info_test = HprimString((3, 5))
    value = HprimString(4)


class EtimaxSegmentQ(Segment):
    starting_range_id = HprimString(3)
    universal_test_id = HprimString(5)
    status_code = HprimString(13)


class EtimaxSegmentC(Segment):
    comment_source = HprimString(3)
    # Structure avec des sous champs non documentés
    comment_text = HprimList(4)
    comment_type = HprimString(5)


class EtimaxSegmentL(Segment):
    terminaison_code = HprimString(3, default="N")


class Driver(AstmMessageV2):
    SPEC = {
        "code": "etimax",
        "label": "Etimax"
    }
    SEGMENT_CLASS_MAP = {
        "H": EtimaxSegmentH,
        "P": EtimaxSegmentP,
        "O": EtimaxSegmentO,
        "R": EtimaxSegmentR,
        "Q": EtimaxSegmentQ,
        "C": EtimaxSegmentC,
        "L": EtimaxSegmentL,
    }
    DEFAULT_SEPARATORS = "|\\^&"

    def create(self, **kwargs):
        message = EtimaxSegmentH(type_segment="H", **kwargs)
        message.end = EtimaxSegmentL(type_segment="L")
        return message

    def encode_worklist(self, worklist, context):
        if worklist.context != "order":
            raise NotImplementedError("context %s" % (worklist.context,))
        msg = self.create()
        for patient in worklist.get_astm_worklist(
            unique_together=["action"],
            action={"del": worklist.IGNORE, worklist.ELSE: ""},
        ):
            for sample, _ in patient.samples:
                seg_p = msg.patients.add(
                    laboratory_patient_id=sample.sample_id,
                    patient_name=patient.last_name,
                    first_name=patient.first_name,
                    birthdate=patient.birthdate,
                    sex=patient.sex,
                )
                for order in sample.orders:
                    seg_p.orders.add(
                        universal_test_id=order.code,
                    )
        if len(msg.patients):
            return msg
        elif context["query"] and context["query"].queries:
            for seg_q in context["query"].queries:
                msg.end.terminaison_code = "I"
            return msg

    def decode_message(self, message):
        if message.queries:
            worklist = Worklist(context="query")
            for seg_q in message.queries:
                worklist.get_or_create(
                    type="patient",
                    sample_id=seg_q.starting_range_id
                )
        else:
            worklist = Worklist(context="result")
            for seg_p in message.patients:
                for seg_o in seg_p.orders:
                    sample = worklist.get_or_create(
                        sample_id=seg_p.laboratory_patient_id,
                        type="patient",
                    )
                    for seg_r in seg_o.results:
                        test = "%s^%s" % (seg_r.test, seg_r.info_test,)
                        sample.results.add(
                            code=test,
                            value=seg_r.value,
                        )
        return worklist
