# -*- coding: utf8 -*-
from __future__ import unicode_literals

import datetime

from driverng import Message
from driverng.exceptions import ProgrammingError, ParsingError


class AbstractGTDField(object):

    def __init__(self, tag, required=False, default=None):
        self.tag = tag
        self.required = required
        self.default = default

    def from_value(self, value):
        raise NotImplementedError

    def to_string(self, value):
        raise NotImplementedError

    def validate(self, value):
        raise NotImplementedError


class GDTField(AbstractGTDField):

    def from_value(self, value):
        return value

    def to_string(self, value):
        return value.strip()

    def validate(self, value):
        if self.required and (not value or not value.strip()):
            raise ParsingError("Field %s is required" % (self.tag,))
        return True


class GDTDateField(GDTField):

    def __init__(self, tag, fmt="%d%m%Y", **kwargs):
        super(GDTDateField, self).__init__(tag, **kwargs)
        self.fmt = fmt

    def from_value(self, value):
        try:
            return datetime.datetime.strptime(value, self.fmt)
        except ValueError:
            raise ParsingError("Cannot parse date '%s'" % (value))

    def to_string(self, value):
        return value.strftime(self.fmt)

    def validate(self, value):
        if self.required and not value:
            raise ParsingError("Field %s is required" % (self.tag,))
        if value is not None and not isinstance(value, datetime.datetime):
            raise ParsingError(
                "'%s' is not a datetime.datetime instance" % (value,))
        return True


class GDTMessage(object):
    DATE_FMT = "%d%m%Y"
    FIELDS = ()
    MULTIPLE_FIELDS = ()

    def __init__(self, **kwargs):
        self.field_dict = dict((k, v) for k, v in self.FIELDS)
        self.tag_dict = dict((v.tag, k) for k, v in self.FIELDS)

        for attr, value in kwargs.items():
            if attr in self.field_dict:
                self.field_dict[attr].validate(value)
                setattr(self, attr, value)

        for attr, field in self.FIELDS:
            if not hasattr(self, attr):
                if attr in self.MULTIPLE_FIELDS:
                    setattr(self, attr, [])
                else:
                    setattr(self, attr, field.default)

        super(GDTMessage, self).__init__()

    def validate(self):
        for attr, field in self.FIELDS:
            if attr in self.MULTIPLE_FIELDS:
                values = getattr(self, attr, [])
                if not isinstance(values, list):
                    raise ParsingError(
                        "Field %s must be a list instance" % (field,))
                for value in values:
                    field.validate(value)
            else:
                field.validate(getattr(self, attr))

    @classmethod
    def parse(cls, data):
        obj = cls()

        for line in data.splitlines():

            try:
                length = int(line[:3])
            except ValueError:
                raise ParsingError("Cannot parse line length '%s'" % (line,))

            key = line[3:][:4]

            raw_value = line[7:].strip()

            if len(raw_value) > length - 9:
                pass
                # raise ParsingError("len('%s') > %d char" % (raw_value, length
                # - 9))

            try:
                attr = obj.tag_dict[key]
            except KeyError:
                raise ParsingError("Field tag %s Unknown" % (key,))

            value = obj.field_dict[attr].from_value(raw_value)

            if attr in obj.MULTIPLE_FIELDS:
                getattr(obj, attr).append(value)
            else:
                setattr(obj, attr, value)

        obj.validate()
        return obj

    def get_values(self, attr):

        attr_value = getattr(self, attr)

        if attr in self.MULTIPLE_FIELDS:
            values = attr_value
        else:
            values = [attr_value]

        raw_values = []
        for value in values:
            if not self.field_dict[attr].required and value is None:
                continue
            raw_value = self.field_dict[attr].to_string(value)
            raw_values.append(raw_value)

        return raw_values

    def to_string(self):
        if "set_length" in self.field_dict:
            self.set_length = "XXXXX"
        total_length = 0
        self.validate()
        lines = []

        for attr, field in self.FIELDS:

            for value in self.get_values(attr):
                data = "%s%s" % (field.tag, value)
                # <length (3) ><data><cr><lf>
                total_length = len(data) + 5
                length = "%3.3s" % (len(data) + 5)
                length = length.replace(" ", "0")
                lines.append("%s%s" % (length, data))

        # TODO: Too many hacks
        if "set_length" in self.field_dict:
            self.set_length = "%5.5s" % total_length
            self.set_length = self.set_length.replace(" ", "0")
            for i, line in enumerate(lines):
                if line[3:][:4] == "8100":
                    lines[i] = line.replace("XXXXX", self.set_length)
                    break

        return "\r\n".join(lines)


class EurolabImport(GDTMessage):
    FIELDS = (
        ("set_type", GDTField("8000", required=True, default="6302")),
        ("set_length", GDTField("8100", required=True)),
        ("receiving_software_id", GDTField("8315")),
        ("sending_software_id", GDTField("8316")),
        ("character_set", GDTField("9206", default="3")),
        ("gdt_version", GDTField("9218", required=True, default="02.10")),
        ("lis_company", GDTField("0102")),
        ("lis_name", GDTField("0103")),
        ("lis_version", GDTField("0132")),
        ("customer_short_name", GDTField("0203")),
        ("patient_number", GDTField("3000", required=True)),
        ("patient_surname", GDTField("3101", required=True)),
        ("patient_firstname", GDTField("3102")),
        ("patient_birthdate", GDTDateField("3103")),  # required = True
        ("health_insurance_membership_no", GDTField("3105")),
        # 24556 Testcity
        ("postcode_and_city", GDTField("3106")),
        # Street 42
        ("street_and_number", GDTField("3107")),
        # 1 = Male
        # 2 = Female
        ("sex", GDTField("3110")),
        # 1+2 = statutory health insurance
        # 3 = private insurance (Not GDT standard)
        ("biling_type", GDTField("4121")),
        # Date de prélevement ou de demande.
        ("date", GDTDateField("8432")),
        ("device_type", GDTField("8402")),  # not used
        # Code barre
        ("sample_id", GDTField("8310")),
        ("tests", GDTField("8410")),
    )
    MULTIPLE_FIELDS = ("tests")


class EurolabExportLHeader(GDTMessage):
    FIELDS = (
        ("set_type", GDTField("8000", required=True, default="8220")),
        ("set_length", GDTField("8100", required=True)),
        ("ldt_version", GDTField("9212", required=True,
         default="LDT1001.02")),
        ("nbsnr", GDTField("0201")),
        # German unique identifier of laboratories
        ("nbsnr_name", GDTField("0203", required=True)),
        ("nbsnr_address", GDTField("0205")),
        ("customer_name", GDTField("0211")),
        ("lifetime_doctor_number", GDTField("0212")),
        ("postal_code_of_nbsnr_address", GDTField("0215")),
        ("place_of_nbsnr_address", GDTField("0216")),
        ("laboratory", GDTField("8300")),
        ("laboratory_name", GDTField("8320", required=True)),
        ("laboratory_street", GDTField("8321")),
        ("laboratory_postalcode", GDTField("8322")),
        ("laboratory_place", GDTField("8323")),
        ("kbv_test_number", GDTField("0101")),
        # An unique identifier for the customer
        ("character_set", GDTField("9106", required=True, default="3")),
        ("date", GDTDateField("9103", required=True)),
    )


class EurolabResult(GDTMessage):
    FIELDS = (
        ("id", GDTField("8410", required=True)),
        ("name", GDTField("8411")),
        ("aspect", GDTField("8412")),
        ("status", GDTField("8418")),  # required but not used.
        ("value", GDTField("8420")),
        ("unit", GDTField("8421")),
        ("threshold_indicator", GDTField("8422")),
        ("test_material_identification", GDTField("8428")),
        ("test_material_index", GDTField("8429")),
        ("test_material_name", GDTField("8430")),
        ("test_material_specification", GDTField("8431")),
        ("standard_value_text", GDTField("8460")),
        ("standard_value_lower_limit", GDTField("8461")),
        ("standard_value_upper_limit", GDTField("8462")),
        ("test_notes", GDTField("8470")),
        ("result_text", GDTField("8480")),
        ("order_based_notes", GDTField("8490")),
    )
    MULTIPLE_FIELDS = ("value", "test_notes", "aspect")


class EurolabResults(list):

    def add(self, **kwargs):
        resu = EurolabResult(**kwargs)
        self.append(resu)
        return resu

    def parse(self, data):
        resu = EurolabResult.parse(data)
        self.append(resu)
        return resu


class EurolabExportLGReport(GDTMessage):
    FIELDS = (
        ("set_type", GDTField("8000", required=True, default="8202")),
        ("set_length", GDTField("8100", required=True)),
        ("date_of_entry_of_the_order_in_the_laboratory", GDTDateField("8301")),
        ("raw_report_date", GDTDateField("8302", required=True)),
        ("raw_report_time", GDTField("8303", required=True)),  # %H%M
        ("sample_id", GDTField("8310", required=True)),
        ("order_number", GDTField("8311")),
        ("fee_number", GDTField("5001")),
        ("costs_in_cent", GDTField("5002")),
        ("multiplier", GDTField("5005")),
        ("explanatory_statement", GDTField("5009")),
        ("patient_surname", GDTField("3101", required=True)),
        ("patient_firstname", GDTField("3102")),
        ("patient_birthdate", GDTDateField("3103")),
        ("report_format", GDTField("8401", required=True, default="T")),
        ("billing_type", GDTField("8403")),
        ("patient_information", GDTField("8405")),
        ("other_costs_in_cents", GDTField("8406")),
        # Pas décrit mais on suppose
        # 1 = male
        # 2 = female
        ("gender", GDTField("8407")),
        ("other_billing_type", GDTField("8609")),
        ("billed_by", GDTField("8614")),
        ("customer_id", GDTField("8615")),

    )
    MULTIPLE_FIELDS = ("fee_number")

    def __init__(self, *args, **kwargs):
        self.results = EurolabResults()
        self.report_date = None
        super(EurolabExportLGReport, self).__init__(*args, **kwargs)

    @classmethod
    def parse(cls, data):
        # TODO: Too many hacks
        payload = []
        results = []
        results_codes = [f.tag for _, f in EurolabResult.FIELDS]
        for line in data.splitlines():
            code = line[3:][:4]
            if code in results_codes:
                if code == "8410":  # Nouveau block de resultats
                    results.append([line])
                else:
                    results[-1].append(line)
            else:
                payload.append(line)

        obj = super(EurolabExportLGReport, cls).parse("\r\n".join(payload))
        if obj.raw_report_date and obj.raw_report_time:
            obj.report_date = datetime.datetime.strptime(
                "%s %s" % (obj.raw_report_date.strftime(
                    "%d%m%Y"), obj.raw_report_time),
                "%d%m%Y %H%S")

        for result in results:
            obj.results.parse("\r\n".join(result))

        return obj

    def to_string(self):

        if self.report_date:
            self.raw_report_date = self.report_date
            self.raw_report_time = self.report_date.strftime("%H%M")

        # TODO: Too many hacks
        report_data = super(EurolabExportLGReport, self).to_string()
        resu_data = "\r\n".join([r.to_string() for r in self.results])
        data = "\r\n".join([report_data, resu_data])
        self.set_length = "%5.5s" % len(data)
        self.set_length = self.set_length.replace(" ", "0")
        lines = data.splitlines()
        for i, line in enumerate(lines):
            if line[3:][:4] == "8100":  # set length
                lines[i] = "%s%s" % (line[:7], self.set_length)
                break

        return "\r\n".join(lines)


class EurolabExportLConclusion(GDTMessage):
    FIELDS = (
        ("set_type", GDTField("8000", required=True, default="8221")),
        ("set_length", GDTField("8100", required=True)),
        ("total_length_of_data_packets", GDTField("9202", required=True)),
    )


class EurolabExportLGReports(list):
    def add(self, **kwargs):
        report = EurolabExportLGReport(**kwargs)
        self.append(report)
        return report

    def parse(self, data):
        report = EurolabExportLGReport.parse(data)
        self.append(report)
        return report


class EurolabExport(object):
    def __init__(self, **kwargs):
        self.header = EurolabExportLHeader(**kwargs.get("header_kwargs", {}))
        self.conclusion = EurolabExportLConclusion(
            **kwargs.get("conclusion_kwargs", {}))
        # Liste de LGReport
        self.reports = EurolabExportLGReports()

    @classmethod
    def parse(cls, data):
        obj = cls()
        obj.header = obj.conclusion = None

        def get_next_ldt_block(data, pattern):
            lines = data.splitlines()
            buf = []

            for line in lines:
                if line.startswith(pattern):
                    if not buf:
                        buf.append(line)
                    else:
                        yield "\r\n".join(buf)
                        buf = [line]
                else:
                    buf.append(line)
            yield "\r\n".join(buf)

        for ldt_block in get_next_ldt_block(data, "0138000"):
            set_type = ldt_block[7:][:4]
            if set_type == "8220":
                assert obj.header is None
                assert obj.conclusion is None
                assert len(obj.reports) == 0
                obj.header = EurolabExportLHeader.parse(ldt_block)
            elif set_type == "8202":
                assert obj.header is not None
                assert obj.conclusion is None
                obj.reports.parse(ldt_block)
            elif set_type == "8221":
                assert obj.header is not None
                assert obj.conclusion is None
                assert len(obj.reports) > 0
                obj.conclusion = EurolabExportLConclusion.parse(ldt_block)
            else:
                raise ParsingError("set type (%s) unknown" % (set_type,))

        return obj

    def to_string(self):
        # TODO: Too many hacks...
        self.conclusion.total_length_of_data_packets = "XXXXX"
        blocks = [self.header]
        blocks.extend(self.reports)
        blocks.append(self.conclusion)
        data = "\r\n".join([b.to_string() for b in blocks])
        length = "%5.5s" % len(data)
        self.conclusion.total_length_of_data_packets = length.replace(" ", "0")
        lines = data.splitlines()
        for i, line in enumerate(lines):
            if line[3:][:4] == "9202":  # total length of data packet
                lines[i] = "%s%s" % (
                    line[:7], self.conclusion.total_length_of_data_packets)
                break

        return "\r\n".join(lines)


class Driver(Message):
    SPEC = {
        "code": "eurolab",
        "label": "Eurolab IMMUN (GDT/LDT)",
    }

    _MESSAGE_TYPE_MAP = {
        Message.ORDER: EurolabImport,
        Message.RESULT: EurolabExport,
    }

    def create(self, message_type, *args, **kwargs):
        try:
            cls = self._MESSAGE_TYPE_MAP[message_type]
        except KeyError:
            raise ProgrammingError("Unknow message type")

        return cls(*args, **kwargs)

    def get_message_type(self, message):
        return self.ORDER

    def to_string(self, message):
        return message.to_string()

    def parse(self, data):
        data = data.replace("\x00", "")
        c_map = {
            "8220": self.RESULT,
            "6302": self.ORDER,
        }
        code = data[7:][:4]
        if code in c_map:
            return self._MESSAGE_TYPE_MAP[c_map[code]].parse(data)
        raise ParsingError("Unknown message type")

    def get_order_message(self, worklist, context):
        msg = self.create(self.ORDER)
        if len(worklist.samples) == 0:
            return
        elif len(worklist.samples) > 1:
            raise RuntimeError("Multi-sample not implemented")
        else:
            sample = worklist.samples[0]
            msg.sending_software_id = "MCA"
            msg.receiving_software_id = "ELO"
            msg.patient_number = sample["patient_id"]
            msg.patient_surname = sample["last_name"]
            msg.patient_firstname = sample["first_name"]
            msg.patient_birthdate = sample["birthdate"]
            msg.sex = "2" if sample["sex"] == "F" else "1"
            msg.sample_id = sample["sample_id"]
            msg.date = sample["collection_date"]
            msg.customer_short_name = sample["lab"]

            for order in sample["orders"]:
                msg.tests.append(order["code"])

            if msg.tests:
                return msg

    def get_result_dict(self, message):
        resultats = {}
        for report in message.reports:
            resultats.setdefault(report.sample_id, {"resultats": []})
            for resu in report.results:
                # Resultat multiples
                # NUCL
                if resu.aspect:

                    for index, aspect in enumerate(resu.aspect):
                        if resu.value:
                            try:
                                value = resu.value[index]
                            except IndexError:
                                break
                            resultats[report.sample_id]["resultats"].append({
                                "test": aspect,
                                "resu": value,
                                "date": report.report_date,
                            })

                    # Conclusion
                    if resu.result_text:
                        resultats[report.sample_id]["resultats"].append({
                            "test": resu.id,
                            "resu": resu.result_text,
                            "date": report.report_date,
                        })
                else:
                    resultats[report.sample_id]["resultats"].append({
                        "test": resu.id,
                        "resu": resu.value[0],
                        "date": report.report_date,
                    })

        return resultats
