# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng import Worklist
from driverng.astm import (
    AstmMessageV2,
    HeaderSegment,
    Segment,
    HprimList,
    HprimMultiple,
    HprimString,
    HprimDate,
    HprimForeign,
)


class GenexpertSegmentH(HeaderSegment):
    message_id = HprimDate(3)
    sender_name = HprimString(5)
    receiver = HprimMultiple(
        10,
        receiver_id=HprimString(1),
        receiver_name=HprimString(2),
        receiver_version=HprimString(3)
    )
    processing_id = HprimString(12, default="P")
    version = HprimString(13, default="1394-97")
    date = HprimDate(14)

    patients = HprimForeign("P", "header")
    queries = HprimForeign("Q", "header")
    end = HprimForeign("L", "header", one2one=True, required=True)


class GenexpertSegmentP(Segment):
    patient_id = HprimString(5)

    orders = HprimForeign("O", "patient")


class GenexpertSegmentO(Segment):
    specimen_id = HprimString(3)
    instrument_specimen_id = HprimString(4)
    test = HprimString((5, 4))
    priority = HprimString(6)
    ordered_date = HprimDate(7)
    action_code = HprimString(12)
    specimen_descriptor = HprimString(16)
    report_type = HprimString(26)

    results = HprimForeign("R", "order")


class GenexpertSegmentR(Segment):
    universal_test_id = HprimMultiple(
        3,
        test_panel=HprimString(2),
        test_id=HprimString(4),
        test_name=HprimString(5),
        test_version=HprimString(6),
        analyte_name=HprimString(7),
        result_name=HprimString(8),

    )
    qualitative_result = HprimString((4, 1))
    quantitative_result = HprimString((4, 2))
    units = HprimString(5)
    abnormal_flags = HprimList(7)
    result_status = HprimString(9)
    operator_identification = HprimString(11)
    date_test_started = HprimDate(12)
    date_test_completed = HprimDate(13)
    lot_reactif = HprimString((14, 5))
    date_expiration = HprimDate((14, 6), fmt_date="%Y%m%d")


class GenexpertSegmentC(Segment):
    comment_source = HprimString(3)
    comment_text = HprimString(4)
    comment_type = HprimString(5)


class GenexpertSegmentQ(Segment):
    patient_id = HprimString((3, 1))
    specimen_id = HprimString((3, 2))
    request_info = HprimString(13)


class GenexpertSegmentL(Segment):
    # N: Normal
    # I: Invalid
    terminaison_code = HprimString(3, default="N")


class Driver(AstmMessageV2):
    SPEC = {
        "code": "genexpert",
        "label": "Genexpert"
    }
    SEGMENT_CLASS_MAP = {
        "H": GenexpertSegmentH,
        "P": GenexpertSegmentP,
        "O": GenexpertSegmentO,
        "R": GenexpertSegmentR,
        "C": GenexpertSegmentC,
        "Q": GenexpertSegmentQ,
        "L": GenexpertSegmentL,
    }
    DEFAULT_SEPARATORS = "|@^&"

    def __init__(self,
                 code_emetteur="",
                 code_recepteur="GeneXpert PC^GeneXpert^4.6",
                 **kwargs):
        self.code_emetteur = code_emetteur
        self.code_recepteur = code_recepteur
        super(Driver, self).__init__(**kwargs)

    def create(self, **kwargs):
        message = GenexpertSegmentH(type_segment="H", **kwargs)
        message.end = GenexpertSegmentL(type_segment="L")
        return message

    def encode_worklist(self, worklist, context):
        if worklist.context != "order":
            raise NotImplementedError("context %s" % (worklist.context,))
        msg = self.create(
            message_id=context["now"],
            date=context["now"]
        )
        msg.set_raw("sender_name", self.code_emetteur)
        msg.set_raw("receiver", self.code_recepteur)
        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority"],
            priority={"S": "S", worklist.ELSE: "R"},
            action={"del": "C", "add": "A", "new": "A"},
        ):
            seg_p = msg.patients.add(
                patient_id=patient.patient_id,
            )
            for sample, action, priority in patient.samples:
                for order in sample.orders:
                    seg_p.orders.add(
                        specimen_id=sample.sample_id,
                        test=order.code,
                        priority=priority,
                        action_code=action,
                        ordered_date=sample.request_date,
                        specimen_descriptor=sample.specimen,
                        report_type="Q",
                    )
        if msg.patients:
            return msg
        elif context["query"] and context["query"].queries:
            msg.end.terminaison_code = "I"
            return msg

    def decode_message(self, message):
        if message.queries:
            worklist = Worklist(context="query")
            for seg_q in message.queries:
                worklist.get_or_create(
                    type="patient",
                    sample_id=seg_q.patient_id
                )
        else:
            worklist = Worklist(context="result")
            for seg_p in message.patients:
                for seg_o in seg_p.orders:
                    if seg_o.action_code == "Q":
                        sample = worklist.get_or_create(
                            control_lot=seg_o.specimen_id,
                            type="control",
                        )
                    else:
                        sample = worklist.get_or_create(
                            sample_id=seg_o.specimen_id,
                            type="patient"
                        )
                    for seg_r in seg_o.results:
                        test = "%s%s%s%s%s" % (
                            seg_r.universal_test_id.test_id,
                            seg_r.universal_test_id.test_name,
                            seg_r.universal_test_id.test_version,
                            seg_r.universal_test_id.analyte_name,
                            seg_r.universal_test_id.result_name,
                        )
                        commentaires = []
                        alarmes = []
                        if seg_r.abnormal_flags:
                            alarmes = seg_r.abnormal_flags
                        for seg_c in seg_r.commentaires:
                            if seg_c.comment_type == "I":
                                alarmes.append(seg_c.comment_text)
                            else:
                                commentaires.append(seg_c.comment_text)
                        result = sample.results.add(
                            code=test,
                            value=seg_r.qualitative_result or seg_r.quantitative_result,
                            flags=alarmes,
                            comments=commentaires,
                            date=seg_r.date_test_completed,
                        )
                        if seg_r.lot_reactif:
                            result.lots.add(
                                type="reagent",
                                lot=seg_r.lot_reactif,
                                expiration=seg_r.date_expiration,
                            )
        return worklist
