# -*- coding: utf8 -*-
from __future__ import unicode_literals

import datetime

from driverng import Message
from driverng.fields import String, Date


class ResultContainer(object):
    def __init__(self, code, value):
        self.code, self.value = code, value
        super(ResultContainer, self).__init__()

    def to_string(self):
        return ":".join([self.code, self.value])


class ResultList(list):
    def add(self, code, value):
        obj = ResultContainer(code, value)
        self.append(obj)
        return obj


class GentianeResult(object):
    FIELDS = (
        String("header"),
        String("identifiant"),
        String("nom_nomjf"),
        String("prenom"),
        Date("_naissance_j", fmt="%d"),
        Date("_naissance_m", fmt="%m"),
        Date("_naissance_a", fmt="%Y"),
        String("sexe"),  # M/F
        Date("_prelevement_time", fmt="%H:%M"),
        Date("_prelevement_date", fmt="%d/%m/%Y"),
        String("numero_dossier"),
        String("addresse1"),
        String("addresse2"),
        String("code_postal"),
        String("ville"),
        String("medecin"),
        String("commentaire"),
        String("nb_param"),
    )
    SEP = ";"

    def __init__(self, **kwargs):
        for field in self.FIELDS:
            setattr(self, field.name, kwargs.get(field.name, None))
        self.results = ResultList()
        self.date_prelevement = kwargs.get("date_prelevement", None)
        self.date_naissance = kwargs.get("date_naissance", None)
        super(GentianeResult, self).__init__()

    def parse(self, data):
        for i, value in enumerate(data.split(self.SEP)):
            try:
                field = self.FIELDS[i]
            except IndexError:
                self.results.add(*value.split(":"))
            else:
                setattr(self, field.name, field.to_python(value))
        if all([getattr(self, "_naissance_%s" % k) for k in ("j", "m", "a")]):
            self.date_naissance = datetime.datetime(
                self._naissance_a.year,
                self._naissance_m.month,
                self._naissance_j.day,
                0, 0, 0)
        if all([self._prelevement_date, self._prelevement_time]):
            self.date_prelevement = datetime.datetime(
                *(self._prelevement_date.timetuple()[:3] +
                    self._prelevement_time.timetuple()[3:5]))
        return self

    def to_string(self):
        if self.date_prelevement:
            self._prelevement_time = self.date_prelevement
            self._prelevement_date = self.date_prelevement
        if self.date_naissance:
            self._naissance_a = self.date_naissance
            self._naissance_m = self.date_naissance
            self._naissance_j = self.date_naissance
        values = [f.to_string(getattr(self, f.name)) for f in self.FIELDS]
        values.extend([r.to_string() for r in self.results])
        return self.SEP.join(values)


class Gentiane(list):

    def add(self, *args, **kwargs):
        obj = GentianeResult(*args, **kwargs)
        self.append(obj)
        return obj

    def parse(self, data):
        for line in data.splitlines():
            if not line:
                continue
            self.append(GentianeResult().parse(line))
        return self

    def to_string(self):
        return "\n".join([e.to_string() for e in self])


class Driver(Message):
    SPEC = {
        "code": "gentiane",
        "label": "Gentiane",
    }

    def create(self):
        return Gentiane()

    def parse(self, data):
        return Gentiane().parse(data)

    def to_string(self, message):
        return message.to_string()

    def get_order_message(self, worklist, context):
        msg = self.create()
        for sample in worklist.samples:
            message = msg.add(
                header="0000",
                identifiant=sample["patient_id"],
                nom_nomjf="%(last_name)s/%(maiden_name)s" % sample,
                prenom=sample["first_name"],
                date_naissance=sample["birthdate"],
                sexe=sample["sex"],
                date_prelevement=sample["collection_date"],
                numero_dossier=sample["sample_id"],
                addresse1=sample["address"],
                code_postal=sample["postal_code"],
                ville=sample["city"],
                medecin=sample["physician"],
            )
            message.nb_param = "%s" % (len(sample["results"]),)
            for result in sample["results"]:
                message.results.add(
                    code=result["code"],
                    value=result['value']
                )
        if len(msg):
            return msg
        else:
            return None
