# -*- coding: utf8 -*-
from __future__ import unicode_literals

import json

from driverng import GraphContainer
from driverng import JSONGraphContainer
from driverng.astm import (
    AstmMessage,
    HeaderSegment,
    Segment,
    HprimString,
    HprimMultiple,
    HprimList,
    HprimDate,
    HprimForeign,
)


class Ha8180GraphContainer(GraphContainer):
    def __init__(self, data):
        self.data = data
        super(Ha8180GraphContainer, self).__init__(type_fichier="json")

    def get_bytes(self):
        return json.dumps(self.data)


class Ha8180SegmentH(HeaderSegment):
    sender_name = HprimMultiple(
        5,
        instrument_name=HprimString(1),
        instrument_no=HprimString(2),
        version=HprimString(3),
        surplus=HprimString(4, default="    "),
    )
    date = HprimDate(14)

    patients = HprimForeign("P", "header")
    queries = HprimForeign("Q", "header")
    end = HprimForeign(
        "L", "header", one2one=True, required=True
    )


class Ha8180SegmentP(Segment):
    orders = HprimForeign("O", "patient")


class Ha8180SegmentO(Segment):
    patient_id = HprimMultiple(
        3,
        barcode_id=HprimString(1),
        port_no=HprimString(2),
        rack_id=HprimString(3),
    )
    measurement = HprimList(5, basefield=HprimMultiple(
        0,
        # code demande par défaut HbA1c
        measure=HprimString(4),
    ))
    # R: Normal sample
    # S: STAT
    # C: Control
    priority = HprimString(6)
    sample_type = HprimString(16)
    additional_comment = HprimMultiple(
        20,
        measurement_error=HprimString(1, default="00000000"),
        reserve=HprimString(2, default="00000000"),
        degradation_info=HprimString(3),
        mode=HprimString(4),
        eluant_a_lot=HprimString(5),
        eluant_a_date=HprimDate(6),
        eluant_b_lot=HprimString(7),
        eluant_b_date=HprimString(8),
        eluant_cv_lot=HprimString(9),
        eluant_cv_date=HprimDate(10),
        hemorisys_lot=HprimString(11),
        hemorisys_date=HprimDate(12),
        calibrator_lot=HprimString(13),
        calibrator_date=HprimString(14),
    )
    results = HprimForeign("R", "order")


class Ha8180SegmentR(Segment):
    test = HprimString((3, 4))
    value = HprimMultiple(
        4,
        **dict(("value%d" % i, HprimString(i)) for i in range(1, 31))
    )
    unit = HprimMultiple(
        5,
        unit1=HprimString(1),
        unit2=HprimString(2),
        unit3=HprimString(3),
    )
    abnormal_flag = HprimString(7)
    result_status = HprimString(9)
    date = HprimDate(12)


class Ha8180SegmentQ(Segment):
    patient_id = HprimMultiple(
        3,
        barcode_id=HprimString(1),
        port_no=HprimString(2),
        rack_id=HprimString(3),
    )
    query_code = HprimString(13)


class Ha8180SegmentL(Segment):
    code = HprimString(3)


class Driver(AstmMessage):
    SPEC = {
        "code": "ha8180",
        "label": "Ha8180 en direct"
    }
    SEGMENT_CLASS_MAP = {
        "H": Ha8180SegmentH,
        "P": Ha8180SegmentP,
        "O": Ha8180SegmentO,
        "Q": Ha8180SegmentQ,
        "R": Ha8180SegmentR,
        "L": Ha8180SegmentL,
    }
    DEFAULT_SEPARATORS = "|\\^&"

    def create(self, **kwargs):
        message = Ha8180SegmentH(type_segment="H", **kwargs)
        message.end = Ha8180SegmentL(type_segment="L")
        return message

    def get_message_type(self, message):
        if message.queries:
            return self.QUERY
        else:
            return self.ORDER

    def get_queries(self, message):
        queries = []
        for seg_q in message.queries:
            queries.append(seg_q.patient_id.barcode_id.replace("-", ""))
        return queries

    def get_order_message(self, worklist, context):
        msg = self.create(date=context["now"])
        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority"],
            action={"del": worklist.IGNORE, worklist.ELSE: ""},
            priority={"S": "S", worklist.ELSE: "R"},
        ):
            seg_p = msg.patients.add()
            for sample, action, priority in patient["samples"]:
                seg_o = seg_p.orders.add(
                    priority=priority,
                )
                seg_o.patient_id.barcode_id = sample["sample_id"]
                if context["query"] is not None:
                    for seg_q in context["query"].queries:
                        if seg_q.patient_id.barcode_id.replace(
                                "-", "").strip() == sample["sample_id"]:
                            seg_o.patient_id.barcode_id = seg_q.patient_id.barcode_id
                            seg_o.patient_id.port_no = seg_q.patient_id.port_no
                            seg_o.patient_id.rack_id = seg_q.patient_id.rack_id
                for order in sample["orders"]:
                    test = seg_o.measurement.add()
                    test.measure = order["code"]

        if msg.patients:
            return msg
        elif context["query"]:
            for seg_q in context["query"].queries:
                seg_p = msg.patients.add()
                seg_o = seg_p.orders.add(
                    priority="R",
                    sample_type="NULL",
                )
                seg_o.patient_id.barcode_id = seg_q.patient_id.barcode_id
                seg_o.patient_id.port_no = seg_q.patient_id.port_no
                seg_o.patient_id.rack_id = seg_q.patient_id.rack_id
                seg_o.measurement.measure = "NULL"
            return msg

    @staticmethod
    def _get_results(seg_o):
        results = []
        for seg_r in seg_o.results:
            if seg_r.result_status != "F":
                continue
            if seg_r.abnormal_flag:
                alarmes = [seg_r.abnormal_flag]
            else:
                alarmes = []
            if seg_r.test[:4] == "Peak":
                continue
            elif seg_r.test[:7] == "ADCount":
                continue
            else:
                results.append({
                    "test": seg_r.test,
                    "resu": seg_r.value.value1,
                    "alarmes": alarmes,
                    "date": seg_r.date,
                })
        return results

    def get_result_dict(self, message):
        results = {}
        for seg_p in message.patients:
            for seg_o in seg_p.orders:

                if seg_o.priority == "C":
                    # QC
                    continue

                tube = seg_o.patient_id.barcode_id.replace('-', '')

                if seg_o.patient_id.port_no and seg_o.patient_id.rack_id:
                    rack = "%s-%s" % (
                        seg_o.patient_id.rack_id,
                        seg_o.patient_id.port_no,
                    )
                else:
                    rack = None

                results.setdefault(tube, {
                    "resultats": [],
                    "rack": rack,
                })

                lots_reactifs = []
                for attr in (
                    "eluant_a_lot", "eluant_b_lot", "eluant_cv_lot",
                    "hemorisys_lot", "calibrator_lot",
                ):
                    ref = getattr(seg_o.additional_comment, attr)
                    if ref:
                        lots_reactifs.append({"ref": ref})

                for r in self._get_results(seg_o):
                    r.update(lots_reactifs=lots_reactifs)
                    results[tube]["resultats"].append(r)

                points = []
                peaks = []
                for seg_r in seg_o.results:
                    if seg_r.result_status != "F":
                        continue
                    if seg_r.test[:4] == "Peak":
                        peaks.append({
                            "name": seg_r.test[4:],
                            "top": int(seg_r.value.value1) * 10,
                            "area": float(seg_r.value.value2),
                            "area_percent": float(seg_r.value.value3),
                        })
                    elif seg_r.test[:7] == "ADCount":
                        for i in range(1, 31):
                            point = getattr(seg_r.value, "value%d" % i)
                            points.append(int((point), 16))

                if points:
                    while points[-20:] == 20 * [0]:
                        points.pop()
                    peaks = sorted(peaks, cmp=lambda a, b: a["top"] > b["top"])
                    for i, peak in enumerate(peaks):
                        if i == 0:
                            peak["start"] = 0

                        if i == len(peaks) - 1:
                            peak["end"] = len(points)
                        else:
                            portion = points[peak["top"]:peaks[i + 1]["top"]]
                            peaks[i + 1]["start"] = peak["end"] = portion.index(min(portion)) + peak["top"]

                    results[tube]["resultats"].append({
                        "test": "ADCount",
                        "date": seg_r.date,
                        "resu": "GRAPHE",
                        "graph": Ha8180GraphContainer({
                            "graph_type": "area",
                            "points": points,
                        }),
                    })
                    results[tube]["resultats"].append({
                        "test": "GRAPHE_A1C",
                        "resu": "GRAPHE",
                        "graph": JSONGraphContainer({
                            "graph_type": "a1c",
                            "peaks": peaks,
                            "points": points,
                        }),
                    })
        return results

    def get_qc_result_list(self, message):
        results = []
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                if seg_o.priority != "C":
                    # Non QC
                    continue
                lot = seg_o.patient_id.barcode_id.replace('-', '')
                for r in self._get_results(seg_o):
                    results.append({
                        "lot": lot,
                        "test": r["test"],
                        "resu": r["resu"],
                        "date": r["date"],
                    })
        return results
