# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng import JSONGraphContainer
from driverng.astm import (
    AstmMessage,
    HeaderSegment,
    Segment,
    HprimString,
    HprimMultiple,
    HprimList,
    HprimDate,
    HprimForeign,
)
from driverng import utils


class HelenaSegmentH(HeaderSegment):
    institution = HprimString(4)
    emetteur = HprimMultiple(
        5,
        code=HprimString(1),
        libelle=HprimString(2),
        version=HprimString(3),
    )
    address = HprimString(5)
    institution_telephone = HprimString(7)
    version = HprimString(9)
    recepteur = HprimString(10)
    patients = HprimForeign("P", "header")
    queries = HprimForeign("Q", "header")
    end = HprimForeign(
        "L", "header", one2one=True, required=True
    )


class HelenaSegmentP(Segment):
    # numéro d'échantillon
    patient_id = HprimString(3)
    last_name = HprimString(4)
    first_name = HprimString(5)
    birthdate = HprimDate(6, fmt_date="%d/%m/%Y")
    location = HprimString(7)
    sex = HprimString(8)
    orders = HprimForeign("O", "patient")


class HelenaSegmentQ(Segment):
    sample_id = HprimString((3, 1))
    sample_type = HprimString(5)


class HelenaSegmentO(Segment):
    specimen_id = HprimString(3)
    analysis_type = HprimList(5, basefield=HprimMultiple(
        0,
        code=HprimString(4),
    ))
    results = HprimForeign("R", "order")


class HelenaSegmentC(Segment):
    comment_type = HprimString(3)
    comment_text = HprimString(4)


class HelenaSegmentR(Segment):
    test = HprimMultiple(
        3,
        analysis_type=HprimString(4),
        test_name=HprimString(5),
        test_type=HprimString(6),
    )
    value = HprimString(4)
    units = HprimString(5)
    range_value = HprimString(6)
    abnormal_flag = HprimList(7)
    status = HprimString(9)
    date = HprimDate(12)


class HelenaSegmentL(Segment):
    code = HprimString(3)


class Driver(AstmMessage):
    SPEC = {
        "code": "helena",
        "label": "Helena",
    }
    SEGMENT_CLASS_MAP = {
        "H": HelenaSegmentH,
        "P": HelenaSegmentP,
        "Q": HelenaSegmentQ,
        "O": HelenaSegmentO,
        "C": HelenaSegmentC,
        "R": HelenaSegmentR,
        "L": HelenaSegmentL,
    }
    DEFAULT_SEPARATORS = "|\\^&"

    def create(self, **kwargs):
        message = HelenaSegmentH(type_segment="H", **kwargs)
        message.end = HelenaSegmentL(type_segment="L")
        return message

    def get_message_type(self, message):
        if message.queries:
            return self.QUERY
        else:
            return self.ORDER

    def get_queries(self, message):
        return [q.sample_id for q in message.queries]

    def get_order_message(self, worklist, context):
        msg = self.create()
        for patient in worklist.get_astm_worklist(
            unique_together=["action"],
            action={"del": worklist.IGNORE, worklist.ELSE: ""},
        ):
            seg_p = msg.patients.add(
                first_name=patient["first_name"],
                last_name=patient["last_name"],
                birthdate=patient["birthdate"],
                sex=patient["sex"],
            )
            for sample, action in patient["samples"]:
                seg_o = seg_p.orders.add(
                    specimen_id=sample["sample_id"]
                )
                seg_p.patient_id = sample["sample_id"]
                seg_p.location = "%s %s" % (
                    sample["location"],
                    sample["location_name"],
                )
                for order in sample["orders"]:
                    test = seg_o.analysis_type.add()
                    order_code = order["code"].split("^", 1)
                    test.code = order_code[0]
                for result in sample["results"]:
                    if "^" in result["code"]:
                        analysis_type, test_name = result["code"].split("^", 1)
                    else:
                        analysis_type = result["code"]
                        test_name = ""
                    # Envoi du resultat des protides sériques
                    seg_r = seg_o.results.add(
                        value=result["value"],
                        units="g/l",
                        abnormal_flag="N",
                        status="F",
                        date=result["date"],
                    )
                    seg_r.test.analysis_type = analysis_type
                    seg_r.test.test_name = test_name

        if msg.patients or context["query"]:
            return msg

    def get_result_dict(self, message):
        results = {}
        specimen_commentaires = {}
        for seg_p in message.patients:
            tube = seg_p.patient_id.strip()
            for seg_o in seg_p.orders:
                for seg_c in seg_o.commentaires:
                    liste_commentaire = seg_c.comment_text.split('&XD&&XA&')
                    for comment in liste_commentaire:
                        specimen_commentaires.setdefault(
                            tube,
                            []).append(comment)

                for seg_r in seg_o.results:
                    # uniquement avec le statut Final
                    if seg_r.status == "F":
                        if seg_r.test.test_name in (
                                "TraceValues",
                                "Troughs",
                                "BaselineValues",
                                "IFE1",
                                "IFE2"
                        ):
                            points = []
                            points = map(float, seg_r.value)
                            graph = JSONGraphContainer({
                                "graph_type": "area",
                                "points": points,
                            })
                            results[tube]["resultats"].append({
                                "test": "%s%s" % (
                                    seg_r.test.test_name,
                                    seg_r.test.test_type,
                                ),
                                "date": seg_r.date,
                                "resu": "GRAPHE",
                                "graph": graph,
                            })
                        else:
                            results.setdefault(
                                tube,
                                {"resultats": []},
                            )["resultats"].append({
                                "test": "%s%s" % (
                                    seg_r.test.test_name,
                                    seg_r.test.test_type,
                                ),
                                "resu": seg_r.value,
                                "alarmes": seg_r.abnormal_flag,
                                "date": seg_r.date,
                            })

                for sample, values in specimen_commentaires.items():
                    for i, value in enumerate(utils.OrderedSet(values), 1):
                        results[sample]["resultats"].append({
                            "test": "%sComment%d" % (seg_r.test.analysis_type, i),
                            "resu": value
                        })
        return results
