# -*- coding: utf8 -*-
from __future__ import unicode_literals

import logging
from collections import OrderedDict
from driverng import Worklist
from driverng.astm import (
    AstmMessageV2,
    Segment,
    HeaderSegment,
    HprimString,
    HprimMultiple,
    HprimDate,
    HprimPN,
    HprimList,
    HprimAD,
    HprimCNA,
    HprimFloat,
    HprimCE,
    HprimForeign,
)

logger = logging.getLogger("driverng.hprim")


class HprimSegmentH(HeaderSegment):

    nom_fichier = HprimString(3, max_length=12)
    emetteur = HprimMultiple(5,
                             code=HprimString(1, obligatoire=True),
                             nom=HprimString(2, obligatoire=True),
                             )

    adresse = HprimAD(6)

    # ORM: Demande
    # ORU: Resultat
    # ADM: Admission
    # FAC: Facturation
    # REG: Reglement
    # ERR: Erreur
    type_message = HprimString(7, obligatoire=True, max_length=7)
    telephone = HprimList(8)

    recepteur = HprimMultiple(10,
                              code=HprimString(1, obligatoire=True),
                              nom=HprimString(2, obligatoire=True),
                              )

    # P: Production
    # T: Test
    # D: Debug
    mode_traitement = HprimString(12, max_length=1, default="P")
    version = HprimString((13, 1), default="H2.1")
    type_liaison = HprimString((13, 2), default="L")
    date = HprimDate(14, obligatoire=True)

    patients = HprimForeign("P", "header")
    end = HprimForeign(
        "L", "header", one2one=True, required=True)


class HprimSegmentP(Segment):

    patient_demandeur = HprimString((3, 1), max_length=16)
    patient_fusion = HprimString((3, 2))
    patient_executant = HprimString((4, 1), max_length=16)
    no_adm_cli = HprimString((5, 1))
    nom = HprimPN(6, max_length=48)
    nomjf = HprimString(7, max_length=24)
    date_nai = HprimDate(8, fmt_date="%Y%m%d")
    # M/F/U
    sexe = HprimString(9, max_length=1)
    adresse = HprimAD(11)
    telephone = HprimList(13)
    medecins = HprimList(14, basefield=HprimCNA(0))
    traite_lc1 = HprimString(15)
    traite_lc2 = HprimString(16)
    taille = HprimString(17)
    poids = HprimFloat(18)
    diag_pat = HprimString(19)
    diagnostic_patient = HprimList(19, basefield=HprimCE(0))
    trait_medical = HprimString(20)
    traitement_medical = HprimList(20)
    regime = HprimString(21)
    comment_dem1 = HprimString(22)
    comment_dem2 = HprimString(23)
    date_entree = HprimDate((24, 1), fmt_date="%Y%m%d")
    date_sortie = HprimDate((24, 2), fmt_date="%Y%m%d")
    stat_adm = HprimString(25, max_length=2)
    localisation = HprimMultiple(26, max_length=100,
                                 lit=HprimString(1),
                                 chambre=HprimString(2),
                                 corr=HprimString(3),
                                 )
    class_diag = HprimString(27)
    situ_maritale = HprimString(29)
    precaution = HprimString(30)
    langue = HprimString(31)
    stat_conf = HprimString(32)
    date_modif = HprimDate(33, fmt_date="%Y%m%d")
    date_deces = HprimDate(34, fmt_date="%Y%m%d")

    orders = HprimForeign("OBR", "order")


class HprimSegmentOBR(Segment):

    tube_demandeur = HprimString((3, 1))
    no_demandeur = HprimString((3, 2))
    tube_executant = HprimString((4, 1))
    no_executant = HprimString((4, 2))
    prescription = HprimList(
        5, basefield=HprimCE(0), obligatoire=True)

    # S: Très urgent
    # A: Urgent
    # R: Routine
    # C: Rappel
    # P: Pré-opératoire
    priorite = HprimString((6, 1))
    date_demande = HprimDate(7)
    date_prel = HprimDate(8)
    date_fin_prel = HprimDate(9)
    volume = HprimString(10, max_length=10)
    preleveur = HprimCNA(11, max_length=60)

    # A: Ajout
    # C: Annulation
    # N: Nouvelle demande
    # L: Nouvelle demande avec prélèvement à effectuer par l'éxecutant
    # R: Modification de demande
    action = HprimString(12)
    risque = HprimString(13, max_length=60)
    renseignement_clinique = HprimString(14, max_length=300)
    date_recu = HprimDate(15)
    nature = HprimMultiple(16, max_length=300,
                           type_tube=HprimMultiple(1,
                                                   code=HprimString(1),
                                                   label=HprimString(2),
                                                   table=HprimString(3),
                                                   ),
                           additif=HprimString(2),
                           methode_prelevement=HprimString(3),
                           )
    prescripteur = HprimMultiple(17,
                                 code_medecin=HprimString(1),
                                 medecin=HprimPN(2),
                                 type_medecin=HprimString(3),
                                 code_service=HprimString(4),
                                 service=HprimPN(5),
                                 type_service=HprimString(6),

                                 # Legacy attributes
                                 nom_medecin=HprimString(2),
                                 nom_service=HprimString(5),
                                 )
    tel_prescrip = HprimList(18)
    lib1_dem = HprimString(19, max_length=60)
    lib2_dem = HprimString(20, max_length=60)
    lib1_exe = HprimString(21, max_length=60)
    lib2_exe = HprimString(22, max_length=60)
    date_resu = HprimDate(23, fmt_date="%Y%m%d%H%M%S")
    service_executant = HprimMultiple(25,
                                      code=HprimString(1),
                                      label=HprimString(2),
                                      )
    stat_ana = HprimString(26, max_length=1)
    dest_copie = HprimList(29)
    mod_trans_pat = HprimString(31, max_length=20)
    valideur = HprimCNA(33)
    assistant = HprimCNA(34)
    technicien = HprimCNA(35)
    ope_saisie = HprimCNA(36)
    date_prv_rendu = HprimDate(37)

    results = HprimForeign("OBX", "order")

    def get_action(self):
        return {
            "C": "del",
            "A": "add",
        }.get(self.action, "new")

    def get_services(self):
        if self.prescripteur.code_service:
            return [{
                "code": self.prescripteur.code_service,
                "name": self.prescripteur.service.nom,
            }]
        else:
            return []

    def get_comments(self):
        return [e for e in [
            self.renseignement_clinique,
            self.lib1_dem, self.lib2_dem,
        ] if e]


class HprimSegmentOBX(Segment):

    type_resu = HprimString(3, max_length=3)
    code = HprimString((4, 1), max_length=22)
    libele = HprimString((4, 2), max_length=60)
    table_reference = HprimString((4, 3))  # L pour codification locale
    sous_test = HprimString(5, max_length=20)
    resu = HprimList(6, basefield=HprimCE(0))
    unite = HprimCE(7, max_length=20)
    normale = HprimString((8, 1))
    indic_norm = HprimString((9, 1))
    probabilite = HprimString(10, max_length=5)
    nature_norm = HprimString((11, 1))
    stat_resu = HprimString(12, max_length=2)
    date_alt = HprimDate(13)
    droit_acces = HprimString(14, max_length=20)
    date = HprimDate(15)
    secteur_tech = HprimString(
        16, max_length=60, basefield=HprimString(0))  # /!\ est un CE
    valideur = HprimString(17, max_length=60)

    # HORS NORME !!
    analyseur = HprimString(19)
    resultat_controle = HprimString(20)
    # Information demande antérieure clarilab
    no_demande = HprimString((21, 1))
    date_demande = HprimDate((21, 2))


class HprimSegmentL(Segment):

    nombre_segment_patient = HprimString(4, max_length=4)
    nombre_segment = HprimString(5, max_length=10)
    numero_lot = HprimString(6, max_length=12)


class HprimSegmentC(Segment):
    # P => Émis par le demandeur
    # L => Remis par l'éxecutant
    origine = HprimString(3, max_length=1)
    texte = HprimList(4)


class Driver(AstmMessageV2):
    SPEC = {
        "code": "hprim",
        "label": "Hprim 2.1",
    }
    DEFAULT_SEPARATORS = "|~^\&"

    SEGMENT_CLASS_MAP = {
        "H": HprimSegmentH,
        "P": HprimSegmentP,
        "OBR": HprimSegmentOBR,
        "OBX": HprimSegmentOBX,
        "L": HprimSegmentL,
        "C": HprimSegmentC,
    }
    ADDENDUM_SEGMENT_SIZE = 220

    def __init__(
        self, hprim_code_emetteur="MCA", hprim_code_recepteur="",
        **kwargs
    ):
        self.code_emetteur = hprim_code_emetteur
        self.code_recepteur = hprim_code_recepteur
        super(Driver, self).__init__(**kwargs)

    def to_string(self, message, *args, **kwargs):
        message.end.nombre_segment_patient = len(message.patients)
        message.end.nombre_segment = len(message.get_segments())
        return super(Driver, self).to_string(message, *args, **kwargs)

    def create(self, **kwargs):
        message = self.SEGMENT_CLASS_MAP["H"](type_segment="H", **kwargs)
        message.end = self.SEGMENT_CLASS_MAP["L"](type_segment="L")
        return message

    def encode_worklist(self, worklist, context):
        if worklist.context not in ("result", "order"):
            raise NotImplementedError("context %s" % (worklist.context,))

        msg = self.create(
            date=context["now"],
            type_message="ORU" if worklist.context == "result" else "ORM",
        )
        msg.emetteur.code = worklist.sender_id or self.code_emetteur
        msg.recepteur.code = worklist.receiver_id or self.code_recepteur

        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority"],
            action={"del": "C", "add": "A", "new": "N"},
        ):
            seg_p = msg.patients.add(
                patient_demandeur=patient.patient_id,
                no_adm_cli=patient.alternate_patient_id,
                date_nai=patient.birthdate,
                sexe=patient.sex,
                nomjf=patient.maiden_name,
                date_entree=patient.admission_date,
                date_sortie=patient.admission_end_date,
                poids=patient.weight,
                taille=patient.height,
            )
            if patient.phone:
                seg_p.telephone = [p for p in patient.phone.split(";") if p]
            if patient.medications:
                seg_p.traitement_medical = patient.medications
            seg_p.nom.nom = patient.last_name
            seg_p.nom.prenom = patient.first_name
            seg_p.nom.civilite = patient.title
            seg_p.localisation.lit = patient.bed
            seg_p.localisation.chambre = patient.room
            seg_p.localisation.corr = patient.location
            seg_p.adresse.ville = patient.city
            seg_p.adresse.pays = patient.country
            seg_p.adresse.code_postal = patient.postal_code
            if patient.address:
                if "~" in patient.address:
                    seg_p.adresse.ligne1, seg_p.adresse.ligne2 = patient.address.split("~", 1)
                else:
                    seg_p.adresse.ligne1 = patient.address
            for physician in patient.physicians:
                if physician.type == "attending":
                    med = seg_p.medecins.add()
                    med.code = physician.code
                    med.nom.nom = physician.last_name
                    med.nom.prenom = physician.first_name
            for sample, action, priority in patient.samples:
                if not msg.emetteur.code:
                    msg.emetteur.code = sample.lab
                seg_o = seg_p.orders.add(
                    tube_demandeur=sample.sample_id,
                    no_demandeur=sample.request_id or sample.sample_id,
                    priorite=priority,
                    action=action or "N",
                    date_demande=sample.request_date,
                    date_prel=sample.collection_date,
                    date_fin_prel=sample.collection_end_date,
                )
                if sample.services:
                    seg_o.prescripteur.code_service = sample.services[0].code
                    seg_o.prescripteur.service.nom = sample.services[0].name
                for physician in sample.physicians:
                    if physician.type == "prescriber":
                        seg_o.prescripteur.code_medecin = physician.code
                        seg_o.prescripteur.medecin.nom = physician.last_name
                        if physician.phone:
                            seg_o.tel_prescrip = physician.phone.split(";")
                    elif physician.type == "sampler":
                        seg_o.preleveur.code = physician.code
                        seg_o.preleveur.nom.nom = physician.last_name
                        seg_o.preleveur.nom.prenom = physician.first_name

                if sample.transcriptionist:
                    seg_o.ope_saisie.code = sample.transcriptionist
                for code in OrderedDict.fromkeys(
                    [o.code for o in sample.orders] +
                        [r.code for r in sample.results]):
                    test = seg_o.prescription.add()
                    test.code1 = code

                for result in sample.results:
                    if worklist.context == "order" and not result.anterior:
                        date = None
                    else:
                        date = result.date
                    seg_r = seg_o.results.add(
                        code=result.code,
                        sous_test=result.sub_code,
                        date=date,
                    )
                    seg_r.unite.code1 = result.unit
                    for value in result.value.split(";"):
                        resu = seg_r.resu.add()
                        resu.code1 = value

        if msg.patients:
            return msg

    def ignore_seg_p(self, seg_p):
        for seg_obr in seg_p.orders:
            for seg_obx in seg_obr.results:
                if seg_obx.date is None and seg_obx._raw_date:
                    logger.error("Date OBX mal formée '%s' on ignore le segment P", seg_obx._raw_date)
                    return True
        return False

    def add_result(self, worklist, sample, seg_obx):
        if worklist.context == "result" and seg_obx.stat_resu == "X":
            return

        if (
            worklist.context == "order"
            and seg_obx.date is not None
        ):
            anterior = True
        else:
            anterior = False

        if seg_obx.type_resu == "FIC":
            # 10.6.1 -> Nom de l'application éméttrice
            # 10.6.2 -> Identification du fichier
            # 10.6.3 -> Type de fichier
            value = ";".join([r.label1 for r in seg_obx.resu])
        else:
            value = ";".join([r.code1 for r in seg_obx.resu])

        result = sample.results.get_or_create(
            code=seg_obx.code,
            sub_code=seg_obx.sous_test,
            date=seg_obx.date,
            value=value,
            anterior=anterior,
            unit=seg_obx.unite.code1,
        )
        if anterior:
            result.request_id = seg_obx.no_demande
            result.request_date = seg_obx.date_demande

    def decode_message(self, message):
        if message.type_message not in ("ORU", "ORM"):
            raise NotImplementedError("type message %s" % (message.type_message,))

        worklist = Worklist(
            context="result" if message.type_message == "ORU" else "order",
            sender_id=message.emetteur.code,
            receiver_id=message.recepteur.code,
        )
        for seg_p in message.patients:
            if self.ignore_seg_p(seg_p):
                continue
            for seg_obr in seg_p.orders:
                sample = worklist.get_or_create(
                    sample_id=seg_obr.tube_demandeur,
                    request_id=seg_obr.no_demandeur,
                    patient_id=seg_p.patient_demandeur or seg_p.patient_executant or seg_obr.no_demandeur,
                )
                sample.update(
                    alternate_patient_id=seg_p.no_adm_cli,
                    last_name=seg_p.nom.nom,
                    first_name=seg_p.nom.prenom,
                    title=seg_p.nom.civilite,
                    maiden_name=seg_p.nomjf,
                    birthdate=seg_p.date_nai,
                    sex=seg_p.sexe or "U",
                    request_date=(
                        seg_obr.date_demande or
                        seg_obr.date_prel or
                        seg_p.date_entree or
                        message.date),
                    type="patient",
                    diagnostic=";".join([e.code1 for e in seg_p.diagnostic_patient]),
                    location=seg_p.localisation.corr,
                    room=seg_p.localisation.chambre,
                    bed=seg_p.localisation.lit,
                    collection_date=seg_obr.date_prel,
                    collection_end_date=seg_obr.date_fin_prel,
                    admission_date=seg_p.date_entree,
                    admission_end_date=seg_p.date_sortie,
                    height=seg_p.taille,
                    phone=";".join([p for p in seg_p.telephone if p]),
                    transcriptionist=seg_obr.ope_saisie.code,
                    city=seg_p.adresse.ville,
                    postal_code=seg_p.adresse.code_postal,
                    country=seg_p.adresse.pays,
                    address="~".join([
                        l for l in [seg_p.adresse.ligne1, seg_p.adresse.ligne2]
                        if l
                    ]),
                )
                if seg_p.poids:
                    sample.weight = unicode(seg_p.poids)
                for service in seg_obr.get_services():
                    sample.services.get_or_create(**service)
                for type, segments in (
                    ("patient", seg_p.commentaires),
                    ("sample", seg_obr.commentaires),
                ):
                    for seg_c in segments:
                        for item in seg_c.texte:
                            sample.comments.get_or_create(type=type, text=item)
                for item in seg_obr.get_comments():
                    sample.comments.get_or_create(type="sample", text=item)
                for item in [
                    seg_p.traite_lc1, seg_p.traite_lc2,
                    seg_p.comment_dem1, seg_p.comment_dem2,
                ]:
                    if item:
                        sample.comments.get_or_create(type="patient", text=item)
                for item in seg_p.traitement_medical:
                    sample.medications.get_or_create(item)
                for med in seg_p.medecins:
                    if med.code not in [p.code for p in sample.physicians]:
                        sample.physicians.get_or_create(
                            code=med.code,
                            last_name=med.nom.nom,
                            first_name=med.nom.prenom,
                            type="attending",
                        )
                if (
                    seg_obr.prescripteur.code_medecin
                    and seg_obr.prescripteur.code_medecin not in [p.code for p in sample.physicians]
                ):
                    sample.physicians.get_or_create(
                        code=seg_obr.prescripteur.code_medecin,
                        last_name=seg_obr.prescripteur.medecin.nom,
                        type="prescriber",
                        phone=";".join(seg_obr.tel_prescrip),
                    )
                if worklist.context == "order":
                    for code in OrderedDict.fromkeys([
                        p.code1 for p in seg_obr.prescription
                    ] + [r.code for r in seg_obr.results]):
                        sample.orders.get_or_create(
                            code=code,
                            priority=seg_obr.priorite or "R",
                            action=seg_obr.get_action(),
                        )
                for seg_obx in seg_obr.results:
                    self.add_result(worklist, sample, seg_obx)
        return worklist
