# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng.astm import (
    AstmMessage,
    HeaderSegment,
    Segment,
    HprimString,
    HprimMultiple,
    HprimList,
    HprimDate,
    HprimForeign,
)


class IsysTestField(HprimMultiple):

    def __init__(self, *position, **kwargs):
        kwargs.update({
            # Le code test
            "code": HprimString(4),

            # 02 => Contrôle
            # 04 => Patient
            "type_test": HprimString(5),

            # F pour final
            "type_resultat": HprimString(6),

            # 1 à 10
            "no_replicat": HprimString(7),

            # Niveau du controle
            # 24 => niveau 1
            # 22 => niveau 2
            # 23 => niveau 3
            # 24 => niveau 4
            "niveau": HprimString(8),
        })
        super(IsysTestField, self).__init__(*position, **kwargs)


class IsysReactifField(HprimMultiple):

    def __init__(self, *position, **kwargs):
        kwargs.update({
            "reactif": HprimString(1),
            "lot": HprimString(2),
            "serial": HprimString(3),
            "expire": HprimDate(4)
        })
        super(IsysReactifField, self).__init__(*position, **kwargs)


class IsysSegmentH(HeaderSegment):

    # TODO: automate->ident_serv_auto
    # Non utilisé dans le sens LIS => dispositif
    # Fix à IDS-iSYS + Numéro de série dans le sens dispositif vers LIS
    emetteur = HprimString(5, default="MCA")

    # Type de message fixé à P pour Production
    id_traitement = HprimString(12, default="P")

    # version du protocole, toujours égal à 2
    no_version = HprimString(13, default="2")
    date = HprimDate(14)

    patients = HprimForeign("P", "header")
    queries = HprimForeign("Q", "header")
    end = HprimForeign(
        "L", "header", one2one=True, required=True)


class IsysSegmentP(Segment):

    # PID du patient
    no_patient = HprimString(4)

    # CESBRON^Samuel^^^^
    # Champs constitué de 5 composants Nom, prénom, Initiales, Suffixe et titre
    # Seul les deux premiers champs sont utilisés
    # Pour l'envoi des QC, le premier composant du champ sera utilisé
    # pour l'envoi du nom du produit
    nom = HprimString((6, 1))
    prenom = HprimString((6, 2))

    date_naissance = HprimDate(8, fmt_date="%Y%m%d")

    # Valeurs possibles dans le sens LIS vers ISD-iSYS
    # M = Homme
    # F = Femme
    # E = Enfant
    # U = AUtre
    # Dans le sens IDS-iSYS vers LIS
    # M = Homme
    # F = Femme
    # U = Enfant et Autre
    sexe = HprimString(9)

    # Num du medecin associé au bilan
    medecin = HprimString((14, 1))

    # NULL pour un patient
    # 02 pour un controle
    controle = HprimString(15)

    # Commentaire associé au bilan
    commentaire_bilan = HprimString(19)

    # Service associé au bilan
    service = HprimString((33, 1))

    orders = HprimForeign("O", "patient")


class IsysSegmentO(Segment):

    # Type de contenant
    # 01 = Tube Primaire PT_11_4,5 ml
    # 02 = Tube Primaire PT_13_6 ml
    # 03 = Tube secondaire ST_13_5 ml
    # 05 = Cupule 500 ul
    # 06 = Cupule 2 ml
    # 07 = Tube primaire 10 ml PT_15_8,5 ml
    # de 08 à 17 pour les contenants configurables sur demande
    # Pour un fonctionnement correct de l'appareil, il est important
    # d'attribuer le type de contenant correct utilisé pour le spécimen
    type_contenant = HprimString((3, 1))

    tube = HprimString((3, 2))

    tests = HprimList(5, basefield=HprimMultiple(0,
                                                 code=HprimString(4),
                                                 ))

    # S = Stat
    # A = ASAP
    # R = Routine
    # C = Callback
    # P = préoperative
    # Seul S à une influence, les autres valeurs sont ignorées
    urgence = HprimString(6, max_length=1)

    # A = Add
    # C = Cancel
    # N = New
    # P = Pending
    # Q = QC
    # A et N sont gérés de la même façon (nouvelle demande ou nouvelle repasse)
    # P et Q ne sont pas gérés
    code_action = HprimString(12)

    # U = Urine
    # S = Serum
    # O = Autre
    specimen = HprimString((16, 1))

    medecin = HprimString(17)

    # O = Order
    # C = Correction
    # P = Preliminary Results
    # F = Final results
    # X = Cancel
    # I = Instrument in pending
    # Y = No order on record for this test
    # Z = No record for this patient
    # Q = réponse à une requête
    # Dans le dispositif C ne sera pas géré
    report_type = HprimString(26)

    results = HprimForeign("R", "order")


class IsysSegmentQ(Segment):

    # patient ID (optionel)
    pid = HprimString((3, 1))
    tube = HprimString((3, 2), max_length=16)

    # Un test particulier ou tous les tests (ALL)
    request = HprimString(5, default="ALL")

    # Cf IsysSegmentO.report_type
    status_code = HprimString(13, default="O")


class IsysSegmentR(Segment):

    test = IsysTestField(3)

    valeur = HprimString(4)

    # Attention les abréviations doivent être ISO 2955
    unite = HprimString(5)

    # Valeurs de références normales utilisées pour le patient
    # 10.5 to 13.5^Normal\9 to 15^Limite
    valeur_reference = HprimList(
        6,
        basefield=HprimMultiple(
            0,
            valeur=HprimString(1),
            type_valeur=HprimString(2),
        ),
    )

    # L = Below Low Normal
    # H = Above High Normal
    # < = Absolute low (domaine de mesure)
    # > = absolute high (domaine de mesure)
    # N = Normal
    # A = Anormal
    # Tout message associé est envoyé dans un bloc Comment
    flag_abnormal = HprimString(7)

    # Nature of Abnormality Testing
    # A = Age
    # S = Sexe
    # Permet d'indiquer sur quelles bases ont été déterminées les valeurs de
    # normalité
    nature_of_abnormality = HprimString(8)

    # F = Résultat finaux
    # C = Correction
    # P = Résultat préliminaires
    # X = Ne peuvent être obtenus
    # I = En attente dans l'analyseur
    # S = Résultats partiels
    status = HprimString(9)

    date_debut_realisation = HprimDate(12)
    date_fin_realisation = HprimDate(13)

    # Nom de l'instrument + numéro de série (IDS-iSYS01)
    analyseur = HprimString(14)

    info = HprimForeign("M", "result")


class IsysSegmentC(Segment):

    # P = Preactice
    # L pour computer system dans le sens LIS
    # I pour Instrument
    origine = HprimString(3)

    texte = HprimString(4)

    # G = generic
    # T = Test name comment
    # P = Positive test comment
    # N = Negative test comment
    # I = Instument flags comment
    type_commentaire = HprimString(5)


class IsysSegmentM(Segment):

    # Même chose que le segment R
    test = IsysTestField(3)

    # 51 => Immunologie
    # 52 => Biochimie
    assay_type = HprimString(4)

    # Utilisé uniquement pour les contrôles
    # 12,50^8,00^17,00^1142^^20130131
    component_information = HprimMultiple(5,
                                          valeur_cible=HprimString(1),
                                          val_inf=HprimString(2),
                                          val_sup=HprimString(3),
                                          lot=HprimString(4),
                                          no_flacon=HprimString(5),
                                          date_expiration=HprimDate(6),
                                          )

    # 250HD^0952^2095256^02987^20120229
    tracabilite_reactif = HprimList(6, basefield=IsysReactifField(0))

    # System Liquid^72634^00162^20121209\Cuvettes^18121^02177^20120506
    common_ancillaries = HprimList(7, basefield=IsysReactifField(0))

    # Pour un champ Assay type de 51 (Immunologie) ce champ est composé de:
    # TRIGB^Lot^N°flacon^Expire
    # TRIGA^Lot^N°flacon^Expire
    # Wash1^N°laveur
    # Wash2^N°laveur
    # Wash3^N°laveur
    # DSORB^Lot^N°flacon^Expire
    # Pour un champ Assay type de 52 (biochimie) ce champ est composé de:
    # Serum Physiologique^Lot^N°flacon^Expire
    specific_ancillaries = HprimList(8, basefield=IsysReactifField(0))


class IsysSegmentL(Segment):

    # N = Terminaison normale
    # T = Interrompue par l'émetteur
    # R = Interruption demandée par le récepteur
    # E = Erreur système inconnue
    # Q = Erreur dans la dernière demande d'informations
    # I = Pas d'informations disponibles à partir de la dernière demande
    # F = Dernière demande d'information traitée
    code = HprimString(3, max_length=1)


class Driver(AstmMessage):
    SPEC = {
        "code": "isys",
        "label": "Isys",
    }
    SEGMENT_CLASS_MAP = {
        "H": IsysSegmentH,
        "P": IsysSegmentP,
        "O": IsysSegmentO,
        "Q": IsysSegmentQ,
        "R": IsysSegmentR,
        "C": IsysSegmentC,
        "L": IsysSegmentL,
        "M": IsysSegmentM,
    }
    DEFAULT_SEPARATORS = "|\\^_"

    def create(self, **kwargs):
        message = IsysSegmentH(type_segment="H", **kwargs)
        message.end = IsysSegmentL(type_segment="L")
        return message

    def get_message_type(self, message):
        if message.queries:
            return self.QUERY
        return self.ORDER

    def get_queries(self, message):
        queries = []
        for seg_q in message.queries:
            queries.append(seg_q.tube)
        return queries

    def get_order_message(self, worklist, context):
        msg = self.create(date=context["now"])
        msg.end.code = "N"

        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority"],
            action={"del": "C", worklist.ELSE: "A"},
            priority={"S": "S", "A": "S", worklist.ELSE: "R"},
        ):
            seg_p = msg.patients.add(
                no_patient=patient["patient_id"],
                nom=patient["last_name"],
                prenom=patient["first_name"],
                date_naissance=patient["birthdate"],
                sexe=patient["sex"],
            )
            for sample, action, priority in patient["samples"]:
                # En réponse à une requête le LIS doit envoyer une trame
                # contentant un bloc test order avec le bilan à effectuer et
                # dont le champs report type est impérativement à la valeur Q
                seg_o = seg_p.orders.add(
                    tube=sample["sample_id"],
                    urgence=priority,
                    code_action=action,
                    report_type="Q" if context["query"] else "O",
                )
                for order in sample["orders"]:
                    t = seg_o.tests.add()
                    t.code = order["code"]

        if msg.patients:
            return msg
        elif context["query"]:
            msg.end.code = "N"
            for sample_id in self.get_queries(context["query"]):
                seg_p = msg.patients.add()
                seg_o = seg_p.orders.add()
                seg_o.tube = sample_id
                seg_o.report_type = "Y"
            return msg

    def get_result_dict(self, message):
        if self.get_message_type(message) != self.ORDER:
            return {}

        result = {}
        for seg_p in message.patients:

            # On ne prend pas les controles
            if seg_p.controle == "02":
                continue

            for seg_o in seg_p.orders:

                # TODO /!\ /!\ /!\ à verifier
                # La doc dis bien que le tube est en 3.2
                # mais en réalité sur un message de résultat
                # le tube peut être en 3.1
                tube = seg_o.tube or seg_o.type_contenant

                result[tube] = {"resultats": []}

                for seg_r in seg_o.results:

                    if seg_r.status != "F":
                        continue

                    alarmes = []
                    if seg_r.flag_abnormal:
                        alarmes.append(seg_r.flag_abnormal)

                    commentaires = []
                    for seg_c in seg_r.commentaires:
                        if seg_c.type_commentaire == "N":
                            commentaires.append(seg_c.texte)

                    try:
                        float(seg_r.valeur.replace(",", "."))
                        resultat = seg_r.valeur.replace(",", ".")
                    except ValueError:
                        resultat = seg_r.valeur

                    lots_reactifs = []
                    if seg_r.info:
                        seg_m = seg_r.info[0]
                        for reactif in seg_m.tracabilite_reactif:
                            lots_reactifs.append({
                                "ref": reactif.lot,
                                "peremption": reactif.expire,
                            })

                    result[tube]["resultats"].append({
                        "test": seg_r.test.code,
                        "resu": resultat,
                        "alarmes": alarmes,
                        "date": seg_r.date_fin_realisation,
                        "analyseur": seg_r.analyseur,
                        "commentaires": commentaires,
                        "lots_reactifs": lots_reactifs,
                    })

        return result

    def get_qc_result_list(self, message):
        if self.get_message_type(message) != self.ORDER:
            return []

        results = []

        for seg_p in message.patients:

            # on ne prend que les controles
            if seg_p.controle != "02":
                continue

            for seg_o in seg_p.orders:
                for seg_r in seg_o.results:

                    if seg_r.status != "F":
                        continue

                    try:
                        seg_m = seg_r.info[0]
                    except IndexError:
                        raise RuntimeError(
                            "Pas d'informations manufacturer")

                    lots_reactifs = []
                    for reactif in seg_m.tracabilite_reactif:
                        lots_reactifs.append({
                            "ref": reactif.lot,
                            "peremption": reactif.expire,
                        })

                    try:
                        float(seg_r.valeur.replace(",", "."))
                        resultat = seg_r.valeur.replace(",", ".")
                    except ValueError:
                        resultat = seg_r.valeur

                    results.append({
                        "lot": seg_m.component_information.lot,
                        "nom_lot": seg_p.nom,
                        "test": seg_r.test.code,
                        "resu": resultat,
                        "date": seg_r.date_fin_realisation,
                        "analyseur": seg_r.analyseur,
                        "lots_reactifs": lots_reactifs,
                    })

        return results
