# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng.astm import (
    AstmMessage,
    HeaderSegment,
    Segment,
    HprimString,
    HprimMultiple,
    HprimList,
    HprimDate,
    HprimForeign,
)


class KoneSegmentH(HeaderSegment):
    sender_id = HprimMultiple(
        5,
        instrument_type=HprimString(1, default="LIS"),
        instrument_id=HprimString(2),
        software_version=HprimString(3),
    )
    # valeurs possibles:
    #  P: Production
    #  Q: QC
    #  D: Debugging
    #  T: Training
    processing_id = HprimString(12, default="P")
    date = HprimDate(14)
    patients = HprimForeign("P", "header")
    end = HprimForeign(
        "L", "header", one2one=True, required=True
    )


class KoneSegmentP(Segment):
    patient_id = HprimString(3)
    last_name = HprimString((6, 1))
    first_name = HprimString((6, 2))
    birthday = HprimDate(8, fmt_date="%Y%m%d")
    # M/F/U
    sex = HprimString(9)
    orders = HprimForeign("O", "patient")


class KoneSegmentO(Segment):

    specimen_id = HprimMultiple(
        3,
        specimen_id=HprimString(1),
        manual_dilution=HprimString(2),
        segment=HprimString(3),
        position=HprimString(4),
    )
    universal_test_id = HprimList(5, basefield=HprimMultiple(
        0,
        code=HprimString(4),
        dilution=HprimString(5),
    ))
    # S: Stat
    # A: Asap
    # R: Routine
    # C: Callback
    # P: Preoperative
    priority = HprimString(6)
    collection_date = HprimDate(8)
    # C: Cancel
    # A: Add
    # N: New test requests + new specimen
    # P: Pending
    # L: reserved
    # X: Specimen or test in process
    # Q: QC specimen
    action_code = HprimString(12)
    clinical_information = HprimString(14)
    # 1: Serum
    # 2: Plasma
    # 3: Urine
    # 4: CSF
    # 5: Other
    specimen_type = HprimString(16)
    instrument_id = HprimString(25)
    # O: Order
    # C: Correction to previously transmitted results
    # P: Preliminary results
    # F: Final resuls
    # X: Requests cancelled
    # I: Instrument pending
    # Y: No order for test (response to query)
    # Z: No record of this patient (response to query)
    # Q: Response to query
    # En download et en Query le report_type
    # est identique
    report_type = HprimList(26)
    results = HprimForeign("R", "order")


class KoneSegmentR(Segment):
    test = HprimString((3, 4))
    dilution = HprimString((3, 5))
    value = HprimString(4)
    units = HprimString(5)
    reference_ranges = HprimMultiple(
        6,
        low=HprimString(1),
        high=HprimString(2),
        description=HprimString(3),
    )
    abnormal_flag = HprimList(7)
    # C: Correction
    # P: Preliminary
    # F: Final
    # X: Cancelled
    # I: Pending
    # R: Reporte
    # Q: Response to query
    status = HprimList(9)
    operator_identification = HprimString(11)
    date_completed = HprimDate(13)
    instrument_identification = HprimString(14)


class KoneSegmentC(Segment):
    comment_source = HprimString(3)
    comment_text = HprimString(4)
    comment_type = HprimString(5)


class KoneSegmentQ(Segment):
    starting_range_id = HprimMultiple(
        3,
        patient_id=HprimString(1),
        sample_id=HprimString(2),
        segment=HprimString(3),
        position=HprimString(4),
    )
    manufacturer_test_code = HprimString((5, 4))
    nature_request_time = HprimString(6)
    beginning_request_date = HprimDate(7)
    ending_request_date = HprimDate(8)
    request_info_status = HprimString(13)


class KoneSegmentL(Segment):
    code = HprimString(3, default="N")


class Driver(AstmMessage):
    SPEC = {
        "code": "kone",
        "label": "Kone ASTM",
    }
    SEGMENT_CLASS_MAP = {
        "H": KoneSegmentH,
        "P": KoneSegmentP,
        "O": KoneSegmentO,
        "Q": KoneSegmentQ,
        "R": KoneSegmentR,
        "C": KoneSegmentC,
        "L": KoneSegmentL,
    }
    DEFAULT_SEPARATORS = "|\\^&"

    def create(self, **kwargs):
        message = KoneSegmentH(type_segment="H", **kwargs)
        message.end = KoneSegmentL(type_segment="L")
        return message

    def get_message_type(self, message):
        for patient in message.patients:
            for order in patient.orders:
                if order.action_code == "Q":
                    return self.QC
        return self.ORDER

    def get_order_message(self, worklist, context):
        msg = self.create(date=context["now"])
        for patient in worklist.get_astm_worklist(
                unique_together=["action", "priority"],
                action={
                    "del": "C",
                    "add": "A",
                    "new": "N"
                },
                priority={
                    "S": "S",
                    "A": "A",
                    worklist.ELSE: "R"
                }
        ):
            seg_p = msg.patients.add(
                patient_id=patient["patient_id"],
                last_name=patient["last_name"],
                first_name=patient["first_name"],
                birthday=patient["birthdate"],
                sex=patient["sex"],
            )
            for sample, action, priority in patient["samples"]:
                seg_o = seg_p.orders.add(
                    priority=priority,
                    collection_date=sample["collection_date"],
                    action_code=action,
                    specimen_type=sample["specimen"],
                    report_type=["Q"],
                )
                seg_o.report_type.append("O")
                seg_o.specimen_id.specimen_id = sample["sample_id"]
                for order in sample["orders"]:
                    test = seg_o.universal_test_id.add()
                    test.code = order["code"]

        if msg.patients or context["query"]:
            return msg

    def get_result_dict(self, message):
        results = {}
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                # exclusion des QC
                if seg_o.action_code == "Q":
                    continue
                for seg_r in seg_o.results:
                    # uniquement les resultats avec le statut F
                    if "F" not in seg_r.status[0] or len(seg_r.status) > 1:
                        continue
                    alarmes = []
                    if seg_r.abnormal_flag:
                        alarmes = seg_r.abnormal_flag
                    for seg_c in seg_r.commentaires:
                        comment = seg_c.comment_text.split(" ")
                        alarmes.append(comment[0])
                    results.setdefault(
                        seg_o.specimen_id.specimen_id,
                        {"resultats": []})["resultats"].append(
                        {
                            "test": seg_r.test,
                            "date": seg_r.date_completed,
                            "alarmes": alarmes,
                            "resu": seg_r.value,
                        })
        return results

    def get_qc_result_list(self, message):
        result = []
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                # uniquement les QC
                if seg_o.action_code != "Q":
                    continue
                for seg_r in seg_o.results:
                    if "F" not in seg_r.status:
                        continue
                    result.append({
                        "lot": seg_o.specimen_id.specimen_id,
                        "test": seg_r.test,
                        "resu": seg_r.value,
                        "date": seg_r.date_completed,
                    })
        return result
