# -*- coding: utf8 -*-

from __future__ import unicode_literals
from driverng.hl7 import (
    AstmMessage,
    Segment,
    HL7Foreign,
    HL7String,
    HL7CE,
    HL7List,
    HL7HD,
    HL7DTM,
    HL7NM,
    HL7RP,
    HL7CompositePrice,
    HL7CompositeQuantity,
    HL7SegmentMSH,
)


class HL7SegmentMFI(Segment):
    """Master file Identifier
    Identifie le catalogue.
    """
    type_catalogue = HL7String(2,
                               obligatoire=True,
                               max_length=250,
                               default="OMC")
    # champ modifié par le IHE
    # APPLICATION_IDENTIFIER doit être utilisé
    # Exemple: CLARISYS_OMC_FRA dans le sous champ 1
    version_catalogue = HL7HD(3, obligatoire=True, max_length=180)
    code_traitement = HL7String(4,
                                obligatoire=True,
                                max_length=3,
                                default='REP')
    date_effet = HL7DTM(6, obligatoire=True)
    # Available: AL => Always, NE => Never
    response_level = HL7String(7, obligatoire=True, default='NE')

    mfe = HL7Foreign("MFE", "mfi")


class HL7SegmentMFE(Segment):
    """Master File Entry
    Identifie une entrée dans le catalogue.
    """
    code_traitement = HL7String(2, obligatoire=True, default='MAD')
    control_id = HL7String(3)
    date_effet = HL7DTM(4)
    pkey = HL7CE(5, obligatoire=True, max_length=200)
    pkey_type = HL7String(6, obligatoire=True, default="CE")

    om1 = HL7Foreign("OM1", "mfe")
    om4 = HL7Foreign("OM4", "mfe")
    om5 = HL7Foreign("OM5", "mfe")
    zca = HL7Foreign("ZCA", "mfe")


class HL7SegmentOM1(Segment):
    """Identifie une analyse."""
    seq = HL7NM(2, obligatoire=True, default=1)
    analyse = HL7CE(3, obligatoire=True)
    echantillon_requis = HL7String(5, default="Y")
    service_production = HL7CE(6, obligatoire=True)
    synonymes = HL7List(9, basefield=HL7String(0))
    technique = HL7List(15, basefield=HL7CE(0))
    specialite = HL7List(17, basefield=HL7CE(0))
    nature = HL7List(19,
                     basefield=HL7String(0, max_length=20, default="A"))
    # En minutes
    delai_realisation = HL7List(24, basefield=HL7NM(0))
    frequence_realisation = HL7List(41, basefield=HL7String(0))
    commentaire = HL7String(42, max_length=65536)


class HL7SegmentOM4(Segment):
    """Observations That Require Specimens Segment
    Une occurrence de ce segment contient des informations liées
    à un type d’échantillon nécessaire à la réalisation de l’examen.
    """
    seq = HL7NM(2, obligatoire=True)
    container_description = HL7String(4, max_length=60)
    container_volume = HL7NM(5)
    container_unit = HL7CE(6)
    nature = HL7CE(7)
    additif = HL7CE(8)
    conditions_conservation = HL7String(10)
    volume_requis = HL7CompositeQuantity(11)


class HL7SegmentOM5(Segment):
    """Défini un "groupe d'anlyses"
    Pour un analyse unique il faut mettre la meme chose que OM1-2 dans OM5-2.
    """
    seq = HL7NM(2)
    analyses = HL7List(3, obligatoire=True, basefield=HL7CE(0))


class HL7SegmentZCA(Segment):
    """Informations tarifaires. Spécificité IHE france."""
    prix_hn = HL7CompositePrice(1)
    prix_fixe = HL7String(2, default="Y")
    entente_prealable = HL7String(3, default="N")
    consentement = HL7String(4, default="N")
    examen_supplemtaire = HL7String(5)
    code_nabm = HL7List(6, basefield=HL7CE(0))
    url_documentaire = HL7RP(7)
    conditions_tarifaires = HL7String(8, max_length=250)


class Driver(AstmMessage):
    SPEC = {
        "code": "HL7_LCSD",
        "label": "LCSD National Extension 1.2",
    }
    FIRST_SEP_INDEX = 3
    DEFAULT_SEPARATORS = r"|^~\&"

    ENCODING_HL7_MAP = {
        "ISO-8859-15": "8859/15",
        "UTF-8": "UNICODE UTF-8",
    }

    SEGMENT_CLASS_MAP = {
        "MSH": HL7SegmentMSH,
        "MFI": HL7SegmentMFI,
        "MFE": HL7SegmentMFE,
        "OM1": HL7SegmentOM1,
        "OM4": HL7SegmentOM4,
        "OM5": HL7SegmentOM5,
        "ZCA": HL7SegmentZCA,
    }
    ADDENDUM_SEGMENT_SIZE = None

    def create(self, **kwargs):
        msg = self.SEGMENT_CLASS_MAP["MSH"](type_segment="MSH", **kwargs)
        msg.encoding = self.ENCODING_HL7_MAP[
            kwargs.get("encoding", "ISO-8859-15")]
        msg.type_message.message_type = "MFM"
        msg.type_message.event_type = "M10"
        msg.type_message.structure = "MFM_M10"
        msg.mode_traitement = "D"
        return msg

    def get_message_type(self, message):
        """TODO: C'est quoi ce get_message_type et self.ORDER ?"""
        return self.ORDER

    def get_result_dict(self, message):
        return {}
