# -*- coding: utf8 -*-
from __future__ import unicode_literals

import json
import logging

from driverng import GraphContainer
from driverng import Message
from driverng.fields import String, Date

logger = logging.getLogger("driverng.drivers.lh")


class LhGraphContainer(GraphContainer):
    def __init__(self, data):
        self.data = data
        super(LhGraphContainer, self).__init__(type_fichier="json")

    def get_bytes(self):
        return json.dumps(self.data)


class LhBase(object):

    FIELDS = ()
    CODES = ()
    FLAGS_PATHO = ()

    def __init__(self, **kwargs):
        for code, field in self.FIELDS:
            setattr(self, field.name, kwargs.get(field.name, None))
        super(LhBase, self).__init__()

    def parse(self, data):
        fields = dict(self.FIELDS)
        flags_patho = self.FLAGS_PATHO
        for elt in data.splitlines():
            if not elt:
                continue
            elt = elt.strip()
            position = elt.find(' ')
            if position > 0 and elt not in flags_patho:
                code = elt[:position]
                if code in fields:
                    field = fields[code]
                    value = field.to_python(elt[position + 1:])
                    setattr(self, field.name, value)
                else:
                    logger.warning("Unknown field '%s'", code)
            elif elt in flags_patho:
                code = elt[0:]
                field = fields[code]
                setattr(self, field.name, elt)

    def to_string(self):
        values = []
        for code, field in self.FIELDS:
            value = field.to_string(getattr(self, field.name))
            if code in self.FLAGS_PATHO and value != "":
                values.append("%s" % (value,))
            elif not value:
                values.append(code)
            elif code not in self.FLAGS_PATHO:
                values.append("%s %s" % (
                    code,
                    value,
                ))
        return "\n".join(values) + "\n"


class Resultat(LhBase):
    CODES = (
        "WBC", "RBC", "HGB", "HCT", "MCV", "MCH", "MCHC",
        "RDW", "PLT", "MPV", "LY#", "MO#", "NE#", "EO#",
        "BA#", "LY%", "MO%", "NE%", "EO%", "BA%", "RBCH",
        "PLTH", "PLTF", "WBCT", "WBCF", "DIFF",
    )

    FLAGS_PATHO = (
        "Abnormal WBC Pop", "Abnormal RBC Pop", "Abnormal PLT Pop",
        "Normal WBC Pop", "Normal RBC Pop", "Normal PLT Pop",
        "Leukopenia", "Leukocytosis", "Neutropenia %", "Neutropenia #",
        "Neutrophilia %", "Neutrophilia #", "Lymphopenia %",
        "Lymphopenia #", "Lymphocytosis %", "Lymphocytosis #",
        "Monocytosis %", "Monocytosis #", "Eosinophilia %",
        "Eosinophilia #", "Basophilia %", "Basophilia #",
        "Anemia", "Anisocytosis", "1+ Anisocytosis",
        "2+ Anisocytosis", "3+ Anisocytosis", "Microcytosis",
        "1+ Macrocytosis", "2+ Macrocytosis", "3+ Macrocytosis",
        "Hypochromia", "1+ Hypochromia", "2+ Hypochromia",
        "3+ Hypochromia", "1+ Poikilocytosis", "2+ Poikilocytosis",
        "3+ Poikilocytosis", "Erythrocytosis", "Pancytopenia",
        "Thrombocytopenia", "Thrombocytosis", "Small Platelets",
        "Large Platenets",
    )

    FIELDS = [
        ("DATE", Date("date", fmt="%d/%m/%y")),
        ("TIME", Date("time", fmt="%H:%M:%S")),
        ("ID1", String("tube", 16)),
        ("ID2", String("tube_secondaire", 16)),
        ("CASS/POS", String("rack", 8)),
        ("SEQUENCE", String("sequence", 7)),
        ("1EC/P", String("cassette", 8)),
        ("ID1STATUS", String("status", 2)),
        ("C/PSTATUS", String("status_cassette", 2)),
        ("WLSTATUS", String("list_status", 9)),
    ]
    for code in CODES:
        FIELDS.append((code, String(code, 5)),)
    FIELDS.extend([
        ("BIRTH", Date("birth_date", fmt="%d/%m/%Y")),
        ("UF1", String("uf1", 17)),
        ("UF2", String("uf2", 17)),
        ("UF3", String("uf3", 17)),
        ("SEX", String("sex", 2)),
        ("LOCATION", String("location", 17)),
        ("PHYSICIAN", String("physician", 22)),
        ("UDATE", Date("update", fmt="%d/%m/%Y")),
        ("UTIME", Date("uptime", fmt="%h:%M:%S"))
    ])
    for flag in FLAGS_PATHO:
        FIELDS.append((flag, String(flag, 0)),)


class Order(LhBase):
    FIELDS = (
        ("WLAD", String("type_message", 0)),
        ("ID", String("tube", 16)),
        ("PL", String("last_name", 16)),
        ("PF", String("first_name", 16)),
        ("BD", Date("birth_date", fmt="%d/%m/%Y")),
        ("SX", String("sex", 1)),
        ("TS", String("test", 64)),
    )


class Driver(Message):
    SPEC = {
        "code": "lh",
        "label": "LH (750, 780, 500)",
    }

    TYPE_MESSAGE_MAP = {
        Message.ORDER: Order,
        Message.RESULT: Resultat,
    }

    def create(self, message_type, **kwargs):
        return self.TYPE_MESSAGE_MAP[message_type](**kwargs)

    def to_string(self, message):
        return message.to_string()

    def get_message_type(self, message):
        return self.ORDER

    def parse(self, data):
        if data[:4] == "WLAD":
            message_type = self.ORDER
        else:
            message_type = self.RESULT
        obj = self.TYPE_MESSAGE_MAP[message_type]()
        obj.parse(data)
        return obj

    def get_order_message(self, worklist, context):
        if len(worklist.samples) == 0:
            return None
        elif len(worklist.samples) > 1:
            raise RuntimeError("Multi-sample not implemented")
        else:
            sample = worklist.samples[0]
            codes = ["", "CBC", "DIFF", "DIFF", "RETIC", "CR", "CDR", "CDR"]
            order_map = {
                "CBC": 1,
                "DIFF": 3,
                "RETIC": 4,
                "CDR": 7,
                "CR": 5,
            }
            test = 0
            for order in sample["orders"]:
                if order["action"] in ("new", "add"):
                    test |= order_map.get(order["code"], 0)

            if test == 0:
                return None

            msg = self.create(
                self.ORDER,
                # WTF ????
                tube="%c%s" % (
                    sample["sample_id"][0],
                    sample["lab_patient_id"],
                ),
                last_name=sample["last_name"],
                first_name=sample["first_name"],
                birth_date=sample["birthdate"],
                sex=sample["sex"],
                # WTF #2 ????
                test="%s,%s," % (
                    codes[test],
                    sample["sample_id"],
                )
            )
            return msg

    def get_results(self, message):
        results = []
        flags = []
        for code, field in message.FIELDS:
            value = getattr(message, field.name)
            if code in message.CODES and value:
                if " " in value:
                    resu, alarme = value.split(" ", 1)
                else:
                    resu, alarme = value, ""
                    if len(resu) > 5:
                        if code in ('RBCH', 'WBCT', 'WBCF', 'PLTH', 'PLTF'):
                            graph = LhGraphContainer({
                                "graph_type": "dxh_scatter",
                                "blob": resu,
                            })
                        elif code in ('DIFF'):
                            points = []
                            p = resu
                            while p:
                                points.append(int(p[0:2], 16))
                                p = p[2:]
                            graph = LhGraphContainer({
                                "graph_type": "area",
                                "points": points,
                            })
                        resu = graph
                results.append((code, resu, [c for c in alarme if c]))
            elif code in message.FLAGS_PATHO and value:
                flags.append(value)
        if flags:
            results.append(("ALARM2", ";".join(flags), []))

        return results

    def get_result_dict(self, message):
        results = {}
        for test, resu, alarmes in self.get_results(message):
            if test in ('RBCH', 'WBCT', 'WBCF', 'PLTH', 'PLTF', 'DIFF'):
                results.setdefault(message.tube.strip(), {
                    "resultats": []})["resultats"].append({
                        "test": test,
                        "resu": "GRAPHE",
                        "graph": resu,
                        "date": message.date,
                    })
            else:
                results.setdefault(message.tube.strip(), {
                    "resultats": []})["resultats"].append({
                        "test": test,
                        "resu": resu,
                        "date": message.date,
                        "alarmes": alarmes,
                    })
        return results
