# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng.astm import (
    AstmMessage,
    HeaderSegment,
    Segment,
    HprimString,
    HprimMultiple,
    HprimList,
    HprimDate,
    HprimForeign,
)


class LiaisonSegmentH(HeaderSegment):

    sender_id = HprimString(5)
    receiver_id = HprimString(10, max_length=20)
    version = HprimString(13, default="1")
    patients = HprimForeign("P", "header")
    date = HprimDate(14)
    queries = HprimForeign("Q", "header")
    end = HprimForeign(
        "L", "header", one2one=True, required=True)


class LiaisonSegmentP(Segment):

    patient_id = HprimString(4, max_length=20)
    name = HprimString((6, 1), max_length=20)
    first_name = HprimString((6, 2), max_length=20)
    birthdate = HprimDate(8, fmt_date="%Y%m%d")
    # M/F/U
    sex = HprimString(9)
    physician_id = HprimString(14, max_length=6)
    orders = HprimForeign("O", "patient")


class LiaisonSegmentQ(Segment):
    range_id = HprimList(3)
    test_id = HprimString(5, default="ALL")
    status = HprimString(13, default="O")


class LiaisonSegmentO(Segment):

    specimen_id = HprimString(3, max_length=20)
    tests = HprimList(5, basefield=HprimMultiple(
        0,
        code=HprimString(4),
        dilution_ratio=HprimString(5),
    ))

    # S: Stat
    # N: Normal
    priority = HprimString(6)

    request_date = HprimDate(7)

    # O:
    # F:
    # I:
    # C:
    # X:
    report_type = HprimString(26)

    results = HprimForeign("R", "order")


class LiaisonSegmentR(Segment):

    test = HprimString((3, 4))
    value = HprimString(4)
    units = HprimString(5)

    # L: Below normal range
    # H: Above normal range
    # <: Below assay range
    # >: Above assay range
    # N: Within normal range
    abnormal_flag = HprimList(7)

    # F:
    # I:
    # C:
    # X:
    status = HprimString(9)
    date = HprimDate(13)
    instrument_id = HprimString(14)


class LiaisonSegmentC(Segment):

    # I: Clinical instrument system
    source = HprimString(3)
    text = HprimList(4)
    # I: Instrument flag comments
    comment_type = HprimString(5)


class LiaisonSegmentL(Segment):

    # N: Normal termination
    # I: No info available
    code = HprimString(3, default="N")


class Driver(AstmMessage):
    SPEC = {
        "code": "liaison",
        "label": "Liaison",
    }
    SEGMENT_CLASS_MAP = {
        "H": LiaisonSegmentH,
        "P": LiaisonSegmentP,
        "O": LiaisonSegmentO,
        "Q": LiaisonSegmentQ,
        "C": LiaisonSegmentC,
        "R": LiaisonSegmentR,
        "L": LiaisonSegmentL,
    }
    DEFAULT_SEPARATORS = "|\\^&"

    def create(self, **kwargs):
        message = LiaisonSegmentH(type_segment="H", **kwargs)
        message.end = LiaisonSegmentL(type_segment="L")
        return message

    def get_message_type(self, message):
        if message.queries:
            return self.QUERY
        else:
            return self.ORDER

    def get_queries(self, message):
        queries = []
        for seg_q in message.queries:
            queries.extend(seg_q.range_id)
        return queries

    def get_order_message(self, worklist, context):
        msg = self.create(date=context["now"])
        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority"],
            action={"del": worklist.IGNORE, worklist.ELSE: "N"},
            priority={"S": "S", worklist.ELSE: "N"},
        ):
            seg_p = msg.patients.add(
                patient_id=patient["patient_id"],
                name=patient["last_name"],
                first_name=patient["first_name"],
                birthdate=patient["birthdate"],
                sex=patient["sex"],
            )
            for sample, action, priority in patient["samples"]:
                seg_o = seg_p.orders.add(
                    specimen_id=sample["sample_id"],
                    priority=priority,
                    action_code=action,
                    request_date=sample["request_date"],
                    report_type="O",
                )
                for order in sample["orders"]:
                    t = seg_o.tests.add()
                    t.code = order["code"]
                    t.dilution_ratio = order["dilution"]

        if msg.patients:
            return msg
        elif context["query"]:
            for sample_id in self.get_queries(context["query"]):
                seg_p = msg.patients.add()
                seg_p.orders.add(
                    specimen_id=sample_id,
                    report_type="X",
                )
            return msg

    def get_result_dict(self, message):
        results = {}
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                if not seg_o.specimen_id.isdigit():
                    continue
                for seg_r in seg_o.results:
                    if seg_r.status != "F":
                        continue
                    alarmes = []
                    if seg_r.abnormal_flag:
                        alarmes = seg_r.abnormal_flag
                    for seg_c in seg_r.commentaires:
                        alarmes.extend(seg_c.text)
                    results.setdefault(
                        seg_o.specimen_id,
                        {"resultats": []})["resultats"].append(
                        {
                            "test": seg_r.test,
                            "date": seg_r.date,
                            "alarmes": alarmes,
                            "resu": seg_r.value,
                        })
        return results

    def get_qc_result_list(self, message):
        results = []
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                if seg_o.specimen_id.isdigit():
                    continue
                for seg_r in seg_o.results:
                    if seg_r.status != "F":
                        continue
                    results.append({
                        "lot": seg_o.specimen_id,
                        "test": seg_r.test,
                        "date": seg_r.date,
                        "resu": seg_r.value,
                    })
        return results
