# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng.astm import (
    AstmMessage,
    HeaderSegment,
    Segment,
    HprimString,
    HprimList,
    HprimDate,
    HprimForeign,
)
from driverng.utils import get_date, format_date


class MPLSegmentH(HeaderSegment):

    emetteur = HprimString(5)
    type_message = HprimString(7)
    recepteur = HprimString(10)
    version = HprimString(13, default="A2.2")
    date = HprimDate(14)

    patients = HprimForeign("P", "header")
    end = HprimForeign("L", "header", one2one=True, required=True)


class MPLSegmentP(Segment):

    no_patient = HprimString(3, max_length=20)

    # nom patient OU nom controle
    nom = HprimString((6, 1), max_length=30)

    # prenom patient OU numéro de lot
    prenom = HprimString((6, 2), max_length=30)
    nomjf = HprimString(7, max_length=30)
    date_naissance = HprimDate(8, fmt_date="%Y%m%d")
    # M/F/U
    sexe = HprimString(9)

    # Q pour les QC
    controle = HprimString(12, max_length=1)

    orders = HprimForeign("OBR", "patient")


class MPLSegmentOBR(Segment):

    # tube OU nom du controle
    tube = HprimString((3, 1), max_length=16)
    # Numéro SIL de liaison ou numéro de lot
    access_numbrer = HprimString((3, 2), max_length=20)

    tests = HprimList(5)

    # Priorité
    # A: urgent
    # S: imédiat (même effet que A)
    # R: Routine
    urgence = HprimString(6, max_length=1)

    # Date et jeure de rendu des résultats
    # Même que rendu_date ??
    result_date = HprimDate(7)

    # Code action
    # A ou N ou vide: Ajout de tests
    # C: suppréssion de tests
    code_action = HprimString(12, max_length=1)

    # Date et heure d'accueil de la demande sur le SIL
    order_date = HprimDate(15)

    # Prescripteur
    prescripteur = HprimString(17, max_length=25)

    # Service prescripteur (UF dans MPL.net)
    service = HprimString(19, max_length=12)

    # Emetteur, facultatif
    # remplace le service prescripteur si il est renseigné
    emetteur = HprimString(21, max_length=25)

    # Flag << résultat du tube téléphonés >>
    # TEL ou vide
    telephone = HprimString(22, max_length=3)

    # Date et heure de rendu des résultats.
    # Même que result_date ??
    rendu_date = HprimDate(24)

    results = HprimForeign("OBX", "order")


class MPLSegmentOBX(Segment):

    # NM: Numérique, type par défaut,
    #     Il est possible de forcer une valeur non numérique à rendre NM
    # CE: Type par défaut si le résultat est alpha-numérique
    # DT: Retourné si le nom de l'analyse dans MPL.net
    #     commence par HEURE ou DATE
    # ST: Utilisé avec un système préanalytique,
    #     retourné uniquement si le resultat vaut TSMTRI
    # FIC: Fichier
    type_resultat = HprimString(3, max_length=3)
    # ALU^PTUR => pas conforme a la doc mais c'est ce qui se passe en prod !
    # Le premier semble être le code test et le deuxième le "sous test"
    test = HprimString((4, 1), max_length=15)
    sous_test = HprimString((4, 2))
    # le resultat
    valeur = HprimString(6, max_length=8)

    flags = HprimList(9)

    # A: antério si ORM
    # R: antério si ORU
    # F: final
    status = HprimString(12, max_length=1)

    date = HprimDate(13)

    numero_analyseur = HprimString(14)

    # valideur max_length = 10
    # Initiales du valideur, MPLauto pour validation automatique
    valideur = HprimString((17, 1))
    # B: Biologiste
    # T: technicien
    type_valideur = HprimString((17, 2))

    infos = HprimForeign("M", "result")


class MPLSegmentM(Segment):

    # L pour lot réactif
    type_lot = HprimString(3, max_length=1)

    # R1, R2 etc
    numero = HprimString((4, 1), max_length=2)

    # lot réactif
    lot = HprimString((4, 2), max_length=12)

    # maximum 8 lot par résultat


class MPLSegmentC(Segment):
    # C|1|L|^XXXXXX

    commentaire = HprimString((4, 2))


class MPLSegmentL(Segment):

    nombre_segment_patient = HprimString(4)
    nombre_segment = HprimString(5)


class Driver(AstmMessage):
    SPEC = {
        "code": "mpl",
        "label": "Mpl",
    }
    SEGMENT_CLASS_MAP = {
        "H": MPLSegmentH,
        "P": MPLSegmentP,
        "OBR": MPLSegmentOBR,
        "OBX": MPLSegmentOBX,
        "M": MPLSegmentM,
        "C": MPLSegmentC,
        "L": MPLSegmentL,
    }
    DEFAULT_SEPARATORS = "|^~\&"

    def create(self, **kwargs):
        message = MPLSegmentH(type_segment="H", **kwargs)
        message.end = MPLSegmentL(type_segment="L")
        return message

    def to_string(self, message, *args, **kwargs):
        message.end.nombre_segment_patient = len(message.patients)
        message.end.nombre_segment = len(message.get_segments())
        return super(Driver, self).to_string(message, *args, **kwargs)

    def get_message_type(self, message):
        for patient in message.patients:
            if patient.controle == "Q":
                return self.QC
        return self.ORDER

    def get_order_message(self, worklist, context):
        msg = self.create(
            emetteur="MCA",
            type_message="ORM",
            date=context["now"],
        )
        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority"],
            action={"del": "C", worklist.ELSE: ""},
            priority={"A": "A", "S": "S", worklist.ELSE: "R"},
        ):
            seg_p = msg.patients.add(
                no_patient=patient["patient_id"],
                nom=patient["last_name"],
                prenom=patient["first_name"],
                nomjf=patient["maiden_name"],
                date_naissance=patient["birthdate"],
                sexe=patient["sex"],
            )
            for sample, action, priority in patient["samples"]:
                seg_o = seg_p.orders.add(
                    tube=sample["sample_id"],
                    urgence=priority,
                    prescripteur=sample["physician"],
                    service=sample["service"],
                    code_action=action,
                    order_date=sample["collection_date"],
                )
                for order in sample["orders"]:
                    seg_o.tests.append(order["code"])
                if action == "":
                    for result in sample["results"]:
                        resu = result["value"]
                        date = get_date(result["value"])
                        if date is not None:
                            resu = format_date(date, fmt="%Y%m%d")
                            type_resultat = "NM"
                        else:
                            try:
                                float(result["value"])
                                type_resultat = "NM"
                            except ValueError:
                                type_resultat = "CE"
                        seg_o.results.add(
                            type_resultat=type_resultat,
                            test=result["code"],
                            valeur=resu,
                            status="A",
                            date=result["date"],
                        )

        if msg.patients:
            return msg

    def get_result_dict(self, message):
        result = {}
        for seg_p in message.patients:

            if seg_p.controle == "Q":
                continue

            for seg_o in seg_p.orders:
                result[seg_o.tube] = {"resultats": []}

                for seg_r in seg_o.results:

                    if seg_r.status != "F":
                        # On ne prend pas les anterio
                        continue

                    lots_reactifs = []
                    for seg_m in seg_r.infos:
                        if seg_m.type_lot == "L":  # réactifs
                            if seg_m.lot and seg_m.lot not in lots_reactifs:
                                lots_reactifs.append(seg_m.lot)

                    commentaires = [seg_c.commentaire for seg_c in
                                    seg_r.commentaires]

                    result[seg_o.tube]["resultats"].append({
                        "test": seg_r.test,
                        "resu": seg_r.valeur,
                        "date": seg_r.date,
                        "lots_reactifs": lots_reactifs,
                        "commentaires": commentaires,
                        "analyseur": seg_r.numero_analyseur,
                        "alarmes": seg_r.flags,
                    })

        return result

    def get_qc_result_list(self, message):
        result = []
        for seg_p in message.patients:

            if seg_p.controle != "Q":
                continue

            for seg_o in seg_p.orders:

                for seg_r in seg_o.results:
                    result.append({
                        "lot": seg_p.prenom,
                        "test": seg_r.test,
                        "resu": seg_r.valeur,
                        "date": seg_r.date,
                        "analyseur": seg_r.numero_analyseur,
                    })

        return result
