# -*- coding: utf8 -*-
from __future__ import unicode_literals

import xml.etree.cElementTree as ET
import datetime

from driverng import Message
from driverng import utils


class Driver(Message):
    SPEC = {
        "code": "mylaxml",
        "label": "Myla XML",
    }
    ENCODING = "utf8"

    def __init__(self, bmr="", default_cmi="-", **kwargs):
        self.bmr = bmr
        self.default_cmi = default_cmi
        return super(Driver, self).__init__(**kwargs)

    def parse(self, data):
        return ET.fromstring(data.encode("utf8"))

    def to_string(self, message):
        return utils.prettify_xml(ET.tostring(message, 'utf-8'))

    def get_message_type(self, message):
        return self.ORDER

    def get_order_message(self, worklist, context):
        root = ET.Element("lisMessage")
        header = ET.SubElement(root, "header")
        ET.SubElement(header, "version").text = "1394-97"
        ET.SubElement(
            header,
            "dateTime"
        ).text = context["now"].strftime("%Y%m%d%H%M%S")

        empty = True
        for sample, _ in worklist.get_worklist(
            unique_together=["action"],
            action={"del": worklist.IGNORE, worklist.ELSE: None},
        ):
            request = ET.SubElement(root, "request")
            patient = ET.SubElement(request, "patientInformation")
            ET.SubElement(
                patient,
                "patientIdentifier"
            ).text = sample["patient_id"]
            ET.SubElement(patient, "firstName").text = sample["first_name"]
            ET.SubElement(patient, "lastName").text = sample["last_name"]
            if sample["birthdate"]:
                ET.SubElement(
                    patient,
                    "birthdate"
                ).text = sample["birthdate"].strftime("%Y%m%d")
            ET.SubElement(
                patient, "sex").text = sample["sex"]
            ET.SubElement(
                patient, "physicianIdentifier").text = sample["physician"]
            ET.SubElement(
                patient, "physicianName").text = sample["physician_name"]
            if sample.get("admission_date"):
                ET.SubElement(
                    patient,
                    "admissionDate"
                ).text = sample.get("admission_date").strftime("%Y%m%d")
            ET.SubElement(
                patient, "location").text = sample["location"]

            for order in sample["orders"]:
                test_order = ET.SubElement(request, "testOrder")
                specimen = ET.SubElement(test_order, "specimen")
                ET.SubElement(
                    specimen,
                    "specimenIdentifier"
                ).text = sample["sample_id"]
                ET.SubElement(
                    specimen,
                    "specimenType"
                ).text = order["code"]
                ET.SubElement(
                    specimen,
                    "specimenNature"
                ).text = order["code"]
                if sample["collection_date"]:
                    ET.SubElement(
                        specimen,
                        "collectedDateTime"
                    ).text = sample["collection_date"].strftime("%Y%m%d%H%M%S")
                test = ET.SubElement(test_order, "test")
                universal_identifier = ET.SubElement(
                    test, "universalIdentifier")
                # Type de prélèvement d'hémoculture
                if sample["specimen"] == "SU":
                    ET.SubElement(
                        universal_identifier,
                        "testGroupCode"
                    ).text = sample["specimen"]
                else:
                    ET.SubElement(
                        universal_identifier,
                        "testIdentifier"
                    ).text = order["code"]
                empty = False

        if not empty:
            return root

    def get_result_dict(self, message):
        samples_results = {}
        for request in message.findall("requestResult"):
            for test_order in request.findall("testOrder"):
                sample_id = test_order.findtext(
                    "./specimen/specimenIdentifier")
                sample_results = samples_results.setdefault(sample_id, {})
                specimen_type = test_order.findtext("./specimen/specimenType")

                # Bloodculture bottle
                if test_order.findtext(
                    "./test/result/bmxResultData/disposable/disposableType"
                ) == "BOTTLE":
                    reference = test_order.findtext("./test/result/bmxResultData/disposable/reference")

                    # Identification bloodculture bottle
                    test = reference + "_IDENT"
                    sample_results[test] = {
                        "test": test,
                        "resu": test_order.findtext(
                            "./test/result/bmxResultData/disposable/disposableIdentifier"
                        )
                    }
                    # Dates
                    for suffix, value in (
                        ("_START_DATETIME", test_order.findtext("./test/result/startDateTime")),
                        ("_DISPOSABLE_START_DATETIME", test_order.findtext("./test/result/bmxResultData/disposable/disposableStartDateTime")),
                        ("_DISPOSABLE_FINAL_DATETIME", test_order.findtext("./test/result/bmxResultData/disposable/disposableFinalDateTime")),
                    ):
                        if value is not None:
                            test = reference + suffix
                            sample_results[test] = {
                                "test": test,
                                "resu": value,
                            }

                    # Ignoré dans les messages d'acquittement du tube
                    if test_order.findtext("./test/result/resultStatus") != "I":
                        # BloodCulture result
                        test = reference + "_RESULT"
                        sample_results[test] = {
                            "test": test,
                            "resu": test_order.findtext(
                                "./test/result/value/bloodCulture/result"
                            ),
                            "date": datetime.datetime.strptime(
                                test_order.findtext("./test/result/value/bloodCulture/dateResultDetermined"),
                                "%Y%m%d%H%M%S"
                            )
                        }
                        # Dates
                        for suffix, value in (
                            ("_TIME_TO_DETECTION", test_order.findtext("./test/result/value/bloodCulture/timeToDetectionInHours")),
                            ("_DETERMINED_DATETIME", test_order.findtext("./test/result/value/bloodCulture/dateResultDetermined")),
                        ):
                            if value is not None:
                                test = reference + suffix
                                sample_results[test] = {
                                    "test": test,
                                    "resu": value,
                                }
                elif specimen_type:
                    germe = "G_%s" % (
                        test_order.findtext(
                            "./test/result/value/identification/significantTaxon/code"
                        ).upper(),)
                    result = {
                        "atb": [],
                        "seq": test_order.findtext("isolateNumber"),
                        "test": specimen_type,
                        "date": datetime.datetime.strptime(
                            test_order.findtext("./test/result/resultDateTime"),
                            "%Y%m%d%H%M%S"
                        ),
                        "resu": germe,
                    }
                    # Gestion des bacteries multi resistantes On le fait passer
                    # en antibiotique dans le même test mais ça n'a rien à voir
                    for resistance in test_order.findall(
                        "./test/result/value/expertFinding/resistanceMechanism"
                    ):
                        if resistance.findtext("drugFamily") == self.bmr:
                            bmrs = [e.text for e in resistance.findall("phenotype")]
                            # TODO: À supprimer après avoir repensé la gestion
                            # des BMR.  à voir avec damien@
                            if germe != "G_STAAUR" and "MODIFICATION DES PLP (mecA)" in bmrs:
                                bmrs.remove("MODIFICATION DES PLP (mecA)")
                            if bmrs:
                                result["atb"].append((
                                    ("ANTIBIO", ";".join(["BMR_%s" % (e,) for e in bmrs])),
                                    ("CMI", "#"),
                                    ("SENSI", "BMR"),
                                ))
                    for antibiotic in test_order.findall(
                        "./test/result/value/ast/antibiotic"
                    ):
                        code = "A_%s" % (antibiotic.findtext("code").upper(),)
                        cmi = "%s%s" % (
                            antibiotic.findtext("./highMic/micSign") or "",
                            antibiotic.findtext("./highMic/micValue") or "",
                        )
                        sensi = antibiotic.findtext("category")
                        result["atb"].append((
                            ("ANTIBIO", code),
                            ("CMI", cmi or self.default_cmi),
                            ("SENSI", sensi)))

                    if not result["atb"]:
                        # identification de germes
                        result["test"] = "IDENT_" + result["test"]
                        result["resu"] = {
                            "type_resultat": "identification_germes",
                            "germes": [],
                        }
                        if result["test"] in sample_results:
                            result["resu"]["germes"] = (
                                sample_results[result["test"]]["resu"].get("germes", [])
                            )
                        result["resu"]["germes"].append({
                            "id": germe,
                            "rang": int(result["seq"]),
                        })
                    elif result["test"] in sample_results:
                        raise NotImplementedError(
                            "Plusieurs resultats pour un meme test avec ATB."
                        )
                    sample_results[result["test"]] = result

        results = {}
        for sample_id, data in samples_results.iteritems():
            results.setdefault(sample_id, {"resultats": []})["resultats"].extend(
                data.values())
        return results
