# -*- coding: utf8 -*-

from __future__ import unicode_literals

import json
import datetime
from collections import OrderedDict

from driverng import (
    Message,
    GraphContainer,
)
from driverng.utils import splitlines, RS
from driverng.fixed import (
    FixedLengthFormat,
    String,
    Date,
)


class PhoresysGraphContainer(GraphContainer):
    def __init__(self, points):
        self.points = points
        super(PhoresysGraphContainer, self).__init__(type_fichier="json")

    def get_bytes(self):
        return json.dumps({"points": self.points})


class PhoresysBase(FixedLengthFormat):
    pass


class PhoresysQuery(PhoresysBase):
    FMT_DESCRIPTION = [
        String("code", 1),
        String("patient_id", 15),
    ]


class PhoresysDemande(PhoresysBase):
    FMT_DESCRIPTION_HEADER = [
        String("code", 1),
        String("sample", 4),
        String("patient_id", 15),
        String("patient_name", 30),
        Date("birthdate", 8, fmt_date="%d%m%Y"),
        String("sex", 1),  # M/F
        String("age", 3, right_justified=True, empty="0"),
        String("department", 20),
        Date("sample_date", 8, fmt_date="%d%m%Y"),
        String("concentration", 5),
    ]
    FMT_DESCRIPTION = FMT_DESCRIPTION_HEADER + [
        String("free%d" % i, 30) for i in range(1, 6)
    ]


class PhoresysExtraData(PhoresysBase):
    FMT_DESCRIPTION = [
        "**",
    ] + [String("programmable_field%d" % i, 15) for i in range(1, 41)] + [
        String("card_comment", 256),
    ]


class PhoresysResultData(PhoresysBase):
    FMT_DESCRIPTION = [
        String("ndots", 4, right_justified=True, empty="0"),
        # A: Automatic scale
        # M: Manual scale
        String("curve_scale_flag", 1),
        String("curve_scale_factor", 4, right_justified=True, empty="0"),
    ]
    for i in range(1, 5):
        FMT_DESCRIPTION += [
            String("peak%d_start" % (i,), 4, right_justified=True, empty="0"),
            String("peak%d_end" % (i,), 4, right_justified=True, empty="0"),
        ]

    def __init__(self, *args, **kwargs):
        self._locked = False
        self.points = kwargs.pop("points", "")
        self.extra_data = PhoresysExtraData()
        self.extended_comment = ""
        super(PhoresysResultData, self).__init__(*args, **kwargs)

    def parse(self, data):
        remain = super(PhoresysResultData, self).parse(data)
        if "|" in remain:
            self.points, remain = remain.split("|", 1)
            if remain:
                self.extended_comment = self.extra_data.parse(remain)
            else:
                self.extended_comment = ""
        else:
            self.points = ""
            self.extended_comment = ""
        return remain

    def to_string(self):
        return "%s%s|%s" % (
            super(PhoresysResultData, self).to_string(),
            self.points,
            self.extra_data.to_string(),
            self.extended_comment,
        )


class PhoresysResult(PhoresysBase):
    FMT_DESCRIPTION = PhoresysDemande.FMT_DESCRIPTION_HEADER + [
        String("measurement_unit", 8),
    ] + [String("free%d" % i, 30) for i in range(1, 6)] + [
        String("operator_id", 3),
        Date("date", 8, fmt_date="%d%m%Y"),
        String("fractions", 2, right_justified=True, empty="0"),
    ] + [
        String("fraction%d_name" % i, 10) for i in range(1, 11)
    ] + [
        String("fraction%d_value" % i, 5) for i in range(1, 11)
    ] + [
        String("fraction%d_conc" % i, 5) for i in range(1, 11)
    ] + [
        String("area%d_name" % i, 10) for i in range(1, 5)
    ] + [
        String("area%d_value" % i, 5) for i in range(1, 5)
    ] + [
        String("area%d_conc" % i, 5) for i in range(1, 5)
    ] + [
        # 0: Normal
        # 1: Pathological
        String("pathological_flag", 1),
        String("ratio1", 5),
        String("ratio2", 5),
        String("comment", 230, right_justified=True),
        # 0: Normal pattern
        # 1: Reference pattern
        String("reference_pattern_flag", 1),
        # 0: Normal sample
        # 1: QC sample
        String("qc_sample_flag", 1),
    ]

    def __init__(self, *args, **kwargs):
        # TODO
        self._locked = False
        self.data = None
        super(PhoresysResult, self).__init__(*args, **kwargs)

    def parse(self, data):
        remain = super(PhoresysResult, self).parse(data)
        if remain:
            self.data = PhoresysResultData()
            return self.data.parse(remain)
        return remain

    def to_string(self):
        return "%s%s" % (
            super(PhoresysResult, self).to_string(),
            self.data.to_string(),
        )


class PhoresysMultiple(list):
    QUERY, ORDER, RESULT = list(range(3))

    def add(self, type_message, **kwargs):
        m = {
            self.QUERY: PhoresysQuery,
            self.ORDER: PhoresysDemande,
            self.RESULT: PhoresysResult,
        }[type_message](**kwargs)
        self.append(m)
        return m


class Driver(Message):
    SPEC = {
        "code": "phoresys",
        "label": "Phoresys (Capilaris)",
    }

    def __init__(self, qc_legacy=False, **kwargs):
        self.qc_legacy = qc_legacy
        super(Driver, self).__init__(**kwargs)

    def create(self):
        return PhoresysMultiple()

    def parse(self, data):
        messages = PhoresysMultiple()
        for line in splitlines(data):
            m = {
                16: PhoresysQuery,
                245: PhoresysDemande,
            }.get(len(data), PhoresysResult)()
            m.parse(data)
            messages.append(m)
        return messages

    def to_string(self, message):
        return "\n".join([m.to_string() for m in message])

    def get_message_type(self, message):
        for m in message:
            if isinstance(m, PhoresysQuery):
                return self.QUERY
            else:
                return self.ORDER

    def get_queries(self, message):
        queries = []
        for m in message:
            if isinstance(m, PhoresysQuery):
                queries.append(m.patient_id)
        return queries

    def get_order_message(self, worklist, context):

        if not context["query"]:
            return None

        msg = self.create()
        known_samples = []
        # sample_id -> profil
        profils = OrderedDict()

        for m in context["query"]:
            if isinstance(m, PhoresysQuery):
                profils[m.patient_id] = m.code

        for sample, action, priority in worklist.get_worklist(
            unique_together=["action", "priority"],
            action={"del": worklist.IGNORE, worklist.ELSE: ""},
            priority={"A": "Urgent", "S": "Urgent", worklist.ELSE: ""},
        ):
            concentration = None
            for r in sample["results"]:
                if r["code"] == "PROTEINE":
                    concentration = r["value"]
            sex = {
                "M": "M",
                "F": "F",
            }.get(sample["sex"], "")
            patient_name = "%s %s" % (
                sample["last_name"],
                sample["first_name"]
            )
            msg.add(
                msg.ORDER,
                code=profils[sample["sample_id"]],
                patient_id=sample["sample_id"],
                patient_name=patient_name[:30],
                birthdate=sample["birthdate"],
                sex=sex,
                sample_date=sample["collection_date"],
                department=sample["service"],
                concentration=concentration,
                free1=sample["patient_id"],
                free5=priority,
            )
            known_samples.append(sample["sample_id"])

        for sample_id, profil in profils.items():
            if sample_id not in known_samples:
                msg.add(
                    msg.ORDER,
                    code=profil,
                    patient_id=sample_id,
                    patient_name="Inconnu",
                )

        if len(msg):
            return msg

    def get_resultats(self, message):
        r = []
        for i in range(1, 11):
            name, value, conc = [
                getattr(message, "fraction%d_%s" % (i, k)) for k in (
                    "name", "value", "conc")]
            if name and value:
                r.append({
                    "test": "%s%%" % (name,),
                    "resu": value,
                })
            if name and conc:
                r.append({
                    "test": "%s#" % (name,),
                    "resu": conc,
                })
        return r

    @staticmethod
    def get_analyseur(message):
        if (
            message.data
            and message.data.extra_data
            and message.data.extra_data.programmable_field1
        ):
            return message.data.extra_data.programmable_field1
        else:
            return None

    def get_result_dict(self, message):
        results = {}
        for m in message:
            if m.qc_sample_flag == "1" or 'CONTROL' in m.patient_id:  # QC
                return {}

            analyseur = self.get_analyseur(m)

            results.setdefault(m.patient_id, {"resultats": []})
            resultats = self.get_resultats(m)
            alarmes = [m.pathological_flag]
            for i in range(1, 5):
                name, value, conc = [
                    getattr(m, "area%d_%s" % (i, k)) for k in (
                        "name", "value", "conc")]
                if name and value:
                    resultats.append({
                        "test": "%s%%" % (name,),
                        "resu": value,
                    })
                if name and conc:
                    resultats.append({
                        "test": "%s#" % (name,),
                        "resu": conc,
                    })
            for r in resultats:
                results[m.patient_id]["resultats"].append({
                    "test": r["test"],
                    "resu": r["resu"],
                    "date": m.date,
                    "alarmes": alarmes,
                    "analyseur": analyseur,
                })
            if m.data and m.data.points:
                points = []
                p = m.data.points
                while len(p) >= 4:
                    points.append(int(p[1:4], 16))
                    p = p[4:]
                if points:
                    results[m.patient_id]["resultats"].append({
                        "test": "%sGRAPHE" % (m.code,),
                        "resu": "GRAPHE",
                        "date": m.date,
                        "alarmes": alarmes,
                        "graph": PhoresysGraphContainer(points),
                        "analyseur": analyseur,
                    })
            if m.comment:
                results[m.patient_id]["resultats"].append({
                    "test": "%sCOMMENT" % (m.code,),
                    "resu": m.comment,
                    "date": m.date,
                    "alarmes": alarmes,
                    "analyseur": analyseur,
                })
            if m.data and m.data.extended_comment:
                results[m.patient_id]["resultats"].append({
                    "test": "%sCOMMENT_EXTENDED" % (m.code,),
                    "resu": m.data.extended_comment.replace(RS, "\n"),
                    "date": m.date,
                    "alarmes": alarmes,
                    "analyseur": analyseur,
                })
            if m.ratio1:
                results[m.patient_id]["resultats"].append({
                    "test": "%sRATIO1" % (m.code,),
                    "resu": m.ratio1,
                    "date": m.date,
                    "alarmes": alarmes,
                    "analyseur": analyseur,
                })
            if m.ratio2:
                results[m.patient_id]["resultats"].append({
                    "test": "%sRATIO2" % (m.code,),
                    "resu": m.ratio2,
                    "date": m.date,
                    "alarmes": alarmes,
                    "analyseur": analyseur,
                })
        return results

    def get_qc_result_list(self, message):
        for m in message:
            if m.qc_sample_flag == "0" and 'CONTROL' not in m.patient_id:
                return []
            results = []
            now = datetime.datetime.now()
            date = datetime.datetime(*(
                m.date.timetuple()[:3] + now.timetuple()[3:5]))
            if self.qc_legacy:
                lot = m.patient_id
            else:
                lot = m.patient_name.rsplit("-", 1)[1]
            analyseur = self.get_analyseur(m)
            for r in self.get_resultats(m):
                results.append({
                    "lot": lot,
                    "analyseur": analyseur,
                    "test": r["test"],
                    "date": date,
                    "resu": r["resu"],
                })
            return results
