# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng import Message
from driverng.astm import (
    AstmMessage,
    HeaderSegment,
    Segment,
    HprimString,
    HprimList,
    HprimDate,
    HprimForeign,
    HprimMultiple,
)


class PianoSegmentH(HeaderSegment):

    # les champs ci-dessous ne doivent pas etre transmis lors de la demande
    # code emetteur
    emetteur = HprimString(5)
    # date emission du resultat
    date = HprimDate(14)

    patients = HprimForeign("P", "header")
    queries = HprimForeign("Q", "header")
    end = HprimForeign("L", "header", one2one=True, required=True)


class PianoSegmentP(Segment):

    # Practice Assigned Patient ID
    no_patient = HprimString(3, max_length=14)

    # nom patient
    nom = HprimString((6, 1), max_length=16)

    # prenom patient
    prenom = HprimString((6, 2), max_length=16)
    date_naissance = HprimDate(8, fmt_date="%Y%m%d")
    # M/F/U
    sexe = HprimString(9)

    orders = HprimForeign("O", "patient")


class PianoSegmentO(Segment):

    # Specimen ID
    tube = HprimString((3, 1), max_length=16)

    # Liste des codes tests
    # ^^^Test^Request.DilutionCoefficient
    tests = HprimList(5, basefield=HprimMultiple(0,
                                                 code=HprimString(4),
                                                 coeff_dilution=HprimString(5),
                                                 ))

    # type de specimen d'echantillon Sp1, Sp2 ou Control
    specimen_type = HprimString(16)

    results = HprimForeign("R", "order")


class PianoSegmentQ(Segment):
    # Le segment Q n'est pas clairement décrit dans
    # la doc, mais il y a des exemples

    tube = HprimString((3, 2))

    request = HprimString(5, default="ALL")


class PianoSegmentR(Segment):

    # code test
    # ^^^Test
    test = HprimString((3, 4), max_length=3)

    # le resultat
    valeur = HprimString(4)

    # unite
    unite = HprimString(5, max_length=6)

    #  reference ranges
    ref_ranges = HprimString(6)

    #  result abnormal flags
    flag = HprimString(7, max_length=2)

    # statut du resultat
    # F: final (manual validation)
    # P: Pending (automated transfer)
    # X: Rejected
    # I: Not measured
    # Attention ce champ n'est pas systèmatiquement envoyé
    status = HprimString(9)

    # code operator
    operator = HprimString(11, max_length=32)

    # date du resultat ou du QC
    date = HprimDate(13)


class PianoSegmentL(Segment):
    # Code de terminaison
    # N: Terminaison normale
    code = HprimString(3)


class AstmPiano(AstmMessage):
    SEGMENT_CLASS_MAP = {
        "H": PianoSegmentH,
        "P": PianoSegmentP,
        "Q": PianoSegmentQ,
        "O": PianoSegmentO,
        "R": PianoSegmentR,
        "L": PianoSegmentL,
    }
    DEFAULT_SEPARATORS = "|\^&"

    def create(self, **kwargs):
        message = PianoSegmentH(type_segment="H", **kwargs)
        message.end = PianoSegmentL(type_segment="L")
        return message

    def get_queries(self, message):
        return [q.tube for q in message.queries]


class MultipleAstmPiano(list):

    def __init__(self):
        self.parser = AstmPiano()
        super(MultipleAstmPiano, self).__init__()

    def add(self):
        msg = self.parser.create()
        self.append(msg)
        return msg

    @staticmethod
    def parse(data):

        message = MultipleAstmPiano()
        block = []

        for line in data.splitlines():
            if line[0] == "L" and block:
                block.append(line)
                msg = message.parser.parse("\r\n".join(block))
                message.append(msg)
                block = []
            else:
                block.append(line)

        assert block == [], "Unparsed data %s" % (block,)

        return message

    @staticmethod
    def to_string(message):
        return "\r\n".join([message.parser.to_string(msg) for msg in message])

    @staticmethod
    def get_queries(message):
        queries = []
        for msg in message:
            queries.extend(message.parser.get_queries(msg))

        return queries


class Driver(Message):
    SPEC = {
        "code": "piano",
        "label": "Piano",
    }

    def create(self):
        return MultipleAstmPiano()

    def get_message_type(self, message):
        for msg in message:
            if msg.queries:
                # TODO: Que faire en cas de QUERY + ORDER ?
                return self.QUERY

        return self.ORDER

    def to_string(self, message):
        return MultipleAstmPiano.to_string(message)

    def parse(self, data):
        return MultipleAstmPiano.parse(data)

    def get_queries(self, message):
        return MultipleAstmPiano.get_queries(message)

    def get_order_message(self, worklist, context):
        message = self.create()
        msg = message.add()
        for patient in worklist.get_astm_worklist(
            unique_together=["action"],
            action={"del": worklist.IGNORE, worklist.ELSE: ""},
        ):
            seg_p = msg.patients.add(
                no_patient=patient["patient_id"],
                nom=patient["last_name"],
                prenom=patient["first_name"],
                date_naissance=patient["birthdate"],
                sexe=patient["sex"],
            )
            for sample, action in patient["samples"]:
                seg_o = seg_p.orders.add(
                    tube=sample["sample_id"],
                )
                for order in sample["orders"]:
                    # TODO: À migrer dans MCA
                    for code_demande in order["code"].split(";"):
                        if code_demande and code_demande not in [
                            t.code for t in seg_o.tests
                        ]:
                            t = seg_o.tests.add()
                            t.code = code_demande

        if msg.patients or context["query"]:
            return message

    def get_result_dict(self, message):

        result = {}
        for msg in message:
            for seg_p in msg.patients:
                for seg_o in seg_p.orders:

                    if seg_o.specimen_type == "Control":
                        continue

                    result.setdefault(seg_o.tube, {"resultats": []})

                    for seg_r in seg_o.results:

                        if seg_r.status not in ("F", "P", ""):
                            continue

                        if seg_r.flag:
                            alarmes = [seg_r.flag]
                        else:
                            alarmes = []

                        result[seg_o.tube]["resultats"].append({
                            "test": seg_r.test,
                            "resu": seg_r.valeur,
                            "alarmes": alarmes,
                            "date": seg_r.date,
                            "analyseur": msg.emetteur,
                        })

        return result

    def get_qc_result_list(self, message):
        result = []
        for msg in message:
            for seg_p in msg.patients:

                for seg_o in seg_p.orders:

                    if seg_o.specimen_type != "Control":
                        continue

                    for seg_r in seg_o.results:

                        if seg_r.status not in ("F", "P", ""):
                            continue

                        result.append({
                            "lot": seg_o.tube,
                            "test": seg_r.test,
                            "resu": seg_r.valeur,
                            "date": seg_r.date,
                            "analyseur": msg.emetteur,
                        })

        return result

    def enregistrer(self, data, message=None):
        if data:
            lines = data.splitlines()
            if len(lines) == 3 and lines[1] == "Q|1|^ALL||ALL":
                return False
        return True
