# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng.astm import (
    AstmMessage,
    HeaderSegment,
    Segment,
    HprimString,
    HprimList,
    HprimDate,
    HprimForeign,
    HprimMultiple,
)


class PochSegmentH(HeaderSegment):
    sender_id = HprimMultiple(
        5,
        instrument_name=HprimString(1),
        software_version=HprimString(2),
        user_instrument=HprimString(6),
        ps_code=HprimString(7),
    )
    version = HprimString(13, default="E1394-97")

    patients = HprimForeign("P", "header")
    end = HprimForeign("L", "header", one2one=True, required=True)


class PochSegmentP(Segment):
    # le segment P est vide
    orders = HprimForeign("O", "patient")


class PochSegmentO(Segment):
    instrument_specimen_id = HprimMultiple(
        4,
        sample_id=HprimString(3),
        sample_id_attributes=HprimString(4),
    )
    analysis_parameter_id = HprimList(5, basefield=HprimMultiple(
        0,
        parameter_name=HprimString(5),
    ))
    action_code = HprimString(12, default="N")
    report_type = HprimString(26)
    results = HprimForeign("R", "order")


class PochSegmentR(Segment):
    parameter_name = HprimString((3, 5))
    dilution_ratio = HprimString((3, 6))
    value = HprimString(4)
    units = HprimString(5)
    ranges = HprimString(6)
    flag_error = HprimString(7)
    date = HprimDate(13)


class PochSegmentL(Segment):
    code = HprimString(3, default="N")


class Driver(AstmMessage):
    SPEC = {
        "code": "poch",
        "label": "Poch Sysmex",
    }
    SEGMENT_CLASS_MAP = {
        "H": PochSegmentH,
        "P": PochSegmentP,
        "O": PochSegmentO,
        "R": PochSegmentR,
        "L": PochSegmentL,
    }
    DEFAULT_SEPARATORS = "|\\^&"

    def create(self, **kwargs):
        message = PochSegmentH(type_segment="H", **kwargs)
        message.end = PochSegmentL(type_segment="L")
        return message

    def get_message_type(self, message):
        return self.ORDER

    def get_result_dict(self, message):
        results = {}
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                # exclusion des QC
                if seg_o.action_code == 'Q':
                    continue
                tube = seg_o.instrument_specimen_id.sample_id.strip()
                for seg_r in seg_o.results:
                    results.setdefault(
                        tube,
                        {"resultats": []})["resultats"].append(
                        {
                            "test": seg_r.parameter_name,
                            "date": seg_r.date,
                            "alarmes": seg_r.flag_error,
                            "resu": seg_r.value,
                        })
        return results

    def get_qc_result_list(self, message):
        result = []
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                # uniquement les QC
                if seg_o.action_code != "Q":
                    continue
                for seg_r in seg_o.results:
                    result.append({
                        "lot": seg_o.instrument_specimen_id.sample_id.strip(),
                        "test": seg_r.parameter_name,
                        "resu": seg_r.value,
                        "date": seg_r.date,
                    })
        return result
