# -*- coding: utf8 -*-
from __future__ import unicode_literals

import datetime
import xml.etree.cElementTree as ET

from driverng import Message
from driverng import utils


class Driver(Message):
    SPEC = {
        "code": "previisolaxml",
        "label": "Previ Isola XML",
    }
    ENCODING = "utf8"

    def parse(self, data):
        return ET.fromstring(data.encode("utf8"))

    def to_string(self, message):
        return utils.prettify_xml(ET.tostring(message, 'utf-8'))

    def get_message_type(self, message):
        return self.ORDER

    def get_order_message(self, worklist, context):
        root = ET.Element("lisMessage")
        root.set("xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance")
        root.set("xsi:noNamespaceSchemaLocation", "ASTM.xsd")
        header = ET.SubElement(root, "header")
        ET.SubElement(header, "version").text = "1394-97"
        ET.SubElement(header, "dateTime").text = context["now"].strftime(
            "%Y%m%d%H%M%S")
        empty = True
        for sample, _ in worklist.get_worklist(
            unique_together=["action"],
            action={"del": worklist.IGNORE, worklist.ELSE: None},
        ):
            request = ET.SubElement(root, "request")
            patient_information = ET.SubElement(request, "patientInformation")
            ET.SubElement(
                patient_information, "patientIdentifier"
            ).text = sample["patient_id"]
            ET.SubElement(
                patient_information, "lastName"
            ).text = sample["last_name"]
            ET.SubElement(
                patient_information, "firstName"
            ).text = sample["first_name"]

            if sample["birthdate"]:
                birthdate = sample["birthdate"].strftime('%Y%m%d')
            else:
                birthdate = ""
            ET.SubElement(patient_information, "birthDate").text = birthdate

            ET.SubElement(patient_information, "sex").text = sample["sex"]
            ET.SubElement(
                patient_information, "location"
            ).text = sample["location"]

            for order in sample["orders"]:
                test_order = ET.SubElement(request, "testOrder")
                specimen = ET.SubElement(test_order, "specimen")
                ET.SubElement(
                    specimen, "specimenIdentifier"
                ).text = sample["sample_id"]
                ET.SubElement(
                    specimen, "collectedDateTime"
                ).text = sample["collection_date"].strftime("%Y%m%d%H%M%S")
                ET.SubElement(
                    specimen, "specimenType"
                ).text = order["code"]

                ET.SubElement(
                    ET.SubElement(
                        ET.SubElement(test_order, "test"),
                        "universalIdentifier"
                    ),
                    "testIdentifier",
                ).text = "CULTURE"
                empty = False

        if not empty:
            return root

    def get_result_dict(self, message):
        # (sample_id, specimen_type) => {
        #   "BOITE": [],
        #   "INFO": [],
        # }
        resultats = {}
        for request in message.findall("requestResult"):
            for test_order in request.findall("testOrder"):
                specimen = test_order.find("specimen")
                sample_id = specimen.find("specimenIdentifier").text
                specimen_type = specimen.find("specimenType").text
                for test in test_order.findall("test"):
                    if test.find("universalIdentifier").find("testIdentifier").text != "CULTURE":
                        continue
                    result = test.find("result")
                    for disposable in result.find("bmxResultData").findall("disposable"):

                        date = datetime.datetime.strptime(
                            result.find("resultDateTime").text,
                            "%Y%m%d%H%M%S")

                        lot = disposable.findtext("lotNumber")
                        expire = disposable.findtext("expirationDate")
                        dispo_type = disposable.findtext("disposableType")
                        conso = disposable.findtext(
                            "disposableRelevantInformation")

                        n_boite = disposable.findtext("disposableNumber").split("/", 1)[0]

                        try:
                            n_boite = int(n_boite)
                        except ValueError:
                            code = "INFO"
                            resu = dispo_type
                            comm = "%s: lot %s expiration %s" % (
                                resu, lot, expire)
                        else:
                            code = "BOITE"
                            resu = disposable.findtext("reference")
                            comm = "%s (%d): lot %s expiration %s. %s %s" % (
                                resu, n_boite, lot, expire, conso, dispo_type)

                        resultats.setdefault((sample_id, specimen_type), {}).setdefault(
                            code, []).append({
                                "resu": resu,
                                "date": date,
                                "comm": comm,
                            })

        results = {}
        for (sample_id, specimen_type), values in resultats.items():
            # LEGACY
            for code, elements in values.items():
                date = None
                resu = []
                commentaires = []
                for e in elements:
                    if date is None:
                        date = e["date"]
                    resu.append(e["resu"])
                    commentaires.append(e["comm"])
                results.setdefault(
                    sample_id, {"resultats": []},
                )["resultats"].append({
                    "test": "%s_%s" % (code, specimen_type),
                    "resu": ";".join(resu),
                    "date": date,
                    "commentaires": commentaires,
                })

            date = None
            resu = []
            commentaires = []
            for code, elements in sorted(values.items()):
                for e in elements:
                    if date is None:
                        date = e["date"]
                    resu.append(e["resu"])
                    commentaires.append(e["comm"])
            results.setdefault(
                sample_id, {"resultats": []},
            )["resultats"].append({
                "test": "CULTURE_%s" % (specimen_type,),
                "resu": ";".join(resu),
                "date": date,
                "commentaires": commentaires,
            })
        return results
