# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng import Worklist
from driverng.astm import (
    AstmMessageV2,
    HeaderSegment,
    Segment,
    HprimString,
    HprimList,
    HprimDate,
    HprimForeign,
)
from driverng.utils import get_date, format_date


class PriscaSegmentH(HeaderSegment):
    emetteur = HprimString((5, 1), default='CLARISYS')
    emetteur_nom = HprimString((5, 2))
    type_message = HprimString(7, default='ORM')
    recepteur = HprimString((10, 1), default='SIEMENS')
    recepteur_nom = HprimString((10, 2))
    version = HprimString(13, default="A2.2")
    date = HprimDate(14)

    patients = HprimForeign("P", "header")
    end = HprimForeign("L", "header", required=True)


class PriscaSegmentP(Segment):
    no_patient = HprimString(3)
    id_patient = HprimString(4)
    nom = HprimString((6, 1))
    prenom = HprimString((6, 2))
    nomjf = HprimString(7)
    date_naissance = HprimDate(8, fmt_date="%Y%m%d")

    # M/F/U
    sexe = HprimString(9)

    orders = HprimForeign("OBR", "patient")


class PriscaSegmentOBR(Segment):
    tube = HprimString((3, 1), max_length=14)
    # tube +date
    tube_unique = HprimString((3, 2), max_length=14)
    tests = HprimList(5)
    # S: Stat
    # R: Routine
    urgence = HprimString(6, max_length=1)

    collection_date = HprimDate(8)
    action = HprimString(12, default="N")
    info_clinique = HprimString(14)
    date_prel = HprimDate(15)
    specimen_code = HprimString(16)
    medecin = HprimString((17, 2))
    service = HprimString((17, 5))

    results = HprimForeign("OBX", "order")


class PriscaSegmentOBX(Segment):

    test_type = HprimString((3, 1))
    test = HprimString((4, 1))
    test_name = HprimString((4, 2))
    result = HprimString(6)
    resu_status = HprimString(12)
    # date du contenant
    date = HprimDate(14)


class PriscaSegmentL(Segment):

    # terminaison
    code = HprimString(3, default="N")


class Driver(AstmMessageV2):
    SPEC = {
        "code": "prisca",
        "label": "Prisca Connect",
    }

    SEGMENT_CLASS_MAP = {
        "H": PriscaSegmentH,
        "P": PriscaSegmentP,
        "OBR": PriscaSegmentOBR,
        "OBX": PriscaSegmentOBX,
        "L": PriscaSegmentL,
    }
    DEFAULT_SEPARATORS = "|^~\\&"

    def create(self, **kwargs):
        message = PriscaSegmentH(type_segment="H", **kwargs)
        message.end = PriscaSegmentL(type_segment="L")
        return message

    def encode_worklist(self, worklist, context):
        if worklist.context != "order":
            raise NotImplementedError("context %s" % (worklist.context,))

        message = self.create(date=context["now"])
        for patient in worklist.get_astm_worklist(
            unique_together=["action"],
            action={"del": worklist.IGNORE, worklist.ELSE: "N"},
        ):
            seg_p = message.patients.add(
                no_patient=patient.patient_id,
                id_patient=patient.lab_patient_id,
                nom=patient.last_name,
                prenom=patient.first_name,
                nomjf=patient.maiden_name,
                date_naissance=patient.birthdate,
                sexe=patient.sex,
            )
            for sample, _ in patient.samples:
                seg_o = seg_p.orders.add(
                    tube=sample.sample_id,
                    tube_unique="%s%s" % (
                        sample.sample_id,
                        format_date(sample.collection_date, fmt="%Y%m%d"),
                    ),
                    specimen_code=sample.specimen,
                    collection_date=sample.collection_date,
                    date_prel=sample.collection_date,
                    info_clinique=sample.diagnostic,
                )
                if sample.physicians:
                    for physician in sample.physicians:
                        if physician.type == "prescriber":
                            seg_o.medecin = physician.code
                            break
                if sample.services:
                    seg_o.service = sample.services[0].code
                for order in sample.orders:
                    seg_o.tests.append(order.code)
                seg_o.results.add(
                    test_type="ST",
                    test="REF1",
                    result=seg_o.medecin,
                    resu_status="F",
                )
                for result in sample.results:
                    if not result.anterior:
                        resu = result.value
                        date = get_date(result.value)
                        if date is not None:
                            resu = format_date(date, fmt="%Y%m%d")
                            type_resultat = "DT"
                        else:
                            try:
                                float(result.value)
                                type_resultat = "NM"
                            except ValueError:
                                type_resultat = "CE"
                        seg_o.results.add(
                            test_type=type_resultat,
                            test=result.code,
                            result=resu,
                            resu_status="F",
                            date=result.date,
                        )
        if len(message.patients):
            return message

    def decode_message(self, message):
        if message.type_message != "ORU":
            raise NotImplementedError("type message %s" % (message.type_message,))

        worklist = Worklist(context="result")

        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                sample = worklist.get_or_create(type="patient", sample_id=seg_o.tube)
                for seg_r in seg_o.results:
                    if seg_r.test_type == "DT" and len(seg_r.result) == 8:
                        # Date sans heure il faut rajouter l'heure
                        # pour clarilab...
                        value = seg_r.result + "000000"
                    else:
                        value = seg_r.result
                    sample.results.add(
                        code=seg_r.test,
                        value=value,
                        date=seg_r.date or message.date,
                    )
        return worklist
