# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng import MessageV2
from driverng import Worklist
from driverng.exceptions import ParsingError
from driverng.fields import String


class PvsBase(object):
    FIELDS = ()

    def __init__(self, **kwargs):
        for code, field in self.FIELDS:
            setattr(self, field.name, kwargs.get(field.name, None))
        super(PvsBase, self).__init__()

    def parse(self, data):
        fields = dict(self.FIELDS)
        for elt in data.split("|"):
            if not elt:
                continue
            code = elt[:3]
            try:
                field = fields[code]
            except KeyError:
                pass
            else:
                value = field.to_python(elt[4:])
                setattr(self, field.name, value)

    def to_string(self):
        values = []
        for code, field in self.FIELDS:
            values.append("%s:%s" % (
                code,
                field.to_string(getattr(self, field.name)),
            ))
        return "%s|" % ("|".join(values), )


class PvsOrder(PvsBase):
    FIELDS = (
        ("TYP", String("typ", 3)),
        ("SID", String("tube", 16)),
        # a vérifier la taille
        ("NAM", String("nom", 16)),
        ("MAT", String("material_code", 4)),
        ("TST", String("tests", 64)),
    )


class PvsQuery(PvsBase):
    FIELDS = (
        ("TYP", String("typ", 3)),
        ("SID", String("tube", 16)),
    )


class PvsResult(PvsBase):
    FIELDS = (
        ("TYP", String("typ", 3)),
        ("SID", String("tube", 16)),
        ("POS", String("rack", 8)),
    )


class PvsMultipleMessage(list):
    _MESSAGE_TYPE_MAP = {
        "RQ": PvsOrder,
        "LA": PvsQuery,
        "WP": PvsResult,
    }

    def add(self, message_type=None, **kwargs):
        obj = self._MESSAGE_TYPE_MAP[message_type](
            message_type=message_type, **kwargs)
        self.append(obj)
        return obj

    def parse(self, data):
        for line in data.splitlines():
            if line:
                if line[:3] != "TYP":
                    raise ParsingError("Missing message type")
                obj = self._MESSAGE_TYPE_MAP[line[4:][:2]]()
                obj.parse(line)
                self.append(obj)

    def to_string(self):
        return "\n".join([
            e.to_string() for e in self
        ])


class Driver(MessageV2):
    SPEC = {
        "code": "pvs",
        "label": "Pvs Sarstedt",
    }

    MESSAGE_TYPE_MAP = {
        "RQ": PvsOrder,
        "LA": PvsQuery,
        "WP": PvsResult,
    }

    def create(self, message_type, **kwargs):
        return self.MESSAGE_TYPE_MAP[message_type](**kwargs)

    def to_string(self, message):
        return message.to_string()

    def parse(self, data):
        obj = PvsMultipleMessage()
        obj.parse(data)
        return obj

    def encode_worklist(self, worklist, context):
        samples = worklist.get_worklist(
            unique_together=["action"],
            action={"del": worklist.IGNORE, worklist.ELSE: ""},
        )
        if len(samples) == 0:
            return None
        elif len(samples) > 1:
            raise RuntimeError("Multi-sample not implemented")
        else:
            sample = samples[0][0]
            nom_patient = "%s_%s" % (sample.last_name, sample.first_name)
            msg = self.create(
                "RQ",
                typ="RQ",
                tube=sample.sample_id,
                nom=nom_patient,
                material_code=sample.specimen
            )
            # liste des tests
            tests = ",".join([order["code"] for order in sample.orders])
            msg.tests = "%s,%s" % (sample.specimen, tests)
            return msg

    def decode_message(self, message):
        for m in message:
            if m.typ == "LA":
                worklist = Worklist(context="query")
                if m.tube:
                    worklist.get_or_create(type="patient", sample_id=m.tube)
            elif m.typ == "WP":
                # pas de résultat réceptionné en production
                # c'est la position qui devrait être récupérée
                # manque d'exemple
                worklist = Worklist(context="result")
                worklist.get_or_create(
                    sample_id=m.tube,
                    rack=m.rack,
                    type="patient",
                )
            else:
                raise NotImplementedError("Type %s inconnu" % (m.typ,))

        return worklist
