# -*- coding: utf8 -*-
from __future__ import unicode_literals

import datetime

from driverng import MessageV2
from driverng import Worklist
from driverng.splitted import Splitted, String, Multiple, Date


class SedimaxBase(Splitted):
    SEPARATOR = "|"


class SedimaxResultSediment(SedimaxBase):
    BASE_RESULT_FIELDS = [
        String("type"),
        String("analyzer_id"),
        String("uid"),
        String("barcode"),
        String("barcode_flag"),
        String("patient_name"),
        String("rack"),
        String("tube"),
        String("operator"),
        Date("date", fmt_date="%Y/%m/%d"),
        Date("time", fmt_date="%H:%M:%S"),
        String("valid_status"),
        String("crowded_status"),
        String("dilution_factor"),
        String("comment"),
    ]
    RESULT4 = [
        Multiple("results", (
            String("code"),
            String("result1"),
            String("units1"),
            String("result2"),
            String("units2"),
            # resultat à l'unité micro litre
            String("result3"),
            String("units3"),
            String("result4"),
            String("units4"),
            String("mod_status"),
        ), separator="^")
    ]
    FIELDS = BASE_RESULT_FIELDS + RESULT4


class SedimaxResultSediment2(SedimaxBase):
    FIELDS = SedimaxResultSediment.BASE_RESULT_FIELDS + [
        # 0: Normal sample level
        # 1: Low sample level
        String("low_level"),
        # MEAS_TYPE^QC_LEVEL
        # MEAS_TYPE 0: Normal measurement
        # MEAS_TYPE 1: QC measurement
        # QC_LEVEL -1: Normal measurement
        # QC_LEVEL 0: Low level QC
        # QC_LEVEL 1: High level QC
        String("meas_type"),
        String("validator"),
    ] + [
        String("reserved%d" % (i,)) for i in range(7)
    ] + SedimaxResultSediment.RESULT4


class SedimaxResultChemical(SedimaxBase):
    FIELDS = (
        String("type"),
        String("analyzer_id"),
        String("uid"),
        String("barcode"),
        String("barcode_flag"),
        String("rack"),
        String("tube"),
        String("operator"),
        String("date"),
        String("time"),
        Multiple("results", (
            String("code"),
            # resultat de chimie
            String("result1"),
            String("units1"),
            String("result2"),
            String("units2"),
            String("result3"),
            String("units3"),
            String("mod_status"),
        ), separator="^")
    )


class SedimaxOrder(SedimaxBase):
    FIELDS = (
        String("type"),
        String("analyzer_id"),
        String("patient"),
        String("comment"),
    )


class SedimaxQS2(SedimaxBase):
    FIELDS = (
        String("type"),
        String("analyzer_id"),
        String("from_date"),
        String("to_date"),
    )


class Driver(MessageV2):
    SPEC = {
        "code": "sedimax",
        "label": "Sedimax",
    }

    MESSAGE_TYPE_MAP = {
        "W": SedimaxOrder,
        "S": SedimaxResultSediment,
        "S2": SedimaxResultSediment2,
        "CH": SedimaxResultChemical,
        "QS2": SedimaxQS2,
    }

    def __init__(self, analyzer_id="-", **kwargs):
        self.analyzer_id = analyzer_id
        super(Driver, self).__init__(**kwargs)

    def create(self, message_type, **kwargs):
        return self.MESSAGE_TYPE_MAP[message_type](**kwargs)

    def to_string(self, message):
        return message.to_string()

    def parse(self, data):
        message_type = data.split("|", 1)[0]
        obj = self.MESSAGE_TYPE_MAP[message_type]()
        obj.parse(data)
        return obj

    def encode_worklist(self, worklist, context):
        samples = worklist.get_worklist(
            unique_together=["action"],
            action={"del": worklist.IGNORE, worklist.ELSE: ""},
        )
        if len(samples) == 0:
            return None
        elif len(samples) > 1:
            raise RuntimeError("Multi-sample not implemented")
        else:
            sample = samples[0][0]
            info_patient = "%s^%s %s" % (sample.sample_id, sample.last_name, sample.first_name)
            msg = self.create(
                "W",
                type="W",
                analyzer_id=self.analyzer_id,
                patient=info_patient,
            )
        return msg

    def decode_message(self, message):
        if message.type not in ("S", "S2"):
            raise NotImplementedError("Type %s not implemented" % (message.type,))

        worklist = Worklist(context="result")
        if message.patient_name.startswith("QC_"):
            sample = worklist.get_or_create(
                type="control",
                control_lot=message.barcode,
                sample_id=message.barcode,
            )
        else:
            sample = worklist.get_or_create(
                type="patient",
                sample_id=message.barcode,
            )

        flags = []
        if message.valid_status == "0":
            flags.append("INVALID")
        if message.crowded_status == "1":
            flags.append("CROWDED")
        if message.type == "S2" and message.low_level == "1":
            flags.append("LOW_LEVEL")

        sample.rack = "%s-%s" % (message.rack, message.tube)
        date = datetime.datetime(*(
            message.date.timetuple()[:3] + message.time.timetuple()[3:5]))
        if message.comment and sample.type == "patient":
            sample.results.add(
                code="COMMENT",
                date=date,
                value=message.comment,
                flags=flags,
            )

        cristal = set(["-"])
        caox = ""
        for result in message.results:
            if (
                result.code in (".CaOxm", ".CaOxd")
                and result.result1 not in ("-", "N/A")
                and len(result.result1) > len(caox)
            ):
                caox = result.result1

        if caox:
            cristal.add(".CaOx%s" % (caox,))

        cylindre = set(["-"])

        for result in message.results:
            if result.code in ("CRY", ".CRY", "PAT", ".PAT", ".CaOxm", ".CaOxd"):
                continue
            if result.code in (
                ".TRI", ".URI", ".U-AMO", ".P-AMO",
            ):
                if result.result1 not in ("-", "N/A"):
                    cristal.add("%s%s" % (result.code, result.result1))
            elif result.code in (
                "HYA", ".C-GRA", ".C-WBC", ".C-RBC"
            ):
                if result.result1 not in ("-", "N/A"):
                    cylindre.add("%s%s" % (result.code, result.result1))
            else:
                for i, value in (
                    (1, result.result1),
                    (3, result.result3),
                ):
                    if result.code:
                        sample.results.add(
                            code="%s%s" % (result.code, i),
                            value=value,
                            date=date,
                            flags=flags,
                        )

        if sample.type == "patient":
            if len(cristal) > 1:
                cristal.remove("-")
            if len(cylindre) > 1:
                cylindre.remove("-")
            sample.results.add(
                code="CRISTAL",
                value=";".join(sorted(cristal)),
                date=date,
                flags=flags,
            )
            sample.results.add(
                code="CYLINDRE",
                value=";".join(sorted(cylindre)),
                date=date,
                flags=flags,
            )
        return worklist
