# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng.astm import (
    AstmMessage,
    HeaderSegment,
    Segment,
    HprimString,
    HprimList,
    HprimDate,
    HprimForeign,
    HprimMultiple,
)


class StreamlabSegmentH(HeaderSegment):
    message_control_id = HprimString(3)
    access_password = HprimString(4)
    sender_id = HprimString(5, default="NEO")
    receiver_id = HprimString(10, default="STREAMLAB")
    # P: Production
    # D: Debugging
    # Q: QC
    processing_id = HprimString(12, default="P")
    version = HprimString(13, default="DB 100102")
    date = HprimDate(14)
    patients = HprimForeign("P", "header")
    end = HprimForeign("L", "header", one2one=True, required=True)


class StreamlabSegmentP(Segment):
    practice_id = HprimString(3)
    last_name = HprimString((6, 1))
    first_name = HprimString((6, 2))
    mother_maiden_name = HprimString(7)
    birth_date = HprimDate(8, fmt_date="%Y%m%d")
    # M/F/U
    sex = HprimString(9)
    attending_physician = HprimString(14)
    orders = HprimForeign("O", "patient")


class StreamlabSegmentO(Segment):
    specimen_id = HprimString(3)
    universal_test_id = HprimList(5, basefield=HprimMultiple(
        0,
        code=HprimString(4),
        instrument_code=HprimString(5),
    ))
    priority = HprimString(6)
    ordered_date = HprimDate(7)
    collection_date = HprimDate(8)
    action_code = HprimString(12)
    specimen_descriptor = HprimString(16)
    sample_dilution = HprimString(19, default="1")
    total_number_tests = HprimString(20)
    # O: Order
    # F: Final
    # P: Preliminary
    # Q: Response to Query
    report_type = HprimString(26)


class StreamlabSegmentC(Segment):
    # seule valeur admise:
    #  I: Instrument/device
    source = HprimString(3)
    # pour la position du tube sur le rack pour le type sample, le format est:
    #   <code>^<tube>^<rack>
    comment = HprimMultiple(
        4,
        code=HprimString(1),
        # tube
        comment1=HprimString(2),
        # rack et position du tube
        comment2=HprimString(3),
    )
    # Les differents types possibles sont:
    #  I: Instrument flags
    #  G: General text
    #  E: Error comments
    #  T: Test Name comment
    #  A: Auxiliary information
    #  R: Reagent lot information
    #  S: Sample storage information
    #  C: Calibrator information
    #  Q: Quality control information
    comment_type = HprimString(5)


class StreamlabSegmentL(Segment):
    code = HprimString(3, default="N")


class Driver(AstmMessage):
    SPEC = {
        "code": "streamlab",
        "label": "Streamlab",
    }
    SEGMENT_CLASS_MAP = {
        "H": StreamlabSegmentH,
        "P": StreamlabSegmentP,
        "O": StreamlabSegmentO,
        "C": StreamlabSegmentC,
        "L": StreamlabSegmentL,
    }
    DEFAULT_SEPARATORS = "|\\^&"

    def create(self, **kwargs):
        message = StreamlabSegmentH(type_segment="H", **kwargs)
        message.end = StreamlabSegmentL(type_segment="L")
        return message

    def get_message_type(self, message):
        return self.ORDER

    def get_order_message(self, worklist, context):
        msg = self.create(date=context["now"])
        for patient in worklist.get_astm_worklist(
                unique_together=["action", "priority"],
                action={"del": "C", worklist.ELSE: "A"},
                priority={
                    "S": "S",
                    "A": "S",
                    worklist.ELSE: "R"
                },
        ):
            seg_p = msg.patients.add(
                practice_id=patient["patient_id"],
                last_name=patient["last_name"],
                first_name=patient["first_name"],
                mother_maiden_name=patient["maiden_name"],
                birth_date=patient["birthdate"],
                sex=patient["sex"],
                attending_physician=patient["physician"],
            )
            for sample, action, priority in patient["samples"]:
                seg_o = seg_p.orders.add(
                    specimen_id=sample["sample_id"],
                    priority=priority,
                    ordered_date=sample["request_date"],
                    collection_date=sample["collection_date"],
                    action_code=action,
                    specimen_descriptor=sample["specimen"],
                    report_type="O",
                )
                for order in sample["orders"]:
                    test = seg_o.universal_test_id.add()
                    if "^" in order["code"]:
                        test.code, test.instrument_code = order["code"].split(
                            "^", 1)
                    else:
                        test.code = order["code"]
                seg_o.total_number_tests = len(sample["orders"])

        if msg.patients:
            return msg

    def get_result_dict(self, message):
        results = {}
        if message.commentaires:
            if message.commentaires[0].comment.comment2:
                rack = message.commentaires[0].comment.comment2
            else:
                rack = " "
            results = {
                message.commentaires[0].comment.comment1: {
                    "resultats": [],
                    "rack": rack,
                }}
        return results
