# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng import (
    Message,
)
from driverng.utils import get_date, format_date


class Field(object):
    def __init__(self, name):
        self.name = name
        super(Field, self).__init__()

    def to_python(self, value):
        raise NotImplementedError

    def to_string(self, value):
        raise NotImplementedError


class String(Field):
    def __init__(self, name):
        super(String, self).__init__(name)

    def to_python(self, value):
        return value.strip()

    def to_string(self, value):
        if value is None:
            return ""
        else:
            return value


class Date(Field):
    def __init__(self, name, fmt):
        self.fmt = fmt
        super(Date, self).__init__(name)

    def to_python(self, value):
        return get_date(value, fmt=self.fmt)

    def to_string(self, value):
        return format_date(value, fmt=self.fmt)


class SwingBase(object):
    FIELDS = ()

    def __init__(self, **kwargs):
        for field in self.FIELDS:
            setattr(self, field.name, kwargs.get(field.name, None))
        super(SwingBase, self).__init__()

    def parse(self, data):
        fields = self.FIELDS
        nbMaxElementsLigne = len(fields)
        for idx, elt in enumerate(data.split("|")[:-1]):
            if idx >= nbMaxElementsLigne:
                break
            field = fields[idx]
            value = field.to_python(elt)
            setattr(self, field.name, value)

    def to_string(self):
        ligne = []
        for field in self.FIELDS:
            ligne.append("%s" % field.to_string(getattr(self, field.name)))
        return "%s|" % ("|".join(ligne))


class Order(SwingBase):
    FIELDS = (
        String("code_test"),
        String("no_echantillon"),
        String("no_donneur"),
        String("patient_id"),
        String("nom"),
        String("prenom"),
        String("nom_jeune_fille"),
        Date("date_naissance", fmt="%d/%m/%Y"),
        # F=0/M=1
        String("sexe"),
        # anterios facultatives
        # TODO: voir lors de la qualif s'il faut les générer
        # deux manieres pour les generer:
        # soit dans un même champ, le resultat brut code separe par des
        # espaces, par exemple:
        # '0000 1001 3102 4102 5002'
        # soit dans chaque champ, le resultat formaté comme souhaité par
        # l'automate, exemple:
        # |groupe A|Rh(D) négatif|CC ee|Kell positif|
        String("anterio_groupe"),
        String("anterio_rhesus"),
        String("anterio_pheno"),
        String("anterio_kell"),
    )


class Resultat(SwingBase):
    FIELDS = (
        String("no_echantillon"),
        String("code_test"),
        String("code_resultat"),
    )


class SwingMultiple(list):

    _MESSAGE_TYPE_MAP = {
        Message.ORDER: Order,
        Message.RESULT: Resultat,
    }

    def add(self, type_message=None, **kwargs):
        obj = self._MESSAGE_TYPE_MAP[type_message](**kwargs)
        self.append(obj)
        return obj

    def parse(self, data):
        for line in data.splitlines():
            if not len(line):
                continue
            if len(line.split("|")) == 14:
                type_message = Message.ORDER
            else:
                type_message = Message.RESULT
            obj = self.add(type_message=type_message)
            obj.parse(line)

    def to_string(self):
        return "\r\n".join([e.to_string() for e in self])


class Driver(Message):
    SPEC = {
        "code": "swing",
        "label": "Swing",
    }

    def create(self):
        return SwingMultiple()

    def get_message_type(self, message):
        return self.ORDER

    def to_string(self, message):
        return message.to_string()

    def parse(self, data):
        obj = SwingMultiple()
        obj.parse(data)
        return obj

    def get_order_message(self, worklist, context):
        msg = self.create()
        for sample in worklist.samples:
            if sample["sex"] == "F":
                sexe = 0
            else:
                # Default M
                sexe = 1
            for order in sample["orders"]:
                if order["action"] == "del":
                    continue
                msg.add(
                    type_message=self.ORDER,
                    code_test=order["code"],
                    no_echantillon=sample["sample_id"],
                    no_donneur="",
                    patient_id=sample["patient_id"],
                    nom=sample["last_name"],
                    prenom=sample["first_name"],
                    nom_jeune_fille=sample["maiden_name"],
                    date_naissance=sample["birthdate"],
                    sexe=sexe,
                    anterio_groupe="",
                    anterio_rhesus="",
                    anterio_pheno="",
                    anterio_kell="",
                )

        if len(msg):
            return msg

    def get_result_dict(self, message):
        results = {}
        for resultat in message:
            results.setdefault(
                resultat.no_echantillon,
                {"resultats": []},
            )["resultats"].append({
                "test": resultat.code_test,
                "resu": resultat.code_resultat,
            })
        return results
