# -*- coding: utf8 -*-
from __future__ import unicode_literals

import logging

from driverng.astm import (
    AstmMessage,
    HeaderSegment,
    Segment,
    HprimString,
    HprimDate,
    HprimList,
    HprimMultiple,
    HprimForeign,
)

logger = logging.getLogger("driverng.drivers.sysmex")


class SysmexSegmentH(HeaderSegment):

    version = HprimString(13, default="A.2")
    date = HprimDate(14, fmt_date="%Y%m%d%H%M")

    patients = HprimForeign("P", "header")
    controls = HprimForeign("S", "header")
    queries = HprimForeign("Q", "header")
    end = HprimForeign("L", "header", one2one=True, required=True)


class SysmexSegmentP(Segment):

    # Patient ID
    pid = HprimString(3, max_length=16)

    # Alternative Patiet ID
    # Use as Public insurance ID
    apid = HprimString(4, max_length=16)

    prenom = HprimString((6, 1), max_length=20)
    nom = HprimString((6, 2), max_length=20)
    nomjf = HprimString(7, max_length=20)
    date_naissance = HprimDate(8, fmt_date="%Y%m%d")

    # M/F/U
    sexe = HprimString(9, max_length=1)

    # Attending physician ID
    # Primary Doctore Code/Doctor Name
    doctor_code = HprimString((14, 1), max_length=6)
    doctor_name = HprimString((14, 2), max_length=20, trunc=True)

    # Patient height
    # cm (w/ decimals) or feet^inch
    taille = HprimString((17, 1), max_length=6)

    # Patient weight
    # kg (w/ decimals) or pound^ounce
    poids = HprimString((18, 1), max_length=6)

    # In/Out patient differentiation
    # OP: Out patient
    # IP: In patient
    status_admission = HprimString(25, max_length=2)

    # Location
    clinic_code = HprimString((26, 1), max_length=6)
    ward_code = HprimString((26, 2), max_length=6)

    # Date / Time registered
    # Data Registration Date
    # (Patient Data/ Latest Update Date)
    last_updated = HprimDate(33, fmt_date="%Y%m%d")

    orders = HprimForeign("OBR", "patient")


class SysmexSegmentOBR(Segment):

    # Requester specimen ID (HOST side)
    tube_demandeur = HprimString(3, max_length=29)

    # Producer speciment ID (SUIT side)
    tube_executant = HprimString(4, max_length=29)

    tests = HprimList(
        5,
        max_length=200,
        basefield=HprimMultiple(
            0,
            code=HprimString(1),
            nom=HprimString(2),
        ),
    )

    # Processing Priority level:
    # S: STAT/Urgent
    # TODO: peu clair...
    urgence = HprimString(6, max_length=1)

    # Specimen collection or observation date
    cldt = HprimDate(8, fmt_date="%Y%m%d%H%M")

    # Collection volume (Urine etc) Unit ml
    cvlm = HprimString(10, max_length=5)

    # Order Process Code
    # For new samples:
    # A: Add the ordered
    # L: Lab. to obtain a sample
    # For registered samples
    # A: Clear registred orders, register the order as new
    # L: Add orders
    code_action = HprimString(12, max_length=1)

    # Relevent Clinical Information
    rcinf = HprimString(14, max_length=48)

    # Registration Date
    order_date = HprimDate(15, fmt_date="%Y%m%d%H%M")

    # Date and time observation reported or status changed
    dtr = HprimDate(23, fmt_date="%Y%m%d%H%M")

    # Quantity timing / Order Comment Code
    qt = HprimString(28, max_length=6)

    results = HprimForeign("OBX", "order")


class SysmexSegmentOBX(Segment):

    # Test Result Date Format
    # ST: Characters
    # NM: Numeric value
    # CE: Code
    type_resultat = HprimString(3, max_length=2)

    # Test item
    # Code^Name
    test_code = HprimString((4, 1), max_length=30)
    test_name = HprimString((4, 2), max_length=30)

    valeur = HprimString((6, 1), max_length=100)

    # W is set:
    # - When WBC, LYMPH# and LYMPH% are compensated by NRBC
    # - When NEUT# and NEUT% are compensated by IG or NRBC
    # - When WBC is reported by using WBC-D and PLT is reported by using PLT-O
    comment_code = HprimString((6, 2), max_length=3)

    dilution_ratio = HprimString((6, 3), max_length=1)

    # Upper/Lower Flag ~ Delta Check Flag
    # L: Low
    # H: High
    # LL: Panic Low
    # HH: Panic High
    # >: out of linearity
    # W: low reliability
    # A: Abnormal (Except numeric value data)
    abnormal_flags = HprimList(9, max_length=5)

    # P: Preliminary Report
    # F: Final Result
    # I: Pending
    # C: Revision Report
    status = HprimString((12, 1), max_length=1)
    # Edit: input or edit
    # Validate: manual validate
    # Count: count at pad menu
    # Manual Send: Send manually
    # (N/A): Others
    latest_operation = HprimString((12, 2), max_length=20)

    # Sending date
    date = HprimDate(13, fmt_date="%Y%m%d%H%M")

    # Latest operator ID
    operateur = HprimString(17, max_length=6)


class SysmexSegmentQ(Segment):

    tubes = HprimList(4, max_length=200)
    date = HprimDate(7, fmt_date="%Y%m%d%H%M")


class SysmexSegmentS(Segment):

    # Analytical method (Manual or Closed)
    mode = HprimString(3, max_length=10)

    # analyzer Name or Instrument ID
    instrument_id = HprimString((4, 1), max_length=10)

    # Quality Control ("QC" fixed)
    qc = HprimString(7, max_length=2, default="QC")

    # Lot Number of Control Reagent
    ref_lot = HprimString(11, max_length=10)

    # Analysis Item Name
    test = HprimString(12, max_length=8)

    # QC Data
    result = HprimString(13, max_length=8)

    # Analysis Date and Time
    date = HprimDate(16, fmt_date="%Y%m%d%H%M")

    results = HprimForeign("OBX", "control")


class SysmexSegmentC(Segment):

    code1 = HprimString((3, 1))
    code2 = HprimString((3, 2))
    commentaire = HprimString((4, 1), max_length=100)

    def to_string(self, separators=None):
        segment_str = super(
            SysmexSegmentC, self).to_string(separators=separators)
        if self.commentaire[:3] == "PNG":
            # Cf note plus bas
            # segment_str = segment_str.replace("\\\\", "&R&")
            segment_str = segment_str.replace("/", "&R&")

        return segment_str


class SysmexSegmentL(Segment):

    nombre_segment_patient = HprimString(4)
    nombre_segment = HprimString(5)


class AstmSysmexBase(AstmMessage):
    SEGMENT_CLASS_MAP = {
        "H": SysmexSegmentH,
        "P": SysmexSegmentP,
        "OBR": SysmexSegmentOBR,
        "OBX": SysmexSegmentOBX,
        "Q": SysmexSegmentQ,
        "S": SysmexSegmentS,
        "C": SysmexSegmentC,
        "L": SysmexSegmentL,
    }
    DEFAULT_SEPARATORS = "|^~\\&"

    def __init__(self, legacy_mode_open=True, inverser_nom_prenom=False, **kwargs):
        self.legacy_mode_open = legacy_mode_open
        self.inverser_nom_prenom = inverser_nom_prenom
        super(AstmSysmexBase, self).__init__(**kwargs)

    def to_string(self, message, *args, **kwargs):
        message.end.nombre_segment_patient = len(message.patients)
        message.end.nombre_segment = len(message.get_segments())
        return super(AstmSysmexBase, self).to_string(message, *args, **kwargs)

    def parse_segment(self, line, separators):
        # \ et & ignorés
        line = line.replace("&R&", "/")
        line = line.replace(separators[3], "{0}E{0}".format(separators[3]))
        line = line.replace(separators[4], "{0}T{0}".format(separators[3]))
        return super(AstmSysmexBase, self).parse_segment(line, separators)

    def create(self, **kwargs):
        message = self.SEGMENT_CLASS_MAP["H"](type_segment="H", **kwargs)
        message.end = self.SEGMENT_CLASS_MAP["L"](type_segment="L")
        return message

    def get_message_type(self, message):
        if message.controls:
            return self.QC
        if message.queries:
            return self.QUERY
        return self.ORDER

    def get_queries(self, message):
        queries = []
        for seg_q in message.queries:
            queries.extend(seg_q.tubes)
        return queries

    def get_order_message(self, worklist, context):
        msg = self.create(date=context["now"])
        actions = {worklist.ELSE: "L"}
        if self.HANDLE_DELETED:
            actions["del"] = "C"
        else:
            actions["del"] = worklist.IGNORE
        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority"],
            action=actions,
            priority={"A": "A", "S": "S", worklist.ELSE: "R"},
        ):
            if (patient["physician_name"],
                    patient["physician_first_name"]) != (None, None):
                doctor_name = "%s %s" % (
                    patient["physician_name"] or "",
                    patient["physician_first_name"] or "",
                )
            else:
                doctor_name = None
            seg_p = msg.patients.add(
                pid=patient["patient_id"],
                apid=patient["patient_id"],
                nomjf=patient["maiden_name"],
                date_naissance=patient["birthdate"],
                sexe=patient["sex"],
                doctor_code=patient["physician"],
                doctor_name=doctor_name,
                clinic_code=patient["service"],
            )
            if self.inverser_nom_prenom:
                seg_p.nom = patient["first_name"]
                seg_p.prenom = patient["last_name"]
            else:
                seg_p.nom = patient["last_name"]
                seg_p.prenom = patient["first_name"]
            for sample, action, priority in patient["samples"]:
                seg_o = seg_p.orders.add(
                    tube_demandeur=sample["sample_id"],
                    urgence=priority,
                    code_action=action,

                    # SUIT+ only
                    doctor_code=sample["physician"],
                    clinic_code=sample["service"],
                )
                # TODO: C'est pour gérer un hack du paramêtrage,
                # plusieurs code_demande dans le code_demande.
                for order in sample["orders"]:
                    test = seg_o.tests.add()
                    test.code = order["code"]

        if msg.patients:
            return msg
        elif context["query"]:
            for sample_id in self.get_queries(context["query"]):
                seg_p = msg.patients.add()
                seg_p.orders.add(tube_demandeur=sample_id)
            return msg

    def get_tube_graphe(self, tube_obr, commentaire, tubes):
        """Extrait le tube et le graphe d'un champ commentaire
        @param commentaire: le champ à parser
        @param tubes: la liste des tubes potentiels
        @return (tube, code_test)
        """
        for tube in tubes:
            s = commentaire.split("_%s_" % (tube,))
            if len(s) == 2:
                return (tube, s[1])
        else:
            logger.warning(
                "Impossible d'extraire un tube connu dans '%s'" % (
                    commentaire))
            return (None, None)

    @staticmethod
    def get_lots_reactifs(seg_obx):
        raise NotImplementedError

    @staticmethod
    def ignore_segment_obx(seg_obx):
        raise NotImplementedError

    @staticmethod
    def get_flags(message):
        return {}

    def get_analyseur_rack(self, seg_obr):
        raise NotImplementedError

    def get_analyseur(self, analyseur, seg_obx):
        raise NotImplementedError

    def get_analyseur_qc(self, message):
        raise NotImplementedError

    def get_alarmes(self, seg_obx):
        raise NotImplementedError

    def get_result_dict(self, message):
        result = {}
        flags = self.get_flags(message)
        for seg_p in message.patients:
            for seg_obr in seg_p.orders:
                analyseur, rack = self.get_analyseur_rack(seg_obr)
                result.setdefault(seg_obr.tube_executant, {
                    "rack": rack,
                    "resultats": [],
                })

                for seg_obx in seg_obr.results:
                    if self.ignore_segment_obx(seg_obx):
                        continue
                    alarmes = self.get_alarmes(seg_obx)
                    alarmes.extend(flags.get(
                        seg_obr.tube_executant, {}
                    ).get(seg_obx.test_code, []))
                    result[seg_obr.tube_executant]["resultats"].append({
                        "test": seg_obx.test_code,
                        "resu": seg_obx.valeur,
                        "date": seg_obx.date,
                        "alarmes": alarmes,
                        "lots_reactifs": self.get_lots_reactifs(seg_obx),
                        "analyseur": self.get_analyseur(analyseur, seg_obx),
                    })

                # Gestion des commentaires sur tube (alarmes) et graphes
                for seg_obr in seg_p.orders:
                    commentaire_tube = []
                    for seg_obx in seg_obr.results:
                        for seg_c in seg_obx.commentaires:
                            if seg_c.commentaire[:3] == "PNG":
                                tube, code_test = self.get_tube_graphe(
                                    seg_obr.tube_executant,
                                    seg_c.commentaire,
                                    result.keys())
                                if tube is not None and code_test is not None:
                                    result[tube]["resultats"].append({
                                        "test": code_test,
                                        "resu": seg_c.commentaire,
                                    })
                            elif seg_c.commentaire:  # commentaire tube
                                commentaire_tube.append(seg_c.commentaire)

                    if commentaire_tube:
                        result[seg_obr.tube_executant]["resultats"].append({
                            "test": "ALARME",
                            "resu": "",
                            "commentaires": commentaire_tube,
                        })

        return result

    def get_qc_result_list(self, message):
        if message.controls:
            analyseur = self.get_analyseur_qc(message)
        else:
            return []

        result = []
        for seg_s in message.controls:

            # On virre le prefixe QC-
            if seg_s.ref_lot[:3] == "QC-":
                ref_lot = seg_s.ref_lot[3:]
            else:
                ref_lot = seg_s.ref_lot

            if seg_s.test in (
                "H_RACK", "U_RACK", "C_RACK",
                "H_TUBE", "U_TUBE", "C_TUBE",
                "H_INID", "U_INID", "C_INID",
                "H_INST", "U_INST", "C_INST",
            ):
                # On ignore ce segment
                continue

            if self.legacy_mode_open:
                if not analyseur and seg_s.mode == "Manual":
                    instrument = "%s-OPEN" % (seg_s.instrument_id,)
                else:
                    instrument = analyseur or seg_s.instrument_id
            else:
                instrument = analyseur or seg_s.instrument_id
                if seg_s.mode in ("Manual", "Open"):
                    instrument += "-OPEN"

            if len(ref_lot) == 8:
                # nom_lot + control level
                nom_lot = ref_lot[:4]
                control_level = ref_lot[4:]
            else:
                nom_lot = control_level = None

            result.append({
                "lot": ref_lot,
                "test": seg_s.test,
                "resu": seg_s.result,
                "date": seg_s.date,
                "analyseur": instrument,
                "nom_lot": nom_lot,
                "control_level": control_level,
            })

        if analyseur:
            for r in result:
                if not r["analyseur"]:
                    r["analyseur"] = analyseur

        return result
