# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng.astm import (
    HprimString,
    HprimMultiple,
    HprimList,
)
from driverng.drivers.sysmex.base import (
    SysmexSegmentH,
    SysmexSegmentP,
    SysmexSegmentOBR,
    SysmexSegmentOBX,
    AstmSysmexBase,
)


class SysmexPlusSegmentH(SysmexSegmentH):

    version = HprimString(13, default="SUITplus1.0")


class SysmexPlusSegmentP(SysmexSegmentP):

    # Seulement en SUIT+
    diag_code = HprimString((19, 1), max_length=20)
    diag_name = HprimString((19, 2), max_length=80)


class SysmexPlusSegmentOBR(SysmexSegmentOBR):

    # Order Process Code
    # L: Add orders
    # C: Cancel order
    code_action = HprimString(12, max_length=1)

    # CSF: CSF
    # ASCITES: Ascites
    # PLEURAL: Pleural Fluid
    # CAPD: CAPD Fluid
    # SYNOVIA: Synovial Fluid
    body_fluid = HprimString((16, 2))

    doctor_code = HprimString((17, 1), max_length=20)
    doctor_name = HprimString((17, 2), max_length=80)

    clinic_code = HprimString((19, 1), max_length=20)
    clinic_name = HprimString((19, 2), max_length=80)

    ward_code = HprimString((20, 1), max_length=20)
    ward_name = HprimString((20, 2), max_length=80)


class SysmexPlusSegmentOBX(SysmexSegmentOBX):

    # Only in SUIT+
    # R: Research
    # D: Diagnostic
    # S: Service
    parameter_classification = HprimString(5, max_length=1)

    # c is set when a parameter was re-calculated
    # & is set:
    # - When WBC, LYMPH# and LYMPH% are compensated by NRBC
    # - When NEUT# and NEUT% are compensated by IG or NRBC
    # - When WBC is reported by using WBC-D and PLT is reported by using PLT-O
    # or PLT-F
    comment_code = HprimString((6, 2), max_length=3)

    # Only in SUIT+
    reference_ranges = HprimString(8, max_length=100)

    # Reagent information (Only in SUIT+)
    reagents = HprimList(10, basefield=HprimMultiple(
        10,
        code=HprimString(1),
        lot=HprimString(2),
        serial=HprimString(3),
    ))

    # Only in SUIT+
    # In case results are produced by the manual differentiation or a result
    # was changed manually by the user -> OBX-16 is empty
    prod_analyser_type = HprimString((16, 1), max_length=25)
    prod_instrument_id = HprimString((16, 2), max_length=25)
    prod_rack = HprimString((16, 3), max_length=25)
    prod_tube = HprimString((16, 4), max_length=25)


class AstmSysmexPlus(AstmSysmexBase):
    SPEC = {
        "code": "sysmex_suit_plus",
        "label": "Sysmex SUIT+",
    }

    SEGMENT_CLASS_MAP = dict(AstmSysmexBase.SEGMENT_CLASS_MAP.items() + {
        "H": SysmexPlusSegmentH,
        "P": SysmexPlusSegmentP,
        "OBR": SysmexPlusSegmentOBR,
        "OBX": SysmexPlusSegmentOBX,
    }.items())
    HANDLE_DELETED = True

    @staticmethod
    def get_lots_reactifs(seg_obx):
        lots = []
        for reagent in seg_obx.reagents:
            lot = {
                "ref": reagent.lot,
                "nom": reagent.code,
            }
            if lot not in lots:
                lots.append(lot)
        return lots

    @staticmethod
    def ignore_segment_obx(seg_obx):
        # On ne garde que les résultats finaux
        if seg_obx.status == "F":
            return False
        else:
            return True

    def get_analyseur(self, analyseur, seg_obx):
        return "%s-%s" % (
            seg_obx.prod_analyser_type,
            seg_obx.prod_instrument_id,
        )

    def get_analyseur_qc(self, message):
        return None

    def get_analyseur_rack(self, seg_obr):
        rack = None
        for seg_obx in seg_obr.results:
            if seg_obx.prod_rack and seg_obx.prod_tube:
                rack = "%s-%s" % (
                    seg_obx.prod_rack,
                    seg_obx.prod_tube,
                )
        return None, rack

    def get_tube_graphe(self, tube_obr, commentaire, tubes):
        return (tube_obr, commentaire.rsplit('-')[-1].split('_', 1)[1])

    def get_alarmes(self, seg_obx):
        alarmes = seg_obx.abnormal_flags[:]
        if (seg_obx.test_code, seg_obx.comment_code) == ("PLT", "&"):
            alarmes.append("O")
        return alarmes

    def get_result_dict(self, message):
        results = super(AstmSysmexPlus, self).get_result_dict(message)
        for seg_p in message.patients:
            for seg_obr in seg_p.orders:
                indexes = {
                    "RBC": 0,
                    "WBC": 0,
                    "PLT": 0,
                    "WBCRBC": 0,
                }
                for seg_c in seg_obr.commentaires:
                    if (
                        seg_c.code2 == "MORPH"
                        and seg_c.code1 in ("WBC", "RBC", "PLT")
                    ):
                        if seg_c.code1 in ("WBC", "RBC"):
                            tests = (seg_c.code1, "WBCRBC")
                        else:
                            tests = (seg_c.code1,)
                        for test in tests:
                            indexes[test] += 1
                            results.setdefault(seg_obr.tube_executant, {
                                "resultats": []
                            })["resultats"].append({
                                "test": "%s_MORPH%d" % (test, indexes[test]),
                                "resu": seg_c.commentaire,
                            })
        return results
