# -*- coding: utf8 -*-
from __future__ import unicode_literals

import math

from driverng import Message, JSONGraphContainer
from driverng.fixed import (
    FixedLengthFormat,
    String,
    Multiple,
)

DOWN, UP = range(2)


def get_epaulements(points):
    atan = [math.atan(points[i] - points[i - 1]) for i in range(1, len(points))]
    epaulements = []
    for i in range(1, len(atan)):
        diff = atan[i] - atan[i - 1]
        if diff > 0:
            direction = UP
        else:
            direction = DOWN
        if not epaulements or epaulements[-1]["direction"] != direction:
            epaulements.append({
                "direction": direction,
                "x": i,
                "y": atan[i],
            })
    return epaulements


def get_graph_analysis(graph):
    sa1c, a0 = None, None
    for peak in graph["peaks"]:
        if peak["name"] == "SA1C":
            sa1c = peak
        if peak["name"] == "A0":
            a0 = peak

    results = []

    if sa1c:
        # Épaulement gauche
        ep_sa1c_left = 0
        epaulements = get_epaulements(graph["points"][sa1c["start"]:sa1c["top"]])
        for i in range(len(epaulements)):
            if [e["direction"] for e in epaulements[i:i + 3]] == [DOWN, UP, DOWN]:
                ep_sa1c_left = max(
                    ep_sa1c_left,
                    abs(epaulements[i + 1]["y"] - min(epaulements[i]["y"], epaulements[i + 2]["y"])),
                )

        # Épaulement droit
        ep_sa1c_right = 0
        epaulements = get_epaulements(graph["points"][sa1c["top"]:sa1c["end"]])
        for i in range(len(epaulements)):
            if [e["direction"] for e in epaulements[i:i + 3]] == [UP, DOWN, UP]:
                ep_sa1c_right = max(
                    ep_sa1c_right,
                    abs(epaulements[i + 1]["y"] - max(epaulements[i]["y"], epaulements[i + 2]["y"])),
                )

        results.extend([{
            "test": "A1C_EPA1G",
            "resu": unicode(round(math.degrees(ep_sa1c_left), 2))
        }, {
            "test": "A1C_EPA1D",
            "resu": unicode(round(math.degrees(ep_sa1c_right), 2)),
        }])

        # Hauteur en pourcentage du début SA1C par rapport à la
        # hauteur min du début du pic précédent jusqu'à la fin du
        # SA1C.
        # Permet de voir si le début du SA1C est un ceux bien marqué.
        peaks = sorted(graph["peaks"], cmp=lambda a, b: a["start"] - b["start"])
        names = [p["name"] for p in peaks]
        before_sa1c_index = names.index("SA1C") - 1
        if before_sa1c_index >= 0:
            peak = peaks[before_sa1c_index]
            vmin = min(graph["points"])
            points = [v - vmin for v in graph["points"]]
            height = points[sa1c["start"]]
            height_p = (height - min(points[peak["start"]:sa1c["end"]])) / height
            results.append({
                "test": "A1C_START_SA1C_HEIGHT_P",
                "resu": unicode(round(height_p * 100, 2)),
            })

    if sa1c and a0:
        # Aire en pourcentage par rapport à l'aire totale.
        # de la fin de SA1C jusqu'au dernier point avant le top A0 ou
        # la courbe a une tangente avec un angle inférieur à 45°.
        vmin = min(graph["points"])
        points = [v - vmin for v in graph["points"]]
        part = points[sa1c["end"]:a0["top"]]
        last = None
        for i in range(1, len(part)):
            if part[i] - part[i - 1] < 1:  # atan(1) = 45°
                last = sa1c["end"] + i

        total = sum(points)
        ep_a0_area = sum(points[sa1c["end"]:last])

        results.extend([{
            "test": "A1C_EPA0_P",
            "resu": unicode(round((ep_a0_area / total) * 100, 2)),
        }])

    return results


class SampleInformation(FixedLengthFormat):
    FMT_DESCRIPTION = [
        "1",
        # 0 = STD mode
        # 1 = VAR mode
        # 2 = beta-thalassemia mode
        String("operation_mode", 1, default="1"),
        # 0 = STAT
        # 1 = Transport (LA)
        # 2 = Loader
        String("sample_position", 1, default="1"),
        # /!\ Length 0, 13, 18 or 20
        String("sample_id", 20),
    ]


class SampleNumber(FixedLengthFormat):
    FMT_DESCRIPTION = [
        "2",
        # /!\ Length 3, 5 or 8
        # The number of digits used for the sample number is set by the
        # SMPx key in the RS232C screen of the analyzer
        String("sample_number", 8),
    ]


class MeasurementValue(FixedLengthFormat):
    FMT_DESCRIPTION = [
        "3",
        # 00-99
        # 00 == Normal
        String("flag_judgement", 2, default="00"),
        # /!\ Length 5 or 6
        # Spaces or 0.00 ~ 100.00
        # String("total_a1_percent", 6, right_justified=True),
        Multiple("components", [
            # Spaces or 0.00 ~ 100.00
            # Component | STD | VAR | Beta-thalassemia
            # 1 FP% A1a% (HbF%)
            # 2 A1a% A1b% (A0%)
            # 3 A1b% F% (HbA2%)
            # 4 F% LA1c% (HbD+%)
            # 5 LA1c% SA1c% or IFCC(*1) (HbS+%)
            # 6 SA1c% or IFCC(*1) A0% (HbC+%)
            # 7 A0% H-V0% (Reserved)
            # 8 (Reserved) H-V1% (Reserved)
            # 9 (Reserved) H-V2% (Reserved)
            # 10 Reserved or IFCC(*2) Reserved or IFCC(*2) (Reserved)
            # The number of digits used for each percent value obeys
            # (*2) PERCENT DECIMAL of the PARAMETER screen
            String("value", 6, right_justified=True),
        ]),
        # 0 ~ 99999
        # TP ??
        String("reserved", 5, right_justified=True),
    ]


class DataCollectionParameters(FixedLengthFormat):
    FMT_DESCRIPTION = [
        "4",
        # 0 ~ 9999 (x10mS)
        String("collection_start_time", 4, right_justified=True),
        # 0 ~ 9 (x100mS)
        String("collection_pitch", 1),
        # 0 ~ 20
        String("number_of_peaks", 2, right_justified=True),
        # 0 ~ 1000
        String("number_of_raw_data_sets", 4, right_justified=True),
    ]


class PeakData(FixedLengthFormat):
    FMT_DESCRIPTION = [
        "5",
        # 1 ~ 20
        String("number", 2, right_justified=True),
        String("name", 5),
        # B = base peak
        # V = valley peak
        # S = skimming peak
        String("shape", 1),

        # 0 ~ 1000
        String("base_start_point", 4, right_justified=True),
        String("base_end_point", 4, right_justified=True),
        String("peak_start_point", 4, right_justified=True),
        String("peak_top", 4, right_justified=True),
        String("peak_end_point", 4, right_justified=True),

        # 0.00 ~ 99999.99 (mV sec)
        String("peak_area", 8, right_justified=True),

        # /!\ Length 5 or 6
        # 0.00 ~ 100.00
        String("peak_area_percent", 6, right_justified=True),
    ]


class PeakDataEnd(FixedLengthFormat):
    FMT_DESCRIPTION = ["6"]


class RawData(FixedLengthFormat):
    FMT_DESCRIPTION = [
        "7",
        # 1 ~ 999
        String("block_number", 3, right_justified=True),
        Multiple("block_signals", [
            # -9999.999 ~ 9999.999 (mV)
            String("value", 9, right_justified=True),
        ]),
    ]


class CalibrationInformation(FixedLengthFormat):
    FMT_DESCRIPTION = [
        "8",
        # 0000 ~ 9999
        String("rack", 4),
        # 00 ~ 10
        String("position", 2),
        # factor A of element 1
        # factor B of element 1
        # factor A of element 2
        # factor B of element 2
        Multiple("factors", [
            String("value", 8, right_justified=True),
        ]),
    ]


class One(object):
    def __init__(self, name, klass):
        self.name = name
        self.klass = klass

    def get(self, **kwargs):
        return self.klass(**kwargs)


class ManyContainer(list):

    def __init__(self, klass):
        self.klass = klass
        list.__init__(self)

    def add(self, **kwargs):
        obj = self.klass(**kwargs)
        self.append(obj)
        return obj


class Many(One):

    def get(self):
        return ManyContainer(self.klass)


class _TosohMessage(object):
    FMT_MAP = {
        1: One("sample_information", SampleInformation),
        2: One("sample_number", SampleNumber),
        3: One("measurement_value", MeasurementValue),
        4: One("data_collection_parameters", DataCollectionParameters),
        5: Many("peak_datas", PeakData),
        6: One("peak_data_end", PeakDataEnd),
        7: Many("raw_datas", RawData),
        8: One("calibration_information", CalibrationInformation),
    }

    def __init__(self, **kwargs):
        for attr in self.FMT_MAP.values():
            setattr(self, attr.name, attr.get())
        super(_TosohMessage, self).__init__()

    @staticmethod
    def parse(data):
        obj = _TosohMessage()
        for line in data.splitlines():
            attr = obj.FMT_MAP[int(line[0])]
            value = getattr(obj, attr.name)
            if isinstance(value, ManyContainer):
                o = value.add()
                o.parse(line)
            else:
                o = value.parse(line)

        return obj

    def to_string(self):
        lines = []
        for key in range(1, 9):
            attr = self.FMT_MAP[key]
            value = getattr(self, attr.name)
            if isinstance(value, ManyContainer):
                lines.extend([v.to_string() for v in value])
            else:
                lines.append(value.to_string())
        return "\n".join(lines)


class Driver(Message):
    SPEC = {
        "code": "tosoh",
        "label": "Tosoh G8",
    }
    COMPONENT_MAP = {
        "0": (
            "FP%", "A1a%", "A1b%", "F%", "LA1c%", "SA1c%", "A0%", "",
            "", "", "", "TP"),
        "1": (
            "HbA1", "A1a%", "A1b%", "F%", "LA1c%", "SA1c%", "A0%",
            "H-V0%", "H-V1%", "H-V2%", "IFCC", "TP"),
        "2": (
            "(HbF%)", "(A0%)", "(HbA2%)", "(HbD+%)", "(HbS+%)",
            "(HbC+%)"),
    }

    def create(self):
        return _TosohMessage()

    def get_message_type(self, message):
        return self.ORDER

    def to_string(self, message, crlf="\r\n"):
        return message.to_string()

    def parse(self, data):
        return _TosohMessage.parse(data)

    def get_measurement_values(self, message):
        mode = message.sample_information.operation_mode
        values = [c.value for c in message.measurement_value.components]
        values.append(message.measurement_value.reserved)
        results = []
        for index, value in enumerate(values):
            try:
                test = self.COMPONENT_MAP[mode][index]
            except IndexError:
                continue
            if not test:
                continue
            results.append((test, value))
        return results

    def get_graphe(self, message):
        g = {
            "graph_type": "a1c",
            "points": [],
            "peaks": [],
        }
        cp = message.data_collection_parameters
        g["parameters"] = {
            "start_time": int(cp.collection_start_time),
            "pitch": int(cp.collection_pitch),
            "n_peaks": int(cp.number_of_peaks),
        }
        for raw in message.raw_datas:
            for point in raw.block_signals:
                v = float(point.value)
                g["points"].append(v)
        for peak in message.peak_datas:
            g["peaks"].append({
                "name": peak.name,
                "shape": peak.shape,
                "start": int(peak.peak_start_point),
                "top": int(peak.peak_top),
                "end": int(peak.peak_end_point),
                "area": float(peak.peak_area),
                "area_percent": float(peak.peak_area_percent)
            })
        results = get_graph_analysis(g)
        results.append({
            "test": "GRAPHE_A1C",
            "resu": "GRAPHE",
            "graph": JSONGraphContainer(g)
        })
        return results

    def get_legacy_graphe(self, message, alarmes):
        resu_tube = []
        cp = message.data_collection_parameters
        resu_tube.append({
            "test": "NbPick",
            "resu": cp.number_of_peaks,
            "alarmes": alarmes,
        })
        aireTotal = sum([float(p.peak_area) for p in message.peak_datas])
        resu_tube.append({
            "test": "areaTotal",
            "resu": "%.2f" % (aireTotal,),
            "alarmes": alarmes,
        })
        pourcTotal = sum([float(p.peak_area_percent) for p in
                         message.peak_datas])
        resu_tube.append({
            "test": "pourcTotal",
            "resu": "%.2f" % (pourcTotal,),
            "alarmes": alarmes,
        })

        tempsTotal = 0
        for peak in message.peak_datas:
            tr = (float(cp.collection_start_time) + (int(peak.peak_top) *
                  int(cp.collection_pitch))) / 600
            tempsTotal += tr
            resu_tube.append({
                "test": "tpr_%s" % (peak.name,),
                "resu": "%.4f" % (tr,),
                "alarmes": alarmes,
            })
            resu_tube.append({
                "test": "area_%s" % (peak.name,),
                "resu": "%.2f" % (float(peak.peak_area),),
                "alarmes": alarmes,
            })
            resu_tube.append({
                "test": "areap_%s" % (peak.name,),
                "resu": "%.2f" % (float(peak.peak_area_percent),),
                "alarmes": alarmes,
            })

        resu_tube.append({
            "test": "tempsTotal",
            "resu": "%.3f" % (tempsTotal,),
            "alarmes": alarmes,
        })
        return resu_tube

    def get_result_dict(self, message):
        tube = message.sample_information.sample_id[2:]

        if "QC" in tube:
            return {}

        results = {tube: {
            "resultats": [{
                "test": "RACK",
                "resu": message.sample_number.sample_number,
            }],
            "rack": message.sample_number.sample_number,
        }}

        resu_tube = results[tube]["resultats"]

        flag = message.measurement_value.flag_judgement
        alarmes = [flag] if flag != "00" else []

        if flag != "00":
            resu_tube.append({
                "test": "HB_ERR",
                "resu": flag,
            })

        for test, value in self.get_measurement_values(message):
            resu_tube.append({
                "test": test,
                "resu": value,
                "alarmes": alarmes,
            })

        resu_tube.extend(self.get_legacy_graphe(message, alarmes))
        resu_tube.extend(self.get_graphe(message))
        return results

    def get_qc_result_list(self, message):
        lot = message.sample_information.sample_id[2:]
        if "QC" not in lot:
            return []
        results = []

        for test, value in self.get_measurement_values(message):
            results.append({
                "lot": lot,
                "test": test,
                "resu": value,
                "date": None,
            })

        return results
