# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng.astm import (
    AstmMessage,
    HeaderSegment,
    Segment,
    HprimString,
    HprimMultiple,
    HprimList,
    HprimDate,
    HprimForeign,
)


class VariantSegmentH(HeaderSegment):
    sender_id = HprimMultiple(
        5,
        system_name=HprimString(1),
        instrument_name=HprimString(2),
    )
    processing_id = HprimString(12, default="P")
    date = HprimDate(14)
    patients = HprimForeign("P", "header")
    end = HprimForeign(
        "L", "header", one2one=True, required=True
    )


class VariantSegmentP(Segment):
    patient_id = HprimString(3)
    last_name = HprimString((6, 1))
    first_name = HprimString((6, 2))
    mother_maiden_name = HprimString(7)
    birthday = HprimDate(8, fmt_date="%Y%m%d")
    # M/F/U
    sex = HprimString(9)
    physician_id = HprimString(14)
    orders = HprimForeign("O", "patient")


class VariantSegmentO(Segment):

    # numero de tube ou lot
    # Les deux niveaux LC et HC présent sur les deux premiers caractères
    # de ce champ permettent l'identification du QC
    sample_id = HprimString((3, 1))

    # numero de rack du tube
    vial = HprimString((3, 2))

    # position du tube sur le rack
    replicate = HprimString((3, 3))
    universal_test_id = HprimList(5, basefield=HprimMultiple(
        0,
        test_id=HprimString(4),
    ))
    instrument_id = HprimString(25)
    results = HprimForeign("R", "order")


class VariantSegmentR(Segment):
    universal_test_id = HprimMultiple(
        3,
        peak_name=HprimString(4),
        component=HprimString(5),
    )
    value = HprimString(4)
    units = HprimString(5)

    # Not supported dans la doc
    # mais en réalité c'est renseigné alors que date_completed ne l'est pas...
    date_started = HprimDate(12)
    date_completed = HprimDate(13)
    instrument_identification = HprimString(14)


class VariantSegmentL(Segment):
    code = HprimString(3, default="N")


class Driver(AstmMessage):
    SPEC = {
        "code": "variant",
        "label": "Variant",
    }
    SEGMENT_CLASS_MAP = {
        "H": VariantSegmentH,
        "P": VariantSegmentP,
        "O": VariantSegmentO,
        "R": VariantSegmentR,
        "L": VariantSegmentL,
    }
    DEFAULT_SEPARATORS = "|\\^&"

    def create(self, **kwargs):
        message = VariantSegmentH(type_segment="H", **kwargs)
        message.end = VariantSegmentL(type_segment="L")
        return message

    def get_message_type(self, message):
        for patient in message.patients:
            for order in patient.orders:
                if order.sample_id[0:2] in ("HC", "LC"):
                    return self.QC
        return self.ORDER

    def get_order_message(self, worklist, context):
        msg = self.create(date=context["now"])
        # ignorer les suppressions car non gérées par le Variant
        for patient in worklist.get_astm_worklist(
                unique_together=["action"],
                action={"del": worklist.IGNORE},
        ):
            seg_p = msg.patients.add(
                patient_id=patient["patient_id"],
                last_name=patient["last_name"],
                first_name=patient["first_name"],
                birthday=patient["birthdate"],
                sex=patient["sex"],
                physician_id=patient["physician"],
            )
            for sample, action in patient["samples"]:
                seg_o = seg_p.orders.add(
                    sample_id=sample["sample_id"]
                )
                for order in sample["orders"]:
                    test = seg_o.universal_test_id.add()
                    test.test_id = order["code"]

        if msg.patients:
            return msg

    def get_result_dict(self, message):
        results = {}
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                # exclusion des QC
                if seg_o.sample_id[0:2] in ("HC", "LC"):
                    continue
                rack = "%s-%s" % (
                    seg_o.vial,
                    seg_o.replicate
                )
                for seg_r in seg_o.results:
                    test = "%s%s" % (
                        seg_r.universal_test_id.peak_name,
                        seg_r.universal_test_id.component,
                    )
                    results.setdefault(
                        seg_o.sample_id,
                        {"resultats": [], "rack": rack})["resultats"].append(
                        {
                            "test": test,
                            "resu": seg_r.value,
                            "date": seg_r.date_completed or seg_r.date_started,
                        })
        return results

    def get_qc_result_list(self, message):
        result = []
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                # uniquement les QC
                if seg_o.sample_id[0:2] not in ("HC", "LC"):
                    continue
                for seg_r in seg_o.results:
                    test = "%s%s" % (
                        seg_r.universal_test_id.peak_name,
                        seg_r.universal_test_id.component,
                    )
                    result.append({
                        "lot": seg_o.sample_id,
                        "test": test,
                        "resu": seg_r.value,
                        "date": seg_r.date_completed or seg_r.date_started,
                    })
        return result
