# -*- coding: utf8 -*-
from __future__ import unicode_literals

import datetime

from driverng import Message
from driverng.exceptions import ParsingError
from driverng.fields import String, Date


class VidasBCIBase(object):
    FIELDS = (
        ("mt", String("message_type", 3)),
        ("pi", String("no_patient", 16)),
        ("pn", String("nom_patient", 30)),
        ("pb", Date("date_naissance", fmt="%d/%m/%Y")),
        ("ps", String("sexe", 1)),
        ("so", String("origine", 1)),
        ("si", String("separateur", 0)),
        ("ci", String("tube", 12)),
        ("rt", String("code_test", 4)),
    )

    def __init__(self, **kwargs):
        for code, field in self.FIELDS:
            setattr(self, field.name, kwargs.get(field.name, None))
        super(VidasBCIBase, self).__init__()

    def parse(self, data):
        fields = dict(self.FIELDS)
        for elt in data.split("|"):
            if not elt:
                continue
            code = elt[:2]
            try:
                field = fields[code]
            except KeyError:
                pass
            else:
                value = field.to_python(elt[2:])
                setattr(self, field.name, value)

    def to_string(self):
        values = []
        for code, field in self.FIELDS:
            values.append("%s%s" % (
                code,
                field.to_string(getattr(self, field.name)),
            ))
        return "%s|" % ("|".join(values), )


class Order(VidasBCIBase):
    FIELDS = VidasBCIBase.FIELDS + (
        ("qd", String("dilution", 3)),
    )


class Resultat(VidasBCIBase):
    FIELDS = VidasBCIBase.FIELDS + (
        ("rn", String("nom_test", 40)),
        ("tt", Date("_time", fmt="%H:%M")),
        ("td", Date("_date", fmt="%d/%m/%Y")),
        ("ql", String("resultat_qualitatif", 29)),
        ("qn", String("resultat_quantitatif", 29)),
        ("y3", String("unite", 30)),
        ("qd", String("dilution", 3)),

        # certains vidas transmettent le flag de l'alarme
        # dans le champ du message_associe
        # flag alarme possible :
        # - valid : calibration valide
        # - calinc : calibration incomplete
        # - calexp : calibration expiree
        # - lotexp : lot expire
        # - contneff : controle non realise
        # - conthnor : controle hors normes
        # - conthwst : controle hors limites de Westgard
        ("nc", String("message_associe", 8)),
    )

    def __init__(self, date=None, **kwargs):
        if date is not None:
            kwargs.setdefault("_date", date)
            kwargs.setdefault("_time", date)
        super(Resultat, self).__init__(**kwargs)

    def parse(self, data):
        super(Resultat, self).parse(data)
        if self._time is not None and self._date is not None:
            self.date = datetime.datetime(
                *(self._date.timetuple()[:3] +
                    self._time.timetuple()[3:5]))
        elif self._date is not None:
            self.date = self._date
        else:
            self.date = None

    def to_string(self):
        if self.date is not None:
            self._time = self.date
            self._date = self.date
        return super(Resultat, self).to_string()


class VidasMultipleMessage(list):
    MESSAGE_TYPE_MAP = {
        "mpr": Order,
        "rsl": Resultat,
    }

    def add(self, message_type=None, **kwargs):
        obj = self.MESSAGE_TYPE_MAP[message_type](
            message_type=message_type, **kwargs)
        self.append(obj)
        return obj

    def parse(self, data):
        for line in data.splitlines():
            if line:
                if line[:2] != "mt":
                    raise ParsingError("Missing message type")
                obj = self.MESSAGE_TYPE_MAP[line[2:][:3]]()
                obj.parse(line)
                self.append(obj)

    def to_string(self):
        return "\n".join([
            e.to_string() for e in self
        ])


class Driver(Message):
    SPEC = {
        "code": "vidas",
        "label": "Vidas BCI",
    }

    def create(self):
        return VidasMultipleMessage()

    def get_message_type(self, message):
        return self.ORDER

    def to_string(self, message):
        return message.to_string()

    def parse(self, data):
        obj = VidasMultipleMessage()
        obj.parse(data)
        return obj

    def get_order_message(self, worklist, context):
        msg = self.create()

        for sample in worklist.samples:
            patient_name = "%(last_name)s %(first_name)s" % sample
            if sample["sex"] in ("M", "F"):
                sexe = sample["sex"]
            else:
                sexe = "-"
            for order in sample["orders"]:
                if order["action"] == "del":
                    continue
                msg.add(
                    message_type="mpr",
                    no_patient=sample["patient_id"],
                    nom_patient=patient_name,
                    date_naissance=sample["birthdate"],
                    sexe=sexe,
                    tube=sample["sample_id"],
                    code_test=order["code"],
                    dilution="1",
                )

        if len(msg):
            return msg

    @staticmethod
    def get_resultats(message):
        results = []
        for m in message:
            if m.message_associe:
                alarmes = [m.message_associe]
            else:
                alarmes = []
            if m.resultat_qualitatif:
                results.append((
                    m.tube,
                    # I pour interpreté
                    "%sI" % (m.code_test,),
                    m.resultat_qualitatif,
                    alarmes,
                    m.date,
                ))
            if m.resultat_quantitatif:
                # hack gérant la particularité des résultats quantitatif
                # transmis sur des formats variables:
                # exemple: "NN.NN ng/ml 3IS" ou "< N UA/ml"
                splitted = m.resultat_quantitatif.split(" ", 2)
                if splitted[0] in ("<", ">"):
                    resu = "%s %s" % (splitted[0], splitted[1])
                else:
                    resu = splitted[0]
                results.append((
                    m.tube,
                    m.code_test,
                    resu,
                    alarmes,
                    m.date,
                ))
        return results

    def get_result_dict(self, message):
        results = {}
        for tube, test, resu, alarmes, date in self.get_resultats(message):
            if tube[0] == "C":
                continue
            results.setdefault(tube, {"resultats": []})["resultats"].append({
                "test": test,
                "resu": resu,
                "alarmes": alarmes,
                "date": date,
            })
        return results

    def get_qc_result_list(self, message):
        results = []
        # la reference du lot est composée ainsi :
        # <C><niveau><espace><no_de_lot>
        for lot, test, resu, _, date in self.get_resultats(message):
            if lot[0] != "C":
                continue
            results.append({
                "lot": lot,
                "test": test,
                "resu": resu,
                "date": date,
            })
        return results
