# -*- coding: utf8 -*-
from __future__ import unicode_literals

import datetime
import xml.etree.cElementTree as ET

from driverng import Message
from driverng import utils


class Driver(Message):
    SPEC = {
        "code": "vidasxml",
        "label": "Vidas XML",
    }
    ENCODING = "utf8"

    def parse(self, data):
        return ET.fromstring(data.encode(self.ENCODING))

    def to_string(self, message):
        return utils.prettify_xml(ET.tostring(message, 'utf-8'))

    def get_message_type(self, message):
        return self.ORDER

    def get_order_message(self, worklist, context):
        root = ET.Element("lisMessage")
        root.set("xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance")
        root.set("xsi:noNamespaceSchemaLocation", "ASTM.xsd")
        header = ET.SubElement(root, "header")
        ET.SubElement(header, "version").text = "1394-97"
        ET.SubElement(header, "dateTime").text = context["now"].strftime(
            "%Y%m%d%H%M%S")
        empty = True
        for sample in worklist.samples:
            request = ET.SubElement(root, "request")
            patient_information = ET.SubElement(request, "patientInformation")
            ET.SubElement(
                patient_information, "patientIdentifier"
            ).text = sample["patient_id"]
            ET.SubElement(
                patient_information, "lastName"
            ).text = "%s %s" % (sample["last_name"], sample["first_name"])
            test_order = ET.SubElement(request, "testOrder")
            ET.SubElement(
                ET.SubElement(test_order, "specimen"),
                "specimenIdentifier",
            ).text = sample["sample_id"]
            for order in sample["orders"]:
                if order["action"] == "del":
                    continue
                ET.SubElement(
                    ET.SubElement(
                        ET.SubElement(test_order, "test"),
                        "universalIdentifier",
                    ),
                    "testIdentifier",
                ).text = order["code"]
                empty = False

        if not empty:
            return root

    def get_result_dict(self, message):
        results = {}
        for request in message.findall("requestResult"):
            for test_order in request.findall("testOrder"):
                sample_id = test_order.find(
                    "specimen").find("specimenIdentifier").text
                # Exclusion des CQ
                if sample_id[0] == 'C':
                    continue
                for test in test_order.findall("test"):
                    testid = test.find(
                        "universalIdentifier").find("testIdentifier").text
                    result = test.find("result")
                    resu = result.find("value").text
                    if "|" in resu:
                        resu, resui = resu.split("|", 1)
                        # resultat interprété
                        resultats = [
                            (testid, resu),
                            ("%sI" % testid, resui)
                        ]
                    else:
                        resultats = [(testid, resu)]
                    date = datetime.datetime.strptime(
                        result.find("resultDateTime").text,
                        "%Y%m%d%H%M%S",
                    )
                    if result.find("abnormalFlags") is None:
                        alarme = ""
                    else:
                        alarme = result.find("abnormalFlags").text

                    lots_reactifs = []
                    for bmxResultData in result.findall("bmxResultData"):
                        for disposable in bmxResultData.findall("disposable"):
                            lot = disposable.findtext("lotNumber")
                            if lot:
                                lot = {"ref": lot}
                                expire = disposable.findtext("expirationDate")
                                if expire:
                                    lot.update(
                                        peremption=datetime.datetime.strptime(
                                            expire, "%Y%m%d"))
                                lots_reactifs.append(lot)

                    for code, resu in resultats:
                        results.setdefault(
                            sample_id, {"resultats": []},
                        )["resultats"].append({
                            "test": code,
                            "resu": resu,
                            "date": date,
                            "alarmes": [alarme],
                            "lots_reactifs": lots_reactifs,
                        })
        return results

    def get_qc_result_list(self, message):
        results = []
        for request in message.findall("requestResult"):
            for test_order in request.findall("testOrder"):
                sample_id = test_order.find(
                    "specimen").find("specimenIdentifier").text
                # Uniquement les CQ
                if sample_id[0] != 'C':
                    continue
                for test in test_order.findall("test"):
                    testid = test.find(
                        "universalIdentifier").find("testIdentifier").text
                    result = test.find("result")
                    resu = result.find("value").text
                    if "|" in resu:
                        resu, resui = resu.split("|", 1)
                        # resultat interprété
                        resultats = [
                            (testid, resu),
                            ("%sI" % testid, resui)
                        ]
                    else:
                        resultats = [(testid, resu)]
                    date = datetime.datetime.strptime(
                        result.find("resultDateTime").text,
                        "%Y%m%d%H%M%S",
                    )
                    for code, resu in resultats:
                        results.append({
                            "lot": sample_id,
                            "test": code,
                            "resu": resu,
                            "date": date,
                        })
        return results
