# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging
from collections import OrderedDict
from driverng import Message
from driverng.utils import get_date, format_date
from driverng.exceptions import ParsingError


logger = logging.getLogger("driverng.drivers.vitek")


class Field(object):

    def __init__(self, name, default=None):
        self.name = name
        self.default = default
        super(Field, self).__init__()

    def get_default(self):
        return self.default

    def to_python(self):
        raise NotImplementedError

    def to_string(self):
        raise NotImplementedError


class String(Field):

    def __init__(self, name, default=None, max_length=None, required=False):
        self.max_length = max_length
        self.required = required
        super(String, self).__init__(name, default=default)

    def to_python(self, value):
        return value

    def to_string(self, value):
        if self.max_length is not None:
            return value[:self.max_length]
        if value is None:
            return ""
        return value


class UpperString(String):

    def to_python(self, value):
        return value.upper()

    def to_string(self, value):
        return value.upper()


class Number(Field):

    def to_python(self, value):
        return int(value)

    def to_string(self, value):
        return str(value)


class Date(Field):

    def __init__(self, name, default=None, fmt="%d/%m/%Y"):
        self.fmt = fmt
        super(Date, self).__init__(name, default)

    def to_python(self, value):
        return get_date(
            value,
            fmt=self.fmt,
        )

    def to_string(self, value):
        return format_date(value, fmt=self.fmt)


class Container(object):

    def __init__(self, klass):
        self.klass = klass
        self.items = []

    def add(self, message):
        obj = self.klass().parse(message)
        self.items.append(obj)
        return obj

    def create(self, **kwargs):
        obj = self.klass(**kwargs)
        self.items.append(obj)
        return obj

    def to_string(self):
        return BCIMessage.SEPATATOR.join(
            [x.to_string() for x in self.items])

    def __iter__(self):
        for x in self.items:
            yield x


class BCIMessage(object):

    FIELDS = {}

    SEPATATOR = "|"
    END_MESSAGE = "zz"
    RESULT_SEPARATOR = "ra"
    SPECIMEN_SEPARATOR = "si"
    TEST_SEPARATOR = "ta"
    CULTURE_SEPARATOR = "ci"

    def __init__(self, **kwargs):
        for field in self.FIELDS.values():
            setattr(self,
                    field.name,
                    kwargs.get(field.name, field.get_default()))
        super(BCIMessage, self).__init__()

    def parse(self, data):
        splitted = data.split("|")
        for part in splitted:
            key = part[:2]

            field = self.FIELDS.get(key)
            if field:
                setattr(self, field.name, field.to_python(part[2:]))

        return self

    def to_string(self):
        data = []
        for key, field in self.FIELDS.items():
            value = getattr(self, field.name, None)
            if (
                (value is not None and value != "")
                or getattr(field, "required", False)
            ):
                data.append("%s%s" % (
                    key, field.to_string(value),
                ))
        return BCIMessage.SEPATATOR.join(data)

    def _make_string(self, *args):
        return BCIMessage.SEPATATOR.join([x for x in args if x])

    def _parse_with_container(self, data, name, separator, insert_first=False):
        container = getattr(self, name)
        splitted = data.split("|")
        items = []
        for part in splitted:
            key = part[:2]

            if key == separator:
                if items:
                    container.add(BCIMessage.SEPATATOR.join(items))
                items = [part]
                continue

            if key not in self.FIELDS and items:
                items.append(part)
                continue

            field = self.FIELDS.get(key)
            if field:
                setattr(self, field.name, field.to_python(part[2:]))
            else:
                logger.warning("champs inconnu : %s", key)

        if items:
            container.add(BCIMessage.SEPATATOR.join(items))

        return self


class VitekBaseMessage(BCIMessage):

    GENERAL_FIELDS = (
        ("id", String("instrument_id", max_length=3)),
        ("ii", String("instrument_code", max_length=2)),
        ("is", String("instrument_serial_number", max_length=80)),
        ("it", String("test_group_code", max_length=2)),
    )

    # level patient
    PATIENT_FIELDS = (
        ("pi", String("patient_id", max_length=16)),  # FIRST
        ("pn", UpperString("patient_name", max_length=40)),
        ("pb", Date("patient_birthdate", fmt="%d/%m/%Y")),
        ("ps", String("patient_sex", max_length=10)),

        # Service executant
        ("pl", String("patient_location_code", max_length=6)),
        ("p2", String("patient_location_text", max_length=32)),
        ("px", String("patient_service_code", max_length=6)),
        ("p3", String("patient_service_text", max_length=20)),

        ("po", String("patient_diagnosis_code", max_length=6)),
        ("p4", String("patient_diagnosis_text", max_length=40)),

        # Prescripteur
        ("pp", String("patient_physician_code", max_length=8)),
        ("p5", String("patient_physician_text", max_length=30)),

        ("pa", Date("patient_admission_date", fmt="%d/%m/%Y")),
        ("pd", String("patient_dosage_group_code", max_length=6)),
        ("p6", String("patient_dosage_group_text", max_length=40)),
        ("pc", String("patient_comment_code", max_length=6)),
        ("pt", String("patient_comment_text", max_length=800)),
        ("pv", String("patient_alternate_id", max_length=16)),
        ("w1", String("patient_wild_field1", max_length=40)),
        ("w2", String("patient_wild_field2", max_length=40)),
        ("w3", String("patient_wild_field3", max_length=40)),
        ("w4", String("patient_wild_field4", max_length=40)),
        ("w5", String("patient_wild_field5", max_length=40)),
        ("w6", String("patient_wild_field6", max_length=40)),
        ("w7", String("patient_wild_field7", max_length=40)),
        ("w8", String("patient_wild_field8", max_length=40)),
        ("ua", String("patient_user_field_1_code", max_length=6)),
        ("ub", String("patient_user_field_1_text", max_length=6)),
        ("uc", String("patient_user_field_2_code", max_length=6)),
        ("ud", String("patient_user_field_2_text", max_length=6)),
        ("ue", String("patient_user_field_3_code", max_length=6)),
        ("uf", String("patient_user_field_3_text", max_length=6)),
        ("u1", String("patient_user_field_4_code", max_length=6)),
        ("u2", String("patient_user_field_4_text", max_length=6)),
        ("u3", String("patient_user_field_5_code", max_length=6)),
        ("u4", String("patient_user_field_5_text", max_length=6)),
    )

    FIELDS = OrderedDict(GENERAL_FIELDS + PATIENT_FIELDS)

    def __init__(self, **kwargs):
        self.specimen_list = Container(VitekSpecimenMessage)
        super(VitekBaseMessage, self).__init__(**kwargs)

    def parse(self, data):
        return self._parse_with_container(data,
                                          "specimen_list",
                                          BCIMessage.SPECIMEN_SEPARATOR)

    def to_string(self):
        return BCIMessage.SEPATATOR.join([
            self.message_type,
            super(VitekBaseMessage, self).to_string(),
            self.specimen_list.to_string(),
            BCIMessage.END_MESSAGE
        ])


class VitekSpecimenMessage(BCIMessage):
    FIELDS = OrderedDict([
        ("si", String("specimen_separator", required=True)),
        ("s0", String("relative_specimen_number", max_length=8)),
        ("ss", String("specimen_source_code", max_length=6)),
        ("s5", String("specimen_source_text", max_length=26)),
        ("st", String("specimen_site_code", max_length=6)),
        ("s6", String("specimen_site_text", max_length=26)),

        # Prescripteur
        ("sp", String("specimen_physician_code", max_length=8)),
        ("s8", String("specimen_physician_text", max_length=30)),

        # Service executant
        ("sl", String("specimen_location_code", max_length=6)),
        ("s7", String("specimen_location_text", max_length=32)),
        ("sx", String("specimen_service_code", max_length=6)),
        ("s9", String("specimen_service_text", max_length=20)),

        ("sy", String("specimen_status_code", max_length=2)),
        ("sz", String("specimen_status_text", max_length=16)),
        ("s1", Date("specimen_collection_date", fmt="%d/%m/%Y")),
        ("s2", Date("specimen_collection_time", fmt="%H:%M")),
        ("s3", Date("specimen_receipt_date", fmt="%d/%m/%Y")),
        ("s4", Date("specimen_receipt_time", fmt="%H:%M")),
        ("sf", String("patient_temperature", max_length=5)),
        ("so", String("collection_number", max_length=8)),
        ("sc", String("specimen_comment_code", max_length=6)),
        ("sn", String("specimen_comment_text", max_length=800)),
        ("ug", String("specimen_user_field_1_code", max_length=6)),
        ("uh", String("specimen_user_field_1_text", max_length=40)),
        ("ui", String("specimen_user_field_2_code", max_length=6)),
        ("uj", String("specimen_user_field_2_text", max_length=40)),
        ("uk", String("specimen_user_field_3_code", max_length=6)),
        ("ul", String("specimen_user_field_3_text", max_length=40)),
        ("um", String("specimen_user_field_4_code", max_length=6)),
        ("un", String("specimen_user_field_4_text", max_length=40)),
        ("uo", String("specimen_user_field_5_code", max_length=6)),
        ("up", String("specimen_user_field_5_text", max_length=40)),
        ("uq", String("specimen_user_field_6_code", max_length=6)),
        ("ur", String("specimen_user_field_6_text", max_length=40)),
        ("us", String("specimen_user_field_7_code", max_length=6)),
        ("ut", String("specimen_user_field_7_text", max_length=40)),
        ("uu", String("specimen_user_field_8_code", max_length=6)),
        ("uv", String("specimen_user_field_8_text", max_length=40)),
        ("sa", String("physician_address_line1", max_length=40)),
        ("sb", String("physician_address_line2", max_length=40)),
        ("se", String("physician_address_line3", max_length=40)),
        ("sd", String("physician_address_line4", max_length=40)),
        ("wa", String("specimen_wild_field_1", max_length=40)),
        ("wb", String("specimen_wild_field_2", max_length=40)),
        ("wc", String("specimen_wild_field_3", max_length=40)),
        ("wd", String("specimen_wild_field_4", max_length=40)),
        ("we", String("specimen_wild_field_5", max_length=40)),
    ])

    def __init__(self, **kwargs):
        self.culture_list = Container(VitekCultureMessage)
        super(VitekSpecimenMessage, self).__init__(**kwargs)

    def parse(self, data):
        return self._parse_with_container(data,
                                          "culture_list",
                                          BCIMessage.CULTURE_SEPARATOR,
                                          insert_first=True)

    def to_string(self):
        return self._make_string(super(VitekSpecimenMessage, self).to_string(),
                                 self.culture_list.to_string())


class VitekCultureCommentCode(BCIMessage):
    FIELDS = OrderedDict([
        ("na", String("culture_comment_code", max_length=6)),
    ])


class VitekCultureCommentText(BCIMessage):
    FIELDS = OrderedDict([
        ("nb", String("culture_comment_text", max_length=800)),
    ])


class VitekCultureMessage(BCIMessage):
    FIELDS = OrderedDict([
        ("ci", String("culture_id", max_length=20)),  # FIRST
        ("c0", String("exam_number", max_length=8)),
        ("ct", String("culture_type_code", max_length=6)),
        ("cn", String("culture_type_text", max_length=26)),
        ("c1", Date("culture_completion_date", fmt="%d/%m/%Y")),
        ("c2", Date("culture_completion_time", fmt="%H:%M")),
        ("c3", String("culture_status_code", max_length=2)),
        ("c4", String("culture_status_text", max_length=16)),
        ("x1", String("culture_wild_field1", max_length=40)),
        ("x2", String("culture_wild_field2", max_length=40)),
    ])

    def __init__(self, **kwargs):
        self.test_list = Container(VitekTestMessage)
        self.comment_code_list = Container(VitekCultureCommentCode)
        self.comment_text_list = Container(VitekCultureCommentText)
        super(VitekCultureMessage, self).__init__(**kwargs)

    def parse(self, data):
        splitted = data.split("|")
        items = []
        keys = []
        for part in splitted:
            key = part[:2]

            if key == BCIMessage.END_MESSAGE:
                continue

            if key not in self.FIELDS:
                if key == "na":
                    self.comment_code_list.add(part)
                elif key == "nb":
                    self.comment_text_list.add(part)
                else:
                    if key in keys:
                        self.test_list.add(BCIMessage.SEPATATOR.join(items))
                        items = [part]
                        keys = []
                    else:
                        items.append(part)
                    if key in VitekTestMessage.FIELDS:
                        keys.append(key)
                continue

            field = self.FIELDS.get(key)
            if field:
                setattr(self, field.name, field.to_python(part[2:]))

        if items:
            self.test_list.add(BCIMessage.SEPATATOR.join(items))
        else:
            logger.warning("champs inconnu : %s", key)

        return self

    def to_string(self):
        return self._make_string(
            super(VitekCultureMessage, self).to_string(),
            self.comment_code_list.to_string(),
            self.comment_text_list.to_string(),
            self.test_list.to_string())


class VitekTestMessage(BCIMessage):
    FIELDS = OrderedDict([
        ("ta", String("test_separator", required=True)),
        ("rt", String("test_type_code", max_length=8)),
        # Repeatable for blood test
        ("t1", String("isolate_number", max_length=2)),
        ("rn", String("test_type_name", max_length=40)),
        ("rr", String("relative_test_number", max_length=8)),
        ("ti", String("instrument_system_code", max_length=2)),
        ("tj", String("instrument_name", max_length=16)),
        ("tc", String("instrument_comment_code", max_length=8)),  # Repeatable
        # Repeatable
        ("tn", String("instrument_comment_text", max_length=800)),
        ("t2", String("test_group_code", max_length=2)),
        ("t3", String("test_group_name", max_length=16)),
        ("t4", String("test_status_code", max_length=2)),
        ("t5", String("test_status_name", max_length=16)),
        ("m3", String("technician_code", max_length=6)),
        ("m4", String("technician_name", max_length=30)),
        ("nc", String("test_comment_code", max_length=6)),  # Repeatable
        ("nd", String("test_comment_text", max_length=800)),  # Repeatable
        ("y1", String("test_wild_text1", max_length=40)),
        ("y2", String("test_wild_text2", max_length=40)),
        ("y3", String("test_wild_text3", max_length=40)),
        ("y4", String("test_wild_text4", max_length=40)),
        # FIRST pour blood culture
        ("tb", String("bottle_bar_code", max_length=10)),
        ("tt", String("aes_free_text", max_length=10000)),
        ("t6", String("preliminary_organism_code", max_length=6)),
        ("t7", String("preliminary_organism_name", max_length=60)),
        ("t8", String("preliminary_modifier_code", max_length=6)),
        ("t9", String("preliminary_modifier_name", max_length=20)),
        # Repeatable for blood test
        ("o1", String("final_organism_code", max_length=6)),
        ("o2", String("final_organism_name", max_length=60)),
        ("o3", String("final_bionumber", max_length=16)),
        ("o4", String("final_modifier_code", max_length=6)),
        ("o5", String("final_modifier_name", max_length=20)),
        ("o6", String("organism_group_code", max_length=6)),
        ("o7", String("organism_group_name", max_length=30)),
        ("o9", String("percent_probability", max_length=8)),
        ("rd", Date("result_date", fmt="%d/%m/%Y")),
        ("ru", Date("result_time", fmt="%H:%M")),
        ("re", String("elapsed_hours", max_length=8)),
        ("rc", String("instrument_result_code", max_length=2)),
        ("ri", String("instrument_result_name", max_length=20)),
        ("rj", String("blood_test_final_result_code", max_length=2)),
        ("rk", String("blood_test_final_result_name", max_length=20)),
    ])

    def __init__(self, **kwargs):
        self.result_list = Container(VitekAntibioticMessage)
        self.antibiotic_family = Container(VitekAntibioticFamily)
        super(VitekTestMessage, self).__init__(**kwargs)

    def parse(self, data):
        splitted = data.split("|")
        items = []
        keys = []
        for part in splitted:
            key = part[:2]

            if key and key not in self.FIELDS:
                if key == "af":
                    self.antibiotic_family.add(part)
                elif key == "ap":
                    try:
                        self.antibiotic_family.items[-1].phenotype_list.append(part[2:])
                    except IndexError:
                        pass
                else:
                    if key in keys:
                        self.result_list.add(BCIMessage.SEPATATOR.join(items))
                        items = [part]
                        keys = []
                    else:
                        items.append(part)
                    keys.append(key)
                continue

            field = self.FIELDS.get(key)
            if field:
                setattr(self, field.name, field.to_python(part[2:]))

        if items:
            self.result_list.add(BCIMessage.SEPATATOR.join(items))
        else:
            logger.warning("champs inconnu : %s", key)

        return self

    def to_string(self):
        test_string = super(VitekTestMessage, self).to_string()
        result_string = self.result_list.to_string()
        if test_string or result_string:
            return self._make_string(
                test_string,
                self.antibiotic_family.to_string(),
                result_string)
        return ""


class VitekAntibioticFamily(BCIMessage):
    FIELDS = OrderedDict([
        ("af", String("antibiotic_family_name", max_length=40)),
    ])

    def __init__(self, **kwargs):
        self.phenotype_list = []
        super(VitekAntibioticFamily, self).__init__(**kwargs)

    def to_string(self):
        return self._make_string(
            super(VitekAntibioticFamily, self).to_string(),
            BCIMessage.SEPATATOR.join(
                ["ap%s" % x for x in self.phenotype_list]
            ))


class VitekAntibioticMessage(BCIMessage):
    FIELDS = OrderedDict([
        ("ra", String("result_separator", required=True)),
        ("o1", String("final_organism_code", max_length=6)),
        ("o2", String("final_organism_name", max_length=60)),
        ("o3", String("total_count_code", max_length=6)),
        ("o4", String("final_modifier_code", max_length=6)),
        ("o5", String("final_modifier_name", max_length=20)),  # Repeatable
        ("o6", String("organism_group_code", max_length=6)),  # Repeatable
        ("o7", String("organism_group_name", max_length=30)),  # Repeatable
        ("o8", String("total_count_text", max_length=20)),
        ("rs", String("smear_code", max_length=6)),
        ("rm", String("smear_name", max_length=60)),
        ("rh", Date("hours_to_positive", fmt="%H:%M")),
        ("ad", String("deduced_antibiotic_flag")),
        ("a1", String("antibiotic_code", max_length=8)),
        ("a2", String("antibiotic_name", max_length=60)),
        ("a3", String("final_mic", max_length=8)),
        ("a4", UpperString("final_result", max_length=2)),
        ("a5", String("dosage_column1", max_length=25)),
        ("a6", String("dosage_column2", max_length=16)),
        ("a7", String("dosage_column3", max_length=16)),
        ("an", UpperString("not_expertized_interpretation", max_length=16)),
    ])

    CAR_SUPPRESSION_INDICATOR = "ar"


class VitekRequest(VitekBaseMessage):
    message_type = "mtmpr"


class VitekResult(VitekBaseMessage):
    message_type = "mtrsl"


class Driver(Message):
    SPEC = {
        "code": "vitek",
        "label": "Vitek BCI",
    }
    REQUEST = VitekRequest.message_type
    RESULT = VitekResult.message_type
    MESSAGE_MAPPING = {
        REQUEST: VitekRequest,
        RESULT: VitekResult,
    }
    description = "Vitek"

    def __init__(self, default_cmi="", **kwargs):
        self.default_cmi = default_cmi
        super(Driver, self).__init__(**kwargs)

    def create(self, message_type, **kwargs):
        return Driver.MESSAGE_MAPPING[message_type](**kwargs)

    def parse(self, data):
        try:
            klass = Driver.MESSAGE_MAPPING[data[:5]]
        except KeyError:
            raise ParsingError("Type de message inconnu : %s" % (data[:5],))
        return klass().parse(data[6:])

    def to_string(self, message):
        return message.to_string()

    def get_message_type(self, message):
        return self.ORDER

    def get_result_dict(self, message):
        result_dict = {}
        for specimen in message.specimen_list:
            for culture in specimen.culture_list:
                culture_dict = result_dict.setdefault(culture.culture_id, {
                    "rack": None,
                    "resultats": []
                })
                for test in sorted(culture.test_list,
                                   key=lambda x: x.isolate_number):
                    germe = {
                        "test": culture.culture_type_code,
                        "resu": "G_%s" % (test.final_organism_code.upper()),
                        "date": test.result_date,
                        "seq": test.isolate_number,
                        "atb": []
                    }
                    for result in test.result_list:
                        germe["atb"].append((
                            ("ANTIBIO", "A_%s" % (
                                result.antibiotic_code.upper(),
                            )),
                            ("CMI", result.final_mic or self.default_cmi),
                            ("SENSI", result.final_result),
                        ))
                    if germe["atb"]:
                        culture_dict["resultats"].append(germe)
        return result_dict

    def get_order_message(self, worklist, context):
        if len(worklist.samples) == 0:
            return None
        elif len(worklist.samples) > 1:
            raise RuntimeError("Multi-sample not implemented")
        else:
            sample = worklist.samples[0]
            patient_name = "%s_%s" % (
                sample["last_name"],
                sample["first_name"],
            )
            msg = self.create(
                self.REQUEST,
                patient_id=sample["patient_id"],
                patient_name=patient_name,
                patient_birthdate=sample["birthdate"],
                patient_sex=sample["sex"],
                patient_admission_date=sample.get("admission_date"),
                patient_service_code=sample["location"],
                patient_physician_code=sample["physician"],
                patient_location_code=sample["location"],
            )
            for order in sample["orders"]:
                specimen = msg.specimen_list.create(
                    specimen_source_code=order["code"],
                    specimen_location_code=msg.patient_location_code,
                    specimen_service_code=msg.patient_location_code,
                    specimen_physician_code=msg.patient_physician_code,
                    specimen_collection_date=sample["collection_date"],
                    specimen_collection_time=sample["collection_date"])
                specimen.culture_list.create(
                    culture_id=sample["sample_id"],
                    culture_type_code=order["code"])
            return msg
