# -*-
from __future__ import unicode_literals

import datetime

from driverng import Message
from driverng.exceptions import ParsingError
from driverng.fixed import (
    FixedLengthFormat,
    String,
    Date,
)


class Result(FixedLengthFormat):
    FMT_DESCRIPTION = [
        String("test", 1),
        # NO RESULT is reported in this filed if there is a
        # numerical processing error or if the test is not
        # supported
        String("result", 9, right_justified=True),
        # 0: Conventional
        # 1: SI units
        # 2: Alternate conventional
        String("units", 1),
        # 0: No flag
        # 1: Above reference range
        # 2: Below reference range
        # 4: Above dynamic range
        # 5: Below dynamic range
        # 6: Prediction failure (NO RESULT)
        # 7: Above supplementary range
        # 8: Below supplementary range
        # A: Control result if more than 2 SDI from mean
        # B: Result is more than 3 SDI from mean
        # C: No baseline interval mean and SDI or array not supported
        #    in QC database
        # Q: Result text 1
        # R: Result text 2
        # S: Result text 3
        # T: Result text 4
        # U: Result text 5
        String("error_flag", 1),
        # 0: No warning
        # 1: Analyser warning
        # 2: Operator warning
        # 3: Both
        String("warning_flag", 1),
        String("separator", 1, default="}"),
    ]

    def __init__(self, *args, **kwargs):
        self._locked = False
        # AR: User adjusted parameters changed results
        # NF: No fluid
        # CE: Calibration Expired
        # NI: No inventory
        # FR: Flagged replicate
        # NQ: Not in QC data base
        # DE: Drop error
        # NW: No well
        # ED: Edited result
        # NT: No tip
        # EM: Expired maintenance
        # OD: Operator requested dilution
        # EP: Edit patient data
        # OR: Outside range
        # FC: Flagged component
        # PF: Prediction failure
        # IC: Invalid component
        # RC: Reference consistency check
        # ID: Invalid dilution
        # RD: Reflex dilution
        # II: Insufficient inventory
        # RE: Reagent expired
        # IS: Insufficient sample
        # RP: Reflex process
        # IT: Incubator temperature
        # RR: Recalculated result
        # LS: Lot switch
        # SC: Spread check
        # LT: Luminometer temperature out
        # UC: User calibrated
        # ME: Mechanical error occurred
        # WT: Well wash temperature out
        # M1: Category 1 modified values
        # M2: Category 2 modified values
        # ZS: Zero Set
        # NC: Not calibrated
        self.error_codes = []
        super(Result, self).__init__(*args, **kwargs)

    def parse(self, data):
        length = self.get_length(self.FMT_DESCRIPTION)
        line = data[:data.index("}") + 1]
        remain = data[len(line):]
        error_codes = line[length - 1:-1]
        self.error_codes = [
            error_codes[i:i + 2] for i in range(0, len(error_codes), 2)]
        line = line[:length - 1] + "}"
        return super(Result, self).parse(line + remain)

    def to_string(self):
        data = super(Result, self).to_string()
        return data[:-1] + "".join(self.error_codes) + data[-1]


class ResultMessage(FixedLengthFormat):
    FMT_DESCRIPTION = [
        Date("_time", 6, fmt_date="%H%M%S"),
        Date("_date", 4, fmt_date="%m%d"),
        String("tray", 15),
        String("sample_id", 15),
        # 1: Serum
        # 2: CSF
        # 3: Urine
        # 4: Serum
        # 5: Plasma
        # 6: Urine
        # 7: Blood
        # 8: Amnio
        # 9: Reserved 1
        # :: Reserved 2
        # ;: Reserved 3
        # <: Reserved 4
        String("fluid", 1),
        # 0: Non stat
        # 1: stat
        String("stat", 1, default="0"),
        String("cup", 1),
        # 0: Select
        # 1: Batch
        # 4: Downloaded
        # 5: Downloaded and manually edited
        String("mode", 1),
        String("dilution", 5, right_justified=True, empty="0"),
    ]

    def __init__(self, *args, **kwargs):
        self._locked = False
        self.analyzer = ""
        self.results = []
        self.date = None
        super(ResultMessage, self).__init__(*args, **kwargs)

    def parse(self, data):
        remain = super(ResultMessage, self).parse(data)
        if self._date and self._time:
            self.date = datetime.datetime(
                *(datetime.datetime.now().year,) +
                self._date.timetuple()[1:3] +
                self._time.timetuple()[3:6])
        while remain[0] not in ("|", "]"):
            r = Result()
            remain = r.parse(remain)
            self.results.append(r)
        if remain[0] == "|":
            self.analyzer, remain = remain[:11], remain[11:]
        if remain != "]":
            raise ParsingError("Unexpected remaining data '%s'" % remain)
        return self

    def to_string(self):
        data = super(ResultMessage, self).to_string()
        for r in self.results:
            data += r.to_string()
        data += self.analyzer
        data += "]"
        return data


class OrderMessage(FixedLengthFormat):
    FMT_DESCRIPTION = [
        # String("tray_name", 15)  # optional tray name
        String("sample_id", 15),
        String("fluid", 1),
        String("stat", 1, default="0"),
        String("cup", 1),
        String("dilution", 5, default="01.00"),
        # measured test requests
        # derived test requests
        String("separator", 1, default="|"),
        String("last_name", 15, truncable=True),
        String("first_name", 9, truncable=True),
        String("middle", 1),
        String("end", 1, default="]"),
    ]

    def __init__(self, *args, **kwargs):
        self._locked = False
        self.request = []
        super(OrderMessage, self).__init__(*args, **kwargs)

    def parse(self, data):
        # TODO: self.request
        super(OrderMessage, self).parse(data)
        return self

    def to_string(self):
        data = super(OrderMessage, self).to_string()
        return data[:23] + "".join(sorted(self.request)) + data[23:]


class MultipleResultMessage(list):

    def add(self, **kwargs):
        o = ResultMessage(**kwargs)
        self.append(o)
        return o

    def parse(self, data):
        for line in data.splitlines():
            self.append(ResultMessage().parse(line))
        return self


class Driver(Message):
    SPEC = {
        "code": "vitros_kermit",
        "label": "Vitros KERMIT",
    }

    def parse(self, data):
        return MultipleResultMessage().parse(data)

    def get_message_type(self, message):
        return self.ORDER

    def to_string(self, message):
        return message.to_string()

    def get_order_message(self, worklist, context):
        samples = worklist.get_worklist(
            unique_together=["action"],
            action={"del": worklist.IGNORE, worklist.ELSE: ""},
        )
        if len(samples) == 0:
            return None
        elif len(samples) > 1:
            raise RuntimeError("Multi-sample not implemented")
        else:
            sample, _ = samples[0]
            msg = OrderMessage(
                sample_id=sample["sample_id"],
                fluid=sample["specimen"],
                last_name=sample["last_name"].replace(" ", ""),
                first_name=sample["first_name"].replace(" ", ""),
            )
            msg.fluid = sample["specimen"] or "1"
            for order in sample["orders"]:
                msg.request.append(chr(int(order["code"])))
            return msg

    def get_result_dict(self, message):
        results = {}
        for msg in message:
            sample_results = results.setdefault(
                msg.sample_id,
                {"resultats": []},
            )["resultats"]
            for r in msg.results:
                alarmes = []
                if r.error_flag != "0":
                    alarmes.append(r.error_flag)

                if r.error_flag == "4":
                    resu = ">%s" % r.result
                elif r.error_flag == "5":
                    resu = "<%s" % r.result
                else:
                    resu = r.result

                if r.test == "":
                    test = ord(" ")
                else:
                    test = ord(r.test)
                alarmes.extend(r.error_codes)
                sample_results.append({
                    "test": str(test),
                    "resu": resu,
                    "date": msg.date,
                    "alarmes": alarmes,
                })
            return results
