# -*- coding: utf8 -*-
from __future__ import unicode_literals

from unidecode import unidecode

from driverng.astm import (
    AstmMessage,
    HeaderSegment,
    Segment,
    HprimString,
    HprimList,
    HprimDate,
    HprimForeign,
    HprimMultiple,
)


class VitrosSegmentH(HeaderSegment):

    emetteur = HprimString(5, max_length=7, default="MCA")
    version = HprimString(13, default="LIS2-A")
    date = HprimDate(14)

    patients = HprimForeign("P", "header")
    queries = HprimForeign("Q", "header")
    errors = HprimForeign("M", "header")
    end = HprimForeign("L", "header", one2one=True, required=True)


class VitrosSegmentQ(Segment):

    # The specimen ID of interest, only one allowed per record
    computer_specimen_id = HprimString((3, 2), max_length=15)

    # ALL (Values other than ALL are not supported)
    universal_test_id = HprimString(5, max_length=3, default="ALL")

    # A: Abort/Cancel request
    # O: Requesting test orders and demographics only
    request_information_status_codes = HprimString(
        13, max_length=1, default="O")


class VitrosSegmentP(Segment):

    no_patient = HprimString(3, max_length=20)

    # Last^First^Middle. Suffix and title not supported
    last_name = HprimString((6, 1), max_length=20, trunc=True)
    first_name = HprimString((6, 2), max_length=15, trunc=True)
    middle = HprimString((6, 3), max_length=1)

    date_naissance = HprimDate(8, fmt_date="%Y%m%d")

    # Must match one of the configured demographic sexe characters on the
    # system
    sexe = HprimString(9)

    address = HprimMultiple(
        11,
        ligne1=HprimString(1, max_length=20),
        ligne2=HprimString(2, max_length=20),
    )

    attending_physician_id = HprimMultiple(
        14,
        last_name=HprimString(1, max_length=20),
        first_name=HprimString(2, max_length=15),
        middle=HprimString(3, max_length=1),
    )

    # Range attribute, must match one of the configured demographic attribute
    # characters on the system.
    special_field1 = HprimString(15, max_length=1)

    # Patient room number
    location = HprimString(26, max_length=10)

    orders = HprimForeign("O", "patient")


class VitrosSegmentO(Segment):

    sample_id = HprimString((3, 1), max_length=15)
    tray_id = HprimString((3, 2), max_length=2)
    cup = HprimString((3, 3), max_length=2)

    # Liste des codes tests
    # ^^^LocalOrMfctrCode
    # ^^^1.0000+314+1.0\315+1.0\950+1.0\951+1.0\952+1.0
    # ATTENTION: Ce champ est visiblement invalide du point de vue de la norme
    tests = HprimList(5, basefield=HprimMultiple(0,
                                                 code1=HprimString(1),
                                                 code4=HprimString(4),
                                                 ))

    # Priority
    # S, A, R, C, P. S and A considered STAT priority
    priority = HprimString(6, max_length=1)

    # Date and time specimen was collected
    collection_date = HprimDate(8)

    # Action code
    # C, N, A, P, L, X, Q
    # P, L, X and Q are treated as N
    # N: Adds a program for this order
    # A: Add tests overwritting any existing test
    # C: Cancel
    # On upload, system always sends N
    action_code = HprimString(12, max_length=1)

    # Body fluid
    # 1: CSF
    # 3: Urine
    # 4: Whole blood
    # 5: Serum
    # 6: Plasma
    # 7: Amniotic fluid
    # 8: Reserved fluid 1
    # 9: Reserved fluid 2
    # 10: Reserved fluid 3
    specimen_descriptor = HprimString(16)

    # O: For downloading
    # P: For uploading when some tests are still in progress
    # F: For uploading when all tests in sample are complete
    # Other values not supported
    report_type = HprimString(26)

    results = HprimForeign("R", "order")


class VitrosSegmentR(Segment):

    # ^^^LocalOrMfctrCode
    test = HprimString((3, 4), max_length=13)

    # "No Result" will appear when System cannot report numeric value
    value = HprimString(4, max_length=9)

    units = HprimString(5, max_length=12)

    abnormal_flags = HprimList(
        7,
        basefield=HprimMultiple(
            0,
            flag=HprimString(2, max_length=1),
            code=HprimString(3, max_length=10),
        ),
    )

    # V
    # Values other than V not supported
    status = HprimString(9, default="V", max_length=1)

    operator_id = HprimString(11, max_length=15)
    date_test_started = HprimDate(12)
    date_test_completed = HprimDate(13)
    # Analyseur (= JNubmer)
    instrument_id = HprimString(14)

    infos = HprimForeign("M", "result")


class VitrosSegmentM(Segment):

    # ATTENTION: Plusieurs segment M possibles (différenciés par leur
    # subtype_id)
    subtype_id = HprimString(3, max_length=1, default="X")
    reagent = HprimMultiple(4,
                            lot=HprimString(1, max_length=12),
                            expiration_date=HprimDate(2),
                            load_date=HprimDate(3, max_length=5),
                            erf_lot=HprimString(4, max_length=5),
                            srlot=HprimString(5, max_length=4),
                            )
    calibration = HprimMultiple(5,
                                date=HprimDate(1),
                                # Calibration status table (5.1.3)
                                # N: Normal calibration
                                # U: User calibrated
                                # M: User modified
                                # B: User modified/calibrated
                                status=HprimString(2, max_length=1),
                                expiration_date=HprimDate(3),
                                )
    quality_control = HprimMultiple(6,
                                    lot=HprimString(1, max_length=10),
                                    creation_date=HprimDate(2),
                                    expiration_date=HprimDate(3),
                                    )

    diluent_lots = HprimList(7, basefield=HprimString(0, max_length=12))

    # Reprocessing type table (5.1.7)
    # N: None, initial testing
    # R: Retesting
    # I: Retesting due to insufficient inventory
    # D: Reflex dilution/reduction
    # F: Reflex test; reflexed from same of different test
    reprocessing_type = HprimString(8, max_length=1)


class VitrosSegmentC(Segment):

    comment_source = HprimString(3, max_length=1, default="I")
    comment_text = HprimString(4, max_length=60)

    # G: Generic/free text used for patient comments
    # I: Instrument flags used for result comments to report
    # qualitative results.
    # Values other than G are not supported for patient comments
    comment_type = HprimString(5, max_length=1)


class VitrosSegmentL(Segment):
    # Code de terminaison
    # Values other than N, T, Q or I not supported
    code = HprimString(3)


class Driver(AstmMessage):
    SPEC = {
        "code": "vitros",
        "label": "Vitros 5600/3600/Fusion (ASTM)",
    }
    SEGMENT_CLASS_MAP = {
        "H": VitrosSegmentH,
        "P": VitrosSegmentP,
        "Q": VitrosSegmentQ,
        "O": VitrosSegmentO,
        "R": VitrosSegmentR,
        "M": VitrosSegmentM,
        "C": VitrosSegmentC,
        "L": VitrosSegmentL,
    }
    DEFAULT_SEPARATORS = "|\^&"

    def __init__(self, ref_qc_reagent_vitros=False, **kwargs):
        self.ref_qc_reagent_vitros = ref_qc_reagent_vitros
        super(Driver, self).__init__(**kwargs)

    def create(self, **kwargs):
        message = VitrosSegmentH(type_segment="H", **kwargs)
        message.end = VitrosSegmentL(type_segment="L")
        return message

    def get_message_type(self, message):
        if message.queries:
            return self.QUERY
        else:
            return self.ORDER

    def get_queries(self, message):
        return [q.computer_specimen_id for q in message.queries]

    def get_order_message(self, worklist, context):
        msg = self.create(date=context["now"])
        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority"],
            action={"del": "C", worklist.ELSE: "N"},
            priority={"A": "A", "S": "S", worklist.ELSE: "R"},
        ):
            seg_p = msg.patients.add(
                no_patient=patient["patient_id"],
                last_name=unidecode(patient["last_name"]),
                first_name=unidecode(patient["first_name"]),
                date_naissance=patient["birthdate"],
                sexe=patient["sex"] if patient["sex"] in ("M", "F") else "U",
            )
            for sample, action, priority in patient["samples"]:
                seg_o = seg_p.orders.add(
                    sample_id=sample["sample_id"],
                    priority=priority,
                    collection_date=sample["collection_date"],
                    report_type="O",
                    specimen_descriptor=sample["specimen"],
                    action_code=action,
                )
                for i, order in enumerate(sample["orders"]):
                    t = seg_o.tests.add()
                    if i == 0:
                        t.code4 = "1.0000+%s+1.0" % (
                            order["code"],
                        )
                    else:
                        t.code1 = "%s+1.0" % (
                            order["code"],
                        )
        if len(msg.patients):
            return msg
        else:
            return None

    @staticmethod
    def _is_control(seg_r):

        for seg_m in seg_r.infos:
            if seg_m.subtype_id == "X" and seg_m.quality_control.lot:
                return True

        return False

    @staticmethod
    def _get_code_resultat(seg_r):
        return seg_r.test.split('+')[1]

    @staticmethod
    def _get_reactifs(seg_r):

        lots_reactifs = []
        for seg_m in seg_r.infos:
            if seg_m.reagent.lot:
                lots_reactifs.append({
                    "ref": seg_m.reagent.lot,
                    "peremption": seg_m.reagent.expiration_date,
                })

        return lots_reactifs

    def get_result_dict(self, message):

        result = {}

        for seg_p in message.patients:
            for seg_o in seg_p.orders:

                for seg_r in seg_o.results:

                    if self._is_control(seg_r):
                        continue

                    if seg_o.tray_id and seg_o.cup:
                        rack = "%s-%s" % (seg_o.tray_id, seg_o.cup)
                    else:
                        rack = None

                    result.setdefault(seg_o.sample_id,
                                      {"resultats": [], "rack": rack})

                    alarmes = []

                    for i, abf in enumerate(seg_r.abnormal_flags):
                        if i == 0 and abf.flag != "0":
                            alarmes.append(abf.flag)
                        if i == 0 and abf.code:
                            j = 0
                            while True:
                                code = abf.code[j * 2:][:2]
                                if not code:
                                    break
                                alarmes.append(code)
                                j += 1
                        elif i > 0 and i < 4 and abf.flag == "J":
                            alarmes.append("%c%s" % ("HIT"[i - 1], abf.flag,))

                    result[seg_o.sample_id]["resultats"].append({
                        "test": self._get_code_resultat(seg_r),
                        "resu": seg_r.value,
                        "alarmes": alarmes,
                        "date": seg_r.date_test_completed,
                        "analyseur": seg_r.instrument_id or message.emetteur,
                        "lots_reactifs": self._get_reactifs(seg_r),
                    })

        return result

    def get_qc_result_list(self, message):

        result = []

        for seg_p in message.patients:
            for seg_o in seg_p.orders:

                for seg_r in seg_o.results:

                    if not self._is_control(seg_r):
                        continue
                    if self.ref_qc_reagent_vitros:
                        # Concatener la reference du lot avec
                        # les 8 derniers digits du ReagentLot
                        lot_cq = "%s%s" % (
                            seg_r.infos[0].quality_control.lot,
                            seg_r.infos[0].reagent.lot[-8:]
                        )
                    else:
                        lot_cq = seg_r.infos[0].quality_control.lot
                    result.append({
                        "lot": lot_cq,
                        "test": self._get_code_resultat(seg_r),
                        "resu": seg_r.value,
                        "date": seg_r.date_test_completed,
                        "analyseur": seg_r.instrument_id or message.emetteur,
                        "lots_reactifs": self._get_reactifs(seg_r),
                    })

        return result
