# -*- coding: utf8 -*-
from __future__ import unicode_literals

import datetime

from driverng import Message
from driverng.exceptions import ProgrammingError, ParsingError


class BaseVSCubeMessage(object):
    TYPE_MESSAGE = None
    # Indicateur du nombre de bloc dans le message ?
    _NBLOCK = True

    def get_blocks(self, data):
        if data[:2] != "00" or data[4:][:2] != "01":
            raise ParsingError("Unexpected data")

        data_length = int(data[2:][:2], 16)

        if len(data[8:]) != data_length:
            raise ParsingError("Expect %d length and %d found" %
                               (data_length, len(data[8:])))

        if self._NBLOCK:
            n_blocks = int(data[8:][:2], 16)
            return (n_blocks, data[10:])
        else:
            return (None, data[8:])

    def format_message(self, data):
        return "00%02X01%s%s" % (
            len(data),
            self.TYPE_MESSAGE,
            data,
        )

    def to_string(self):
        raise NotImplementedError


class VSCubeTube(object):

    def __init__(self, code_barre, inconnu=False):
        self.code_barre = code_barre
        self.inconnu = inconnu


class VSCubeOrderMessage(BaseVSCubeMessage):
    TYPE_MESSAGE = "50"

    def __init__(self):
        self.tubes = []
        super(VSCubeOrderMessage, self).__init__()

    def add(self, **kwargs):
        tube = VSCubeTube(**kwargs)
        self.tubes.append(tube)
        return tube

    @staticmethod
    def parse(data):
        obj = VSCubeOrderMessage()

        nb_code_barre, blocks = obj.get_blocks(data)

        code_barre = ""
        for c in blocks:
            if c in ("\x10", "\x11"):
                obj.add(code_barre=code_barre,
                        inconnu=True if c == "\x11" else False)
                code_barre = ""
            else:
                code_barre += c

        if code_barre != "":
            raise ParsingError("Remaining data '%s'" % (code_barre,))
        if len(obj.tubes) != nb_code_barre:
            raise ParsingError("Expect %d bare-code and %d found" % (
                nb_code_barre,
                len(obj.tubes),
            ))

        return obj

    def to_string(self):
        data = "%02X" % len(self.tubes)
        for tube in self.tubes:
            data += "%s%s" % (
                tube.code_barre,
                "\x11" if tube.inconnu else "\x10"
            )
        return self.format_message(data)


class VSCubeResult(object):
    FMT_DATETIME = "%d%m%y%H%M"
    FMT_DATE = "%d%m%y"
    _ARGS = {
        # key => max_length
        "code_barre": 15,
        "date": None,
        "esr": 4,
        "flags": 2,
        "rack": 4,
        "position": 2,
    }

    def __init__(self, **kwargs):

        for key in self._ARGS.keys():
            setattr(self, key, kwargs.pop(key, None))

        if kwargs:
            raise ProgrammingError("Remaining kwargs %s" % (kwargs,))

    @staticmethod
    def parse(data):
        obj = VSCubeResult()

        obj.code_barre, buf = data.split("\x10", 1)
        obj.date = datetime.datetime.strptime(buf[:10], obj.FMT_DATETIME)
        obj.esr = buf[10:][:4]
        obj.flags = buf[14:][:2]
        obj.rack = buf[16:][:4]
        obj.position = buf[20:][:2]

        return (obj, buf[22:])

    def to_string(self):
        for key, max_length in self._ARGS.items():
            value = getattr(self, key)
            if not value:
                raise ParsingError("Field %s is empty" % (key,))
            if max_length and len(value) > max_length:
                raise ParsingError("Field %s too long" % (key,))

        return "%s\x10%s%4.4s%2.2s%4.4s%2.2s" % (
            self.code_barre,
            self.date.strftime(self.FMT_DATETIME),
            self.esr,
            self.flags,
            self.rack,
            self.position,
        )


class VSCubeQCResult(VSCubeResult):
    _ARGS = dict(VSCubeResult._ARGS.items() + {
        "lot": 6,
        "date_peremption": None,
        "valref": 4,
    }.items())

    @staticmethod
    def parse(data):
        obj = VSCubeQCResult()
        obj.lot = data[:6]
        obj.date_peremption = datetime.datetime.strptime(
            data[6:][:6], obj.FMT_DATE)
        obj.valref = data[12:][:4]

        obj.code_barre, buf = data[16:].split("\x10", 1)
        obj.date = datetime.datetime.strptime(buf[:10], obj.FMT_DATETIME)
        obj.esr = buf[10:][:4].strip()
        obj.flags = buf[14:][:2]
        obj.rack = buf[16:][:4]
        obj.position = buf[20:][:2]

        return (obj, buf[22:])

    def to_string(self):
        data = super(VSCubeQCResult, self).to_string()
        return "%6.6s%s%4.4s%s" % (
            self.lot,
            self.date_peremption.strftime(self.FMT_DATE),
            self.valref,
            data)


class VSCubeResultMessage(BaseVSCubeMessage):
    TYPE_MESSAGE = "51"

    def __init__(self):
        self.results = []
        super(VSCubeResultMessage, self).__init__()

    def add(self, **kwargs):
        result = VSCubeResult(**kwargs)
        self.results.append(result)
        return result

    @staticmethod
    def parse(data):
        obj = VSCubeResultMessage()

        nb_result, blocks = obj.get_blocks(data)

        while blocks:
            result, blocks = VSCubeResult.parse(blocks)
            obj.results.append(result)

        if len(obj.results) != nb_result:
            raise ParsingError("Expect %d results and %d found" %
                               (nb_result, len(obj.results)))

        return obj

    def to_string(self):
        data = "%02X" % len(self.results)

        for resu in self.results:
            data += resu.to_string()

        return self.format_message(data)


class VSCubeQCMessage(BaseVSCubeMessage):
    TYPE_MESSAGE = "52"
    _NBLOCK = False

    def __init__(self):
        self.results = []
        super(VSCubeQCMessage, self).__init__()

    def add(self, **kwargs):
        result = VSCubeQCResult(**kwargs)
        self.results.append(result)
        return result

    @staticmethod
    def parse(data):
        obj = VSCubeQCMessage()

        _, blocks = obj.get_blocks(data)

        while blocks:
            result, blocks = VSCubeQCResult.parse(blocks)
            obj.results.append(result)

        return obj

    def to_string(self):
        data = ""
        for resu in self.results:
            data += resu.to_string()
        return self.format_message(data)


class Driver(Message):
    SPEC = {
        "code": "vscube",
        "label": "Ves-Matic Cube",
    }

    _MESSAGE_TYPE_MAP = {
        Message.QUERY: VSCubeOrderMessage,
        Message.ORDER: VSCubeOrderMessage,
        Message.RESULT: VSCubeResultMessage,
        Message.QC: VSCubeQCMessage,
    }

    _TYPE_MESSAGE_MAP = {
        "50": Message.QUERY,
        "51": Message.RESULT,
        "52": Message.QC,
    }

    def create(self, message_type, **kwargs):
        try:
            cls = self._MESSAGE_TYPE_MAP[message_type]
        except KeyError:
            raise ProgrammingError("Unknown message type")

        return cls(**kwargs)

    def get_message_type(self, message):
        return self._TYPE_MESSAGE_MAP[message.TYPE_MESSAGE]

    def to_string(self, message):
        return message.to_string()

    def parse(self, data):
        lines = [l for l in data.splitlines() if l]
        if len(lines) != 1:
            raise ParsingError("Cannot parse data")
        line = lines[0]

        command = line[6:][:2]

        return self._MESSAGE_TYPE_MAP[self._TYPE_MESSAGE_MAP[command]].parse(line)

    def get_queries(self, message):
        if self.get_message_type(message) == self.QUERY:
            return [t.code_barre for t in message.tubes]
        else:
            return []

    def get_order_message(self, worklist, context):
        msg = self.create(self.ORDER)
        requested = [
            s["sample_id"] for s in worklist.samples if s["orders"]]

        for sample_id in requested:
            msg.add(code_barre=sample_id)

        if requested or context["query"]:
            return msg

    def get_result_dict(self, message):

        if self.get_message_type(message) != self.RESULT:
            return {}

        results = {}

        for resu in message.results:

            rack = resu.rack.strip() or None
            position = resu.position.strip()
            if rack and position:
                rack += "-%s" % (position,)

            # TODO à migrer dans param_alarme
            comment_map = {
                0x1: "Blood column too high",
                0x2: "Blood column too low",
                0x4: "Tube Empty",
                0x8: "Generic reading error",
                0x10: "The VES of the QC measured is within "
                      "the acceptability range",
                0x20: "The VES of the QC measured is outside"
                      "the acceptability range",
            }
            flags = int(resu.flags, 16)
            commentaires = []
            for key, commentaire in comment_map.items():
                if flags & key:
                    commentaires.append(commentaire)

            results.setdefault(resu.code_barre, {
                "resultats": [], "rack": rack}
            )["resultats"].append({
                "test": "VS1",
                "resu": resu.esr.strip(),
                "date": resu.date,
                "alarmes": ["%d" % flags],
                "commentaires": commentaires,
            })

        return results

    def get_qc_result_list(self, message):
        # param_connexion QC_AVEC_NIVEAU (biorad)

        if self.get_message_type(message) != self.QC:
            return []

        results = []

        for resu in message.results:
            results.append({
                "lot": resu.code_barre,
                "test": "VS1",
                "resu": resu.esr.strip(),
                "date": resu.date,
            })

        return results
