# -*- coding: utf8 -*-
from __future__ import unicode_literals

import logging
import datetime

from driverng import utils
from driverng.exceptions import ParsingError, ProgrammingError

__all__ = ["FixedLengthFormat", "String", "Date", "Multiple"]

logger = logging.getLogger("driverng.fixed")


class FixedLengthField(object):

    def __init__(self, name, length,
                 right_justified=False, truncable=False, empty=" ",
                 default=None):
        self.name = name
        self.length = length
        self.right_justified = right_justified
        self.truncable = truncable
        self.empty = empty
        self.default = default
        super(FixedLengthField, self).__init__()

    def get_default(self):
        raise NotImplementedError

    def to_python(self, value):
        raise NotImplementedError

    def to_string(self, value):
        raise NotImplementedError


class String(FixedLengthField):

    def get_default(self):
        return self.default or ""

    def to_python(self, value):
        if len(value) > self.length:
            logger.warning(
                "Warning key '%s' with value '%s' too long" % (
                    self.name,
                    value,
                ))

        return value.strip()

    def to_string(self, value):
        if value is None:
            value = ""
        value = "%s" % value
        if len(value) > self.length and not self.truncable:
            raise ParsingError(
                "Cannot format key '%s' with value '%s', too long" % (
                    self.name,
                    value,
                ))
        v = value[:self.length]

        if self.right_justified:
            fmt = "%%%d.%ds" % (self.length, self.length)
        else:
            fmt = "%%-%d.%ds" % (self.length, self.length)

        raw_value = fmt % v

        if self.empty != " ":
            # On remplace les espaces de fin et de début
            # par self.empty
            new_value = list(raw_value)
            i = 0
            while i < len(new_value):
                if new_value[i] == " ":
                    new_value[i] = self.empty
                else:
                    break
                i += 1
            i = 0
            while i < len(new_value):
                if new_value[-i] == " ":
                    new_value[-i] = self.empty
                else:
                    break
                i += 1

            raw_value = "".join(new_value)

        return raw_value


class Date(String):

    def __init__(self, *args, **kwargs):

        self.fmt_map = kwargs.pop("fmt_map", utils.FMT_MAP)
        self.fmt_date = kwargs.pop("fmt_date", "%Y%m%d%H%M%S")
        self.auto_now = kwargs.pop("auto_now", False)
        super(Date, self).__init__(*args, **kwargs)

    def get_default(self):
        if self.auto_now:
            return datetime.datetime.now()
        else:
            return self.default

    def to_python(self, value):
        return utils.get_date(value, fmt_map=self.fmt_map, fmt=self.fmt_date)

    def to_string(self, value):
        return super(Date, self).to_string(
            utils.format_date(value, fmt=self.fmt_date))


class MultipleContainer(list):

    def __init__(self, fmt_description):
        self.fmt_description = fmt_description
        super(MultipleContainer, self).__init__()

    def add(self, *args, **kwargs):
        obj = FixedLengthFormat(fmt_description=self.fmt_description,
                                *args, **kwargs)
        self.append(obj)
        return obj


class Multiple(object):

    def __init__(self, name, fmt_description):

        self.name = name
        self.fmt_description = fmt_description
        self.length = FixedLengthFormat.get_length(self.fmt_description)
        super(Multiple, self).__init__()

    def get_default(self):
        return MultipleContainer(self.fmt_description)

    def to_python(self, data):
        values = MultipleContainer(self.fmt_description)

        remain = data
        while len(remain) >= self.length:
            obj = values.add()
            remain = obj.parse(remain)

        return (remain, values)

    def to_string(self, values):
        return "".join([v.to_string() for v in values])


class FixedLengthFormat(object):
    # ["Q" = fixed string, N = N spaces, field]
    FMT_DESCRIPTION = []

    def __init__(self, *args, **kwargs):

        self._locked = False
        self._fields = {}
        self._fmt_description = kwargs.get("fmt_description",
                                           self.FMT_DESCRIPTION)

        for field in self._fmt_description:
            if isinstance(field, (FixedLengthField, Multiple)):
                self._fields[field.name] = field

        for name, field in self._fields.items():
            setattr(self, name,
                    kwargs.get(name, field.get_default()))

        super(FixedLengthFormat, self).__init__()
        self._locked = True

    def __setattr__(self, attr, value):
        # On empèche la création d'attributs en dehors de __init__
        if attr != "_locked" and self._locked and not hasattr(self, attr):
            raise AttributeError("Object %s has not attribute '%s'" % (
                self, attr,))
        else:
            return super(FixedLengthFormat, self).__setattr__(attr, value)

    @staticmethod
    def get_length(fmt_description):
        length = 0
        for field in fmt_description:
            if isinstance(field, (FixedLengthField, Multiple)):
                length += field.length
            elif isinstance(field, basestring):
                length += len(field)
            elif isinstance(field, int):
                length += field
        return length

    def parse(self, data):
        """Parse les données et renvoie ce qui n'a pas été parsé."""

        cur = data
        for field in self._fmt_description:

            if isinstance(field, FixedLengthField):
                setattr(self, field.name, field.to_python(cur[:field.length]))
                cur = cur[field.length:]
            elif isinstance(field, Multiple):
                cur, values = field.to_python(cur)
                setattr(self, field.name, values)
            elif isinstance(field, basestring):
                if cur[:len(field)] != field:
                    raise ParsingError("Key '%s' not found in '%s'" % (
                        field,
                        cur[:len(field)],
                    ))
                cur = cur[len(field):]
            elif isinstance(field, int):
                # Nombre d'espaces à sauter.
                # On verifie que ce sont bien des espaces
                if cur[:field] != field * " ":
                    raise ParsingError("No enough spaces here '%s' !" % (
                                       cur[:field]))
                cur = cur[field:]
            else:
                raise ProgrammingError("Unknown key type '%s'" % (field,))

        return cur

    def to_string(self):

        data = ""
        for field in self._fmt_description:
            if isinstance(field, (FixedLengthField, Multiple)):
                data += field.to_string(getattr(self, field.name))
            elif isinstance(field, basestring):
                data += field
            elif isinstance(field, int):
                data += field * " "
            else:
                raise ProgrammingError("Unknown key type '%s'" % (field,))

        return data
