# -*- coding: utf8 -*-
from __future__ import unicode_literals

import datetime
import json

from driverng.worklist import Worklist


class GraphContainer(object):

    def __init__(self, type_fichier=None):
        self.type_fichier = type_fichier

    def get_bytes(self):
        """Méthode qui doit retourner le graphe."""
        raise NotImplementedError


class JSONGraphContainer(dict):

    @property
    def type_fichier(self):
        return "json"

    def get_bytes(self):
        return json.dumps(self)


class Message(object):
    _ImplementedByChild = NotImplementedError(
        "Must be implemented in child classes")
    QUERY, ORDER, QC, RESULT = list(range(4))
    ENCODING = "ISO-8859-15"

    def __init__(self, *args, **kwargs):
        super(Message, self).__init__()

    def create(self, *args, **kwargs):
        raise Message._ImplementedByChild

    def parse(self, data):
        """Parse le message `data` et renvoie le message."""
        raise Message._ImplementedByChild

    def get_queries(self, message):
        """Retourne la liste des tube en query."""
        return []

    def to_string(self, message, **kwargs):
        raise Message._ImplementedByChild

    def get_message_type(self, message):
        raise Message._ImplementedByChild

    def get_order_message(self, worklist, context):
        pass

    def get_void_order_msg(self, message, dem_auto_found=False, context=None):
        """DEPRECATED."""
        if context is None:
            context = {}
        context["query"] = message
        context.setdefault("now", datetime.datetime.now())
        return self.get_order_message(Worklist(), context)

    def get_result_dict(self, message):
        """Retourne la liste des resultats patients
        {tube: {
            rack: "rack-pos_rack",
            resultats: [{
                "test": code_test,
                "resu": resultat,
                "alarmes": liste d'alarmes,
                "commentaires": Liste de commentaires,
                "lots_reactifs": Liste de lots réactifs,
                "lots_calibrations": Liste de lots de calibration,
                "lots_controles": Liste de lots de controles,
                "date": date du resultat,
                "analyseur": Analyseur,
                "zero": [liste de valeurs qui sont à ignorer quand
                         param_auto.resu_azero vaut 0],
                "graph": Doit retourner un object GraphContainer.
                "type_prlv": Le code fluide (param_auto),
            }]
        }}
        """
        return {}

    def get_qc_result_list(self, message):
        """Retourne la liste des resultats QC
        [{
            "lot": lot,
            "test": code_test,
            "resu": resultat,
            "date": date,
            "sample_id": reference réele,
            "control_level": niveau,
            "type_prlv": type_prlv
            "lots_reactifs": Liste de lots réactifs,
            "lots_calibrations": Liste de lots de calibration,
            "lots_controles": Liste de lots de controles,
        }]
        """
        return []

    def enregistrer(self, data, message=None):
        """Cette méthode est appellée pour savoir s'il faut enregistrer les
        données dans la table integration.

        Permet de ne pas enregistrer certains messages spécifiques

        @param data: Les données brutes recues
        @param message: Si la connexion est configurée en query c'est
                        l'objet message parsé.
        """
        return True

    def encode_worklist(self, worklist, context):
        if worklist.context != "order":
            raise NotImplementedError
        for sample in worklist.samples:
            if sample.services:
                sample.service = sample.services[0].code
            else:
                sample.service = ""
            if sample.physicians:
                sample.physician = sample.physicians[0].code
                sample.physician_name = sample.physicians[0].last_name
                sample.physician_first_name = sample.physicians[0].first_name
            else:
                sample.physician = ""
                sample.physician_name = ""
                sample.physician_first_name = ""
        return self.get_order_message(worklist, context)

    @staticmethod
    def _lots_from_legacy(r):
        lots = []
        for lot_type, legacy_lots in (
            ("reagent", r.get("lots_reactifs", [])),
            ("calibration", r.get("lots_calibrations", [])),
            ("control", r.get("lots_controles", [])),
        ):
            for lot in legacy_lots:
                if isinstance(lot, basestring):
                    lots.append({
                        "type": lot_type,
                        "lot": lot,
                    })
                else:
                    lots.append({
                        "type": lot_type,
                        "lot": lot["ref"],
                        "name": lot.get("nom"),
                        "expiration": lot.get("peremption"),
                    })
        return lots

    @staticmethod
    def _lots_to_legacy(lots):
        legacy_lots = {
            "lots_reactifs": [],
            "lots_controles": [],
            "lots_calibrations": [],
        }
        for lot in lots:
            legacy_lots[{
                "reagent": "lots_reactifs",
                "calibration": "lots_calibrations",
                "control": "lots_controles",
            }].append({
                "ref": lot["lot"],
                "nom": lot["name"],
                "peremption": lot["expiration"],
            })
        return legacy_lots

    def decode_message(self, message):
        if self.get_message_type(message) == self.QUERY:
            wk = Worklist(context="query", type="patient")
            for sample_id in self.get_queries(message):
                wk.samples.add(sample_id=sample_id)
            return wk

        result_dict = self.get_result_dict(message)
        qc_result_list = self.get_qc_result_list(message)
        wk = Worklist(context="result")
        for sample_id, values in result_dict.items():
            specimens = set([
                r.get("type_prlv")
                for r in values["resultats"]
            ])
            if len(specimens) > 1:
                raise RuntimeError
            elif len(specimens) == 1:
                specimen = list(specimens)[0]
            else:
                specimen = None
            sample = wk.samples.add(
                sample_id=sample_id, rack=values.get("rack"),
                specimen=specimen,
                type="patient",
            )
            for r in values["resultats"]:
                sample.results.add(
                    code=r["test"],
                    value=r["resu"],
                    date=r.get("date"),
                    flags=r.get("alarmes", []),
                    comments=r.get("commentaires", []),
                    lots=self._lots_from_legacy(r),
                    instrument=r.get("analyseur"),
                )
        qcs = {}
        for r in qc_result_list:
            key = (r["lot"], r.get("sample_id"))
            if key in qcs:
                sample = qcs[key]
            else:
                sample = wk.samples.add(
                    control_lot=r["lot"],
                    control_name=r.get("control_level"),
                    sample_id=r.get("sample_id"),
                    specimen=r.get("type_prlv"),
                    type="control",
                )
                qcs[key] = sample
            sample.results.add(
                code=r["test"],
                value=r["resu"],
                date=r["date"],
                lots=self._lots_from_legacy(r),
                instrument=r.get("analyseur"),
            )
        return wk

    def encode(self, worklist, context=None):
        worklist.validate()
        context = context or {}
        context.setdefault("now", datetime.datetime.now())
        context.setdefault("query", None)
        message = self.encode_worklist(worklist, context)
        if message is not None:
            return self.to_string(message)

    def decode(self, data):
        worklist = self.decode_message(self.parse(data))
        if worklist is not None:
            worklist.validate()
        return worklist


class MessageV2(Message):

    def encode_worklist(self, worklist, context):
        raise Message._ImplementedByChild

    def decode_message(self, message):
        raise Message._ImplementedByChild

    def get_order_message(self, worklist, context):
        worklist.context = "order"
        for sample in worklist.samples:
            if sample.service:
                sample.services = [{"code": sample.service}]
            if sample.physician:
                sample.physicians = [{
                    "code": sample.physician,
                    "last_name": sample.physician_name,
                    "first_name": sample.physician_first_name,
                    "type": "prescriber",
                }]
        return self.encode_worklist(worklist, context)

    def get_message_type(self, message):
        worklist = self.decode_message(message)
        if worklist is None:
            raise RuntimeError("No worklist")
        if worklist.context == "query":
            return self.QUERY
        else:
            return self.ORDER

    def get_result_dict(self, message):
        results = {}
        worklist = self.decode_message(message)
        worklist.validate()
        if worklist.context != "result":
            return {}
        for sample in worklist.samples:
            if sample.type != "patient":
                continue
            resultats = results.setdefault(sample.sample_id, {
                "rack": sample.rack,
                "resultats": [],
            })["resultats"]
            for r in sample.results:
                res = {
                    "test": r.code,
                    "resu": r.value,
                    "date": r.date,
                    "alarmes": r.flags,
                    "commentaires": r.comments,
                    "analyseur": r.instrument,
                    "type_prlv": sample.specimen,
                }
                res.update(**self._lots_to_legacy(r["lots"]))
                resultats.append(res)
        return results

    def get_qc_result_list(self, message):
        worklist = self.decode_message(message)
        worklist.validate()
        if worklist.context != "result":
            return []

        results = []
        for sample in worklist.samples:
            if sample.type != "control":
                continue
            for r in sample.results:
                res = {
                    "lot": sample.control_lot,
                    "control_level": sample.control_name,
                    "test": r.code,
                    "resu": r.value,
                    "date": r.date,
                    "analyseur": r.instrument,
                    "type_prlv": sample.specimen,
                }
                res.update(**self._lots_to_legacy(r["lots"]))
                results.append(res)
        return results

    def get_queries(self, message):
        worklist = self.decode_message(message)
        queries = []
        if worklist.context == "query":
            for sample in worklist.samples:
                queries.append(sample.sample_id)
        return queries
