# -*- coding: utf8 -*-
from __future__ import unicode_literals

import logging
import Queue
import threading

from driverng.transport.base import TransportException


class BaseProtocol(object):

    def __init__(
        self, is_automate=False, logger=None, encoding=None, **kwargs
    ):
        """:param is_automate: True si utilisé pour simuler l'automate."""
        self.connexion = None
        self.is_automate = is_automate
        # Nombre d'essais d'envoi de trames
        self.tries = 0
        self.rtimeout = 0.05
        self.connected = False
        if logger is None:
            logger = logging.getLogger("driverng.protocol")
        self.logger = logger
        self.encoding = encoding
        super(BaseProtocol, self).__init__()

    def send_trames(self, trames):
        raise NotImplementedError

    def build_trames(self, message):
        raise NotImplementedError

    def recv_trames(self):
        raise NotImplementedError

    def connect(self, connexion):
        self.connexion = connexion
        self.connected = True

    def write(self, data):
        return self.connexion.write(data)

    def read(self, size=1, timeout=None):
        if timeout is not None:
            self.connexion.set_timeout(timeout)
        return self.connexion.read(size)

    def readline(self, timeout=None, **kwargs):
        if timeout is not None:
            self.connexion.set_timeout(timeout)
        return self.connexion.readline(**kwargs)

    def can_send_trames(self, wqueue):
        """Méthode appellée avant d'appeller send_trames
        Utile pour les protocoles à états.
        """
        return True

    def handle_reinit(self):
        """Méthode appellée pour reinitialiser la communication
        quand self.connected est False
        """
        self.connected = True

    def _start(self, rqueue, wqueue):
        data = trames = uuid = None
        while True:
            try:
                if not self.connected:
                    self.handle_reinit()

                if self.connexion.poll(self.rtimeout, wqueue):
                    self.connexion.pre_recv()
                    recv_trames = self.recv_trames()
                    self.connexion.post_recv()
                    if recv_trames:
                        # On peut recevoir plusieurs messages
                        # dans la même session
                        if isinstance(recv_trames, list):
                            messages = recv_trames
                        else:
                            messages = [recv_trames]
                        for message in messages:
                            rqueue.put({
                                "data": message.decode(self.encoding),
                                "uuid": None
                            })
                    self.rtimeout = 0.05
                    continue

                if trames is None:
                    if not self.can_send_trames(wqueue):
                        continue
                    try:
                        data = wqueue.get(True, 0.05)
                        uuid = data["uuid"]
                        trames = self.build_trames(
                            data["data"].encode(self.encoding)
                        )
                        meta = data["meta"]
                    except Queue.Empty:
                        continue

                self.tries = 0
                self.connexion.pre_send(meta=meta)
                if self.send_trames(trames):
                    rqueue.put({"data": None, "uuid": uuid})
                    uuid = trames = data = None
                self.connexion.post_send()
            except TransportException:
                if data is not None:
                    self.logger.warning(
                        "TransportException for uuid %s", uuid)
                    # Retransmission
                    wqueue.put(data)
                raise

    def start(self, rqueue, wqueue, background=True):
        if background:
            t = threading.Thread(target=self._start, args=(rqueue, wqueue))
            t.daemon = True
            t.start()
        else:
            self._start(rqueue, wqueue)
