# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from driverng.protocol.base import BaseProtocol
from driverng.utils import (
    ACK, NACK, SYN, STX, ETX, ENQ, DC1, DLE, SOH,
    byte2log,
)


class LHProtocol(BaseProtocol):
    MAX_RECORD_LENGTH = 255
    MAX_TRY = 3

    def wait_for_ctrl(self, ctrl, timeout=10):
        data = self.read(timeout=timeout)
        if ctrl == ACK:
            expected = (ACK, NACK)
        else:
            expected = (ctrl,)
        if not data or data[0] not in expected:
            self.logger.info(
                "%s expected in data '%s'",
                byte2log(ctrl),
                byte2log(data),
            )
        return data[0] == ctrl

    def send_trames(self, trames):
        # -> <enq>
        # <- <enq>
        # -> <# data blocks>
        # <- <ack>
        # -> <stx><block number><payload><checksum><etx>
        # <- <ack>
        # -> <enq>
        # <- <ack/nack/enq>
        # <- <dle>
        # <- <accept/reject>
        if self.is_automate:
            ctrl = SYN
        else:
            ctrl = ENQ

        self.write(ctrl)
        if not self.wait_for_ctrl(ctrl):
            return False

        self.write(b"01")

        if not self.wait_for_ctrl(ACK):
            return False

        self.write(trames)
        if not self.wait_for_ctrl(ACK):
            return False

        self.write(ctrl)
        self.wait_for_ctrl(ACK)

        if not self.is_automate:
            # <dle><accept/reject>
            self.read(2)

        return True

    def build_trames(self, message):
        # 1G.1 RETICS FORMAT
        if self.is_automate:
            trame = b"\r\n" * 6 + 14 * b"-" + b"\r\n" + b"%c%02X\r\n%s\r\n" % (
                DC1,
                len(message.splitlines()),
                b"\r\n".join(message.splitlines()),
            ) + DC1 + 2 * b"\r\n" + 14 * b"-" + b"\r\n"
        else:
            trame = b"%c%02X\r\n%s\r\n" % (
                SOH,
                len(message.splitlines()),
                b"\r\n".join(message.splitlines()),
            )
            trame = b"%-256s" % (trame,)
        return b"%c01%s%04X%c" % (
            STX,
            trame,
            self.checksum(trame),
            ETX,
        )

    def recv_trames(self):
        if self.is_automate:
            ctrl = ENQ
        else:
            ctrl = SYN

        if not self.wait_for_ctrl(ctrl):
            return False

        self.write(ctrl)
        n_blocks = int(self.read(2), 16)
        self.write(ACK)
        message = b""
        self.logger.info("will read %d blocks", n_blocks)
        for _ in range(n_blocks):
            data = self.readline(break_func=lambda x: x and x[-1] == ETX)
            # Debut du message
            if data[0] != STX:
                self.logger.warning("<stx> expected in %s", byte2log(data))
                return False

            if self.check_trame(data):
                self.write(ACK)
                message += self.get_payload(data)
            else:
                self.write(NACK)

        if not self.wait_for_ctrl(ctrl):
            return False

        self.write(ACK)

        if self.is_automate:
            # A: Transmission accepted ready for next
            # B: Transmission accepted, do NOT send more
            # C: Transmission rejected, retry
            # D: Transmission rejected, abandon
            # E: Transmission rejected, at leat one duplicate pending test
            #    request in sample request
            # F: Transmission rejected, patient ID already exists in the LH
            #    database
            self.write(b"%cA" % (DLE,))

        return b"\n".join([
            l for l in message.splitlines()
            if l and l[0] not in (DC1, b"-")])

    def check_trame(self, trame):
        try:
            checksum = int(trame[-5:-1], 16)
        except ValueError:
            self.logger.warning("Missing checksum for %s", byte2log(trame))
            return False

        expected = self.checksum(trame[3:-5])

        if checksum != expected:
            self.logger.warning(
                "Bad checksum for %s: (%X != %X)",
                byte2log(trame),
                checksum,
                expected,
            )
            return False

        return True

    def get_payload(self, data):
        # <stx><nblocks><payload><checksum><etx>
        return data[3:-5]

    @staticmethod
    def checksum(data):
        lsb = 0xFF
        msb = 0xFF
        for i, x in enumerate([ord(c) for c in data]):
            x ^= msb
            x ^= ((x >> 4) % 256)
            msb = lsb ^ ((x >> 3) % 256) ^ ((x << 4) % 256)
            lsb = x ^ ((x << 5) % 256)
        lsb ^= 0xFF
        msb ^= 0xFF
        return int(b"%02X%02X" % (msb, lsb), 16)
