# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng.utils import ACK, NACK, STX, ETX, ETB, EOT, byte2log
from driverng.protocol.base import BaseProtocol


class RapidlabProtocol(BaseProtocol):

    def send_trames(self, trame):
        self.write(trame)
        payload = self.recv_trames()
        return (payload == ACK)

    def send_ack(self, ack):
        self.write(self.build_trames(ack))

    def recv_trames(self):
        # <stx><payload><etx><checksum><eot>
        data = self.read(timeout=1)
        if not data or data[0] != STX:
            self.logger.warning("<stx> expected in %s", byte2log(data))
            return False

        data += self.readline(break_func=lambda s: s and s[-1] == EOT)
        if not self.check_trame(data):
            self.send_ack(NACK)
            return False
        payload = self.get_payload(data)
        if payload in (ACK, NACK):
            return payload
        else:
            self.write(self.build_trames(ACK))
        # Pour ces messages il faut répondre les mêmes données
        # en changeant le message type... :D
        for req, res in (
            (b"ID_REQ", b"ID_DATA"),
            (b"CAL_NEW_AV", b"CAL_REQ"),
            (b"SMP_NEW_AV", b"SMP_REQ"),
            (b"QC_NEW_AV", b"QC_REQ"),
        ):
            if data[1:].startswith(req):
                self.write(self.build_trames(b"%s%s" % (
                    res, data[len(req) + 1:-4])))
                self.recv_trames()
                return False

        # Seuls ces messages peuvents remonter sur MCA
        for allowed in (
            b"PAT_DEMOG_REQ", b"SMP_NEW_DATA", b"CAL_NEW_DATA", b"QC_NEW_DATA",
        ):
            if data[1:].startswith(allowed):
                return payload
        # On ignore le reste
        return False

    def check_trame(self, trame):
        try:
            checksum = int(trame[-3:-1], 16)
            checksum_expected = self.checksum(trame[0:-3])
        except ValueError:
            self.logger.warning(
                "Missing checksum for %s",
                byte2log(trame),
            )
            return False

        if checksum != checksum_expected:
            self.logger.warning(
                "Bad checksum for %s: (%02X != %02X)",
                byte2log(trame),
                checksum,
                checksum_expected,
            )
            return False

        return True

    @staticmethod
    def checksum(data):
        """Fonction de calcul du checksum, appelé par `build_trames`

        :param data: les data à checksumer
        :return: le checksum des data (int)
        """
        return (sum(ord(c) for c in data) & 0xFF) % 0x100

    def get_payload(self, data):
        # <stx><payload><etx><checksum><eot>
        return data[1:-4].replace(ETB, '')

    def build_trames(self, message):
        # <stx><payload><etx><checksum><eot>
        reponse = b"%c%s%c" % (STX, message, ETX,)
        return b"%s%02X%c" % (
            reponse,
            self.checksum(reponse),
            EOT,
        )
