# -*- coding: utf8 -*-
from __future__ import unicode_literals

import time
import os
import fnmatch
import datetime
import shutil
import fcntl
from tempfile import NamedTemporaryFile

from driverng.utils import byte2log
from driverng.transport.base import BaseTransport


class FileTransport(BaseTransport):

    def __init__(self, input_dir, output_dir, fmt, glob, append=False, *args, **kwargs):
        """FileTransport:
        @param input_dir: Répertoire d'entrée
        @param output_dir: Répertoire de sortie
        @param prefix: Préfixe des fichiers.
        @param fmt: Format du nom du fichier en sortie
        @param glob: Format du nom du fichier en entrée
        """
        super(FileTransport, self).__init__(*args, **kwargs)
        self.input_dir = input_dir
        self.output_dir = output_dir
        self.fmt = fmt
        self.glob = glob
        self.append = append
        self._rfile = None
        self._wfile = None
        self._rpath = None
        self._wpath = None
        self._count = 0

    def fileno(self):
        return None

    def _write(self, data):
        self._wfile.write(data)
        return len(data)

    def _read(self, size=None):
        if size:
            return self._rfile.read(size)
        else:
            return self._rfile.read()

    def readline(self, break_func=None):
        data = self._rfile.readline()
        self.logger.debug("recv %s", byte2log(data))
        return data

    def pre_recv(self):
        self.logger.info("Read from %s (lock until available)", self._rpath)
        self._rfile = open(self._rpath, "r")
        fcntl.lockf(self._rfile, fcntl.LOCK_SH)

    def post_recv(self):
        if self._rfile is not None:
            self.logger.info("Remove %s", self._rpath)
            fcntl.lockf(self._rfile, fcntl.LOCK_UN)
            self._rfile.close()
            os.unlink(self._rpath)
            self._rfile = None
            self._rpath = None

    def get_path(self):
        if self._wfile is not None and "filename" in self.meta:
            filename = self.meta["filename"]
        else:
            self._count = (self._count + 1) % 1000
            filename = self.fmt.format(
                now=datetime.datetime.now(),
                count=self._count,
            )
        return os.path.join(self.output_dir, filename)

    def pre_send(self, meta):
        self._wfile = NamedTemporaryFile(delete=False)
        super(FileTransport, self).pre_send(meta)

    def post_send(self):
        if self._wfile is not None:
            while True:
                path = self.get_path()
                if self.append or not os.path.exists(path):
                    break
                self.logger.warning("%s already exists, sleep 1 second", path)
                time.sleep(1)
            self._wfile.close()
            with open(self._wfile.name, "r") as src, open(path, "a") as dst:
                shutil.copyfileobj(src, dst)
            os.unlink(self._wfile.name)
            os.chmod(path, 0o660)
            self.logger.info("Write to %s", path)
        super(FileTransport, self).post_send()

    def poll(self, timeout, wqueue):
        if self.input_dir:
            for filename in os.listdir(self.input_dir):
                if fnmatch.fnmatch(filename, self.glob):
                    self._rpath = os.path.join(self.input_dir, filename)
                    return True
        if wqueue.empty():
            time.sleep(10)
        return False
