# -*- coding: utf8 -*-
from __future__ import unicode_literals

import os
import time
import ftplib
import StringIO
import fnmatch
import datetime

from driverng.utils import byte2log
from driverng.transport.base import BaseTransport


class FTP(object):

    def __init__(self, hostname, username=None, password=None, logger=None):
        self.hostname = hostname
        self.username = username
        self.password = password
        self._ttl = time.time()
        self._ftp = None
        self.logger = logger

    @property
    def _expired(self):
        return (time.time() - self._ttl) > 120

    @property
    def ftp(self):
        if self._ftp is None or self._expired:
            if self._ftp is not None:
                self._ftp.close()
            self._ftp = ftplib.FTP(self.hostname)
            self._ftp.login(self.username, self.password)
        self._ttl = time.time()
        return self._ftp

    def cwd(self, path):
        self.ftp.cwd(path)

    def list(self, path=None):
        files = []
        for filepath in self.ftp.nlst(path):
            files.append(os.path.basename(filepath))
        return files

    def put(self, data, filename):
        return self.ftp.storbinary('STOR ' + filename, StringIO.StringIO(data))

    def get(self, filename):
        buf = StringIO.StringIO()
        self.ftp.retrbinary('RETR ' + filename, buf.write)
        return buf.getvalue()

    def delete(self, filename):
        return self.ftp.delete(filename)


class FTPTransport(BaseTransport):
    def __init__(
        self, hostname, username, password, input_dir, output_dir, fmt,
        glob, *args, **kwargs
    ):
        super(FTPTransport, self).__init__(*args, **kwargs)
        self.ftp = FTP(hostname, username, password, self.logger)
        self.input_dir = input_dir
        self.output_dir = output_dir
        self.fmt = fmt
        self.glob = glob
        self._rpath = None
        self._wfile = None
        self._count = 0
        self._ttl = None
        self._files = []

    def list_files(self):
        if self.input_dir and (
            self._ttl is None or (time.time() - self._ttl) > 10
        ):
            self._files = self.ftp.list(self.input_dir)
            self._ttl = time.time()
        return self._files

    def fileno(self):
        return None

    def _write(self, data):
        self._wfile.write(data)
        return len(data)

    def _read(self, size=None):
        self.logger.info("Read from %s", self._rpath)
        data = self.ftp.get(self._rpath)
        self.logger.debug("recv %s", byte2log(data))
        return data

    def readline(self, break_func=None):
        return self._read()

    def post_recv(self):
        if self._rpath is not None:
            self.logger.info("Remove %s", self._rpath)
            self.ftp.delete(self._rpath)

    def get_filename(self):
        self._count = (self._count + 1) % 1000
        return self.fmt.format(
            now=datetime.datetime.now(),
            count=self._count,
        )

    def pre_send(self, meta):
        self._wfile = StringIO.StringIO()
        super(FTPTransport, self).pre_send(meta)

    def post_send(self):
        if self._wfile is not None:
            if "filename" in self.meta:
                path = os.path.join(self.output_dir, self.meta["filename"])
            else:
                filename = self.get_filename()
                path = os.path.join(self.output_dir, filename)
            self.logger.info("Write to %s", path)
            self.ftp.put(self._wfile.getvalue(), path)
        super(FTPTransport, self).post_send()

    def poll(self, timeout, wqueue):
        for filename in self.list_files():
            if fnmatch.fnmatch(filename, self.glob):
                self.logger.info("input_dir: %s, filename: %s", self.input_dir, filename)
                self._rpath = os.path.join(self.input_dir, filename)
                self._files.remove(filename)
                return True
        return False
