# -*- coding: utf8 -*-
from __future__ import unicode_literals

import socket
import select

from driverng.transport.base import BaseTransport, TransportException
from errno import (
    ENOTCONN,
    EBADF,
    ECONNRESET,
    ESHUTDOWN,
    ECONNABORTED,
    ECONNREFUSED,
)


class TCPTransport(BaseTransport):

    def __init__(self, host, port, server, *args, **kwargs):
        super(TCPTransport, self).__init__(*args, **kwargs)

        if server:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            s.bind((host, port))
            s.listen(1)
            self.logger.info("Waiting for connection on %s:%s", host, port)
            conn, addr = s.accept()
            self.logger.info("Connected by %s", addr)
            self.socket = s
            self.connexion = conn
        else:  # client
            try:
                self.connexion = socket.create_connection(
                    (host, port),
                    10,
                )
            except socket.gaierror as e:
                raise TransportException("Disconnected %s" % e)
            except socket.timeout:
                raise TransportException("Timeout")
            except socket.error as e:
                if e.args[0] in (ECONNREFUSED,):
                    raise TransportException("Disconnected (errno=%d)" % (
                        e.args[0],
                    ))
                else:
                    raise
            self.socket = None

    def fileno(self):
        return self.connexion.fileno()

    def set_timeout(self, timeout):
        self.connexion.settimeout(timeout)

    def poll(self, timeout, wqueue):
        rfds = [self.connexion]

        if self.socket is not None:  # serveur
            rfds.append(self.socket)

        r, _, _ = select.select(rfds, [], [], timeout)

        if self.socket is not None and self.socket in r:
            conn, addr = self.socket.accept()
            self.connexion = conn
            self.logger.info("Connexion remplaced by %s", addr)
            return False

        if self.connexion in r:
            return True
        else:
            return False

    def _read(self, size):
        data = None

        r, _, _ = select.select([self.fileno(), ], [], [],
                                self.connexion.gettimeout())

        if not r:
            raise TransportException("Timeout")

        try:
            data = self.connexion.recv(size)
        except socket.error as e:
            if e.args[0] in (
                    EBADF,
                    ECONNRESET,
                    ENOTCONN,
                    ESHUTDOWN,
                    ECONNABORTED,
            ):
                raise TransportException("Disconnected (errno=%d)" % e.args[0])
            else:
                raise

        if not data:
            raise TransportException("Disconnected (recv 0 bytes)")

        return data

    def _write(self, data):

        _, w, _ = select.select([], [self.fileno(), ], [],
                                self.connexion.gettimeout())

        if not w:
            raise TransportException("Timeout")

        totalsent = 0
        msglen = len(data)

        while totalsent < msglen:

            try:
                sent = self.connexion.send(data[totalsent:])
            except socket.error as e:
                if e.args[0] in (
                        EBADF,
                        ECONNRESET,
                        ENOTCONN,
                        ESHUTDOWN,
                        ECONNABORTED,
                ):
                    raise TransportException("Disconnected (errno=%d)" % (
                        e.args[0],
                    ))
                else:
                    raise

            if sent == 0:
                raise TransportException("Disconnected (sent 0 bytes)")

            totalsent += sent

    def close(self):
        try:
            self.connexion.close()
            if self.socket is not None:
                self.socket.close()
        except socket.error as e:
            if e.args[0] not in (ENOTCONN, EBADF):
                raise
