# -*- coding: utf8 -*-
from __future__ import unicode_literals

import json
import types
import datetime
import jsonschema
import jsonschema_model
import driverng.utils as utils
from collections import OrderedDict


ORDER_SCHEMA = {
    "name": "Order",
    "type": "object",
    "properties": {
        "code": {
            "title": "Code",
            "type": "string"
        },
        "priority": {
            "title": "Priority",
            "type": "string"
        },
        "dilution": {
            "title": "Dilution",
            "type": "string"
        },
        "action": {
            "title": "Action",
            "type": "string",
            "enum": ["new", "add", "del"],
        },
    },
    "required": ["code", "action"],
    "additionalProperties": False,
}

LOT_SCHEMA = {
    "name": "Lot",
    "type": "object",
    "properties": {
        "type": {
            "title": "Type",
            "type": "string",
            "enum": ["reagent", "calibration", "control"]
        },
        "lot": {
            "title": "N°",
            "type": "string"
        },
        "name": {
            "title": "Name",
            "type": "string"
        },
        "expiration": {
            "title": "Expiration",
            "type": "date-time",
        },
    },
    "required": ["type", "lot"],
    "additionalProperties": False,
}

RESULT_SCHEMA = {
    "name": "Result",
    "type": "object",
    "properties": {
        "code": {
            "title": "Code",
            "type": "string"
        },
        "sub_code": {
            "title": "Sub code",
            "type": "string",
        },
        "value": {
            "title": "Value",
            "type": "string",
        },
        "date": {
            "title": "Date",
            "type": "date-time",
        },
        "anterior": {
            "title": "Anterior",
            "type": "boolean",
        },
        "request_id": {
            "title": "Anterior request ID",
            "type": "string",
        },
        "request_date": {
            "title": "Anterior request date",
            "type": "date-time",
        },
        "unit": {
            "title": "Unit",
            "type": "string",
        },
        "flags": {
            "title": "Flags",
            "type": "array", "items": {"title": "Flag", "type": "string"}
        },
        "comments": {
            "title": "Comments",
            "type": "array", "items": {"title": "Comment", "type": "string"}},
        "instrument": {
            "title": "Instrument",
            "type": "string"
        },
        # TODO: LEGACY biorad
        "reagent": {"title": "Reagent", "type": "string"},
        "lots": {"type": "array", "items": LOT_SCHEMA},
    },
    "required": ["code", "value"],
    "additionalProperties": False,
}

SAMPLE_SCHEMA = {
    "name": "Sample",
    "type": "object",
    "properties": {
        "sample_id": {
            "title": "Sample ID",
            "type": "string"
        },
        "type": {
            "title": "Sample type",
            "type": "string",
            "enum": ["patient", "control"],
        },
        "request_id": {
            "title": "Request ID",
            "type": "string",
        },
        "control_lot": {
            "title": "Lot ID",
            "type": "string"
        },
        "control_name": {
            "title": "Lot Name",
            "type": "string"
        },
        "control_level": {
            "title": "Lot Level",
            "type": "string"
        },
        "rack": {
            "title": "Rack",
            "type": "string"
        },
        "collection_date": {
            "title": "Collection date",
            "type": "date-time",
        },
        "collection_end_date": {
            "title": "Collection end date",
            "type": "date-time",
        },
        "request_date": {
            "title": "Request date",
            "type": "date-time"
        },
        "admission_date": {
            "title": "Admission date",
            "type": "date-time",
        },
        "admission_end_date": {
            "title": "Admission end date",
            "type": "date-time",
        },
        "services": {
            "title": "Services",
            "type": "array",
            "name": "Services",
            "items": {
                "title": "Service",
                "name": "Service",
                "type": "object",
                "required": ["code"],
                "properties": {
                    "code": {
                        "title": "Code",
                        "type": "string",
                    },
                    "name": {
                        "title": "Name",
                        "type": "string",
                    },
                },
            },
        },
        "physicians": {
            "title": "Physicians",
            "name": "Physicians",
            "type": "array",
            "items": {
                "title": "Physician",
                "name": "Physician",
                "type": "object",
                "required": ["code"],
                "properties": {
                    "code": {
                        "title": "Code",
                        "type": "string",
                    },
                    "type": {
                        "title": "Type",
                        "type": "string",
                        "enum": ["attending", "sampler", "prescriber"],
                    },
                    "last_name": {
                        "title": "Last name",
                        "type": "string",
                    },
                    "first_name": {
                        "title": "First name",
                        "type": "string",
                    },
                    "phone": {
                        "title": "Phone",
                        "type": "string",
                    },
                },
            },
        },
        # LEGACY
        "physician_name": {"type": "string"},
        "physician_first_name": {"type": "string"},
        # END LEGACY
        "transcriptionist": {
            "title": "Transcriptionist",
            "type": "string",
        },
        "location": {
            "title": "Location",
            "type": "string",
        },
        "location_name": {
            "title": "Location name",
            "type": "string",
        },
        "room": {
            "title": "Room",
            "type": "string",
        },
        "bed": {
            "title": "Bed",
            "type": "string",
        },
        "last_name": {
            "title": "Last name",
            "type": "string",
        },
        "first_name": {
            "title": "First name",
            "type": "string"
        },
        "title": {
            "title": "Title",
            "type": "string",
        },
        "birthdate": {
            "title": "Birth date",
            "type": "date-time"
        },
        "maiden_name": {
            "title": "Maiden name",
            "type": "string",
        },
        "postal_code": {
            "title": "Postal code",
            "type": "string",
        },
        "city": {
            "title": "City",
            "type": "string",
        },
        "country": {
            "title": "Country",
            "type": "string",
        },
        "address": {
            "title": "Address",
            "type": "string",
        },
        "lab": {
            "title": "Laboratory",
            "type": "string",
        },
        "sex": {
            "title": "Sex",
            "type": "string", "enum": ["M", "F", "U"]
        },
        "patient_id": {
            "title": "Patient ID",
            "type": "string"
        },
        "lab_patient_id": {
            "title": "Laboratory patient ID",
            "type": "string",
        },
        "alternate_patient_id": {
            "title": "Alternate patient ID",
            "type": "string",
        },
        "specimen": {
            "title": "Specimen code",
            "type": "string",
        },
        "diagnostic": {
            "title": "Diagnostic",
            "type": "string",
        },
        "height": {
            "title": "Height",
            "type": "string",
        },
        "weight": {
            "title": "Weight",
            "type": "string",
        },
        "phone": {
            "title": "Phone",
            "type": "string",
        },
        "medications": {
            "title": "Medications",
            "type": "array",
            "items": {"type": "string"},
        },
        "comments": {
            "title": "Comments",
            "type": "array",
            "items": {
                "title": "Comment",
                "type": "object",
                "name": "Comment",
                "properties": {
                    "text": {
                        "title": "Text",
                        "type": "string",
                    },
                    "type": {
                        "title": "Type",
                        "type": "string",
                        "enum": ["patient", "sample"],
                    },
                },
            },
        },
        "orders": {"type": "array", "items": ORDER_SCHEMA},
        "results": {"type": "array", "items": RESULT_SCHEMA},
    },
    "required": ["sample_id", "type"],
    "additionalProperties": False,
}

PATIENT_ATTRIBUTES = (
    "last_name", "first_name", "birthdate", "maiden_name", "sex",
    "patient_id", "lab", "lab_patient_id",
    "postal_code", "address", "city", "country", "collection_date",
    "services", "physicians", "location", "location_name", "bed", "room",
    "admission_date", "admission_end_date",
    "height", "weight", "title", "alternate_patient_id", "phone",
    # Legacy
    "service", "physician", "physician_name", "physician_first_name",
    "comments", "medications",
)

PATIENT_SCHEMA = {
    "name": "Patient",
    "type": "object",
    "properties": dict([
        (k, v) for k, v in SAMPLE_SCHEMA["properties"].items()
        if k in PATIENT_ATTRIBUTES]),
    "required": ["patient_id"],
    "additionalProperties": True,
}


SCHEMA = {
    "name": "Worklist",
    "type": "object",
    "required": ["context"],
    "properties": {
        "samples": {"type": "array", "items": SAMPLE_SCHEMA},
        "context": {
            "title": "Context",
            "type": "string",
            "enum": ["order", "query", "result"],
        },
        "sender_id": {
            "title": "Sender ID",
            "type": "string",
        },
        "receiver_id": {
            "title": "Receiver ID",
            "type": "string",
        },
        "orig_data": {"type": "string"},
        "data": {"type": "string"},
        "traceback": {"type": "string"},
    },
    "additionalProperties": False,
}


Model = jsonschema_model.model_factory(SCHEMA)
Patient = jsonschema_model.model_factory(PATIENT_SCHEMA)
Sample = jsonschema_model.model_factory(SAMPLE_SCHEMA)
Order = jsonschema_model.model_factory(ORDER_SCHEMA)
Result = jsonschema_model.model_factory(RESULT_SCHEMA)


class Worklist(Model):
    VALIDATOR = jsonschema.Draft4Validator(
        schema={"type": "date-time"},
        types={
            "string": (unicode, str, types.NoneType),
            "date-time": (datetime.datetime, types.NoneType),
        },
    )
    ELSE = "_else_"
    IGNORE = "_ignore_"

    class ValidationError(Exception):
        pass

    def __init__(self, samples=None, **kwargs):
        super(Worklist, self).__init__(samples=(samples or []), **kwargs)

    def get_worklist(self, unique_together=[], **kwargs):
        samples = []
        for sample in self.samples:
            for couple, orders in self.get_orders(
                    sample["orders"], unique_together, **kwargs):
                new_sample = Sample(sample.copy())
                new_sample["orders"] = orders
                samples.append((new_sample,) + couple)
        return samples

    def get_astm_worklist(self, unique_together=[], **kwargs):
        patients = OrderedDict()
        for sample in self.samples:
            patients.setdefault(
                sample["patient_id"], Patient(dict([
                    (k, sample[k])
                    for k in PATIENT_ATTRIBUTES if k in sample])))
            patients[sample["patient_id"]].setdefault("samples", [])
            orphan_results = sample.results[:]
            for couple, orders in self.get_orders(
                    sample["orders"], unique_together, **kwargs):
                new_sample = Sample(sample.copy())
                new_sample["orders"] = orders
                order_codes = [o.code for o in orders]
                new_sample["results"] = []
                for r in sample.results:
                    if r["code"] in order_codes:
                        orphan_results.remove(r)
                        new_sample["results"].append(Result(r))
                patients[sample["patient_id"]]["samples"].append(
                    ((new_sample,) + couple))

            if orphan_results:
                for couple in patients[sample["patient_id"]]["samples"]:
                    couple[0].results.extend([Result(r) for r in orphan_results])

            if not sample.orders and self.context == "result":
                new_sample = Sample(sample.copy())
                patients[sample["patient_id"]]["samples"].append(
                    ((new_sample,) + (None,) * len(unique_together)))
        return [p for p in patients.values() if p.samples]

    @staticmethod
    def get_orders(orders, unique_together=[], **kwargs):
        r = OrderedDict()
        # key => [values to ignore]
        ignore = {}
        for key, values in kwargs.items():
            for k, v in values.items():
                if v == Worklist.IGNORE:
                    ignore.setdefault(key, []).append(k)
        for order in orders:
            for k, values in ignore.items():
                if order[k] in values:
                    break
            else:
                couple = []
                for k in unique_together:
                    v = kwargs.get(k, {})
                    key = v.get(order[k], v.get(Worklist.ELSE, order[k]))
                    couple.append(key)
                r.setdefault(tuple(couple), []).append(Order(order))
        return r.items()

    def get_or_create(self, **kwargs):
        # simple proxy
        return self.samples.get_or_create(**kwargs)

    def validate(self):
        try:
            return self.VALIDATOR.validate(self, SCHEMA)
        except jsonschema.exceptions.ValidationError as error:
            raise self.ValidationError(error)

    def dumps(self):
        return json.dumps(self, cls=utils.JSONEncoder)

    @classmethod
    def loads(cls, data):
        doc = json.loads(data)
        return cls.from_json(doc)

    @classmethod
    def from_json(cls, doc):
        for sample in doc.get("samples", []):
            for k, v in sample.items():
                if k in (
                    "request_date", "admission_date", "birthdate",
                    "collection_date", "collection_end_date",
                ):
                    sample[k] = utils.decode_isoformat(v)
            for result in sample.get("results", []):
                for rk, rv in result.items():
                    if rk in ("date",):
                        result[rk] = utils.decode_isoformat(rv)
                for lot in result.get("lots", []):
                    for lk, lv in lot.items():
                        if lk == "expiration":
                            lot[lk] = utils.decode_isoformat(lv)

        return cls(**doc)
